/**
 *                                  Apache License
 *                            Version 2.0, January 2004
 *                         http://www.apache.org/licenses/
 *
 *    TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *    1. Definitions.
 *
 *       "License" shall mean the terms and conditions for use, reproduction,
 *       and distribution as defined by Sections 1 through 9 of this document.
 *
 *       "Licensor" shall mean the copyright owner or entity authorized by
 *       the copyright owner that is granting the License.
 *
 *       "Legal Entity" shall mean the union of the acting entity and all
 *       other entities that control, are controlled by, or are under common
 *       control with that entity. For the purposes of this definition,
 *       "control" means (i) the power, direct or indirect, to cause the
 *       direction or management of such entity, whether by contract or
 *       otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *       outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *       "You" (or "Your") shall mean an individual or Legal Entity
 *       exercising permissions granted by this License.
 *
 *       "Source" form shall mean the preferred form for making modifications,
 *       including but not limited to software source code, documentation
 *       source, and configuration files.
 *
 *       "Object" form shall mean any form resulting from mechanical
 *       transformation or translation of a Source form, including but
 *       not limited to compiled object code, generated documentation,
 *       and conversions to other media types.
 *
 *       "Work" shall mean the work of authorship, whether in Source or
 *       Object form, made available under the License, as indicated by a
 *       copyright notice that is included in or attached to the work
 *       (an example is provided in the Appendix below).
 *
 *       "Derivative Works" shall mean any work, whether in Source or Object
 *       form, that is based on (or derived from) the Work and for which the
 *       editorial revisions, annotations, elaborations, or other modifications
 *       represent, as a whole, an original work of authorship. For the purposes
 *       of this License, Derivative Works shall not include works that remain
 *       separable from, or merely link (or bind by name) to the interfaces of,
 *       the Work and Derivative Works thereof.
 *
 *       "Contribution" shall mean any work of authorship, including
 *       the original version of the Work and any modifications or additions
 *       to that Work or Derivative Works thereof, that is intentionally
 *       submitted to Licensor for inclusion in the Work by the copyright owner
 *       or by an individual or Legal Entity authorized to submit on behalf of
 *       the copyright owner. For the purposes of this definition, "submitted"
 *       means any form of electronic, verbal, or written communication sent
 *       to the Licensor or its representatives, including but not limited to
 *       communication on electronic mailing lists, source code control systems,
 *       and issue tracking systems that are managed by, or on behalf of, the
 *       Licensor for the purpose of discussing and improving the Work, but
 *       excluding communication that is conspicuously marked or otherwise
 *       designated in writing by the copyright owner as "Not a Contribution."
 *
 *       "Contributor" shall mean Licensor and any individual or Legal Entity
 *       on behalf of whom a Contribution has been received by Licensor and
 *       subsequently incorporated within the Work.
 *
 *    2. Grant of Copyright License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       copyright license to reproduce, prepare Derivative Works of,
 *       publicly display, publicly perform, sublicense, and distribute the
 *       Work and such Derivative Works in Source or Object form.
 *
 *    3. Grant of Patent License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       (except as stated in this section) patent license to make, have made,
 *       use, offer to sell, sell, import, and otherwise transfer the Work,
 *       where such license applies only to those patent claims licensable
 *       by such Contributor that are necessarily infringed by their
 *       Contribution(s) alone or by combination of their Contribution(s)
 *       with the Work to which such Contribution(s) was submitted. If You
 *       institute patent litigation against any entity (including a
 *       cross-claim or counterclaim in a lawsuit) alleging that the Work
 *       or a Contribution incorporated within the Work constitutes direct
 *       or contributory patent infringement, then any patent licenses
 *       granted to You under this License for that Work shall terminate
 *       as of the date such litigation is filed.
 *
 *    4. Redistribution. You may reproduce and distribute copies of the
 *       Work or Derivative Works thereof in any medium, with or without
 *       modifications, and in Source or Object form, provided that You
 *       meet the following conditions:
 *
 *       (a) You must give any other recipients of the Work or
 *           Derivative Works a copy of this License; and
 *
 *       (b) You must cause any modified files to carry prominent notices
 *           stating that You changed the files; and
 *
 *       (c) You must retain, in the Source form of any Derivative Works
 *           that You distribute, all copyright, patent, trademark, and
 *           attribution notices from the Source form of the Work,
 *           excluding those notices that do not pertain to any part of
 *           the Derivative Works; and
 *
 *       (d) If the Work includes a "NOTICE" text file as part of its
 *           distribution, then any Derivative Works that You distribute must
 *           include a readable copy of the attribution notices contained
 *           within such NOTICE file, excluding those notices that do not
 *           pertain to any part of the Derivative Works, in at least one
 *           of the following places: within a NOTICE text file distributed
 *           as part of the Derivative Works; within the Source form or
 *           documentation, if provided along with the Derivative Works; or,
 *           within a display generated by the Derivative Works, if and
 *           wherever such third-party notices normally appear. The contents
 *           of the NOTICE file are for informational purposes only and
 *           do not modify the License. You may add Your own attribution
 *           notices within Derivative Works that You distribute, alongside
 *           or as an addendum to the NOTICE text from the Work, provided
 *           that such additional attribution notices cannot be construed
 *           as modifying the License.
 *
 *       You may add Your own copyright statement to Your modifications and
 *       may provide additional or different license terms and conditions
 *       for use, reproduction, or distribution of Your modifications, or
 *       for any such Derivative Works as a whole, provided Your use,
 *       reproduction, and distribution of the Work otherwise complies with
 *       the conditions stated in this License.
 *
 *    5. Submission of Contributions. Unless You explicitly state otherwise,
 *       any Contribution intentionally submitted for inclusion in the Work
 *       by You to the Licensor shall be under the terms and conditions of
 *       this License, without any additional terms or conditions.
 *       Notwithstanding the above, nothing herein shall supersede or modify
 *       the terms of any separate license agreement you may have executed
 *       with Licensor regarding such Contributions.
 *
 *    6. Trademarks. This License does not grant permission to use the trade
 *       names, trademarks, service marks, or product names of the Licensor,
 *       except as required for reasonable and customary use in describing the
 *       origin of the Work and reproducing the content of the NOTICE file.
 *
 *    7. Disclaimer of Warranty. Unless required by applicable law or
 *       agreed to in writing, Licensor provides the Work (and each
 *       Contributor provides its Contributions) on an "AS IS" BASIS,
 *       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *       implied, including, without limitation, any warranties or conditions
 *       of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *       PARTICULAR PURPOSE. You are solely responsible for determining the
 *       appropriateness of using or redistributing the Work and assume any
 *       risks associated with Your exercise of permissions under this License.
 *
 *    8. Limitation of Liability. In no event and under no legal theory,
 *       whether in tort (including negligence), contract, or otherwise,
 *       unless required by applicable law (such as deliberate and grossly
 *       negligent acts) or agreed to in writing, shall any Contributor be
 *       liable to You for damages, including any direct, indirect, special,
 *       incidental, or consequential damages of any character arising as a
 *       result of this License or out of the use or inability to use the
 *       Work (including but not limited to damages for loss of goodwill,
 *       work stoppage, computer failure or malfunction, or any and all
 *       other commercial damages or losses), even if such Contributor
 *       has been advised of the possibility of such damages.
 *
 *    9. Accepting Warranty or Additional Liability. While redistributing
 *       the Work or Derivative Works thereof, You may choose to offer,
 *       and charge a fee for, acceptance of support, warranty, indemnity,
 *       or other liability obligations and/or rights consistent with this
 *       License. However, in accepting such obligations, You may act only
 *       on Your own behalf and on Your sole responsibility, not on behalf
 *       of any other Contributor, and only if You agree to indemnify,
 *       defend, and hold each Contributor harmless for any liability
 *       incurred by, or claims asserted against, such Contributor by reason
 *       of your accepting any such warranty or additional liability.
 *
 *    END OF TERMS AND CONDITIONS
 *
 *    APPENDIX: How to apply the Apache License to your work.
 *
 *       To apply the Apache License to your work, attach the following
 *       boilerplate notice, with the fields enclosed by brackets "{}"
 *       replaced with your own identifying information. (Don't include
 *       the brackets!)  The text should be enclosed in the appropriate
 *       comment syntax for the file format. We also recommend that a
 *       file or class name and description of purpose be included on the
 *       same "printed page" as the copyright notice for easier
 *       identification within third-party archives.
 *
 *    Copyright 2014 Edgar Espina
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package org.jooby;

import com.google.common.base.Splitter;
import com.google.common.collect.ImmutableList;
import com.google.inject.Key;
import com.google.inject.name.Names;
import com.typesafe.config.Config;
import static java.util.Objects.requireNonNull;
import org.jooby.funzy.Throwing;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Objects;
import java.util.Optional;
import java.util.function.BiFunction;
import java.util.function.BinaryOperator;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.function.Supplier;

/**
 * Allows to optimize, customize or apply defaults values for application services.
 *
 * <p>
 * A env is represented by it's name. For example: <code>dev</code>, <code>prod</code>, etc... A
 * <strong>dev</strong> env is special and a module provider could do some special configuration for
 * development, like turning off a cache, reloading of resources, etc.
 * </p>
 * <p>
 * Same is true for not <strong>dev</strong> environments. For example, a module provider might
 * create a high performance connection pool, caches, etc.
 * </p>
 * <p>
 * By default env is set to <code>dev</code>, but you can change it by setting the
 * <code>application.env</code> property to anything else.
 * </p>
 *
 * @author edgar
 * @since 0.1.0
 */
public interface Env extends LifeCycle {

  /**
   * Property source for {@link Resolver}
   *
   * @author edgar
   * @since 1.1.0
   */
  interface PropertySource {

    /**
     * Get a property value or throw {@link NoSuchElementException}.
     *
     * @param key Property key/name.
     * @return Value or throw {@link NoSuchElementException}.
     * @throws NoSuchElementException If property is missing.
     */
    @Nonnull
    String get(String key) throws NoSuchElementException;
  }

  /**
   * {@link PropertySource} for {@link Config}.
   *
   * @author edgar
   * @since 1.1.0
   */
  class ConfigSource implements PropertySource {

    private Config source;

    public ConfigSource(final Config source) {
      this.source = source;
    }

    @Override
    public String get(final String key) throws NoSuchElementException {
      if (source.hasPath(key)) {
        return source.getString(key);
      }
      throw new NoSuchElementException(key);
    }

  }

  /**
   * {@link PropertySource} for {@link Map}.
   *
   * @author edgar
   * @since 1.1.0
   */
  class MapSource implements PropertySource {

    private Map<String, Object> source;

    public MapSource(final Map<String, Object> source) {
      this.source = source;
    }

    @Override
    public String get(final String key) throws NoSuchElementException {
      Object value = source.get(key);
      if (value != null) {
        return value.toString();
      }
      throw new NoSuchElementException(key);
    }

  }

  /**
   * Template literal implementation, replaces <code>${expression}</code> from a String using a
   * {@link Config} object.
   *
   * @author edgar
   */
  class Resolver {
    private String startDelim = "${";

    private String endDelim = "}";

    private PropertySource source;

    private boolean ignoreMissing;

    /**
     * Set property source.
     *
     * @param source Source.
     * @return This resolver.
     */
    public Resolver source(final Map<String, Object> source) {
      return source(new MapSource(source));
    }

    /**
     * Set property source.
     *
     * @param source Source.
     * @return This resolver.
     */
    public Resolver source(final PropertySource source) {
      this.source = source;
      return this;
    }

    /**
     * Set property source.
     *
     * @param source Source.
     * @return This resolver.
     */
    public Resolver source(final Config source) {
      return source(new ConfigSource(source));
    }

    /**
     * Set start and end delimiters.
     *
     * @param start Start delimiter.
     * @param end End delimiter.
     * @return This resolver.
     */
    public Resolver delimiters(final String start, final String end) {
      this.startDelim = requireNonNull(start, "Start delimiter required.");
      this.endDelim = requireNonNull(end, "End delmiter required.");
      return this;
    }

    /**
     * Ignore missing property replacement and leave the expression untouch.
     *
     * @return This resolver.
     */
    public Resolver ignoreMissing() {
      this.ignoreMissing = true;
      return this;
    }

    /**
     * Returns a string with all substitutions (the <code>${foo.bar}</code> syntax,
     * see <a href="https://github.com/typesafehub/config/blob/master/HOCON.md">the
     * spec</a>) resolved. Substitutions are looked up using the <code>source</code> param as the
     * root object, that is, a substitution <code>${foo.bar}</code> will be replaced with
     * the result of <code>getValue("foo.bar")</code>.
     *
     * @param text Text to process.
     * @return A processed string.
     */
    public String resolve(final String text) {
      requireNonNull(text, "Text is required.");
      if (text.length() == 0) {
        return "";
      }

      BiFunction<Integer, BiFunction<Integer, Integer, RuntimeException>, RuntimeException> err = (
          start, ex) -> {
        String snapshot = text.substring(0, start);
        int line = Splitter.on('\n').splitToList(snapshot).size();
        int column = start - snapshot.lastIndexOf('\n');
        return ex.apply(line, column);
      };

      StringBuilder buffer = new StringBuilder();
      int offset = 0;
      int start = text.indexOf(startDelim);
      while (start >= 0) {
        int end = text.indexOf(endDelim, start + startDelim.length());
        if (end == -1) {
          throw err.apply(start, (line, column) -> new IllegalArgumentException(
              "found '" + startDelim + "' expecting '" + endDelim + "' at " + line + ":"
                  + column));
        }
        buffer.append(text.substring(offset, start));
        String key = text.substring(start + startDelim.length(), end);
        Object value;
        try {
          value = source.get(key);
        } catch (NoSuchElementException x) {
          if (ignoreMissing) {
            value = text.substring(start, end + endDelim.length());
          } else {
            throw err.apply(start, (line, column) -> new NoSuchElementException(
                "Missing " + startDelim + key + endDelim + " at " + line + ":" + column));
          }
        }
        buffer.append(value);
        offset = end + endDelim.length();
        start = text.indexOf(startDelim, offset);
      }
      if (buffer.length() == 0) {
        return text;
      }
      if (offset < text.length()) {
        buffer.append(text.substring(offset));
      }
      return buffer.toString();
    }
  }

  /**
   * Utility class for generating {@link Key} for named services.
   *
   * @author edgar
   */
  class ServiceKey {
    private Map<Object, Integer> instances = new HashMap<>();

    /**
     * Generate at least one named key for the provided type. If this is the first call for the
     * provided type then it generates an unnamed key.
     *
     * @param type Service type.
     * @param name Service name.
     * @param keys Key callback. Invoked once with a named key, and optionally again with an unamed
     *        key.
     * @param <T> Service type.
     */
    public <T> void generate(final Class<T> type, final String name, final Consumer<Key<T>> keys) {
      Integer c = instances.put(type, instances.getOrDefault(type, 0) + 1);
      if (c == null) {
        // def key
        keys.accept(Key.get(type));
      }
      keys.accept(Key.get(type, Names.named(name)));
    }
  }

  /**
   * Build an jooby environment.
   *
   * @author edgar
   */
  interface Builder {

    /**
     * Build a new environment from a {@link Config} object. The environment is created from the
     * <code>application.env</code> property. If such property is missing, env's name must be:
     * <code>dev</code>.
     *
     * Please note an environment created with this method won't have a {@link Env#router()}.
     *
     * @param config A config instance.
     * @return A new environment.
     */
    @Nonnull
    default Env build(final Config config) {
      return build(config, null, Locale.getDefault());
    }

    /**
     * Build a new environment from a {@link Config} object. The environment is created from the
     * <code>application.env</code> property. If such property is missing, env's name must be:
     * <code>dev</code>.
     *
     * @param config A config instance.
     * @param router Application router.
     * @param locale App locale.
     * @return A new environment.
     */
    @Nonnull
    Env build(Config config, @Nullable Router router, Locale locale);
  }

  /**
   * Default builder.
   */
  Env.Builder DEFAULT = (config, router, locale) -> {
    requireNonNull(config, "Config required.");
    String name = config.hasPath("application.env") ? config.getString("application.env") : "dev";
    return new Env() {

      private ImmutableList.Builder<Throwing.Consumer<Registry>> start = ImmutableList.builder();

      private ImmutableList.Builder<Throwing.Consumer<Registry>> started = ImmutableList.builder();

      private ImmutableList.Builder<Throwing.Consumer<Registry>> shutdown = ImmutableList.builder();

      private Map<String, Function<String, String>> xss = new HashMap<>();

      private Map<Object, Object> globals = new HashMap<>();

      private ServiceKey key = new ServiceKey();

      public <T> Env set(Key<T> key, T value) {
        globals.put(key, value);
        return this;
      }

      public <T> T unset(Key<T> key) {
        return (T) globals.remove(key);
      }

      public <T> Optional<T> get(Key<T> key) {
        T value = (T) globals.get(key);
        return Optional.ofNullable(value);
      }

      @Override
      public String name() {
        return name;
      }

      @Override
      public ServiceKey serviceKey() {
        return key;
      }

      @Override
      public Router router() {
        if (router == null) {
          throw new UnsupportedOperationException();
        }
        return router;
      }

      @Override
      public Config config() {
        return config;
      }

      @Override
      public Locale locale() {
        return locale;
      }

      @Override
      public String toString() {
        return name();
      }

      @Override
      public List<Throwing.Consumer<Registry>> stopTasks() {
        return shutdown.build();
      }

      @Override
      public Env onStop(final Throwing.Consumer<Registry> task) {
        this.shutdown.add(task);
        return this;
      }

      @Override
      public Env onStart(final Throwing.Consumer<Registry> task) {
        this.start.add(task);
        return this;
      }

      @Override
      public LifeCycle onStarted(final Throwing.Consumer<Registry> task) {
        this.started.add(task);
        return this;
      }

      @Override
      public List<Throwing.Consumer<Registry>> startTasks() {
        return this.start.build();
      }

      @Override
      public List<Throwing.Consumer<Registry>> startedTasks() {
        return this.started.build();
      }

      @Override
      public Map<String, Function<String, String>> xss() {
        return Collections.unmodifiableMap(xss);
      }

      @Override
      public Env xss(final String name, final Function<String, String> escaper) {
        xss.put(requireNonNull(name, "Name required."),
            requireNonNull(escaper, "Function required."));
        return this;
      }
    };
  };

  /**
   * @return Env's name.
   */
  @Nonnull
  String name();

  /**
   * Application router.
   *
   * @return Available {@link Router}.
   * @throws UnsupportedOperationException if router isn't available.
   */
  @Nonnull
  Router router() throws UnsupportedOperationException;

  /**
   * @return environment properties.
   */
  @Nonnull
  Config config();

  /**
   * @return Default locale from <code>application.lang</code>.
   */
  @Nonnull
  Locale locale();

  /**
   * @return Utility method for generating keys for named services.
   */
  @Nonnull
  default ServiceKey serviceKey() {
    return new ServiceKey();
  }

  /**
   * Returns a string with all substitutions (the <code>${foo.bar}</code> syntax,
   * see <a href="https://github.com/typesafehub/config/blob/master/HOCON.md">the
   * spec</a>) resolved. Substitutions are looked up using the {@link #config()} as the root object,
   * that is, a substitution <code>${foo.bar}</code> will be replaced with
   * the result of <code>getValue("foo.bar")</code>.
   *
   * @param text Text to process.
   * @return A processed string.
   */
  @Nonnull
  default String resolve(final String text) {
    return resolver().resolve(text);
  }

  /**
   * Creates a new environment {@link Resolver}.
   *
   * @return A resolver object.
   */
  @Nonnull
  default Resolver resolver() {
    return new Resolver().source(config());
  }

  /**
   * Runs the callback function if the current env matches the given name.
   *
   * @param name A name to test for.
   * @param fn A callback function.
   * @param <T> A resulting type.
   * @return A resulting object.
   */
  @Nonnull
  default <T> Optional<T> ifMode(final String name, final Supplier<T> fn) {
    if (name().equals(name)) {
      return Optional.of(fn.get());
    }
    return Optional.empty();
  }

  /**
   * @return XSS escape functions.
   */
  @Nonnull
  Map<String, Function<String, String>> xss();

  /**
   * Get or chain the required xss functions.
   *
   * @param xss XSS to combine.
   * @return Chain of required xss functions.
   */
  @Nonnull
  default Function<String, String> xss(final String... xss) {
    Map<String, Function<String, String>> fn = xss();
    BinaryOperator<Function<String, String>> reduce = Function::andThen;
    return Arrays.asList(xss)
        .stream()
        .map(fn::get)
        .filter(Objects::nonNull)
        .reduce(Function.identity(), reduce);
  }

  /**
   * Set/override a XSS escape function.
   *
   * @param name Escape's name.
   * @param escaper Escape function.
   * @return This environment.
   */
  @Nonnull
  Env xss(String name, Function<String, String> escaper);

  /**
   * @return List of start tasks.
   */
  @Nonnull
  List<Throwing.Consumer<Registry>> startTasks();

  /**
   * @return List of start tasks.
   */
  @Nonnull
  List<Throwing.Consumer<Registry>> startedTasks();

  /**
   * @return List of stop tasks.
   */
  @Nonnull
  List<Throwing.Consumer<Registry>> stopTasks();

  /**
   * Add a global object.
   *
   * @param key Object key.
   * @param value Object value.
   * @param <T> Object type.
   * @return This environment.
   */
  @Nonnull
  <T> Env set(Key<T> key, T value);

  /**
   * Add a global object.
   *
   * @param key Object key.
   * @param value Object value.
   * @param <T> Object type.
   * @return This environment.
   */
  @Nonnull
  default <T> Env set(Class<T> key, T value) {
    return set(Key.get(key), value);
  }

  /**
   * Remove a global object.
   *
   * @param key Object key.
   * @param <T> Object type.
   * @return Object value might be null.
   */
  @Nullable <T> T unset(Key<T> key);

  /**
   * Get an object by key or empty when missing.
   *
   * @param key Object key.
   * @param <T> Object type.
   * @return Object valur or empty.
   */
  @Nonnull
  <T> Optional<T> get(Key<T> key);

  /**
   * Get an object by key or empty when missing.
   *
   * @param key Object key.
   * @param <T> Object type.
   * @return Object valur or empty.
   */
  @Nonnull
  default <T> Optional<T> get(Class<T> key) {
    return get(Key.get(key));
  }
}
