/**
 *                                  Apache License
 *                            Version 2.0, January 2004
 *                         http://www.apache.org/licenses/
 *
 *    TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *    1. Definitions.
 *
 *       "License" shall mean the terms and conditions for use, reproduction,
 *       and distribution as defined by Sections 1 through 9 of this document.
 *
 *       "Licensor" shall mean the copyright owner or entity authorized by
 *       the copyright owner that is granting the License.
 *
 *       "Legal Entity" shall mean the union of the acting entity and all
 *       other entities that control, are controlled by, or are under common
 *       control with that entity. For the purposes of this definition,
 *       "control" means (i) the power, direct or indirect, to cause the
 *       direction or management of such entity, whether by contract or
 *       otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *       outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *       "You" (or "Your") shall mean an individual or Legal Entity
 *       exercising permissions granted by this License.
 *
 *       "Source" form shall mean the preferred form for making modifications,
 *       including but not limited to software source code, documentation
 *       source, and configuration files.
 *
 *       "Object" form shall mean any form resulting from mechanical
 *       transformation or translation of a Source form, including but
 *       not limited to compiled object code, generated documentation,
 *       and conversions to other media types.
 *
 *       "Work" shall mean the work of authorship, whether in Source or
 *       Object form, made available under the License, as indicated by a
 *       copyright notice that is included in or attached to the work
 *       (an example is provided in the Appendix below).
 *
 *       "Derivative Works" shall mean any work, whether in Source or Object
 *       form, that is based on (or derived from) the Work and for which the
 *       editorial revisions, annotations, elaborations, or other modifications
 *       represent, as a whole, an original work of authorship. For the purposes
 *       of this License, Derivative Works shall not include works that remain
 *       separable from, or merely link (or bind by name) to the interfaces of,
 *       the Work and Derivative Works thereof.
 *
 *       "Contribution" shall mean any work of authorship, including
 *       the original version of the Work and any modifications or additions
 *       to that Work or Derivative Works thereof, that is intentionally
 *       submitted to Licensor for inclusion in the Work by the copyright owner
 *       or by an individual or Legal Entity authorized to submit on behalf of
 *       the copyright owner. For the purposes of this definition, "submitted"
 *       means any form of electronic, verbal, or written communication sent
 *       to the Licensor or its representatives, including but not limited to
 *       communication on electronic mailing lists, source code control systems,
 *       and issue tracking systems that are managed by, or on behalf of, the
 *       Licensor for the purpose of discussing and improving the Work, but
 *       excluding communication that is conspicuously marked or otherwise
 *       designated in writing by the copyright owner as "Not a Contribution."
 *
 *       "Contributor" shall mean Licensor and any individual or Legal Entity
 *       on behalf of whom a Contribution has been received by Licensor and
 *       subsequently incorporated within the Work.
 *
 *    2. Grant of Copyright License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       copyright license to reproduce, prepare Derivative Works of,
 *       publicly display, publicly perform, sublicense, and distribute the
 *       Work and such Derivative Works in Source or Object form.
 *
 *    3. Grant of Patent License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       (except as stated in this section) patent license to make, have made,
 *       use, offer to sell, sell, import, and otherwise transfer the Work,
 *       where such license applies only to those patent claims licensable
 *       by such Contributor that are necessarily infringed by their
 *       Contribution(s) alone or by combination of their Contribution(s)
 *       with the Work to which such Contribution(s) was submitted. If You
 *       institute patent litigation against any entity (including a
 *       cross-claim or counterclaim in a lawsuit) alleging that the Work
 *       or a Contribution incorporated within the Work constitutes direct
 *       or contributory patent infringement, then any patent licenses
 *       granted to You under this License for that Work shall terminate
 *       as of the date such litigation is filed.
 *
 *    4. Redistribution. You may reproduce and distribute copies of the
 *       Work or Derivative Works thereof in any medium, with or without
 *       modifications, and in Source or Object form, provided that You
 *       meet the following conditions:
 *
 *       (a) You must give any other recipients of the Work or
 *           Derivative Works a copy of this License; and
 *
 *       (b) You must cause any modified files to carry prominent notices
 *           stating that You changed the files; and
 *
 *       (c) You must retain, in the Source form of any Derivative Works
 *           that You distribute, all copyright, patent, trademark, and
 *           attribution notices from the Source form of the Work,
 *           excluding those notices that do not pertain to any part of
 *           the Derivative Works; and
 *
 *       (d) If the Work includes a "NOTICE" text file as part of its
 *           distribution, then any Derivative Works that You distribute must
 *           include a readable copy of the attribution notices contained
 *           within such NOTICE file, excluding those notices that do not
 *           pertain to any part of the Derivative Works, in at least one
 *           of the following places: within a NOTICE text file distributed
 *           as part of the Derivative Works; within the Source form or
 *           documentation, if provided along with the Derivative Works; or,
 *           within a display generated by the Derivative Works, if and
 *           wherever such third-party notices normally appear. The contents
 *           of the NOTICE file are for informational purposes only and
 *           do not modify the License. You may add Your own attribution
 *           notices within Derivative Works that You distribute, alongside
 *           or as an addendum to the NOTICE text from the Work, provided
 *           that such additional attribution notices cannot be construed
 *           as modifying the License.
 *
 *       You may add Your own copyright statement to Your modifications and
 *       may provide additional or different license terms and conditions
 *       for use, reproduction, or distribution of Your modifications, or
 *       for any such Derivative Works as a whole, provided Your use,
 *       reproduction, and distribution of the Work otherwise complies with
 *       the conditions stated in this License.
 *
 *    5. Submission of Contributions. Unless You explicitly state otherwise,
 *       any Contribution intentionally submitted for inclusion in the Work
 *       by You to the Licensor shall be under the terms and conditions of
 *       this License, without any additional terms or conditions.
 *       Notwithstanding the above, nothing herein shall supersede or modify
 *       the terms of any separate license agreement you may have executed
 *       with Licensor regarding such Contributions.
 *
 *    6. Trademarks. This License does not grant permission to use the trade
 *       names, trademarks, service marks, or product names of the Licensor,
 *       except as required for reasonable and customary use in describing the
 *       origin of the Work and reproducing the content of the NOTICE file.
 *
 *    7. Disclaimer of Warranty. Unless required by applicable law or
 *       agreed to in writing, Licensor provides the Work (and each
 *       Contributor provides its Contributions) on an "AS IS" BASIS,
 *       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *       implied, including, without limitation, any warranties or conditions
 *       of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *       PARTICULAR PURPOSE. You are solely responsible for determining the
 *       appropriateness of using or redistributing the Work and assume any
 *       risks associated with Your exercise of permissions under this License.
 *
 *    8. Limitation of Liability. In no event and under no legal theory,
 *       whether in tort (including negligence), contract, or otherwise,
 *       unless required by applicable law (such as deliberate and grossly
 *       negligent acts) or agreed to in writing, shall any Contributor be
 *       liable to You for damages, including any direct, indirect, special,
 *       incidental, or consequential damages of any character arising as a
 *       result of this License or out of the use or inability to use the
 *       Work (including but not limited to damages for loss of goodwill,
 *       work stoppage, computer failure or malfunction, or any and all
 *       other commercial damages or losses), even if such Contributor
 *       has been advised of the possibility of such damages.
 *
 *    9. Accepting Warranty or Additional Liability. While redistributing
 *       the Work or Derivative Works thereof, You may choose to offer,
 *       and charge a fee for, acceptance of support, warranty, indemnity,
 *       or other liability obligations and/or rights consistent with this
 *       License. However, in accepting such obligations, You may act only
 *       on Your own behalf and on Your sole responsibility, not on behalf
 *       of any other Contributor, and only if You agree to indemnify,
 *       defend, and hold each Contributor harmless for any liability
 *       incurred by, or claims asserted against, such Contributor by reason
 *       of your accepting any such warranty or additional liability.
 *
 *    END OF TERMS AND CONDITIONS
 *
 *    APPENDIX: How to apply the Apache License to your work.
 *
 *       To apply the Apache License to your work, attach the following
 *       boilerplate notice, with the fields enclosed by brackets "{}"
 *       replaced with your own identifying information. (Don't include
 *       the brackets!)  The text should be enclosed in the appropriate
 *       comment syntax for the file format. We also recommend that a
 *       file or class name and description of purpose be included on the
 *       same "printed page" as the copyright notice for easier
 *       identification within third-party archives.
 *
 *    Copyright 2014 Edgar Espina
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package org.jooby;

import java.io.IOException;
import java.io.OutputStream;
import java.util.Iterator;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Optional;
import java.util.function.Function;

import org.jooby.internal.parser.BeanParser;

import com.google.inject.Key;
import com.google.inject.TypeLiteral;

/**
 * Parse a request param (path, query, form) or body to something else.
 *
 * <h1>Registering a parser</h1>
 * <p>
 * There are two ways of registering a parser:
 * </p>
 *
 * <ol>
 * <li>Using the {@link Jooby#parser(Parser)} method</li>
 * <li>From a Guice module:
 *
 * <pre>
 *   Multibinder&lt;Parser&gt; pcb = Multibinder
        .newSetBinder(binder, Parser.class);
     pcb.addBinding().to(MyParser.class);
 * </pre>
 * </li>
 * </ol>
 * Parsers are executed in the order they were registered. The first converter that resolved the
 * type: wins!.
 *
 * <h2>Built-in parsers</h2>
 * <p>
 * These are the built-in parsers:
 * </p>
 * <ol>
 * <li>Primitives and String: convert to int, double, char, string, etc...</li>
 * <li>Enums (case-sensitive)</li>
 * <li>{@link java.util.Date}: It parses a date using the <code>application.dateFormat</code>
 * property.</li>
 * <li>{@link java.time.LocalDate}: It parses a date using the <code>application.dateFormat</code>
 * property.</li>
 * <li>{@link java.util.Locale}</li>
 * <li>Classes with a static method: <code>valueOf</code></li>
 * <li>Classes with a static method: <code>fromName</code></li>
 * <li>Classes with a static method: <code>fromString</code></li>
 * <li>Classes with a public constructor with one <code>String</code> argument</li>
 * <li>It is an Optional&lt;T&gt;, List&lt;T&gt;, Set&lt;T&gt; or SortedSet&lt;T&gt; where T
 * satisfies one of previous rules</li>
 * </ol>
 *
 * @author edgar
 * @see Jooby#parser(Parser)
 * @since 0.6.0
 */
public interface Parser {

  /**
   * A parser callback.
   *
   * @author edgar
   *
   * @param <T> Type of data to parse.
   * @since 0.6.0
   */
  interface Callback<T> {

    /**
     * Parse a raw value to something else.
     *
     * @param data Data to parse.
     * @return A parsed value
     * @throws Exception If something goes wrong.
     */
    Object invoke(T data) throws Throwable;

  }

  /**
   * Expose HTTP params from path, query, form url encoded or multipart request as a raw string.
   *
   * @author edgar
   * @since 0.6.0
   */
  interface ParamReference<T> extends Iterable<T> {

    /**
     * @return Descriptive type: parameter, header, cookie, etc...
     */
    String type();

    /**
     * @return Parameter name.
     */
    String name();

    /**
     * @return Return the first param or throw {@link Err} with a bad request code when missing.
     */
    T first();

    /**
     * @return Return the last param or throw {@link Err} with a bad request code when missing.
     */
    T last();

    /**
     * Get the param at the given index or throw {@link Err} with a bad request code when missing.
     *
     * @param index Param index.
     * @return Param at the given index or throw {@link Err} with a bad request code when missing.
     */
    T get(int index);

    @Override
    Iterator<T> iterator();

    /**
     * @return Number of values for this parameter.
     */
    int size();

  }

  /**
   * Expose the HTTP body as a series of bytes or text.
   *
   * @author edgar
   * @since 0.6.0
   */
  interface BodyReference {
    /**
     * Returns the HTTP body as a byte array.
     *
     * @return HTTP body as byte array.
     * @throws IOException If reading fails.
     */
    byte[] bytes() throws IOException;

    /**
     * Returns the HTTP body as text.
     *
     * @return HTTP body as text.
     * @throws IOException If reading fails.
     */
    String text() throws IOException;

    /**
     * @return Body length.
     */
    long length();

    /**
     * Write the content to the given output stream. This method won't close the
     * {@link OutputStream}.
     *
     * @param output An output stream.
     * @throws Exception If write fails.
     */
    void writeTo(final OutputStream output) throws Exception;
  }

  /**
   * A parser can be executed against a simply HTTP param, a set of HTTP params, an file
   * {@link Upload} or HTTP {@link BodyReference}.
   *
   * This class provides utility methods for selecting one of the previous source. It is possible to
   * write a parser and apply it against multiple sources, like HTTP param and HTTP body.
   *
   * Here is an example that will parse text to an int, provided as a HTTP param or body:
   *
   * <pre>
   * {
   *   parser((type, ctx) {@literal ->} {
   *     if (type.getRawType() == int.class) {
   *       return ctx
   *           .param(values {@literal ->} Integer.parseInt(values.get(0))
   *           .body(body {@literal ->} Integer.parseInt(body.text()));
   *     }
   *     return ctx.next();
   *   });
   *
   *   get("/", req {@literal ->} {
   *     // use the param strategy
   *     return req.param("p").intValue();
   *   });
   *
   *   post("/", req {@literal ->} {
   *     // use the body strategy
   *     return req.body().intValue();
   *   });
   * }
   * </pre>
   *
   * @author edgar
   * @since 0.6.0
   */
  interface Builder {

    /**
     * Add a HTTP body callback. The Callback will be executed when current context is bound to the
     * HTTP body via {@link Request#body()}.
     *
     * If current {@link Context} isn't a HTTP body a call to {@link Context#next()} is made.
     *
     * @param callback A body parser callback.
     * @return This builder.
     */
    Builder body(Callback<BodyReference> callback);

    /**
     * Like {@link #body(Callback)} but it skip the callback if the requested type is an
     * {@link Optional}.
     *
     * @param callback A body parser callback.
     * @return This builder.
     */
    Builder ifbody(Callback<BodyReference> callback);

    /**
     * Add a HTTP param callback. The Callback will be executed when current context is bound to a
     * HTTP param via {@link Request#param(String)}.
     *
     * If current {@link Context} isn't a HTTP param a call to {@link Context#next()} is made.
     *
     * @param callback A param parser callback.
     * @return This builder.
     */
    Builder param(Callback<ParamReference<String>> callback);

    /**
     * Like {@link #param(Callback)} but it skip the callback if the requested type is an
     * {@link Optional}.
     *
     * @param callback A param parser callback.
     * @return This builder.
     */
    Builder ifparam(Callback<ParamReference<String>> callback);

    /**
     * Add a HTTP params callback. The Callback will be executed when current context is bound to a
     * HTTP params via {@link Request#params()}.
     *
     * If current {@link Context} isn't a HTTP params a call to {@link Context#next()} is made.
     *
     * @param callback A params parser callback.
     * @return This builder.
     */
    Builder params(Callback<Map<String, Mutant>> callback);

    /**
     * Like {@link #params(Callback)} but it skip the callback if the requested type is an
     * {@link Optional}.
     *
     * @param callback A params parser callback.
     * @return This builder.
     */
    Builder ifparams(Callback<Map<String, Mutant>> callback);

  }

  /**
   * Allows you to access to parsing strategies, content type view {@link #type()} and invoke next
   * parser in the chain via {@link #next()} methods.
   *
   * @author edgar
   * @since 0.6.0
   */
  interface Context extends Builder {

    /**
     * Requires a service with the given type.
     *
     * @param type Service type.
     * @param <T> Service type.
     * @return A service.
     */
    <T> T require(final Class<T> type);

    /**
     * Requires a service with the given type.
     *
     * @param type Service type.
     * @param <T> Service type.
     * @return A service.
     */
    <T> T require(final TypeLiteral<T> type);

    /**
     * Requires a service with the given type.
     *
     * @param key Service key.
     * @param <T> Service type.
     * @return A service.
     */
    <T> T require(final Key<T> key);

    /**
     * Content Type header, if current context was bind to a HTTP body via {@link Request#body()}.
     * If current context was bind to a HTTP param, media type is set to <code>text/plain</code>.
     *
     * @return Current type.
     */
    MediaType type();

    /**
     * Invoke next parser in the chain.
     *
     * @return A parsed value.
     * @throws Exception An err with a 400 status.
     */
    Object next() throws Throwable;

    /**
     * Invoke next parser in the chain and switch/change the target type we are looking for. Useful
     * for generic containers classes, like collections or optional values.
     *
     * @param type A new type to use.
     * @return A parsed value.
     * @throws Exception An err with a 400 status.
     */
    Object next(TypeLiteral<?> type) throws Throwable;

    /**
     * Invoke next parser in the chain and switch/change the target type we are looking for but also
     * the current value. Useful for generic containers classes, like collections or optional
     * values.
     *
     * @param type A new type to use.
     * @param data Data to be parsed.
     * @return A parsed value.
     * @throws Exception An err with a 400 status.
     */
    Object next(TypeLiteral<?> type, Object data) throws Throwable;

  }

  /** Utility function to handle empty values as {@link NoSuchElementException}. */
  static Function<String, String> NOT_EMPTY = v -> {
    if (v.length() == 0) {
      throw new NoSuchElementException();
    }
    return v;
  };

  /**
   * <p>
   * Parse one or more values to the required type. If the parser doesn't support the required type
   * a call to {@link Context#next(TypeLiteral, Object)} must be done.
   * </p>
   *
   * Example:
   *
   * <pre>
   *  Parser converter = (type, ctx) {@literal ->} {
   *    if (type.getRawType() == MyType.class) {
   *      // convert to MyType
   *      return ctx.param(values {@literal ->} new MyType(values.get(0)));
   *    }
   *    // no luck! move next
   *    return ctx.next();
   *  }
   * </pre>
   *
   * It's also possible to create generic/parameterized types too:
   *
   * <pre>
   *  public class MyContainerType&lt;T&gt; {}
   *
   *  ParamConverter converter = (type, ctx) {@literal ->} {
   *    if (type.getRawType() == MyContainerType.class) {
   *      // Creates a new type from current generic type
   *      TypeLiterale&lt;?&gt; paramType = TypeLiteral
   *        .get(((ParameterizedType) toType.getType()).getActualTypeArguments()[0]);
   *
   *      // Ask param converter to resolve the new/next type.
   *      Object result = ctx.next(paramType);
   *      return new MyType(result);
   *    }
   *    // no luck! move next
   *    return ctx.next();
   *  }
   * </pre>
   *
   * @param type Requested type.
   * @param ctx Execution context.
   * @return A parsed value.
   * @throws Throwable If conversion fails.
   */
  Object parse(TypeLiteral<?> type, Context ctx) throws Throwable;

  /**
   * Overwrite the default bean parser with <code>empty/null</code> supports. The default bean
   * parser doesn't allow <code>null</code>, so if a parameter is optional you must declare it as
   * {@link Optional} otherwise parsing fails with a <code>404/500</code> status code.
   *
   * For example:
   * <pre>{@code
   *
   * public class Book {
   *
   *   public String title;
   *
   *   public Date releaseDate;
   *
   *   public String toString() {
   *     return title + ":" + releaseDate;
   *   }
   * }
   *
   * {
   *   parser(Parser.bean(true));
   *
   *   post("/", req -> {
   *     return req.params(Book.class).toString();
   *   });
   * }
   * }</pre>
   *
   * <p>
   * With <code>/?title=Title&amp;releaseDate=</code> prints <code>Title:null</code>.
   * </p>
   * <p>
   * Now, same call with <code>lenient=false</code> results in <code>Bad Request: 400</code>
   * because <code>releaseDate</code> if required and isn't present in the HTTP request.
   * </p>
   *
   * <p>
   * This feature is useful while submitting forms.
   * </p>
   *
   * @param lenient Enabled null/empty supports while parsing HTTP params as Java Beans.
   * @return A new bean parser.
   */
  static Parser bean(final boolean lenient) {
    return new BeanParser(lenient);
  }
}
