/**
 *                                  Apache License
 *                            Version 2.0, January 2004
 *                         http://www.apache.org/licenses/
 *
 *    TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *    1. Definitions.
 *
 *       "License" shall mean the terms and conditions for use, reproduction,
 *       and distribution as defined by Sections 1 through 9 of this document.
 *
 *       "Licensor" shall mean the copyright owner or entity authorized by
 *       the copyright owner that is granting the License.
 *
 *       "Legal Entity" shall mean the union of the acting entity and all
 *       other entities that control, are controlled by, or are under common
 *       control with that entity. For the purposes of this definition,
 *       "control" means (i) the power, direct or indirect, to cause the
 *       direction or management of such entity, whether by contract or
 *       otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *       outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *       "You" (or "Your") shall mean an individual or Legal Entity
 *       exercising permissions granted by this License.
 *
 *       "Source" form shall mean the preferred form for making modifications,
 *       including but not limited to software source code, documentation
 *       source, and configuration files.
 *
 *       "Object" form shall mean any form resulting from mechanical
 *       transformation or translation of a Source form, including but
 *       not limited to compiled object code, generated documentation,
 *       and conversions to other media types.
 *
 *       "Work" shall mean the work of authorship, whether in Source or
 *       Object form, made available under the License, as indicated by a
 *       copyright notice that is included in or attached to the work
 *       (an example is provided in the Appendix below).
 *
 *       "Derivative Works" shall mean any work, whether in Source or Object
 *       form, that is based on (or derived from) the Work and for which the
 *       editorial revisions, annotations, elaborations, or other modifications
 *       represent, as a whole, an original work of authorship. For the purposes
 *       of this License, Derivative Works shall not include works that remain
 *       separable from, or merely link (or bind by name) to the interfaces of,
 *       the Work and Derivative Works thereof.
 *
 *       "Contribution" shall mean any work of authorship, including
 *       the original version of the Work and any modifications or additions
 *       to that Work or Derivative Works thereof, that is intentionally
 *       submitted to Licensor for inclusion in the Work by the copyright owner
 *       or by an individual or Legal Entity authorized to submit on behalf of
 *       the copyright owner. For the purposes of this definition, "submitted"
 *       means any form of electronic, verbal, or written communication sent
 *       to the Licensor or its representatives, including but not limited to
 *       communication on electronic mailing lists, source code control systems,
 *       and issue tracking systems that are managed by, or on behalf of, the
 *       Licensor for the purpose of discussing and improving the Work, but
 *       excluding communication that is conspicuously marked or otherwise
 *       designated in writing by the copyright owner as "Not a Contribution."
 *
 *       "Contributor" shall mean Licensor and any individual or Legal Entity
 *       on behalf of whom a Contribution has been received by Licensor and
 *       subsequently incorporated within the Work.
 *
 *    2. Grant of Copyright License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       copyright license to reproduce, prepare Derivative Works of,
 *       publicly display, publicly perform, sublicense, and distribute the
 *       Work and such Derivative Works in Source or Object form.
 *
 *    3. Grant of Patent License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       (except as stated in this section) patent license to make, have made,
 *       use, offer to sell, sell, import, and otherwise transfer the Work,
 *       where such license applies only to those patent claims licensable
 *       by such Contributor that are necessarily infringed by their
 *       Contribution(s) alone or by combination of their Contribution(s)
 *       with the Work to which such Contribution(s) was submitted. If You
 *       institute patent litigation against any entity (including a
 *       cross-claim or counterclaim in a lawsuit) alleging that the Work
 *       or a Contribution incorporated within the Work constitutes direct
 *       or contributory patent infringement, then any patent licenses
 *       granted to You under this License for that Work shall terminate
 *       as of the date such litigation is filed.
 *
 *    4. Redistribution. You may reproduce and distribute copies of the
 *       Work or Derivative Works thereof in any medium, with or without
 *       modifications, and in Source or Object form, provided that You
 *       meet the following conditions:
 *
 *       (a) You must give any other recipients of the Work or
 *           Derivative Works a copy of this License; and
 *
 *       (b) You must cause any modified files to carry prominent notices
 *           stating that You changed the files; and
 *
 *       (c) You must retain, in the Source form of any Derivative Works
 *           that You distribute, all copyright, patent, trademark, and
 *           attribution notices from the Source form of the Work,
 *           excluding those notices that do not pertain to any part of
 *           the Derivative Works; and
 *
 *       (d) If the Work includes a "NOTICE" text file as part of its
 *           distribution, then any Derivative Works that You distribute must
 *           include a readable copy of the attribution notices contained
 *           within such NOTICE file, excluding those notices that do not
 *           pertain to any part of the Derivative Works, in at least one
 *           of the following places: within a NOTICE text file distributed
 *           as part of the Derivative Works; within the Source form or
 *           documentation, if provided along with the Derivative Works; or,
 *           within a display generated by the Derivative Works, if and
 *           wherever such third-party notices normally appear. The contents
 *           of the NOTICE file are for informational purposes only and
 *           do not modify the License. You may add Your own attribution
 *           notices within Derivative Works that You distribute, alongside
 *           or as an addendum to the NOTICE text from the Work, provided
 *           that such additional attribution notices cannot be construed
 *           as modifying the License.
 *
 *       You may add Your own copyright statement to Your modifications and
 *       may provide additional or different license terms and conditions
 *       for use, reproduction, or distribution of Your modifications, or
 *       for any such Derivative Works as a whole, provided Your use,
 *       reproduction, and distribution of the Work otherwise complies with
 *       the conditions stated in this License.
 *
 *    5. Submission of Contributions. Unless You explicitly state otherwise,
 *       any Contribution intentionally submitted for inclusion in the Work
 *       by You to the Licensor shall be under the terms and conditions of
 *       this License, without any additional terms or conditions.
 *       Notwithstanding the above, nothing herein shall supersede or modify
 *       the terms of any separate license agreement you may have executed
 *       with Licensor regarding such Contributions.
 *
 *    6. Trademarks. This License does not grant permission to use the trade
 *       names, trademarks, service marks, or product names of the Licensor,
 *       except as required for reasonable and customary use in describing the
 *       origin of the Work and reproducing the content of the NOTICE file.
 *
 *    7. Disclaimer of Warranty. Unless required by applicable law or
 *       agreed to in writing, Licensor provides the Work (and each
 *       Contributor provides its Contributions) on an "AS IS" BASIS,
 *       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *       implied, including, without limitation, any warranties or conditions
 *       of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *       PARTICULAR PURPOSE. You are solely responsible for determining the
 *       appropriateness of using or redistributing the Work and assume any
 *       risks associated with Your exercise of permissions under this License.
 *
 *    8. Limitation of Liability. In no event and under no legal theory,
 *       whether in tort (including negligence), contract, or otherwise,
 *       unless required by applicable law (such as deliberate and grossly
 *       negligent acts) or agreed to in writing, shall any Contributor be
 *       liable to You for damages, including any direct, indirect, special,
 *       incidental, or consequential damages of any character arising as a
 *       result of this License or out of the use or inability to use the
 *       Work (including but not limited to damages for loss of goodwill,
 *       work stoppage, computer failure or malfunction, or any and all
 *       other commercial damages or losses), even if such Contributor
 *       has been advised of the possibility of such damages.
 *
 *    9. Accepting Warranty or Additional Liability. While redistributing
 *       the Work or Derivative Works thereof, You may choose to offer,
 *       and charge a fee for, acceptance of support, warranty, indemnity,
 *       or other liability obligations and/or rights consistent with this
 *       License. However, in accepting such obligations, You may act only
 *       on Your own behalf and on Your sole responsibility, not on behalf
 *       of any other Contributor, and only if You agree to indemnify,
 *       defend, and hold each Contributor harmless for any liability
 *       incurred by, or claims asserted against, such Contributor by reason
 *       of your accepting any such warranty or additional liability.
 *
 *    END OF TERMS AND CONDITIONS
 *
 *    APPENDIX: How to apply the Apache License to your work.
 *
 *       To apply the Apache License to your work, attach the following
 *       boilerplate notice, with the fields enclosed by brackets "{}"
 *       replaced with your own identifying information. (Don't include
 *       the brackets!)  The text should be enclosed in the appropriate
 *       comment syntax for the file format. We also recommend that a
 *       file or class name and description of purpose be included on the
 *       same "printed page" as the copyright notice for easier
 *       identification within third-party archives.
 *
 *    Copyright 2014 Edgar Espina
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package org.jooby;

import static java.util.Objects.requireNonNull;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.nio.charset.Charset;
import java.util.Optional;

import org.jooby.Cookie.Definition;

import com.google.common.collect.ImmutableList;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;

/**
 * Give you access to the actual HTTP response. You can read/write headers and write HTTP body.
 *
 * @author edgar
 * @since 0.1.0
 */
public interface Response {

  /**
   * A forwarding response.
   *
   * @author edgar
   * @since 0.1.0
   */
  class Forwarding implements Response {

    /** The target response. */
    protected final Response rsp;

    /**
     * Creates a new {@link Forwarding} response.
     *
     * @param response A response object.
     */
    public Forwarding(final Response response) {
      this.rsp = requireNonNull(response, "A response is required.");
    }

    @Override
    public void download(final String filename, final InputStream stream) throws Throwable {
      rsp.download(filename, stream);
    }

    @Override
    public void download(final File file) throws Throwable {
      rsp.download(file);
    }

    @Override
    public void download(final String filename, final File file) throws Throwable {
      rsp.download(filename, file);
    }

    @Override
    public void download(final String filename) throws Throwable {
      rsp.download(filename);
    }

    @Override
    public void download(final String filename, final String location) throws Throwable {
      rsp.download(filename, location);
    }

    @Override
    public Response cookie(final String name, final String value) {
      rsp.cookie(name, value);
      return this;
    }

    @Override
    public Response cookie(final Cookie cookie) {
      rsp.cookie(cookie);
      return this;
    }

    @Override
    public Response cookie(final Definition cookie) {
      rsp.cookie(cookie);
      return this;
    }

    @Override
    public Response clearCookie(final String name) {
      rsp.clearCookie(name);
      return this;
    }

    @Override
    public Mutant header(final String name) {
      return rsp.header(name);
    }

    @Override
    public Response header(final String name, final Object value) {
      rsp.header(name, value);
      return this;
    }

    @Override
    public Response header(final String name, final Object... values) {
      rsp.header(name, values);
      return this;
    }

    @Override
    public Response header(final String name, final Iterable<Object> values) {
      rsp.header(name, values);
      return this;
    }

    @Override
    public Charset charset() {
      return rsp.charset();
    }

    @Override
    public Response charset(final Charset charset) {
      rsp.charset(charset);
      return this;
    }

    @Override
    public Response length(final long length) {
      rsp.length(length);
      return this;
    }

    @Override
    public Optional<MediaType> type() {
      return rsp.type();
    }

    @Override
    public Response type(final MediaType type) {
      rsp.type(type);
      return this;
    }

    @Override
    public Response type(final String type) {
      rsp.type(type);
      return this;
    }

    @Override
    public void send(final Object result) throws Throwable {
      // Special case: let the default response to deal with Object refs.
      // once resolved it will call the Result version.
      Response.super.send(result);
    }

    @Override
    public void send(final Result result) throws Throwable {
      rsp.send(result);
    }

    @Override
    public void end() {
      rsp.end();
    }

    @Override
    public void redirect(final String location) throws Throwable {
      rsp.redirect(location);
    }

    @Override
    public void redirect(final Status status, final String location) throws Throwable {
      rsp.redirect(status, location);
    }

    @Override
    public Optional<Status> status() {
      return rsp.status();
    }

    @Override
    public Response status(final Status status) {
      rsp.status(status);
      return this;
    }

    @Override
    public Response status(final int status) {
      rsp.status(status);
      return this;
    }

    @Override
    public boolean committed() {
      return rsp.committed();
    }

    @Override
    public void after(final Route.After handler) {
      rsp.after(handler);
    }

    @Override
    public void complete(final Route.Complete handler) {
      rsp.complete(handler);
    }

    @Override
    public String toString() {
      return rsp.toString();
    }

    @Override public boolean isResetHeadersOnError() {
      return rsp.isResetHeadersOnError();
    }

    @Override public void setResetHeadersOnError(boolean value) {
      this.setResetHeadersOnError(value);
    }

    /**
     * Unwrap a response in order to find out the target instance.
     *
     * @param rsp A response.
     * @return A target instance (not a {@link Response.Forwarding}).
     */
    public static Response unwrap(final Response rsp) {
      requireNonNull(rsp, "A response is required.");
      Response root = rsp;
      while (root instanceof Forwarding) {
        root = ((Forwarding) root).rsp;
      }
      return root;
    }
  }

  /**
   * Transfer the file at path as an "attachment". Typically, browsers will prompt the user for
   * download. The <code>Content-Disposition</code> "filename=" parameter (i.e. the one that will
   * appear in the browser dialog) is set to filename.
   *
   * @param filename A file name to use.
   * @param stream A stream to attach.
   * @throws Exception If something goes wrong.
   */
  void download(String filename, InputStream stream) throws Throwable;

  /**
   * Transfer the file at path as an "attachment". Typically, browsers will prompt the user for
   * download. The <code>Content-Disposition</code> "filename=" parameter (i.e. the one that will
   * appear in the browser dialog) is set to filename by default.
   *
   * @param location Classpath location of the file.
   * @throws Exception If something goes wrong.
   */
  default void download(final String location) throws Throwable {
    download(location, location);
  }

  /**
   * Transfer the file at path as an "attachment". Typically, browsers will prompt the user for
   * download. The <code>Content-Disposition</code> "filename=" parameter (i.e. the one that will
   * appear in the browser dialog) is set to filename by default.
   *
   * @param filename A file name to use.
   * @param location classpath location of the file.
   * @throws Exception If something goes wrong.
   */
  void download(final String filename, final String location) throws Throwable;

  /**
   * Transfer the file at path as an "attachment". Typically, browsers will prompt the user for
   * download. The <code>Content-Disposition</code> "filename=" parameter (i.e. the one that will
   * appear in the browser dialog) is set to filename by default.
   *
   * @param file A file to use.
   * @throws Exception If something goes wrong.
   */
  default void download(final File file) throws Throwable {
    download(file.getName(), file);
  }

  /**
   * Transfer the file at path as an "attachment". Typically, browsers will prompt the user for
   * download. The <code>Content-Disposition</code> "filename=" parameter (i.e. the one that will
   * appear in the browser dialog) is set to filename.
   *
   * @param filename A file name to use.
   * @param file A file to use.
   * @throws Exception If something goes wrong.
   */
  default void download(final String filename, final File file) throws Throwable {
    length(file.length());
    download(filename, new FileInputStream(file));
  }

  /**
   * Adds the specified cookie to the response.
   *
   * @param name A cookie's name.
   * @param value A cookie's value.
   * @return This response.
   */
  @Nonnull
  default Response cookie(final String name, final String value) {
    return cookie(new Cookie.Definition(name, value));
  }

  /**
   * Adds the specified cookie to the response.
   *
   * @param cookie A cookie definition.
   * @return This response.
   */
  @Nonnull
  Response cookie(final Cookie.Definition cookie);

  /**
   * Adds the specified cookie to the response.
   *
   * @param cookie A cookie.
   * @return This response.
   */
  @Nonnull
  Response cookie(Cookie cookie);

  /**
   * Discard a cookie from response. Discard is done by setting maxAge=0.
   *
   * @param name Cookie's name.
   * @return This response.
   */
  @Nonnull
  Response clearCookie(String name);

  /**
   * Get a header with the given name.
   *
   * @param name A name.
   * @return A HTTP header.
   */
  @Nonnull
  Mutant header(String name);

  /**
   * Sets a response header with the given name and value. If the header had already been set,
   * the new value overwrites the previous one.
   *
   * @param name Header's name.
   * @param value Header's value.
   * @return This response.
   */
  @Nonnull
  Response header(String name, Object value);

  /**
   * Sets a response header with the given name and value. If the header had already been set,
   * the new value overwrites the previous one.
   *
   * @param name Header's name.
   * @param values Header's value.
   * @return This response.
   */
  @Nonnull
  default Response header(final String name, final Object... values) {
    return header(name, ImmutableList.builder().add(values).build());
  }

  /**
   * Sets a response header with the given name and value. If the header had already been set,
   * the new value overwrites the previous one.
   *
   * @param name Header's name.
   * @param values Header's value.
   * @return This response.
   */
  @Nonnull
  Response header(String name, Iterable<Object> values);

  /**
   * If charset is not set this method returns charset defined in the request body. If the request
   * doesn't specify a character encoding, this method return the global charset:
   * <code>application.charset</code>.
   *
   * @return A current charset.
   */
  @Nonnull
  Charset charset();

  /**
   * Set the {@link Charset} to use and set the <code>Content-Type</code> header with the current
   * charset.
   *
   * @param charset A charset.
   * @return This response.
   */
  @Nonnull
  Response charset(Charset charset);

  /**
   * Set the length of the response and set the <code>Content-Length</code> header.
   *
   * @param length Length of response.
   * @return This response.
   */
  @Nonnull
  Response length(long length);

  /**
   * @return Get the response type.
   */
  @Nonnull
  Optional<MediaType> type();

  /**
   * Set the response media type and set the <code>Content-Type</code> header.
   *
   * @param type A media type.
   * @return This response.
   */
  @Nonnull
  Response type(MediaType type);

  /**
   * Set the response media type and set the <code>Content-Type</code> header.
   *
   * @param type A media type.
   * @return This response.
   */
  @Nonnull
  default Response type(final String type) {
    return type(MediaType.valueOf(type));
  }

  /**
   * Responsible of writing the given body into the HTTP response.
   *
   * @param result The HTTP body.
   * @throws Exception If the response write fails.
   */
  default void send(final @Nullable Object result) throws Throwable {
    if (result instanceof Result) {
      send((Result) result);
    } else if (result != null) {
      // wrap body
      Result b = Results.with(result);
      status().ifPresent(b::status);
      type().ifPresent(b::type);
      send(b);
    } else {
      throw new NullPointerException("Response required.");
    }
  }

  /**
   * Responsible of writing the given body into the HTTP response.
   *
   * @param result A HTTP response.
   * @throws Exception If the response write fails.
   */
  void send(Result result) throws Throwable;

  /**
   * Redirect to the given url with status code defaulting to {@link Status#FOUND}.
   *
   * <pre>
   *  rsp.redirect("/foo/bar");
   *  rsp.redirect("http://example.com");
   *  rsp.redirect("http://example.com");
   *  rsp.redirect("../login");
   * </pre>
   *
   * Redirects can be a fully qualified URI for redirecting to a different site:
   *
   * <pre>
   *   rsp.redirect("http://google.com");
   * </pre>
   *
   * Redirects can be relative to the root of the host name. For example, if you were
   * on <code>http://example.com/admin/post/new</code>, the following redirect to /admin would
   * land you at <code>http://example.com/admin</code>:
   *
   * <pre>
   *   rsp.redirect("/admin");
   * </pre>
   *
   * Redirects can be relative to the current URL. A redirection of post/new, from
   * <code>http://example.com/blog/admin/</code> (notice the trailing slash), would give you
   * <code>http://example.com/blog/admin/post/new.</code>
   *
   * <pre>
   *   rsp.redirect("post/new");
   * </pre>
   *
   * Redirecting to post/new from <code>http://example.com/blog/admin</code> (no trailing slash),
   * will take you to <code>http://example.com/blog/post/new</code>.
   *
   * <p>
   * If you found the above behavior confusing, think of path segments as directories (have trailing
   * slashes) and files, it will start to make sense.
   * </p>
   *
   * Pathname relative redirects are also possible. If you were on
   * <code>http://example.com/admin/post/new</code>, the following redirect would land you at
   * <code>http//example.com/admin</code>:
   *
   * <pre>
   *   rsp.redirect("..");
   * </pre>
   *
   * A back redirection will redirect the request back to the <code>Referer</code>, defaulting to
   * <code>/</code> when missing.
   *
   * <pre>
   *   rsp.redirect("back");
   * </pre>
   *
   * @param location Either a relative or absolute location.
   * @throws Throwable If redirection fails.
   */
  default void redirect(final String location) throws Throwable {
    redirect(Status.FOUND, location);
  }

  /**
   * Redirect to the given url with status code defaulting to {@link Status#FOUND}.
   *
   * <pre>
   *  rsp.redirect("/foo/bar");
   *  rsp.redirect("http://example.com");
   *  rsp.redirect("http://example.com");
   *  rsp.redirect("../login");
   * </pre>
   *
   * Redirects can be a fully qualified URI for redirecting to a different site:
   *
   * <pre>
   *   rsp.redirect("http://google.com");
   * </pre>
   *
   * Redirects can be relative to the root of the host name. For example, if you were
   * on <code>http://example.com/admin/post/new</code>, the following redirect to /admin would
   * land you at <code>http://example.com/admin</code>:
   *
   * <pre>
   *   rsp.redirect("/admin");
   * </pre>
   *
   * Redirects can be relative to the current URL. A redirection of post/new, from
   * <code>http://example.com/blog/admin/</code> (notice the trailing slash), would give you
   * <code>http://example.com/blog/admin/post/new.</code>
   *
   * <pre>
   *   rsp.redirect("post/new");
   * </pre>
   *
   * Redirecting to post/new from <code>http://example.com/blog/admin</code> (no trailing slash),
   * will take you to <code>http://example.com/blog/post/new</code>.
   *
   * <p>
   * If you found the above behavior confusing, think of path segments as directories (have trailing
   * slashes) and files, it will start to make sense.
   * </p>
   *
   * Pathname relative redirects are also possible. If you were on
   * <code>http://example.com/admin/post/new</code>, the following redirect would land you at
   * <code>http//example.com/admin</code>:
   *
   * <pre>
   *   rsp.redirect("..");
   * </pre>
   *
   * A back redirection will redirect the request back to the <code>Referer</code>, defaulting to
   * <code>/</code> when missing.
   *
   * <pre>
   *   rsp.redirect("back");
   * </pre>
   *
   * @param status A redirect status.
   * @param location Either a relative or absolute location.
   * @throws Throwable If redirection fails.
   */
  void redirect(Status status, String location) throws Throwable;

  /**
   * @return A HTTP status or empty if status was not set yet.
   */
  @Nonnull
  Optional<Status> status();

  /**
   * Set the HTTP response status.
   *
   * @param status A HTTP status.
   * @return This response.
   */
  @Nonnull
  Response status(Status status);

  /**
   * Set the HTTP response status.
   *
   * @param status A HTTP status.
   * @return This response.
   */
  @Nonnull
  default Response status(final int status) {
    return status(Status.valueOf(status));
  }

  /**
   * Returns a boolean indicating if the response has been committed. A committed response has
   * already had its status code and headers written.
   *
   * @return a boolean indicating if the response has been committed
   */
  boolean committed();

  /**
   * Ends current request/response cycle by releasing any existing resources and committing the
   * response into the channel.
   *
   * This method is automatically call it from a send method, so you are not force to call this
   * method per each request/response cycle.
   *
   * It's recommended for quickly ending the response without any data:
   *
   * <pre>
   *   rsp.status(304).end();
   * </pre>
   *
   * Keep in mind that an explicit call to this method will stop the execution of handlers. So,
   * any handler further in the chain won't be executed once end has been called.
   */
  void end();

  /**
   * Append an after handler, will be execute before sending response.
   *
   * @param handler A handler
   * @see Route.After
   */
  void after(Route.After handler);

  /**
   * Append complete handler, will be execute after sending response.
   *
   * @param handler A handler
   * @see Route.After
   */
  void complete(Route.Complete handler);

  /**
   * Indicates if headers are cleared/reset on error.
   *
   * @return Indicates if headers are cleared/reset on error. Default is true.
   */
  boolean isResetHeadersOnError();

  /**
   * Indicates if headers are cleared/reset on error.
   *
   * @param value True to reset.
   */
  void setResetHeadersOnError(boolean value);
}
