/**
 *                                  Apache License
 *                            Version 2.0, January 2004
 *                         http://www.apache.org/licenses/
 *
 *    TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *    1. Definitions.
 *
 *       "License" shall mean the terms and conditions for use, reproduction,
 *       and distribution as defined by Sections 1 through 9 of this document.
 *
 *       "Licensor" shall mean the copyright owner or entity authorized by
 *       the copyright owner that is granting the License.
 *
 *       "Legal Entity" shall mean the union of the acting entity and all
 *       other entities that control, are controlled by, or are under common
 *       control with that entity. For the purposes of this definition,
 *       "control" means (i) the power, direct or indirect, to cause the
 *       direction or management of such entity, whether by contract or
 *       otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *       outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *       "You" (or "Your") shall mean an individual or Legal Entity
 *       exercising permissions granted by this License.
 *
 *       "Source" form shall mean the preferred form for making modifications,
 *       including but not limited to software source code, documentation
 *       source, and configuration files.
 *
 *       "Object" form shall mean any form resulting from mechanical
 *       transformation or translation of a Source form, including but
 *       not limited to compiled object code, generated documentation,
 *       and conversions to other media types.
 *
 *       "Work" shall mean the work of authorship, whether in Source or
 *       Object form, made available under the License, as indicated by a
 *       copyright notice that is included in or attached to the work
 *       (an example is provided in the Appendix below).
 *
 *       "Derivative Works" shall mean any work, whether in Source or Object
 *       form, that is based on (or derived from) the Work and for which the
 *       editorial revisions, annotations, elaborations, or other modifications
 *       represent, as a whole, an original work of authorship. For the purposes
 *       of this License, Derivative Works shall not include works that remain
 *       separable from, or merely link (or bind by name) to the interfaces of,
 *       the Work and Derivative Works thereof.
 *
 *       "Contribution" shall mean any work of authorship, including
 *       the original version of the Work and any modifications or additions
 *       to that Work or Derivative Works thereof, that is intentionally
 *       submitted to Licensor for inclusion in the Work by the copyright owner
 *       or by an individual or Legal Entity authorized to submit on behalf of
 *       the copyright owner. For the purposes of this definition, "submitted"
 *       means any form of electronic, verbal, or written communication sent
 *       to the Licensor or its representatives, including but not limited to
 *       communication on electronic mailing lists, source code control systems,
 *       and issue tracking systems that are managed by, or on behalf of, the
 *       Licensor for the purpose of discussing and improving the Work, but
 *       excluding communication that is conspicuously marked or otherwise
 *       designated in writing by the copyright owner as "Not a Contribution."
 *
 *       "Contributor" shall mean Licensor and any individual or Legal Entity
 *       on behalf of whom a Contribution has been received by Licensor and
 *       subsequently incorporated within the Work.
 *
 *    2. Grant of Copyright License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       copyright license to reproduce, prepare Derivative Works of,
 *       publicly display, publicly perform, sublicense, and distribute the
 *       Work and such Derivative Works in Source or Object form.
 *
 *    3. Grant of Patent License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       (except as stated in this section) patent license to make, have made,
 *       use, offer to sell, sell, import, and otherwise transfer the Work,
 *       where such license applies only to those patent claims licensable
 *       by such Contributor that are necessarily infringed by their
 *       Contribution(s) alone or by combination of their Contribution(s)
 *       with the Work to which such Contribution(s) was submitted. If You
 *       institute patent litigation against any entity (including a
 *       cross-claim or counterclaim in a lawsuit) alleging that the Work
 *       or a Contribution incorporated within the Work constitutes direct
 *       or contributory patent infringement, then any patent licenses
 *       granted to You under this License for that Work shall terminate
 *       as of the date such litigation is filed.
 *
 *    4. Redistribution. You may reproduce and distribute copies of the
 *       Work or Derivative Works thereof in any medium, with or without
 *       modifications, and in Source or Object form, provided that You
 *       meet the following conditions:
 *
 *       (a) You must give any other recipients of the Work or
 *           Derivative Works a copy of this License; and
 *
 *       (b) You must cause any modified files to carry prominent notices
 *           stating that You changed the files; and
 *
 *       (c) You must retain, in the Source form of any Derivative Works
 *           that You distribute, all copyright, patent, trademark, and
 *           attribution notices from the Source form of the Work,
 *           excluding those notices that do not pertain to any part of
 *           the Derivative Works; and
 *
 *       (d) If the Work includes a "NOTICE" text file as part of its
 *           distribution, then any Derivative Works that You distribute must
 *           include a readable copy of the attribution notices contained
 *           within such NOTICE file, excluding those notices that do not
 *           pertain to any part of the Derivative Works, in at least one
 *           of the following places: within a NOTICE text file distributed
 *           as part of the Derivative Works; within the Source form or
 *           documentation, if provided along with the Derivative Works; or,
 *           within a display generated by the Derivative Works, if and
 *           wherever such third-party notices normally appear. The contents
 *           of the NOTICE file are for informational purposes only and
 *           do not modify the License. You may add Your own attribution
 *           notices within Derivative Works that You distribute, alongside
 *           or as an addendum to the NOTICE text from the Work, provided
 *           that such additional attribution notices cannot be construed
 *           as modifying the License.
 *
 *       You may add Your own copyright statement to Your modifications and
 *       may provide additional or different license terms and conditions
 *       for use, reproduction, or distribution of Your modifications, or
 *       for any such Derivative Works as a whole, provided Your use,
 *       reproduction, and distribution of the Work otherwise complies with
 *       the conditions stated in this License.
 *
 *    5. Submission of Contributions. Unless You explicitly state otherwise,
 *       any Contribution intentionally submitted for inclusion in the Work
 *       by You to the Licensor shall be under the terms and conditions of
 *       this License, without any additional terms or conditions.
 *       Notwithstanding the above, nothing herein shall supersede or modify
 *       the terms of any separate license agreement you may have executed
 *       with Licensor regarding such Contributions.
 *
 *    6. Trademarks. This License does not grant permission to use the trade
 *       names, trademarks, service marks, or product names of the Licensor,
 *       except as required for reasonable and customary use in describing the
 *       origin of the Work and reproducing the content of the NOTICE file.
 *
 *    7. Disclaimer of Warranty. Unless required by applicable law or
 *       agreed to in writing, Licensor provides the Work (and each
 *       Contributor provides its Contributions) on an "AS IS" BASIS,
 *       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *       implied, including, without limitation, any warranties or conditions
 *       of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *       PARTICULAR PURPOSE. You are solely responsible for determining the
 *       appropriateness of using or redistributing the Work and assume any
 *       risks associated with Your exercise of permissions under this License.
 *
 *    8. Limitation of Liability. In no event and under no legal theory,
 *       whether in tort (including negligence), contract, or otherwise,
 *       unless required by applicable law (such as deliberate and grossly
 *       negligent acts) or agreed to in writing, shall any Contributor be
 *       liable to You for damages, including any direct, indirect, special,
 *       incidental, or consequential damages of any character arising as a
 *       result of this License or out of the use or inability to use the
 *       Work (including but not limited to damages for loss of goodwill,
 *       work stoppage, computer failure or malfunction, or any and all
 *       other commercial damages or losses), even if such Contributor
 *       has been advised of the possibility of such damages.
 *
 *    9. Accepting Warranty or Additional Liability. While redistributing
 *       the Work or Derivative Works thereof, You may choose to offer,
 *       and charge a fee for, acceptance of support, warranty, indemnity,
 *       or other liability obligations and/or rights consistent with this
 *       License. However, in accepting such obligations, You may act only
 *       on Your own behalf and on Your sole responsibility, not on behalf
 *       of any other Contributor, and only if You agree to indemnify,
 *       defend, and hold each Contributor harmless for any liability
 *       incurred by, or claims asserted against, such Contributor by reason
 *       of your accepting any such warranty or additional liability.
 *
 *    END OF TERMS AND CONDITIONS
 *
 *    APPENDIX: How to apply the Apache License to your work.
 *
 *       To apply the Apache License to your work, attach the following
 *       boilerplate notice, with the fields enclosed by brackets "{}"
 *       replaced with your own identifying information. (Don't include
 *       the brackets!)  The text should be enclosed in the appropriate
 *       comment syntax for the file format. We also recommend that a
 *       file or class name and description of purpose be included on the
 *       same "printed page" as the copyright notice for easier
 *       identification within third-party archives.
 *
 *    Copyright 2014 Edgar Espina
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package org.jooby;

import com.google.common.collect.ImmutableList;
import com.google.inject.Injector;
import com.google.inject.Key;
import com.google.inject.TypeLiteral;
import com.google.inject.name.Names;
import static java.util.Objects.requireNonNull;
import org.jooby.Route.Chain;
import org.jooby.internal.SseRenderer;
import org.jooby.funzy.Throwing;
import org.jooby.funzy.Try;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import java.io.IOException;
import java.nio.channels.ClosedChannelException;
import java.nio.charset.StandardCharsets;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Optional;
import java.util.UUID;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicReference;

/**
 * <h1>Server Sent Events</h1>
 * <p>
 * Server-Sent Events (SSE) is a mechanism that allows server to push the data from the server to
 * the client once the client-server connection is established by the client. Once the connection is
 * established by the client, it is the server who provides the data and decides to send it to the
 * client whenever new <strong>chunk</strong> of data is available.
 * </p>
 *
 * <h2>usage</h2>
 *
 * <pre>{@code
 * {
 *   sse("/path", sse -> {
 *     // 1. connected
 *     sse.send("data"); // 2. send/push data
 *   });
 * }
 * }</pre>
 *
 * <p>
 * Simple, effective and easy to use. The callback will be executed once when a new client is
 * connected. Inside the callback we can send data, listen for connection close events, etc.
 * </p>
 *
 * <p>
 * There is a factory method {@link #event(Object)} that let you set event attributes:
 * </p>
 *
 * <pre>{@code
 * {
 *   sse("/path", sse -> {
 *     sse.event("data")
 *         .id("id")
 *         .name("myevent")
 *         .retry(5000L)
 *         .send();
 *   });
 * }
 * }</pre>
 *
 * <h2>structured data</h2>
 * <p>
 * Beside raw/string data you can also send structured data, like <code>json</code>,
 * <code>xml</code>, etc..
 * </p>
 *
 * <p>
 * The next example will send two message one in <code>json</code> format and one in
 * <code>text/plain</code> format:
 * </p>
 * :
 *
 * <pre>{@code
 * {
 *   use(new MyJsonRenderer());
 *
 *   sse("/path", sse -> {
 *     MyObject object = ...
 *     sse.send(object, "json");
 *     sse.send(object, "plain");
 *   });
 * }
 * }</pre>
 *
 * <p>
 * Or if your need only one format, just:
 * </p>
 *
 * <pre>{@code
 * {
 *   use(new MyJsonRenderer());
 *
 *   sse("/path", sse -> {
 *     MyObject object = ...
 *     sse.send(object);
 *   }).produces("json"); // by default always send json
 * }
 * }</pre>
 *
 * <h2>request params</h2>
 * <p>
 * We provide request access via two arguments callback:
 * </p>
 *
 * <pre>{@code
 * {
 *   sse("/events/:id", (req, sse) -> {
 *     String id = req.param("id").value();
 *     MyObject object = findObject(id);
 *     sse.send(object);
 *   });
 * }
 * }</pre>
 *
 * <h2>connection lost</h2>
 * <p>
 * The {@link #onClose(Throwing.Runnable)} callback allow you to clean and release resources on
 * connection close. A connection is closed by calling {@link #close()} or when the client/browser
 * close the connection.
 * </p>
 *
 * <pre>{@code
 * {
 *   sse("/events/:id", sse -> {
 *     sse.onClose(() -> {
 *       // clean up resources
 *     });
 *   });
 * }
 * }</pre>
 *
 * <p>
 * The close event will be generated if you try to send an event on a closed connection.
 * </p>
 *
 * <h2>keep alive time</h2>
 * <p>
 * The keep alive time feature can be used to prevent connections from timing out:
 * </p>
 *
 * <pre>{@code
 * {
 *   sse("/events/:id", sse -> {
 *     sse.keepAlive(15, TimeUnit.SECONDS);
 *   });
 * }
 * }</pre>
 *
 * <p>
 * The previous example will sent a <code>':'</code> message (empty comment) every 15 seconds to
 * keep the connection alive. If the client drop the connection, then the
 * {@link #onClose(Throwing.Runnable)} event will be fired it.
 * </p>
 *
 * <p>
 * This feature is useful when you want to detect {@link #onClose(Throwing.Runnable)} events without
 * waiting for the next time you send a new event. But for example, if your application already
 * generate events every 15s, then the use of keep alive is useless and you can avoid it.
 * </p>
 *
 * <h2>require</h2>
 * <p>
 * The {@link #require(Class)} methods let you access to application services:
 * </p>
 *
 * <pre>{@code
 * {
 *   sse("/events/:id", sse -> {
 *     MyService service = sse.require(MyService.class);
 *   });
 * }
 * }</pre>
 *
 * <h2>example</h2>
 * <p>
 * The next example will generate a new event every 60s. It recovers from a server shutdown by using
 * the {@link #lastEventId()} and clean resources on connection close.
 * </p>
 * <pre>{@code
 * {
 *   // creates an executor service
 *   ScheduledExecutorService executor = Executors.newScheduledThreadPool(1);
 *
 *   sse("/events", sse -> {
 *     // if we go down, recover from last event ID we sent. Otherwise, start from zero.
 *     int lastId = sse.lastEventId(Integer.class).orElse(0);
 *
 *     AtomicInteger next = new AtomicInteger(lastId);
 *
 *     // send events every 60s
 *     ScheduledFuture<?> future = executor.scheduleAtFixedRate(() -> {
 *        Integer id = next.incrementAndGet();
 *        Object data = findDataById(id);
 *
 *        // send data and id
 *        sse.event(data).id(id).send();
 *      }, 0, 60, TimeUnit.SECONDS);
 *
 *      // on connection lost, cancel 60s task
 *      sse.onClose(() -> {
 *       future.cancel(true);
 *      });
 *   });
 * }
 *
 * }</pre>
 *
 * @author edgar
 * @since 1.0.0.CR
 */
public abstract class Sse implements AutoCloseable {

  /**
   * Event representation of Server sent event.
   *
   * @author edgar
   * @since 1.0.0.CR
   */
  public static class Event {
    private Object id;

    private String name;

    private Object data;

    private Long retry;

    private MediaType type;

    private String comment;

    private Sse sse;

    private Event(final Sse sse, final Object data) {
      this.sse = sse;
      this.data = data;
    }

    /**
     * @return Event data (if any).
     */
    public Optional<Object> data() {
      return Optional.ofNullable(data);
    }

    /**
     * Event media type helps to render or format event data.
     *
     * @return Event media type (if any).
     */
    public Optional<MediaType> type() {
      return Optional.ofNullable(type);
    }

    /**
     * Set event media type. Useful for sengin json, xml, etc..
     *
     * @param type Media Type.
     * @return This event.
     */
    public Event type(final MediaType type) {
      this.type = requireNonNull(type, "Type is required.");
      return this;
    }

    /**
     * Set event media type. Useful for sengin json, xml, etc..
     *
     * @param type Media Type.
     * @return This event.
     */
    public Event type(final String type) {
      return type(MediaType.valueOf(type));
    }

    /**
     * @return Event id (if any).
     */
    public Optional<Object> id() {
      return Optional.ofNullable(id);
    }

    /**
     * Set event id.
     *
     * @param id An event id.
     * @return This event.
     */
    public Event id(final Object id) {
      this.id = requireNonNull(id, "Id is required.");
      return this;
    }

    /**
     * @return Event name (a.k.a type).
     */
    public Optional<String> name() {
      return Optional.ofNullable(name);
    }

    /**
     * Set event name (a.k.a type).
     *
     * @param name Event's name.
     * @return This event.
     */
    public Event name(final String name) {
      this.name = requireNonNull(name, "Name is required.");
      return this;
    }

    /**
     * Clients (browsers) will attempt to reconnect every 3 seconds. The retry option allow you to
     * specify the number of millis a browser should wait before try to reconnect.
     *
     * @param retry Retry value.
     * @param unit Time unit.
     * @return This event.
     */
    public Event retry(final int retry, final TimeUnit unit) {
      this.retry = unit.toMillis(retry);
      return this;
    }

    /**
     * Clients (browsers) will attempt to reconnect every 3 seconds. The retry option allow you to
     * specify the number of millis a browser should wait before try to reconnect.
     *
     * @param retry Retry value in millis.
     * @return This event.
     */
    public Event retry(final long retry) {
      this.retry = retry;
      return this;
    }

    /**
     * @return Event comment (if any).
     */
    public Optional<String> comment() {
      return Optional.ofNullable(comment);
    }

    /**
     * Set event comment.
     *
     * @param comment An event comment.
     * @return This event.
     */
    public Event comment(final String comment) {
      this.comment = requireNonNull(comment, "Comment is required.");
      return this;
    }

    /**
     * @return Retry event line (if any).
     */
    public Optional<Long> retry() {
      return Optional.ofNullable(retry);
    }

    /**
     * Send an event and optionally listen for success confirmation or error:
     *
     * <pre>{@code
     * sse.event(data).send().onSuccess(id -> {
     *   // success
     * }).onFailure(cause -> {
     *   // handle error
     * });
     * }</pre>
     *
     * @return A future callback.
     */
    public CompletableFuture<Optional<Object>> send() {
      CompletableFuture<Optional<Object>> future = sse.send(this);
      this.id = null;
      this.name = null;
      this.data = null;
      this.type = null;
      this.sse = null;
      return future;
    }

  }

  /**
   * Server-sent event handler.
   *
   * @author edgar
   * @since 1.0.0.CR
   */
  public interface Handler extends Route.Filter {

    @Override
    default void handle(final Request req, final Response rsp, final Chain chain) throws Throwable {
      Sse sse = req.require(Sse.class);
      String path = req.path();
      rsp.send(new Deferred(deferred -> {
        try {
          sse.handshake(req, () -> {
            Try.run(() -> handle(req, sse))
                .onSuccess(() -> deferred.resolve(null))
                .onFailure(ex -> {
                  deferred.reject(ex);
                  Logger log = LoggerFactory.getLogger(Sse.class);
                  log.error("execution of {} resulted in error", path, ex);
                });
          });
        } catch (Exception ex) {
          deferred.reject(ex);
        }
      }));
    }

    /**
     * Event handler.
     *
     * @param req Current request.
     * @param sse Sse object.
     * @throws Exception If something goes wrong.
     */
    void handle(Request req, Sse sse) throws Exception;
  }

  /**
   * Single argument event handler.
   *
   * @author edgar
   * @since 1.0.0.CR
   */
  public interface Handler1 extends Handler {
    @Override
    default void handle(final Request req, final Sse sse) throws Exception {
      handle(sse);
    }

    void handle(Sse sse) throws Exception;
  }

  /* package */static class KeepAlive implements Runnable {

    /** The logging system. */
    private final Logger log = LoggerFactory.getLogger(Sse.class);

    private Sse sse;

    private long retry;

    public KeepAlive(final Sse sse, final long retry) {
      this.sse = sse;
      this.retry = retry;
    }

    @Override
    public void run() {
      String sseId = sse.id();
      log.debug("running heart beat for {}", sseId);
      Try.run(() -> sse.send(Optional.of(sseId), HEART_BEAT).whenComplete((id, x) -> {
        if (x != null) {
          log.debug("connection lost for {}", sseId, x);
          sse.fireCloseEvent();
          Try.run(sse::close);
        } else {
          log.debug("reschedule heart beat for {}", id);
          // reschedule
          sse.keepAlive(retry);
        }
      }));
    }

  }

  /** Keep alive scheduler. */
  private static final ScheduledExecutorService scheduler = Executors
      .newSingleThreadScheduledExecutor(r -> {
        Thread thread = new Thread(r, "sse-heartbeat");
        thread.setDaemon(true);
        return thread;
      });

  /** Empty comment. */
  static final byte[] HEART_BEAT = ":\n".getBytes(StandardCharsets.UTF_8);

  /** The logging system. */
  protected final Logger log = LoggerFactory.getLogger(Sse.class);

  private Injector injector;

  private List<Renderer> renderers;

  private final String id;

  private List<MediaType> produces;

  private Map<String, Object> locals;

  private AtomicReference<Throwing.Runnable> onclose = new AtomicReference<>(null);

  private Mutant lastEventId;

  private boolean closed;

  private Locale locale;

  public Sse() {
    id = UUID.randomUUID().toString();
  }

  protected void handshake(final Request req, final Runnable handler) throws Exception {
    this.injector = req.require(Injector.class);
    this.renderers = ImmutableList.copyOf(injector.getInstance(Renderer.KEY));
    this.produces = req.route().produces();
    this.locals = req.attributes();
    this.lastEventId = req.header("Last-Event-ID");
    this.locale = req.locale();
    handshake(handler);
  }

  protected abstract void handshake(Runnable handler) throws Exception;

  /**
   * A unique ID (like a session ID).
   *
   * @return Sse unique ID (like a session ID).
   */
  @Nonnull
  public String id() {
    return id;
  }

  /**
   * Server sent event will send a Last-Event-ID header if the server goes down.
   *
   * @return Last event id.
   */
  @Nonnull
  public Optional<String> lastEventId() {
    return lastEventId(String.class);
  }

  /**
   * Server sent event will send a Last-Event-ID header if the server goes down.
   *
   * @param type Last event id type.
   * @param <T> Event id type.
   * @return Last event id.
   */
  @Nonnull
  public <T> Optional<T> lastEventId(final Class<T> type) {
    return lastEventId.toOptional(type);
  }

  /**
   * Listen for connection close (usually client drop the connection). This method is useful for
   * resources cleanup.
   *
   * @param task Task to run.
   * @return This instance.
   */
  @Nonnull
  public Sse onClose(final Throwing.Runnable task) {
    onclose.set(task);
    return this;
  }

  /**
   * Send an event and set media type.
   *
   * <pre>{@code
   *   sse.send(new MyObject(), "json");
   * }</pre>
   *
   * <pre>{@code
   *   sse.send(new MyObject(), "json").whenComplete((id, x) -> {
   *     if (x == null) {
   *       handleSuccess();
   *     } else {
   *       handleError(x);
   *     }
   *   });
   * }</pre>
   *
   * The <code>id</code> of the success callback correspond to the {@link Event#id()}.
   *
   * @param data Event data.
   * @param type Media type, like: json, xml.
   * @return A future. The success callback contains the {@link Event#id()}.
   */
  @Nonnull
  public CompletableFuture<Optional<Object>> send(final Object data, final String type) {
    return send(data, MediaType.valueOf(type));
  }

  /**
   * Send an event and set media type.
   *
   * <pre>{@code
   *   sse.send(new MyObject(), "json");
   * }</pre>
   *
   * <pre>{@code
   *   sse.send(new MyObject(), "json").whenComplete((id, x) -> {
   *     if (x == null) {
   *       handleSuccess();
   *     } else {
   *       handleError(x);
   *     }
   *   });
   * }</pre>
   *
   * The <code>id</code> of the success callback correspond to the {@link Event#id()}.
   *
   * @param data Event data.
   * @param type Media type, like: json, xml.
   * @return A future. The success callback contains the {@link Event#id()}.
   */
  @Nonnull
  public CompletableFuture<Optional<Object>> send(final Object data, final MediaType type) {
    return event(data).type(type).send();
  }

  /**
   * Send an event.
   *
   * <pre>{@code
   *   sse.send(new MyObject());
   * }</pre>
   *
   * <pre>{@code
   *   sse.send(new MyObject(), "json").whenComplete((id, x) -> {
   *     if (x == null) {
   *       handleSuccess();
   *     } else {
   *       handleError(x);
   *     }
   *   });
   * }</pre>
   *
   * The <code>id</code> of the success callback correspond to the {@link Event#id()}.
   *
   * @param data Event data.
   * @return A future. The success callback contains the {@link Event#id()}.
   */
  @Nonnull
  public CompletableFuture<Optional<Object>> send(final Object data) {
    return event(data).send();
  }

  /**
   * Factory method for creating {@link Event} instances.
   *
   * Please note event won't be sent unless you call {@link Event#send()}:
   *
   * <pre>{@code
   *   sse.event(new MyObject()).send();
   * }</pre>
   *
   * The factory allow you to set event attributes:
   *
   * <pre>{@code
   *   // send data
   *   MyObject data = ...;
   *   sse.event(data).send();
   *
   *   // send data with event name
   *   sse.event(data).name("myevent").send();
   *
   *   // send data with event name and id
   *   sse.event(data).name("myevent").id(id).send();
   *
   *   // send data with event name, id and retry interval
   *   sse.event(data).name("myevent").id(id).retry(1500).send();
   * }</pre>
   *
   * @param data Event data.
   * @return A new event.
   */
  @Nonnull
  public Event event(final Object data) {
    return new Event(this, data);
  }

  /**
   * Ask Guice for the given type.
   *
   * @param type A service type.
   * @param <T> Service type.
   * @return A ready to use object.
   */
  @Nonnull
  public <T> T require(final Class<T> type) {
    return require(Key.get(type));
  }

  /**
   * Ask Guice for the given type.
   *
   * @param name A service name.
   * @param type A service type.
   * @param <T> Service type.
   * @return A ready to use object.
   */
  @Nonnull
  public <T> T require(final String name, final Class<T> type) {
    return require(Key.get(type, Names.named(name)));
  }

  /**
   * Ask Guice for the given type.
   *
   * @param type A service type.
   * @param <T> Service type.
   * @return A ready to use object.
   */
  @Nonnull
  public <T> T require(final TypeLiteral<T> type) {
    return require(Key.get(type));
  }

  /**
   * Ask Guice for the given type.
   *
   * @param key A service key.
   * @param <T> Service type.
   * @return A ready to use object.
   */
  @Nonnull
  public <T> T require(final Key<T> key) {
    return injector.getInstance(key);
  }

  /**
   * The keep alive time can be used to prevent connections from timing out:
   *
   * <pre>{@code
   * {
   *   sse("/events/:id", sse -> {
   *     sse.keepAlive(15, TimeUnit.SECONDS);
   *   });
   * }
   * }</pre>
   *
   * <p>
   * The previous example will sent a <code>':'</code> message (empty comment) every 15 seconds to
   * keep the connection alive. If the client drop the connection, then the
   * {@link #onClose(Throwing.Runnable)} event will be fired it.
   * </p>
   *
   * <p>
   * This feature is useful when you want to detect {@link #onClose(Throwing.Runnable)} events without
   * waiting until you send a new event. But for example, if your application already generate
   * events
   * every 15s, then the use of keep alive is useless and you should avoid it.
   * </p>
   *
   * @param time Keep alive time.
   * @param unit Time unit.
   * @return This instance.
   */
  @Nonnull
  public Sse keepAlive(final int time, final TimeUnit unit) {
    return keepAlive(unit.toMillis(time));
  }

  /**
   * The keep alive time can be used to prevent connections from timing out:
   *
   * <pre>{@code
   * {
   *   sse("/events/:id", sse -> {
   *     sse.keepAlive(15, TimeUnit.SECONDS);
   *   });
   * }
   * }</pre>
   *
   * <p>
   * The previous example will sent a <code>':'</code> message (empty comment) every 15 seconds to
   * keep the connection alive. If the client drop the connection, then the
   * {@link #onClose(Throwing.Runnable)} event will be fired it.
   * </p>
   *
   * <p>
   * This feature is useful when you want to detect {@link #onClose(Throwing.Runnable)} events without
   * waiting until you send a new event. But for example, if your application already generate
   * events
   * every 15s, then the use of keep alive is useless and you should avoid it.
   * </p>
   *
   * @param millis Keep alive time in millis.
   * @return This instance.
   */
  @Nonnull
  public Sse keepAlive(final long millis) {
    scheduler.schedule(new KeepAlive(this, millis), millis, TimeUnit.MILLISECONDS);
    return this;
  }

  /**
   * Close the connection and fire an {@link #onClose(Throwing.Runnable)} event.
   */
  @Override
  public final void close() throws Exception {
    closeAll();
  }

  private void closeAll() {
    synchronized (this) {
      if (!closed) {
        closed = true;
        fireCloseEvent();
        closeInternal();
      }
    }
  }

  protected abstract void closeInternal();

  protected abstract CompletableFuture<Optional<Object>> send(Optional<Object> id, byte[] data);

  protected void ifClose(final Throwable cause) {
    if (shouldClose(cause)) {
      closeAll();
    }
  }

  protected void fireCloseEvent() {
    Throwing.Runnable task = onclose.getAndSet(null);
    if (task != null) {
      Try.run(task).onFailure(ex -> log.error("close callback resulted in error", ex));
    }
  }

  protected boolean shouldClose(final Throwable ex) {
    if (ex instanceof IOException) {
      // is there a better way?
      boolean brokenPipe = Optional.ofNullable(ex.getMessage())
          .map(m -> m.toLowerCase().contains("broken pipe"))
          .orElse(false);
      return brokenPipe || ex instanceof ClosedChannelException;
    }
    return false;
  }

  private CompletableFuture<Optional<Object>> send(final Event event) {
    List<MediaType> produces = event.type().<List<MediaType>>map(ImmutableList::of)
        .orElse(this.produces);
    SseRenderer ctx = new SseRenderer(renderers, produces, StandardCharsets.UTF_8, locale, locals);
    return Try.apply(() -> {
      byte[] bytes = ctx.format(event);
      return send(event.id(), bytes);
    }).recover(x -> {
      CompletableFuture<Optional<Object>> future = new CompletableFuture<>();
      future.completeExceptionally(x);
      return future;
    })
        .get();
  }

}
