/**
 *                                  Apache License
 *                            Version 2.0, January 2004
 *                         http://www.apache.org/licenses/
 *
 *    TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *    1. Definitions.
 *
 *       "License" shall mean the terms and conditions for use, reproduction,
 *       and distribution as defined by Sections 1 through 9 of this document.
 *
 *       "Licensor" shall mean the copyright owner or entity authorized by
 *       the copyright owner that is granting the License.
 *
 *       "Legal Entity" shall mean the union of the acting entity and all
 *       other entities that control, are controlled by, or are under common
 *       control with that entity. For the purposes of this definition,
 *       "control" means (i) the power, direct or indirect, to cause the
 *       direction or management of such entity, whether by contract or
 *       otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *       outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *       "You" (or "Your") shall mean an individual or Legal Entity
 *       exercising permissions granted by this License.
 *
 *       "Source" form shall mean the preferred form for making modifications,
 *       including but not limited to software source code, documentation
 *       source, and configuration files.
 *
 *       "Object" form shall mean any form resulting from mechanical
 *       transformation or translation of a Source form, including but
 *       not limited to compiled object code, generated documentation,
 *       and conversions to other media types.
 *
 *       "Work" shall mean the work of authorship, whether in Source or
 *       Object form, made available under the License, as indicated by a
 *       copyright notice that is included in or attached to the work
 *       (an example is provided in the Appendix below).
 *
 *       "Derivative Works" shall mean any work, whether in Source or Object
 *       form, that is based on (or derived from) the Work and for which the
 *       editorial revisions, annotations, elaborations, or other modifications
 *       represent, as a whole, an original work of authorship. For the purposes
 *       of this License, Derivative Works shall not include works that remain
 *       separable from, or merely link (or bind by name) to the interfaces of,
 *       the Work and Derivative Works thereof.
 *
 *       "Contribution" shall mean any work of authorship, including
 *       the original version of the Work and any modifications or additions
 *       to that Work or Derivative Works thereof, that is intentionally
 *       submitted to Licensor for inclusion in the Work by the copyright owner
 *       or by an individual or Legal Entity authorized to submit on behalf of
 *       the copyright owner. For the purposes of this definition, "submitted"
 *       means any form of electronic, verbal, or written communication sent
 *       to the Licensor or its representatives, including but not limited to
 *       communication on electronic mailing lists, source code control systems,
 *       and issue tracking systems that are managed by, or on behalf of, the
 *       Licensor for the purpose of discussing and improving the Work, but
 *       excluding communication that is conspicuously marked or otherwise
 *       designated in writing by the copyright owner as "Not a Contribution."
 *
 *       "Contributor" shall mean Licensor and any individual or Legal Entity
 *       on behalf of whom a Contribution has been received by Licensor and
 *       subsequently incorporated within the Work.
 *
 *    2. Grant of Copyright License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       copyright license to reproduce, prepare Derivative Works of,
 *       publicly display, publicly perform, sublicense, and distribute the
 *       Work and such Derivative Works in Source or Object form.
 *
 *    3. Grant of Patent License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       (except as stated in this section) patent license to make, have made,
 *       use, offer to sell, sell, import, and otherwise transfer the Work,
 *       where such license applies only to those patent claims licensable
 *       by such Contributor that are necessarily infringed by their
 *       Contribution(s) alone or by combination of their Contribution(s)
 *       with the Work to which such Contribution(s) was submitted. If You
 *       institute patent litigation against any entity (including a
 *       cross-claim or counterclaim in a lawsuit) alleging that the Work
 *       or a Contribution incorporated within the Work constitutes direct
 *       or contributory patent infringement, then any patent licenses
 *       granted to You under this License for that Work shall terminate
 *       as of the date such litigation is filed.
 *
 *    4. Redistribution. You may reproduce and distribute copies of the
 *       Work or Derivative Works thereof in any medium, with or without
 *       modifications, and in Source or Object form, provided that You
 *       meet the following conditions:
 *
 *       (a) You must give any other recipients of the Work or
 *           Derivative Works a copy of this License; and
 *
 *       (b) You must cause any modified files to carry prominent notices
 *           stating that You changed the files; and
 *
 *       (c) You must retain, in the Source form of any Derivative Works
 *           that You distribute, all copyright, patent, trademark, and
 *           attribution notices from the Source form of the Work,
 *           excluding those notices that do not pertain to any part of
 *           the Derivative Works; and
 *
 *       (d) If the Work includes a "NOTICE" text file as part of its
 *           distribution, then any Derivative Works that You distribute must
 *           include a readable copy of the attribution notices contained
 *           within such NOTICE file, excluding those notices that do not
 *           pertain to any part of the Derivative Works, in at least one
 *           of the following places: within a NOTICE text file distributed
 *           as part of the Derivative Works; within the Source form or
 *           documentation, if provided along with the Derivative Works; or,
 *           within a display generated by the Derivative Works, if and
 *           wherever such third-party notices normally appear. The contents
 *           of the NOTICE file are for informational purposes only and
 *           do not modify the License. You may add Your own attribution
 *           notices within Derivative Works that You distribute, alongside
 *           or as an addendum to the NOTICE text from the Work, provided
 *           that such additional attribution notices cannot be construed
 *           as modifying the License.
 *
 *       You may add Your own copyright statement to Your modifications and
 *       may provide additional or different license terms and conditions
 *       for use, reproduction, or distribution of Your modifications, or
 *       for any such Derivative Works as a whole, provided Your use,
 *       reproduction, and distribution of the Work otherwise complies with
 *       the conditions stated in this License.
 *
 *    5. Submission of Contributions. Unless You explicitly state otherwise,
 *       any Contribution intentionally submitted for inclusion in the Work
 *       by You to the Licensor shall be under the terms and conditions of
 *       this License, without any additional terms or conditions.
 *       Notwithstanding the above, nothing herein shall supersede or modify
 *       the terms of any separate license agreement you may have executed
 *       with Licensor regarding such Contributions.
 *
 *    6. Trademarks. This License does not grant permission to use the trade
 *       names, trademarks, service marks, or product names of the Licensor,
 *       except as required for reasonable and customary use in describing the
 *       origin of the Work and reproducing the content of the NOTICE file.
 *
 *    7. Disclaimer of Warranty. Unless required by applicable law or
 *       agreed to in writing, Licensor provides the Work (and each
 *       Contributor provides its Contributions) on an "AS IS" BASIS,
 *       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *       implied, including, without limitation, any warranties or conditions
 *       of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *       PARTICULAR PURPOSE. You are solely responsible for determining the
 *       appropriateness of using or redistributing the Work and assume any
 *       risks associated with Your exercise of permissions under this License.
 *
 *    8. Limitation of Liability. In no event and under no legal theory,
 *       whether in tort (including negligence), contract, or otherwise,
 *       unless required by applicable law (such as deliberate and grossly
 *       negligent acts) or agreed to in writing, shall any Contributor be
 *       liable to You for damages, including any direct, indirect, special,
 *       incidental, or consequential damages of any character arising as a
 *       result of this License or out of the use or inability to use the
 *       Work (including but not limited to damages for loss of goodwill,
 *       work stoppage, computer failure or malfunction, or any and all
 *       other commercial damages or losses), even if such Contributor
 *       has been advised of the possibility of such damages.
 *
 *    9. Accepting Warranty or Additional Liability. While redistributing
 *       the Work or Derivative Works thereof, You may choose to offer,
 *       and charge a fee for, acceptance of support, warranty, indemnity,
 *       or other liability obligations and/or rights consistent with this
 *       License. However, in accepting such obligations, You may act only
 *       on Your own behalf and on Your sole responsibility, not on behalf
 *       of any other Contributor, and only if You agree to indemnify,
 *       defend, and hold each Contributor harmless for any liability
 *       incurred by, or claims asserted against, such Contributor by reason
 *       of your accepting any such warranty or additional liability.
 *
 *    END OF TERMS AND CONDITIONS
 *
 *    APPENDIX: How to apply the Apache License to your work.
 *
 *       To apply the Apache License to your work, attach the following
 *       boilerplate notice, with the fields enclosed by brackets "{}"
 *       replaced with your own identifying information. (Don't include
 *       the brackets!)  The text should be enclosed in the appropriate
 *       comment syntax for the file format. We also recommend that a
 *       file or class name and description of purpose be included on the
 *       same "printed page" as the copyright notice for easier
 *       identification within third-party archives.
 *
 *    Copyright 2014 Edgar Espina
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package org.jooby.internal;

import com.google.common.base.Strings;
import com.google.common.base.Throwables;
import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;
import com.google.common.collect.Sets;
import com.google.inject.Injector;
import com.google.inject.Key;
import com.google.inject.name.Names;
import com.typesafe.config.Config;
import static java.util.Objects.requireNonNull;
import org.jooby.Deferred;
import org.jooby.Err;
import org.jooby.Err.Handler;
import org.jooby.MediaType;
import org.jooby.Renderer;
import org.jooby.Request;
import org.jooby.Response;
import org.jooby.Route;
import org.jooby.Session;
import org.jooby.Sse;
import org.jooby.Status;
import org.jooby.WebSocket;
import org.jooby.WebSocket.Definition;
import org.jooby.internal.parser.ParserExecutor;
import org.jooby.spi.HttpHandler;
import org.jooby.spi.NativeRequest;
import org.jooby.spi.NativeResponse;
import org.jooby.spi.NativeWebSocket;
import org.jooby.funzy.Try;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import javax.inject.Named;
import javax.inject.Provider;
import javax.inject.Singleton;
import java.nio.charset.Charset;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.Executor;
import java.util.function.Function;
import java.util.stream.Collectors;

@Singleton
public class HttpHandlerImpl implements HttpHandler {

  private static class RouteKey {

    private final String method;

    private final String path;

    private final MediaType consumes;

    private final List<MediaType> produces;

    private final String key;

    public RouteKey(final String method, final String path, final MediaType consumes,
        final List<MediaType> produces) {
      String c = consumes.name();
      String p = produces.toString();
      key = new StringBuilder(method.length() + path.length() + c.length() + p.length())
          .append(method)
          .append(path)
          .append(c)
          .append(p)
          .toString();
      this.method = method;
      this.path = path;
      this.consumes = consumes;
      this.produces = produces;
    }

    @Override
    public int hashCode() {
      return key.hashCode();
    }

    @Override
    public boolean equals(final Object obj) {
      RouteKey that = (RouteKey) obj;
      return key.equals(that.key);
    }
  }

  private static final String NO_CACHE = "must-revalidate,no-cache,no-store";

  private static final String WEB_SOCKET = "WebSocket";

  private static final String UPGRADE = "Upgrade";

  private static final String REFERER = "Referer";

  private static final String PATH = "path";

  private static final String CONTEXT_PATH = "contextPath";

  private static final Key<Request> REQ = Key.get(Request.class);

  private static final Key<Route.Chain> CHAIN = Key.get(Route.Chain.class);

  private static final Key<Response> RSP = Key.get(Response.class);

  private static final Key<Sse> SSE = Key.get(Sse.class);

  private static final Key<Session> SESS = Key.get(Session.class);

  private static final Key<String> DEF_EXEC = Key.get(String.class, Names.named("deferred"));

  private static final String BYTE_RANGE = "Range";

  /**
   * The logging system.
   */
  private Injector injector;

  private Set<Err.Handler> err;

  private String applicationPath;

  private RequestScope requestScope;

  private Set<Definition> socketDefs;

  private Config config;

  private int port;

  private String _method;

  private Charset charset;

  private List<Renderer> renderers;

  private ParserExecutor parserExecutor;

  private List<Locale> locales;

  private final LoadingCache<RouteKey, Route[]> routeCache;

  private final String redirectHttps;

  private Function<String, String> rpath = null;

  private String contextPath;

  private boolean hasSockets;

  private final Map<String, Renderer> rendererMap;

  private StatusCodeProvider sc;

  /**
   * Global deferred executor.
   */
  private Key<Executor> gexec;

  @Inject
  public HttpHandlerImpl(final Injector injector,
      final RequestScope requestScope,
      final Set<Route.Definition> routes,
      final Set<WebSocket.Definition> sockets,
      final @Named("application.path") String path,
      final ParserExecutor parserExecutor,
      final Set<Renderer> renderers,
      final Set<Err.Handler> err,
      final StatusCodeProvider sc,
      final Charset charset,
      final List<Locale> locale) {
    this.injector = requireNonNull(injector, "An injector is required.");
    this.requestScope = requireNonNull(requestScope, "A request scope is required.");
    this.socketDefs = requireNonNull(sockets, "Sockets are required.");
    this.hasSockets = socketDefs.size() > 0;
    this.applicationPath = normalizeURI(requireNonNull(path, "An application.path is required."));
    this.err = requireNonNull(err, "An err handler is required.");
    this.sc = sc;
    this.config = injector.getInstance(Config.class);
    _method = Strings.emptyToNull(this.config.getString("server.http.Method").trim());
    this.port = config.getInt("application.port");
    this.charset = charset;
    this.locales = locale;
    this.parserExecutor = parserExecutor;
    this.renderers = new ArrayList<>(renderers);
    rendererMap = new HashMap<>();
    this.renderers.forEach(r -> rendererMap.put(r.name(), r));

    // route cache
    routeCache = routeCache(routes, config);
    // force https
    String redirectHttps = config.getString("application.redirect_https").trim();
    this.redirectHttps = redirectHttps.length() > 0 ? redirectHttps : null;

    // custom path?
    if (applicationPath.equals("/")) {
      this.contextPath = "";
    } else {
      this.contextPath = applicationPath;
      this.rpath = rootpath(applicationPath);
    }
    // global deferred executor
    this.gexec = Key.get(Executor.class, Names.named(injector.getInstance(DEF_EXEC)));
  }

  @Override
  public void handle(final NativeRequest request, final NativeResponse response) throws Exception {
    long start = System.currentTimeMillis();

    Map<String, Object> locals = new HashMap<>(16);

    Map<Object, Object> scope = new HashMap<>(16);

    String method = _method == null ? request.method() : method(_method, request);
    String path = normalizeURI(request.path());
    if (rpath != null) {
      path = rpath.apply(path);
    }

    // put request attributes first to make sure we don't override defaults
    Map<String, Object> nativeAttrs = request.attributes();
    if (nativeAttrs.size() > 0) {
      locals.putAll(nativeAttrs);
    }
    // default locals
    locals.put(CONTEXT_PATH, contextPath);
    locals.put(PATH, path);

    Route notFound = RouteImpl.notFound(method, path);

    RequestImpl req = new RequestImpl(injector, request, contextPath, port, notFound, charset,
        locales, scope, locals, start);

    ResponseImpl rsp = new ResponseImpl(req, parserExecutor, response, notFound, renderers,
        rendererMap, locals, req.charset(), request.header(REFERER), request.header(BYTE_RANGE));

    MediaType type = req.type();

    // seed req & rsp
    scope.put(REQ, req);
    scope.put(RSP, rsp);

    // seed sse
    Provider<Sse> sse = () -> Try.apply(() -> request.upgrade(Sse.class)).get();
    scope.put(SSE, sse);

    // seed session
    Provider<Session> session = () -> req.session();
    scope.put(SESS, session);

    boolean deferred = false;
    Throwable x = null;
    try {

      requestScope.enter(scope);

      // force https?
      if (redirectHttps != null) {
        if (!req.secure()) {
          rsp.redirect(MessageFormat.format(redirectHttps, path.substring(1)));
          return;
        }
      }

      // websocket?
      if (hasSockets) {
        if (upgrade(request)) {
          Optional<WebSocket> sockets = findSockets(socketDefs, path);
          if (sockets.isPresent()) {
            NativeWebSocket ws = request.upgrade(NativeWebSocket.class);
            ws.onConnect(() -> ((WebSocketImpl) sockets.get()).connect(injector, req, ws));
            return;
          }
        }
      }

      // usual req/rsp
      Route[] routes = routeCache
          .getUnchecked(new RouteKey(method, path, type, req.accept()));

      RouteChain chain = new RouteChain(req, rsp, routes);
      scope.put(CHAIN, chain);
      chain.next(req, rsp);

    } catch (DeferredExecution ex) {
      deferred = true;
      onDeferred(scope, request, req, rsp, ex.deferred);
    } catch (Throwable ex) {
      x = ex;
    } finally {
      cleanup(req, rsp, true, x, !deferred);
    }
  }

  private boolean upgrade(final NativeRequest request) {
    Optional<String> upgrade = request.header(UPGRADE);
    return upgrade.isPresent() && upgrade.get().equalsIgnoreCase(WEB_SOCKET);
  }

  private void done(final RequestImpl req, final ResponseImpl rsp, final Throwable x,
      final boolean close) {
    // mark request/response as done.
    req.done();
    if (close) {
      rsp.done(Optional.ofNullable(x));
    }
  }

  private void onDeferred(final Map<Object, Object> scope, final NativeRequest request,
      final RequestImpl req, final ResponseImpl rsp, final Deferred deferred) {
    /** Deferred executor. */
    Key<Executor> execKey = deferred.executor()
        .map(it -> Key.get(Executor.class, Names.named(it)))
        .orElse(gexec);

    /** Get executor. */
    Executor executor = injector.getInstance(execKey);

    request.startAsync(executor, () -> {
      try {
        deferred.handler(req, (success, x) -> {
          boolean close = false;
          Optional<Throwable> failure = Optional.ofNullable(x);
          try {
            requestScope.enter(scope);
            if (success != null) {
              close = true;
              rsp.send(success);
            }
          } catch (Throwable exerr) {
            failure = Optional.of(failure.orElse(exerr));
          } finally {
            Throwable cause = failure.orElse(null);
            if (cause != null) {
              close = true;
            }
            cleanup(req, rsp, close, cause, true);
          }
        });
      } catch (Exception ex) {
        handleErr(req, rsp, ex);
      }
    });
  }

  private void cleanup(final RequestImpl req, final ResponseImpl rsp, final boolean close,
      final Throwable x, final boolean done) {
    if (x != null) {
      handleErr(req, rsp, x);
    }
    if (done) {
      done(req, rsp, x, close);
    }
    requestScope.exit();
  }

  private void handleErr(final RequestImpl req, final ResponseImpl rsp, final Throwable ex) {
    Logger log = LoggerFactory.getLogger(HttpHandler.class);
    try {
      log.debug("execution of: {}{} resulted in exception", req.method(), req.path(), ex);
      // execution failed, find status code
      Status status = sc.apply(ex);

      if (status == Status.REQUESTED_RANGE_NOT_SATISFIABLE) {
        String range = rsp.header("Content-Length").toOptional().map(it -> "bytes */" + it)
            .orElse("*");
        rsp.reset();
        rsp.header("Content-Range", range);
      } else {
        rsp.reset();
      }

      rsp.header("Cache-Control", NO_CACHE);
      rsp.status(status);

      Err err = ex instanceof Err ? (Err) ex : new Err(status, ex);

      Iterator<Handler> it = this.err.iterator();
      while (!rsp.committed() && it.hasNext()) {
        Err.Handler next = it.next();
        log.debug("handling err with: {}", next);
        next.handle(req, rsp, err);
      }
    } catch (Throwable errex) {
      log.error("error handler resulted in exception: {}{}\nRoute:\n{}\n\nStacktrace:\n{}\nSource:",
          req.method(), req.path(), req.route().print(6), Throwables.getStackTraceAsString(errex),
          ex);
    }
  }

  private static String normalizeURI(final String uri) {
    int len = uri.length();
    return len > 1 && uri.charAt(len - 1) == '/' ? uri.substring(0, len - 1) : uri;
  }

  private static Route[] routes(final Set<Route.Definition> routeDefs, final String method,
      final String path, final MediaType type, final List<MediaType> accept) {
    List<Route> routes = findRoutes(routeDefs, method, path, type, accept);

    routes.add(RouteImpl.fallback((req, rsp, chain) -> {
      if (!rsp.status().isPresent()) {
        // 406 or 415
        Err ex = handle406or415(routeDefs, method, path, type, accept);
        if (ex != null) {
          throw ex;
        }
        // 405
        ex = handle405(routeDefs, method, path, type, accept);
        if (ex != null) {
          throw ex;
        }
        // favicon.ico
        if (path.equals("/favicon.ico")) {
          // default /favicon.ico handler:
          rsp.status(Status.NOT_FOUND).end();
        } else {
          throw new Err(Status.NOT_FOUND, req.path());
        }
      }
    }, method, path, "err", accept));

    return routes.toArray(new Route[routes.size()]);
  }

  private static List<Route> findRoutes(final Set<Route.Definition> routeDefs, final String method,
      final String path, final MediaType type, final List<MediaType> accept) {

    List<Route> routes = new ArrayList<>();
    for (Route.Definition routeDef : routeDefs) {
      Optional<Route> route = routeDef.matches(method, path, type, accept);
      if (route.isPresent()) {
        routes.add(route.get());
      }
    }
    return routes;
  }

  private static Optional<WebSocket> findSockets(final Set<WebSocket.Definition> sockets,
      final String path) {
    for (WebSocket.Definition socketDef : sockets) {
      Optional<WebSocket> match = socketDef.matches(path);
      if (match.isPresent()) {
        return match;
      }
    }
    return Optional.empty();
  }

  private static Err handle405(final Set<Route.Definition> routeDefs, final String method,
      final String path, final MediaType type, final List<MediaType> accept) {

    if (alternative(routeDefs, method, path).size() > 0) {
      return new Err(Status.METHOD_NOT_ALLOWED, method);
    }

    return null;
  }

  private static List<Route> alternative(final Set<Route.Definition> routeDefs, final String verb,
      final String uri) {
    List<Route> routes = new LinkedList<>();
    Set<String> verbs = Sets.newHashSet(Route.METHODS);
    verbs.remove(verb);
    for (String alt : verbs) {
      findRoutes(routeDefs, alt, uri, MediaType.all, MediaType.ALL)
          .stream()
          // skip glob pattern
          .filter(r -> !r.pattern().contains("*"))
          .forEach(routes::add);

    }
    return routes;
  }

  private static Err handle406or415(final Set<Route.Definition> routeDefs, final String method,
      final String path, final MediaType contentType, final List<MediaType> accept) {
    for (Route.Definition routeDef : routeDefs) {
      Optional<Route> route = routeDef.matches(method, path, MediaType.all, MediaType.ALL);
      if (route.isPresent() && !route.get().pattern().contains("*")) {
        if (!routeDef.canProduce(accept)) {
          return new Err(Status.NOT_ACCEPTABLE, accept.stream()
              .map(MediaType::name)
              .collect(Collectors.joining(", ")));
        }
        if (!contentType.isAny()) {
          return new Err(Status.UNSUPPORTED_MEDIA_TYPE, contentType.name());
        }
      }
    }
    return null;
  }

  private static String method(final String methodParam, final NativeRequest request)
      throws Exception {
    Optional<String> header = request.header(methodParam);
    if (header.isPresent()) {
      return header.get();
    }
    List<String> param = request.params(methodParam);
    return param.size() == 0 ? request.method() : param.get(0);
  }

  private static LoadingCache<RouteKey, Route[]> routeCache(final Set<Route.Definition> routes,
      final Config conf) {
    return CacheBuilder.from(conf.getString("server.routes.Cache"))
        .build(new CacheLoader<RouteKey, Route[]>() {
          @Override
          public Route[] load(final RouteKey key) throws Exception {
            return routes(routes, key.method, key.path, key.consumes, key.produces);
          }
        });
  }

  private static Function<String, String> rootpath(final String applicationPath) {
    return p -> {
      if (applicationPath.equals(p)) {
        return "/";
      } else if (p.startsWith(applicationPath)) {
        return p.substring(applicationPath.length());
      } else {
        // mark as failure
        return Route.errpath(p);
      }
    };
  }

}
