/**
 *                                  Apache License
 *                            Version 2.0, January 2004
 *                         http://www.apache.org/licenses/
 *
 *    TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *    1. Definitions.
 *
 *       "License" shall mean the terms and conditions for use, reproduction,
 *       and distribution as defined by Sections 1 through 9 of this document.
 *
 *       "Licensor" shall mean the copyright owner or entity authorized by
 *       the copyright owner that is granting the License.
 *
 *       "Legal Entity" shall mean the union of the acting entity and all
 *       other entities that control, are controlled by, or are under common
 *       control with that entity. For the purposes of this definition,
 *       "control" means (i) the power, direct or indirect, to cause the
 *       direction or management of such entity, whether by contract or
 *       otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *       outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *       "You" (or "Your") shall mean an individual or Legal Entity
 *       exercising permissions granted by this License.
 *
 *       "Source" form shall mean the preferred form for making modifications,
 *       including but not limited to software source code, documentation
 *       source, and configuration files.
 *
 *       "Object" form shall mean any form resulting from mechanical
 *       transformation or translation of a Source form, including but
 *       not limited to compiled object code, generated documentation,
 *       and conversions to other media types.
 *
 *       "Work" shall mean the work of authorship, whether in Source or
 *       Object form, made available under the License, as indicated by a
 *       copyright notice that is included in or attached to the work
 *       (an example is provided in the Appendix below).
 *
 *       "Derivative Works" shall mean any work, whether in Source or Object
 *       form, that is based on (or derived from) the Work and for which the
 *       editorial revisions, annotations, elaborations, or other modifications
 *       represent, as a whole, an original work of authorship. For the purposes
 *       of this License, Derivative Works shall not include works that remain
 *       separable from, or merely link (or bind by name) to the interfaces of,
 *       the Work and Derivative Works thereof.
 *
 *       "Contribution" shall mean any work of authorship, including
 *       the original version of the Work and any modifications or additions
 *       to that Work or Derivative Works thereof, that is intentionally
 *       submitted to Licensor for inclusion in the Work by the copyright owner
 *       or by an individual or Legal Entity authorized to submit on behalf of
 *       the copyright owner. For the purposes of this definition, "submitted"
 *       means any form of electronic, verbal, or written communication sent
 *       to the Licensor or its representatives, including but not limited to
 *       communication on electronic mailing lists, source code control systems,
 *       and issue tracking systems that are managed by, or on behalf of, the
 *       Licensor for the purpose of discussing and improving the Work, but
 *       excluding communication that is conspicuously marked or otherwise
 *       designated in writing by the copyright owner as "Not a Contribution."
 *
 *       "Contributor" shall mean Licensor and any individual or Legal Entity
 *       on behalf of whom a Contribution has been received by Licensor and
 *       subsequently incorporated within the Work.
 *
 *    2. Grant of Copyright License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       copyright license to reproduce, prepare Derivative Works of,
 *       publicly display, publicly perform, sublicense, and distribute the
 *       Work and such Derivative Works in Source or Object form.
 *
 *    3. Grant of Patent License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       (except as stated in this section) patent license to make, have made,
 *       use, offer to sell, sell, import, and otherwise transfer the Work,
 *       where such license applies only to those patent claims licensable
 *       by such Contributor that are necessarily infringed by their
 *       Contribution(s) alone or by combination of their Contribution(s)
 *       with the Work to which such Contribution(s) was submitted. If You
 *       institute patent litigation against any entity (including a
 *       cross-claim or counterclaim in a lawsuit) alleging that the Work
 *       or a Contribution incorporated within the Work constitutes direct
 *       or contributory patent infringement, then any patent licenses
 *       granted to You under this License for that Work shall terminate
 *       as of the date such litigation is filed.
 *
 *    4. Redistribution. You may reproduce and distribute copies of the
 *       Work or Derivative Works thereof in any medium, with or without
 *       modifications, and in Source or Object form, provided that You
 *       meet the following conditions:
 *
 *       (a) You must give any other recipients of the Work or
 *           Derivative Works a copy of this License; and
 *
 *       (b) You must cause any modified files to carry prominent notices
 *           stating that You changed the files; and
 *
 *       (c) You must retain, in the Source form of any Derivative Works
 *           that You distribute, all copyright, patent, trademark, and
 *           attribution notices from the Source form of the Work,
 *           excluding those notices that do not pertain to any part of
 *           the Derivative Works; and
 *
 *       (d) If the Work includes a "NOTICE" text file as part of its
 *           distribution, then any Derivative Works that You distribute must
 *           include a readable copy of the attribution notices contained
 *           within such NOTICE file, excluding those notices that do not
 *           pertain to any part of the Derivative Works, in at least one
 *           of the following places: within a NOTICE text file distributed
 *           as part of the Derivative Works; within the Source form or
 *           documentation, if provided along with the Derivative Works; or,
 *           within a display generated by the Derivative Works, if and
 *           wherever such third-party notices normally appear. The contents
 *           of the NOTICE file are for informational purposes only and
 *           do not modify the License. You may add Your own attribution
 *           notices within Derivative Works that You distribute, alongside
 *           or as an addendum to the NOTICE text from the Work, provided
 *           that such additional attribution notices cannot be construed
 *           as modifying the License.
 *
 *       You may add Your own copyright statement to Your modifications and
 *       may provide additional or different license terms and conditions
 *       for use, reproduction, or distribution of Your modifications, or
 *       for any such Derivative Works as a whole, provided Your use,
 *       reproduction, and distribution of the Work otherwise complies with
 *       the conditions stated in this License.
 *
 *    5. Submission of Contributions. Unless You explicitly state otherwise,
 *       any Contribution intentionally submitted for inclusion in the Work
 *       by You to the Licensor shall be under the terms and conditions of
 *       this License, without any additional terms or conditions.
 *       Notwithstanding the above, nothing herein shall supersede or modify
 *       the terms of any separate license agreement you may have executed
 *       with Licensor regarding such Contributions.
 *
 *    6. Trademarks. This License does not grant permission to use the trade
 *       names, trademarks, service marks, or product names of the Licensor,
 *       except as required for reasonable and customary use in describing the
 *       origin of the Work and reproducing the content of the NOTICE file.
 *
 *    7. Disclaimer of Warranty. Unless required by applicable law or
 *       agreed to in writing, Licensor provides the Work (and each
 *       Contributor provides its Contributions) on an "AS IS" BASIS,
 *       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *       implied, including, without limitation, any warranties or conditions
 *       of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *       PARTICULAR PURPOSE. You are solely responsible for determining the
 *       appropriateness of using or redistributing the Work and assume any
 *       risks associated with Your exercise of permissions under this License.
 *
 *    8. Limitation of Liability. In no event and under no legal theory,
 *       whether in tort (including negligence), contract, or otherwise,
 *       unless required by applicable law (such as deliberate and grossly
 *       negligent acts) or agreed to in writing, shall any Contributor be
 *       liable to You for damages, including any direct, indirect, special,
 *       incidental, or consequential damages of any character arising as a
 *       result of this License or out of the use or inability to use the
 *       Work (including but not limited to damages for loss of goodwill,
 *       work stoppage, computer failure or malfunction, or any and all
 *       other commercial damages or losses), even if such Contributor
 *       has been advised of the possibility of such damages.
 *
 *    9. Accepting Warranty or Additional Liability. While redistributing
 *       the Work or Derivative Works thereof, You may choose to offer,
 *       and charge a fee for, acceptance of support, warranty, indemnity,
 *       or other liability obligations and/or rights consistent with this
 *       License. However, in accepting such obligations, You may act only
 *       on Your own behalf and on Your sole responsibility, not on behalf
 *       of any other Contributor, and only if You agree to indemnify,
 *       defend, and hold each Contributor harmless for any liability
 *       incurred by, or claims asserted against, such Contributor by reason
 *       of your accepting any such warranty or additional liability.
 *
 *    END OF TERMS AND CONDITIONS
 *
 *    APPENDIX: How to apply the Apache License to your work.
 *
 *       To apply the Apache License to your work, attach the following
 *       boilerplate notice, with the fields enclosed by brackets "{}"
 *       replaced with your own identifying information. (Don't include
 *       the brackets!)  The text should be enclosed in the appropriate
 *       comment syntax for the file format. We also recommend that a
 *       file or class name and description of purpose be included on the
 *       same "printed page" as the copyright notice for easier
 *       identification within third-party archives.
 *
 *    Copyright 2014 Edgar Espina
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package org.jooby.internal;

import com.google.common.collect.ImmutableList;
import static java.util.Objects.requireNonNull;
import org.jooby.Asset;
import org.jooby.Cookie;
import org.jooby.Cookie.Definition;
import org.jooby.Deferred;
import org.jooby.MediaType;
import org.jooby.Mutant;
import org.jooby.Renderer;
import org.jooby.Response;
import org.jooby.Result;
import org.jooby.Results;
import org.jooby.Route;
import org.jooby.Route.After;
import org.jooby.Route.Complete;
import org.jooby.Status;
import org.jooby.internal.parser.ParserExecutor;
import org.jooby.spi.NativeResponse;
import org.jooby.funzy.Try;
import org.slf4j.LoggerFactory;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.net.URLEncoder;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.function.Consumer;
import java.util.stream.Collectors;
import java.util.stream.StreamSupport;

public class ResponseImpl implements Response {

  private static final String LOCATION = "Location";

  /** Char encoded content disposition. */
  private static final String CONTENT_DISPOSITION = "attachment; filename=\"%s\"; filename*=%s''%s";

  private final NativeResponse rsp;

  private final Map<String, Object> locals;

  private Route route;

  private Charset charset;

  private final Optional<String> referer;

  private Status status;

  private MediaType type;

  private long len = -1;

  private Map<String, Cookie> cookies = new HashMap<>();

  private List<Renderer> renderers;

  private ParserExecutor parserExecutor;

  private Map<String, Renderer> rendererMap;

  private List<Route.After> after = new ArrayList<>();

  private List<Route.Complete> complete = new ArrayList<>();

  private RequestImpl req;

  private boolean failure;

  private Optional<String> byteRange;

  private boolean resetHeadersOnError = true;

  public ResponseImpl(final RequestImpl req, final ParserExecutor parserExecutor,
      final NativeResponse rsp, final Route route, final List<Renderer> renderers,
      final Map<String, Renderer> rendererMap, final Map<String, Object> locals,
      final Charset charset, final Optional<String> referer, final Optional<String> byteRange) {
    this.req = req;
    this.parserExecutor = parserExecutor;
    this.rsp = rsp;
    this.route = route;
    this.locals = locals;
    this.renderers = renderers;
    this.rendererMap = rendererMap;
    this.charset = charset;
    this.referer = referer;
    this.byteRange = byteRange;
  }

  @Override public boolean isResetHeadersOnError() {
    return resetHeadersOnError;
  }

  @Override public void setResetHeadersOnError(boolean resetHeadersOnError) {
    this.resetHeadersOnError = resetHeadersOnError;
  }

  @Override
  public void download(final String filename, final InputStream stream) throws Throwable {
    requireNonNull(filename, "A file's name is required.");
    requireNonNull(stream, "A stream is required.");

    // handle type
    type(type().orElseGet(() -> MediaType.byPath(filename).orElse(MediaType.octetstream)));

    Asset asset = new InputStreamAsset(stream, filename, type);
    contentDisposition(filename);
    send(Results.with(asset.stream()));
  }

  @Override
  public void download(final String filename, final String location) throws Throwable {
    URL url = getClass().getResource(location.startsWith("/") ? location : "/" + location);
    if (url == null) {
      throw new FileNotFoundException(location);
    }
    // handle type
    type(type().orElseGet(() -> MediaType.byPath(filename).orElse(MediaType.byPath(location)
        .orElse(MediaType.octetstream))));

    URLAsset asset = new URLAsset(url, location, type);
    length(asset.length());

    contentDisposition(filename);
    send(Results.with(asset));
  }

  @Override
  public Response clearCookie(final String name) {
    requireNonNull(name, "Cookie's name required.");
    return cookie(new Cookie.Definition(name).maxAge(0));
  }

  @Override
  public Response cookie(final Definition cookie) {
    requireNonNull(cookie, "Cookie required.");
    // use default path if none-set
    cookie.path(cookie.path().orElse(Route.normalize(req.contextPath() + "/")));
    return cookie(cookie.toCookie());
  }

  @Override
  public Response cookie(final Cookie cookie) {
    requireNonNull(cookie, "Cookie required.");
    String name = cookie.name();
    // clear cookie?
    if (cookie.maxAge() == 0) {
      // clear previously set cookie, otherwise ignore them
      if (cookies.remove(name) == null) {
        // cookie was set in a previous req, we must send a expire header.
        cookies.put(name, cookie);
      }
    } else {
      cookies.put(name, cookie);
    }
    return this;
  }

  @Override
  public Mutant header(final String name) {
    requireNonNull(name, "A header's name is required.");
    return new MutantImpl(parserExecutor,
        new StrParamReferenceImpl("header", name, rsp.headers(name)));
  }

  @Override
  public Response header(final String name, final Object value) {
    requireNonNull(name, "Header's name is required.");
    requireNonNull(value, "Header's value is required.");

    return setHeader(name, value);
  }

  @Override
  public Response header(final String name, final Iterable<Object> values) {
    requireNonNull(name, "Header's name is required.");
    requireNonNull(values, "Header's values are required.");

    return setHeader(name, values);
  }

  @Override
  public Charset charset() {
    return charset;
  }

  @Override
  public Response charset(final Charset charset) {
    this.charset = requireNonNull(charset, "A charset is required.");
    type(type().orElse(MediaType.html));
    return this;
  }

  @Override
  public Response length(final long length) {
    len = length;
    rsp.header("Content-Length", Long.toString(length));
    return this;
  }

  @Override
  public Optional<MediaType> type() {
    return Optional.ofNullable(type);
  }

  @Override
  public Response type(final MediaType type) {
    this.type = type;
    if (type.isText()) {
      setHeader("Content-Type", type.name() + ";charset=" + charset.name());
    } else {
      setHeader("Content-Type", type.name());
    }
    return this;
  }

  @Override
  public void redirect(final Status status, final String location) throws Throwable {
    requireNonNull(status, "Status required.");
    requireNonNull(location, "Location required.");

    send(Results.with(status).header(LOCATION, location));
  }

  @Override
  public Optional<Status> status() {
    return Optional.ofNullable(status);
  }

  @Override
  public Response status(final Status status) {
    this.status = requireNonNull(status, "Status required.");
    rsp.statusCode(status.value());
    failure = status.isError();
    return this;
  }

  @Override
  public boolean committed() {
    return rsp.committed();
  }

  public void done(final Optional<Throwable> cause) {
    if (complete.size() > 0) {
      for (Route.Complete h : complete) {
        Try.run(() -> h.handle(req, this, cause))
            .onFailure(x -> LoggerFactory.getLogger(Response.class)
                .error("complete listener resulted in error", x));
      }
      complete.clear();
    }
    end();
  }

  @Override
  public void end() {
    if (!rsp.committed()) {
      if (status == null) {
        status(rsp.statusCode());
      }

      writeCookies();

      /**
       * Do we need to figure it out Content-Length?
       */
      boolean lenSet = rsp.header("Content-Length").isPresent()
          || rsp.header("Transfer-Encoding").isPresent();
      if (!lenSet) {
        int statusCode = status.value();
        boolean hasBody = true;
        if (statusCode >= 100 && statusCode < 200) {
          hasBody = false;
        } else if (statusCode == 204 || statusCode == 304) {
          hasBody = false;
        }
        if (hasBody) {
          rsp.header("Content-Length", "0");
        }
      }
    }
    rsp.end();
  }

  @Override
  public void send(final Result result) throws Throwable {
    if (result instanceof Deferred) {
      throw new DeferredExecution((Deferred) result);
    }

    Result finalResult = result;

    if (!failure) {
      // after filter
      for (int i = after.size() - 1; i >= 0; i--) {
        finalResult = after.get(i).handle(req, this, finalResult);
      }
    }

    Optional<MediaType> rtype = finalResult.type();
    if (rtype.isPresent()) {
      type(rtype.get());
    }

    Optional<Status> finalStatus = finalResult.status();
    if (finalStatus.isPresent()) {
      status(finalStatus.get());
    } else if (this.status == null) {
      status(Status.OK);
    }

    Map<String, Object> headers = finalResult.headers();
    if (headers.size() > 0) {
      headers.forEach(this::setHeader);
    }

    writeCookies();

    if (Route.HEAD.equals(route.method())) {
      end();
      return;
    }

    /**
     * Do we need to figure it out Content-Length?
     */
    List<MediaType> produces = this.type == null ? route.produces() : ImmutableList.of(type);
    Object value = finalResult.get(produces);

    if (value != null) {
      Consumer<Long> setLen = len -> {
        if (this.len == -1 && len >= 0) {
          length(len);
        }
      };

      Consumer<MediaType> setType = type -> {
        if (this.type == null) {
          type(type);
        }
      };

      HttpRendererContext ctx = new HttpRendererContext(
          renderers,
          rsp,
          setLen,
          setType,
          locals,
          produces,
          charset,
          req.locale(),
          byteRange);

      // explicit renderer?
      Renderer renderer = rendererMap.get(route.renderer());
      if (renderer != null) {
        renderer.render(value, ctx);
      } else {
        ctx.render(value);
      }
    }
    // end response
    end();
  }

  @Override
  public void after(final After handler) {
    after.add(handler);
  }

  @Override
  public void complete(final Complete handler) {
    complete.add(handler);
  }

  private void writeCookies() {
    if (cookies.size() > 0) {
      List<String> setCookie = cookies.values().stream()
          .map(Cookie::encode)
          .collect(Collectors.toList());
      rsp.header("Set-Cookie", setCookie);
      cookies.clear();
    }
  }

  public void reset() {
    if (resetHeadersOnError) {
      status = null;
      this.cookies.clear();
      rsp.reset();
    }
  }

  void route(final Route route) {
    this.route = route;
  }

  private void contentDisposition(final String filename) throws IOException {
    List<String> headers = rsp.headers("Content-Disposition");
    if (headers.isEmpty()) {
      String basename = filename;
      int last = filename.lastIndexOf('/');
      if (last >= 0) {
        basename = basename.substring(last + 1);
      }

      String cs = charset.name();
      String ebasename = URLEncoder.encode(basename, cs).replaceAll("\\+", "%20");
      header("Content-Disposition", String.format(CONTENT_DISPOSITION, basename, cs, ebasename));
    }
  }

  @SuppressWarnings("unchecked")
  private Response setHeader(final String name, final Object value) {
    if (!committed()) {
      if (value instanceof Iterable) {
        List<String> values = StreamSupport.stream(((Iterable<Object>) value).spliterator(), false)
            .map(Headers::encode)
            .collect(Collectors.toList());
        rsp.header(name, values);
      } else {
        if (LOCATION.equalsIgnoreCase(name)) {
          String location = value.toString();
          String cpath = req.contextPath();
          if ("back".equalsIgnoreCase(location)) {
            location = referer.orElse(cpath + "/");
          } else if (location.startsWith("/") && !location.startsWith(cpath)) {
            location = cpath + location;
          }
          rsp.header(LOCATION, location);
        } else {
          if ("Content-Type".equalsIgnoreCase(name)) {
            // keep type reference
            this.type = MediaType.valueOf(value.toString());
          }
          rsp.header(name, Headers.encode(value));
        }
      }
    }

    return this;
  }

}
