/**
 *                                  Apache License
 *                            Version 2.0, January 2004
 *                         http://www.apache.org/licenses/
 *
 *    TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *    1. Definitions.
 *
 *       "License" shall mean the terms and conditions for use, reproduction,
 *       and distribution as defined by Sections 1 through 9 of this document.
 *
 *       "Licensor" shall mean the copyright owner or entity authorized by
 *       the copyright owner that is granting the License.
 *
 *       "Legal Entity" shall mean the union of the acting entity and all
 *       other entities that control, are controlled by, or are under common
 *       control with that entity. For the purposes of this definition,
 *       "control" means (i) the power, direct or indirect, to cause the
 *       direction or management of such entity, whether by contract or
 *       otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *       outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *       "You" (or "Your") shall mean an individual or Legal Entity
 *       exercising permissions granted by this License.
 *
 *       "Source" form shall mean the preferred form for making modifications,
 *       including but not limited to software source code, documentation
 *       source, and configuration files.
 *
 *       "Object" form shall mean any form resulting from mechanical
 *       transformation or translation of a Source form, including but
 *       not limited to compiled object code, generated documentation,
 *       and conversions to other media types.
 *
 *       "Work" shall mean the work of authorship, whether in Source or
 *       Object form, made available under the License, as indicated by a
 *       copyright notice that is included in or attached to the work
 *       (an example is provided in the Appendix below).
 *
 *       "Derivative Works" shall mean any work, whether in Source or Object
 *       form, that is based on (or derived from) the Work and for which the
 *       editorial revisions, annotations, elaborations, or other modifications
 *       represent, as a whole, an original work of authorship. For the purposes
 *       of this License, Derivative Works shall not include works that remain
 *       separable from, or merely link (or bind by name) to the interfaces of,
 *       the Work and Derivative Works thereof.
 *
 *       "Contribution" shall mean any work of authorship, including
 *       the original version of the Work and any modifications or additions
 *       to that Work or Derivative Works thereof, that is intentionally
 *       submitted to Licensor for inclusion in the Work by the copyright owner
 *       or by an individual or Legal Entity authorized to submit on behalf of
 *       the copyright owner. For the purposes of this definition, "submitted"
 *       means any form of electronic, verbal, or written communication sent
 *       to the Licensor or its representatives, including but not limited to
 *       communication on electronic mailing lists, source code control systems,
 *       and issue tracking systems that are managed by, or on behalf of, the
 *       Licensor for the purpose of discussing and improving the Work, but
 *       excluding communication that is conspicuously marked or otherwise
 *       designated in writing by the copyright owner as "Not a Contribution."
 *
 *       "Contributor" shall mean Licensor and any individual or Legal Entity
 *       on behalf of whom a Contribution has been received by Licensor and
 *       subsequently incorporated within the Work.
 *
 *    2. Grant of Copyright License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       copyright license to reproduce, prepare Derivative Works of,
 *       publicly display, publicly perform, sublicense, and distribute the
 *       Work and such Derivative Works in Source or Object form.
 *
 *    3. Grant of Patent License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       (except as stated in this section) patent license to make, have made,
 *       use, offer to sell, sell, import, and otherwise transfer the Work,
 *       where such license applies only to those patent claims licensable
 *       by such Contributor that are necessarily infringed by their
 *       Contribution(s) alone or by combination of their Contribution(s)
 *       with the Work to which such Contribution(s) was submitted. If You
 *       institute patent litigation against any entity (including a
 *       cross-claim or counterclaim in a lawsuit) alleging that the Work
 *       or a Contribution incorporated within the Work constitutes direct
 *       or contributory patent infringement, then any patent licenses
 *       granted to You under this License for that Work shall terminate
 *       as of the date such litigation is filed.
 *
 *    4. Redistribution. You may reproduce and distribute copies of the
 *       Work or Derivative Works thereof in any medium, with or without
 *       modifications, and in Source or Object form, provided that You
 *       meet the following conditions:
 *
 *       (a) You must give any other recipients of the Work or
 *           Derivative Works a copy of this License; and
 *
 *       (b) You must cause any modified files to carry prominent notices
 *           stating that You changed the files; and
 *
 *       (c) You must retain, in the Source form of any Derivative Works
 *           that You distribute, all copyright, patent, trademark, and
 *           attribution notices from the Source form of the Work,
 *           excluding those notices that do not pertain to any part of
 *           the Derivative Works; and
 *
 *       (d) If the Work includes a "NOTICE" text file as part of its
 *           distribution, then any Derivative Works that You distribute must
 *           include a readable copy of the attribution notices contained
 *           within such NOTICE file, excluding those notices that do not
 *           pertain to any part of the Derivative Works, in at least one
 *           of the following places: within a NOTICE text file distributed
 *           as part of the Derivative Works; within the Source form or
 *           documentation, if provided along with the Derivative Works; or,
 *           within a display generated by the Derivative Works, if and
 *           wherever such third-party notices normally appear. The contents
 *           of the NOTICE file are for informational purposes only and
 *           do not modify the License. You may add Your own attribution
 *           notices within Derivative Works that You distribute, alongside
 *           or as an addendum to the NOTICE text from the Work, provided
 *           that such additional attribution notices cannot be construed
 *           as modifying the License.
 *
 *       You may add Your own copyright statement to Your modifications and
 *       may provide additional or different license terms and conditions
 *       for use, reproduction, or distribution of Your modifications, or
 *       for any such Derivative Works as a whole, provided Your use,
 *       reproduction, and distribution of the Work otherwise complies with
 *       the conditions stated in this License.
 *
 *    5. Submission of Contributions. Unless You explicitly state otherwise,
 *       any Contribution intentionally submitted for inclusion in the Work
 *       by You to the Licensor shall be under the terms and conditions of
 *       this License, without any additional terms or conditions.
 *       Notwithstanding the above, nothing herein shall supersede or modify
 *       the terms of any separate license agreement you may have executed
 *       with Licensor regarding such Contributions.
 *
 *    6. Trademarks. This License does not grant permission to use the trade
 *       names, trademarks, service marks, or product names of the Licensor,
 *       except as required for reasonable and customary use in describing the
 *       origin of the Work and reproducing the content of the NOTICE file.
 *
 *    7. Disclaimer of Warranty. Unless required by applicable law or
 *       agreed to in writing, Licensor provides the Work (and each
 *       Contributor provides its Contributions) on an "AS IS" BASIS,
 *       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *       implied, including, without limitation, any warranties or conditions
 *       of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *       PARTICULAR PURPOSE. You are solely responsible for determining the
 *       appropriateness of using or redistributing the Work and assume any
 *       risks associated with Your exercise of permissions under this License.
 *
 *    8. Limitation of Liability. In no event and under no legal theory,
 *       whether in tort (including negligence), contract, or otherwise,
 *       unless required by applicable law (such as deliberate and grossly
 *       negligent acts) or agreed to in writing, shall any Contributor be
 *       liable to You for damages, including any direct, indirect, special,
 *       incidental, or consequential damages of any character arising as a
 *       result of this License or out of the use or inability to use the
 *       Work (including but not limited to damages for loss of goodwill,
 *       work stoppage, computer failure or malfunction, or any and all
 *       other commercial damages or losses), even if such Contributor
 *       has been advised of the possibility of such damages.
 *
 *    9. Accepting Warranty or Additional Liability. While redistributing
 *       the Work or Derivative Works thereof, You may choose to offer,
 *       and charge a fee for, acceptance of support, warranty, indemnity,
 *       or other liability obligations and/or rights consistent with this
 *       License. However, in accepting such obligations, You may act only
 *       on Your own behalf and on Your sole responsibility, not on behalf
 *       of any other Contributor, and only if You agree to indemnify,
 *       defend, and hold each Contributor harmless for any liability
 *       incurred by, or claims asserted against, such Contributor by reason
 *       of your accepting any such warranty or additional liability.
 *
 *    END OF TERMS AND CONDITIONS
 *
 *    APPENDIX: How to apply the Apache License to your work.
 *
 *       To apply the Apache License to your work, attach the following
 *       boilerplate notice, with the fields enclosed by brackets "{}"
 *       replaced with your own identifying information. (Don't include
 *       the brackets!)  The text should be enclosed in the appropriate
 *       comment syntax for the file format. We also recommend that a
 *       file or class name and description of purpose be included on the
 *       same "printed page" as the copyright notice for easier
 *       identification within third-party archives.
 *
 *    Copyright 2014 Edgar Espina
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package org.jooby.internal.mvc;

import com.google.common.base.CaseFormat;
import com.google.common.collect.ImmutableSet;
import org.jooby.Env;
import org.jooby.MediaType;
import org.jooby.Route;
import org.jooby.Route.Definition;
import org.jooby.funzy.Try;
import org.jooby.internal.RouteMetadata;
import org.jooby.mvc.CONNECT;
import org.jooby.mvc.Consumes;
import org.jooby.mvc.DELETE;
import org.jooby.mvc.GET;
import org.jooby.mvc.HEAD;
import org.jooby.mvc.OPTIONS;
import org.jooby.mvc.PATCH;
import org.jooby.mvc.POST;
import org.jooby.mvc.PUT;
import org.jooby.mvc.Path;
import org.jooby.mvc.Produces;
import org.jooby.mvc.TRACE;

import java.lang.annotation.Annotation;
import java.lang.reflect.AnnotatedElement;
import java.lang.reflect.Array;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.function.BiConsumer;
import java.util.function.Consumer;
import java.util.function.Function;

public class MvcRoutes {

  private static final String[] EMPTY = new String[0];

  private static final Set<Class<? extends Annotation>> VERBS = ImmutableSet.of(GET.class,
      POST.class, PUT.class, DELETE.class, PATCH.class, HEAD.class, OPTIONS.class, TRACE.class,
      CONNECT.class);

  private static final Set<Class<? extends Annotation>> IGNORE = ImmutableSet
      .<Class<? extends Annotation>>builder()
      .addAll(VERBS)
      .add(Path.class)
      .add(Produces.class)
      .add(Consumes.class)
      .build();

  public static List<Route.Definition> routes(final Env env, final RouteMetadata classInfo,
      final String rpath, boolean caseSensitiveRouting, final Class<?> routeClass) {

    // check and fail fast
    methods(routeClass, methods -> {
      routes(methods, (m, a) -> {
        if (!Modifier.isPublic(m.getModifiers())) {
          throw new IllegalArgumentException("Not a public method: " + m);
        }
      });
    });

    RequestParamProvider provider = new RequestParamProviderImpl(
        new RequestParamNameProviderImpl(classInfo));

    String[] rootPaths = path(routeClass);
    String[] rootExcludes = excludes(routeClass, EMPTY);

    // we are good, now collect them
    Map<Method, List<Class<?>>> methods = new HashMap<>();
    routes(routeClass.getMethods(), methods::put);

    List<Definition> definitions = new ArrayList<>();
    Map<String, Object> attrs = attrs(routeClass.getAnnotations());
    methods
        .keySet()
        .stream()
        .sorted((m1, m2) -> {
          int l1 = classInfo.startAt(m1);
          int l2 = classInfo.startAt(m2);
          return l1 - l2;
        })
        .forEach(method -> {
          /**
           * Param provider: dev vs none dev
           */
          RequestParamProvider paramProvider = provider;
          if (!env.name().equals("dev")) {
            List<RequestParam> params = provider.parameters(method);
            paramProvider = (h) -> params;
          }

          List<Class<?>> verbs = methods.get(method);
          List<MediaType> produces = produces(method);
          List<MediaType> consumes = consumes(method);
          Map<String, Object> localAttrs = new HashMap<>(attrs);
          localAttrs.putAll(attrs(method.getAnnotations()));

          for (String path : expandPaths(rootPaths, method)) {
            for (Class<?> verb : verbs) {
              String name = routeClass.getSimpleName() + "." + method.getName();

              String[] excludes = excludes(method, rootExcludes);

              Definition definition = new Route.Definition(
                  verb.getSimpleName(), rpath + "/" + path,
                  new MvcHandler(method, routeClass, paramProvider), caseSensitiveRouting)
                  .produces(produces)
                  .consumes(consumes)
                  .excludes(excludes)
                  .declaringClass(routeClass.getName())
                  .line(classInfo.startAt(method) - 1)
                  .name(name);

              localAttrs.forEach((n, v) -> definition.attr(n, v));
              definitions.add(definition);
            }
          }
        });

    return definitions;
  }

  private static void methods(final Class<?> clazz, final Consumer<Method[]> callback) {
    if (clazz != Object.class) {
      callback.accept(clazz.getDeclaredMethods());
      methods(clazz.getSuperclass(), callback);
    }
  }

  @SuppressWarnings({"rawtypes", "unchecked"})
  private static void routes(final Method[] methods,
      final BiConsumer<Method, List<Class<?>>> consumer) {
    for (Method method : methods) {
      List<Class<?>> annotations = new ArrayList<>();
      for (Class annotationType : VERBS) {
        Annotation annotation = method.getAnnotation(annotationType);
        if (annotation != null) {
          annotations.add(annotationType);
        }
      }
      if (annotations.size() > 0) {
        consumer.accept(method, annotations);
      } else if (method.isAnnotationPresent(Path.class)) {
        consumer.accept(method, Arrays.asList(GET.class));
      }
    }
  }

  private static Map<String, Object> attrs(final Annotation[] annotations) {
    Map<String, Object> result = new LinkedHashMap<>();
    for (Annotation annotation : annotations) {
      result.putAll(attrs(annotation));
    }
    return result;
  }

  private static Map<String, Object> attrs(final Annotation annotation) {
    Map<String, Object> result = new LinkedHashMap<>();
    Class<? extends Annotation> annotationType = annotation.annotationType();
    if (!IGNORE.contains(annotationType)) {
      Method[] attrs = annotation.annotationType().getDeclaredMethods();
      for (Method attr : attrs) {
        Try.apply(() -> attr.invoke(annotation))
            .onSuccess(value -> {
              if (value.getClass().isArray() && Annotation.class
                  .isAssignableFrom(value.getClass().getComponentType())) {
                List<Map<String, Object>> array = new ArrayList<>();
                for(int i = 0; i < Array.getLength(value); i ++) {
                  array.add(attrs((Annotation) Array.get(value, i)));
                }
                result.put(attrName(annotation, attr), array.toArray());
              } else {
                result.put(attrName(annotation, attr), value);
              }
            });
      }
    }
    return result;
  }

  private static String attrName(final Annotation annotation, final Method attr) {
    String name = attr.getName();
    String scope = CaseFormat.UPPER_CAMEL.to(CaseFormat.LOWER_CAMEL, annotation.annotationType().getSimpleName());
    if (name.equals("value")) {
      return scope;
    }
    return scope + "." + name;
  }

  private static List<MediaType> produces(final Method method) {
    Function<AnnotatedElement, Optional<List<MediaType>>> fn = (element) -> {
      Produces produces = element.getAnnotation(Produces.class);
      if (produces != null) {
        return Optional.of(MediaType.valueOf(produces.value()));
      }
      return Optional.empty();
    };

    // method level
    return fn.apply(method)
        // class level
        .orElseGet(() -> fn.apply(method.getDeclaringClass())
            // none
            .orElse(MediaType.ALL));
  }

  private static List<MediaType> consumes(final Method method) {
    Function<AnnotatedElement, Optional<List<MediaType>>> fn = (element) -> {
      Consumes consumes = element.getAnnotation(Consumes.class);
      if (consumes != null) {
        return Optional.of(MediaType.valueOf(consumes.value()));
      }
      return Optional.empty();
    };

    // method level
    return fn.apply(method)
        // class level
        .orElseGet(() -> fn.apply(method.getDeclaringClass())
            // none
            .orElse(MediaType.ALL));
  }

  private static String[] path(final AnnotatedElement owner) {
    Path annotation = owner.getAnnotation(Path.class);
    if (annotation == null) {
      return EMPTY;
    }
    return annotation.value();
  }

  private static String[] excludes(final AnnotatedElement owner, final String[] parent) {
    Path annotation = owner.getAnnotation(Path.class);
    if (annotation == null) {
      return parent;
    }
    String[] excludes = annotation.excludes();
    if (excludes.length == 0) {
      return parent;
    }
    if (parent.length == 0) {
      return excludes;
    }
    // join everything
    int size = parent.length + excludes.length;
    String[] result = new String[size];
    System.arraycopy(parent, 0, result, 0, parent.length);
    System.arraycopy(excludes, 0, result, parent.length, excludes.length);
    return result;
  }

  private static String[] expandPaths(final String[] root, final Method m) {
    String[] path = path(m);
    if (root.length == 0) {
      if (path.length == 0) {
        throw new IllegalArgumentException("No path(s) found for: " + m);
      }
      return path;
    }
    if (path.length == 0) {
      return root;
    }
    String[] result = new String[root.length * path.length];
    int k = 0;
    for (String base : root) {
      for (String element : path) {
        result[k] = base + "/" + element;
        k += 1;
      }
    }
    return result;
  }
}
