/**
 *                                  Apache License
 *                            Version 2.0, January 2004
 *                         http://www.apache.org/licenses/
 *
 *    TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *    1. Definitions.
 *
 *       "License" shall mean the terms and conditions for use, reproduction,
 *       and distribution as defined by Sections 1 through 9 of this document.
 *
 *       "Licensor" shall mean the copyright owner or entity authorized by
 *       the copyright owner that is granting the License.
 *
 *       "Legal Entity" shall mean the union of the acting entity and all
 *       other entities that control, are controlled by, or are under common
 *       control with that entity. For the purposes of this definition,
 *       "control" means (i) the power, direct or indirect, to cause the
 *       direction or management of such entity, whether by contract or
 *       otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *       outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *       "You" (or "Your") shall mean an individual or Legal Entity
 *       exercising permissions granted by this License.
 *
 *       "Source" form shall mean the preferred form for making modifications,
 *       including but not limited to software source code, documentation
 *       source, and configuration files.
 *
 *       "Object" form shall mean any form resulting from mechanical
 *       transformation or translation of a Source form, including but
 *       not limited to compiled object code, generated documentation,
 *       and conversions to other media types.
 *
 *       "Work" shall mean the work of authorship, whether in Source or
 *       Object form, made available under the License, as indicated by a
 *       copyright notice that is included in or attached to the work
 *       (an example is provided in the Appendix below).
 *
 *       "Derivative Works" shall mean any work, whether in Source or Object
 *       form, that is based on (or derived from) the Work and for which the
 *       editorial revisions, annotations, elaborations, or other modifications
 *       represent, as a whole, an original work of authorship. For the purposes
 *       of this License, Derivative Works shall not include works that remain
 *       separable from, or merely link (or bind by name) to the interfaces of,
 *       the Work and Derivative Works thereof.
 *
 *       "Contribution" shall mean any work of authorship, including
 *       the original version of the Work and any modifications or additions
 *       to that Work or Derivative Works thereof, that is intentionally
 *       submitted to Licensor for inclusion in the Work by the copyright owner
 *       or by an individual or Legal Entity authorized to submit on behalf of
 *       the copyright owner. For the purposes of this definition, "submitted"
 *       means any form of electronic, verbal, or written communication sent
 *       to the Licensor or its representatives, including but not limited to
 *       communication on electronic mailing lists, source code control systems,
 *       and issue tracking systems that are managed by, or on behalf of, the
 *       Licensor for the purpose of discussing and improving the Work, but
 *       excluding communication that is conspicuously marked or otherwise
 *       designated in writing by the copyright owner as "Not a Contribution."
 *
 *       "Contributor" shall mean Licensor and any individual or Legal Entity
 *       on behalf of whom a Contribution has been received by Licensor and
 *       subsequently incorporated within the Work.
 *
 *    2. Grant of Copyright License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       copyright license to reproduce, prepare Derivative Works of,
 *       publicly display, publicly perform, sublicense, and distribute the
 *       Work and such Derivative Works in Source or Object form.
 *
 *    3. Grant of Patent License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       (except as stated in this section) patent license to make, have made,
 *       use, offer to sell, sell, import, and otherwise transfer the Work,
 *       where such license applies only to those patent claims licensable
 *       by such Contributor that are necessarily infringed by their
 *       Contribution(s) alone or by combination of their Contribution(s)
 *       with the Work to which such Contribution(s) was submitted. If You
 *       institute patent litigation against any entity (including a
 *       cross-claim or counterclaim in a lawsuit) alleging that the Work
 *       or a Contribution incorporated within the Work constitutes direct
 *       or contributory patent infringement, then any patent licenses
 *       granted to You under this License for that Work shall terminate
 *       as of the date such litigation is filed.
 *
 *    4. Redistribution. You may reproduce and distribute copies of the
 *       Work or Derivative Works thereof in any medium, with or without
 *       modifications, and in Source or Object form, provided that You
 *       meet the following conditions:
 *
 *       (a) You must give any other recipients of the Work or
 *           Derivative Works a copy of this License; and
 *
 *       (b) You must cause any modified files to carry prominent notices
 *           stating that You changed the files; and
 *
 *       (c) You must retain, in the Source form of any Derivative Works
 *           that You distribute, all copyright, patent, trademark, and
 *           attribution notices from the Source form of the Work,
 *           excluding those notices that do not pertain to any part of
 *           the Derivative Works; and
 *
 *       (d) If the Work includes a "NOTICE" text file as part of its
 *           distribution, then any Derivative Works that You distribute must
 *           include a readable copy of the attribution notices contained
 *           within such NOTICE file, excluding those notices that do not
 *           pertain to any part of the Derivative Works, in at least one
 *           of the following places: within a NOTICE text file distributed
 *           as part of the Derivative Works; within the Source form or
 *           documentation, if provided along with the Derivative Works; or,
 *           within a display generated by the Derivative Works, if and
 *           wherever such third-party notices normally appear. The contents
 *           of the NOTICE file are for informational purposes only and
 *           do not modify the License. You may add Your own attribution
 *           notices within Derivative Works that You distribute, alongside
 *           or as an addendum to the NOTICE text from the Work, provided
 *           that such additional attribution notices cannot be construed
 *           as modifying the License.
 *
 *       You may add Your own copyright statement to Your modifications and
 *       may provide additional or different license terms and conditions
 *       for use, reproduction, or distribution of Your modifications, or
 *       for any such Derivative Works as a whole, provided Your use,
 *       reproduction, and distribution of the Work otherwise complies with
 *       the conditions stated in this License.
 *
 *    5. Submission of Contributions. Unless You explicitly state otherwise,
 *       any Contribution intentionally submitted for inclusion in the Work
 *       by You to the Licensor shall be under the terms and conditions of
 *       this License, without any additional terms or conditions.
 *       Notwithstanding the above, nothing herein shall supersede or modify
 *       the terms of any separate license agreement you may have executed
 *       with Licensor regarding such Contributions.
 *
 *    6. Trademarks. This License does not grant permission to use the trade
 *       names, trademarks, service marks, or product names of the Licensor,
 *       except as required for reasonable and customary use in describing the
 *       origin of the Work and reproducing the content of the NOTICE file.
 *
 *    7. Disclaimer of Warranty. Unless required by applicable law or
 *       agreed to in writing, Licensor provides the Work (and each
 *       Contributor provides its Contributions) on an "AS IS" BASIS,
 *       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *       implied, including, without limitation, any warranties or conditions
 *       of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *       PARTICULAR PURPOSE. You are solely responsible for determining the
 *       appropriateness of using or redistributing the Work and assume any
 *       risks associated with Your exercise of permissions under this License.
 *
 *    8. Limitation of Liability. In no event and under no legal theory,
 *       whether in tort (including negligence), contract, or otherwise,
 *       unless required by applicable law (such as deliberate and grossly
 *       negligent acts) or agreed to in writing, shall any Contributor be
 *       liable to You for damages, including any direct, indirect, special,
 *       incidental, or consequential damages of any character arising as a
 *       result of this License or out of the use or inability to use the
 *       Work (including but not limited to damages for loss of goodwill,
 *       work stoppage, computer failure or malfunction, or any and all
 *       other commercial damages or losses), even if such Contributor
 *       has been advised of the possibility of such damages.
 *
 *    9. Accepting Warranty or Additional Liability. While redistributing
 *       the Work or Derivative Works thereof, You may choose to offer,
 *       and charge a fee for, acceptance of support, warranty, indemnity,
 *       or other liability obligations and/or rights consistent with this
 *       License. However, in accepting such obligations, You may act only
 *       on Your own behalf and on Your sole responsibility, not on behalf
 *       of any other Contributor, and only if You agree to indemnify,
 *       defend, and hold each Contributor harmless for any liability
 *       incurred by, or claims asserted against, such Contributor by reason
 *       of your accepting any such warranty or additional liability.
 *
 *    END OF TERMS AND CONDITIONS
 *
 *    APPENDIX: How to apply the Apache License to your work.
 *
 *       To apply the Apache License to your work, attach the following
 *       boilerplate notice, with the fields enclosed by brackets "{}"
 *       replaced with your own identifying information. (Don't include
 *       the brackets!)  The text should be enclosed in the appropriate
 *       comment syntax for the file format. We also recommend that a
 *       file or class name and description of purpose be included on the
 *       same "printed page" as the copyright notice for easier
 *       identification within third-party archives.
 *
 *    Copyright 2014 Edgar Espina
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package org.jooby.internal.parser.bean;

import com.google.common.base.CharMatcher;
import com.google.common.base.Splitter;
import com.google.inject.TypeLiteral;
import org.jooby.Request;
import org.jooby.internal.ParameterNameProvider;
import org.jooby.internal.mvc.RequestParam;
import org.jooby.internal.mvc.RequestParamNameProviderImpl;
import org.jooby.internal.mvc.RequestParamProviderImpl;
import org.jooby.funzy.Throwing;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

@SuppressWarnings("rawtypes")
public class BeanPlan {

  /** The logging system. */
  private final Logger log = LoggerFactory.getLogger(Request.class);

  private Constructor<?> constructor;

  private List<RequestParam> parameters;

  private TypeLiteral beanType;

  private Map<Object, BeanPath> cache = new ConcurrentHashMap<>();

  @SuppressWarnings("unchecked")
  public BeanPlan(final ParameterNameProvider classInfo, final Class beanType) {
    this(classInfo, TypeLiteral.get(beanType));
  }

  public BeanPlan(final ParameterNameProvider classInfo, final TypeLiteral beanType) {
    Constructor<?> inject = null, def = null;
    Class rawType = beanType.getRawType();
    if (rawType == List.class) {
      rawType = ArrayList.class;
    }
    Constructor<?>[] cons = rawType.getDeclaredConstructors();
    if (cons.length == 1) {
      def = cons[0];
    } else {
      for (Constructor<?> c : cons) {
        if (c.isAnnotationPresent(Inject.class)) {
          if (inject != null) {
            throw new IllegalStateException(
                "Ambigous constructor found: " + rawType.getName()
                    + ". Only one @" + Inject.class.getName() + " allowed");
          }
          inject = c;
        } else if (c.getParameterCount() == 0) {
          def = c;
        }
      }
    }
    Constructor<?> constructor = inject == null ? def : inject;
    if (constructor == null) {
      throw new IllegalStateException("Ambigous constructor found: " + rawType.getName()
          + ". Bean/Form type must have a no-args constructor or must be annotated with @"
          + Inject.class.getName());
    }
    this.beanType = beanType;
    this.constructor = constructor;
    this.parameters = new RequestParamProviderImpl(new RequestParamNameProviderImpl(classInfo))
        .parameters(constructor);
  }

  public Object newBean(final Throwing.Function<RequestParam, Object> lookup,
      final Set<String> params) throws Throwable {
    log.debug("instantiating object {}", constructor);

    Object[] args = new Object[parameters.size()];
    List<String> names = new ArrayList<>(params);
    // remove constructor injected params
    for (int i = 0; i < args.length; i++) {
      RequestParam param = parameters.get(i);
      args[i] = lookup.apply(param);
      // skip constructor injected param (don't override)
      names.remove(param.name);
    }
    Object bean = constructor.newInstance(args);

    List<BeanPath> paths = compile(names.stream().sorted().iterator(), beanType);
    for (BeanPath path : paths) {
      String rawpath = path.toString();
      log.debug("  setting {}", rawpath);
      path.set(bean, lookup.apply(new RequestParam(path.setelem(), rawpath, path.settype())));
    }
    return bean;
  }

  private List<BeanPath> compile(final Iterator<String> it, final TypeLiteral beanType) {
    List<BeanPath> result = new ArrayList<>();
    while (it.hasNext()) {
      String path = it.next();
      List<Object> ckey = Arrays.asList(beanType, path);
      BeanPath cached = cache.get(ckey);
      if (cached == null) {
        List<Object[]> segments = segments(path);
        List<BeanPath> chain = new ArrayList<>();
        // traverse path
        TypeLiteral ittype = beanType;
        for (int i = 0; i < segments.size() - 1; i++) {
          Object[] segment = segments.get(i);
          final BeanPath cpath;
          if (segment[1] != null) {
            if (segment[0] == null) {
              cpath = new BeanIndexedPath(null, (Integer) segment[1], ittype);
            } else {
              BeanPath getter = member("get", (String) segment[0], ittype, 0);
              if (getter instanceof BeanMethodPath) {
                ((BeanMethodPath) getter).setter = member("set", (String) segment[0], ittype, 1);
              }
              cpath = new BeanIndexedPath(getter, (Integer) segment[1], ittype);
            }
          } else {
            BeanPath getter = member("get", (String) segment[0], ittype, 0);
            if (getter instanceof BeanMethodPath) {
              ((BeanMethodPath) getter).setter = member("set", (String) segment[0], ittype, 1);
            }
            cpath = getter;
          }
          if (cpath != null) {
            chain.add(cpath);
            ittype = TypeLiteral.get(cpath.type());
          }
        }

        // set path
        Object[] last = segments.get(segments.size() - 1);
        BeanPath cpath = member("set", (String) last[0], ittype, 1);
        if (cpath != null) {
          if (last[1] != null) {
            BeanPath getter = member("get", (String) last[0], ittype, 0);
            if (getter instanceof BeanMethodPath) {
              ((BeanMethodPath) getter).setter = cpath;
            }
            cpath = new BeanIndexedPath(getter, (Integer) last[1], ittype);
          }
          if (chain.size() == 0) {
            cached = cpath;
          } else {
            cached = new BeanComplexPath(chain, cpath, path);
          }
          cache.put(ckey, cached);
        }
      }
      if (cached != null) {
        result.add(cached);
      }
    }
    return result;
  }

  private List<Object[]> segments(final String path) {
    List<String> segments = Splitter.on(CharMatcher.anyOf("[].")).trimResults()
        .omitEmptyStrings()
        .splitToList(path);
    List<Object[]> result = new ArrayList<>(segments.size());
    for (int i = 0; i < segments.size(); i++) {
      String segment = segments.get(i);
      try {
        int idx = Integer.parseInt(segment);
        if (result.size() > 0) {
          result.set(result.size() - 1, new Object[]{result.get(result.size() - 1)[0], idx});
        } else {
          result.add(new Object[]{null, idx});
        }
      } catch (NumberFormatException x) {
        result.add(new Object[]{segment, null});
      }
    }

    return result;
  }

  private BeanPath member(final String prefix, final String name, final TypeLiteral root,
      final int pcount) {
    Class rawType = root.getRawType();
    BeanPath fn = method(prefix, name, rawType.getDeclaredMethods(), pcount);
    if (fn == null) {
      fn = field(name, rawType.getDeclaredFields());
      // superclass lookup?
      if (fn == null) {
        Class<?> superclass = rawType.getSuperclass();
        if (superclass != Object.class) {
          return member(prefix, name, TypeLiteral.get(rawType.getGenericSuperclass()), pcount);
        }
      }
    }
    return fn;
  }

  private BeanFieldPath field(final String name, final Field[] fields) {
    for (Field f : fields) {
      if (f.getName().equals(name)) {
        return new BeanFieldPath(name, f);
      }
    }
    return null;
  }

  private BeanMethodPath method(final String prefix, final String name, final Method[] methods,
      final int pcount) {
    String bname = javaBeanMethod(new StringBuilder(prefix), name);
    for (Method m : methods) {
      String mname = m.getName();
      if ((bname.equals(mname) || name.equals(mname)) && m.getParameterCount() == pcount) {
        return new BeanMethodPath(name, m);
      }
    }
    return null;
  }

  private String javaBeanMethod(final StringBuilder prefix, final String name) {
    return prefix.append(Character.toUpperCase(name.charAt(0))).append(name, 1, name.length())
        .toString();
  }
}
