/**
 *                                  Apache License
 *                            Version 2.0, January 2004
 *                         http://www.apache.org/licenses/
 *
 *    TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *    1. Definitions.
 *
 *       "License" shall mean the terms and conditions for use, reproduction,
 *       and distribution as defined by Sections 1 through 9 of this document.
 *
 *       "Licensor" shall mean the copyright owner or entity authorized by
 *       the copyright owner that is granting the License.
 *
 *       "Legal Entity" shall mean the union of the acting entity and all
 *       other entities that control, are controlled by, or are under common
 *       control with that entity. For the purposes of this definition,
 *       "control" means (i) the power, direct or indirect, to cause the
 *       direction or management of such entity, whether by contract or
 *       otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *       outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *       "You" (or "Your") shall mean an individual or Legal Entity
 *       exercising permissions granted by this License.
 *
 *       "Source" form shall mean the preferred form for making modifications,
 *       including but not limited to software source code, documentation
 *       source, and configuration files.
 *
 *       "Object" form shall mean any form resulting from mechanical
 *       transformation or translation of a Source form, including but
 *       not limited to compiled object code, generated documentation,
 *       and conversions to other media types.
 *
 *       "Work" shall mean the work of authorship, whether in Source or
 *       Object form, made available under the License, as indicated by a
 *       copyright notice that is included in or attached to the work
 *       (an example is provided in the Appendix below).
 *
 *       "Derivative Works" shall mean any work, whether in Source or Object
 *       form, that is based on (or derived from) the Work and for which the
 *       editorial revisions, annotations, elaborations, or other modifications
 *       represent, as a whole, an original work of authorship. For the purposes
 *       of this License, Derivative Works shall not include works that remain
 *       separable from, or merely link (or bind by name) to the interfaces of,
 *       the Work and Derivative Works thereof.
 *
 *       "Contribution" shall mean any work of authorship, including
 *       the original version of the Work and any modifications or additions
 *       to that Work or Derivative Works thereof, that is intentionally
 *       submitted to Licensor for inclusion in the Work by the copyright owner
 *       or by an individual or Legal Entity authorized to submit on behalf of
 *       the copyright owner. For the purposes of this definition, "submitted"
 *       means any form of electronic, verbal, or written communication sent
 *       to the Licensor or its representatives, including but not limited to
 *       communication on electronic mailing lists, source code control systems,
 *       and issue tracking systems that are managed by, or on behalf of, the
 *       Licensor for the purpose of discussing and improving the Work, but
 *       excluding communication that is conspicuously marked or otherwise
 *       designated in writing by the copyright owner as "Not a Contribution."
 *
 *       "Contributor" shall mean Licensor and any individual or Legal Entity
 *       on behalf of whom a Contribution has been received by Licensor and
 *       subsequently incorporated within the Work.
 *
 *    2. Grant of Copyright License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       copyright license to reproduce, prepare Derivative Works of,
 *       publicly display, publicly perform, sublicense, and distribute the
 *       Work and such Derivative Works in Source or Object form.
 *
 *    3. Grant of Patent License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       (except as stated in this section) patent license to make, have made,
 *       use, offer to sell, sell, import, and otherwise transfer the Work,
 *       where such license applies only to those patent claims licensable
 *       by such Contributor that are necessarily infringed by their
 *       Contribution(s) alone or by combination of their Contribution(s)
 *       with the Work to which such Contribution(s) was submitted. If You
 *       institute patent litigation against any entity (including a
 *       cross-claim or counterclaim in a lawsuit) alleging that the Work
 *       or a Contribution incorporated within the Work constitutes direct
 *       or contributory patent infringement, then any patent licenses
 *       granted to You under this License for that Work shall terminate
 *       as of the date such litigation is filed.
 *
 *    4. Redistribution. You may reproduce and distribute copies of the
 *       Work or Derivative Works thereof in any medium, with or without
 *       modifications, and in Source or Object form, provided that You
 *       meet the following conditions:
 *
 *       (a) You must give any other recipients of the Work or
 *           Derivative Works a copy of this License; and
 *
 *       (b) You must cause any modified files to carry prominent notices
 *           stating that You changed the files; and
 *
 *       (c) You must retain, in the Source form of any Derivative Works
 *           that You distribute, all copyright, patent, trademark, and
 *           attribution notices from the Source form of the Work,
 *           excluding those notices that do not pertain to any part of
 *           the Derivative Works; and
 *
 *       (d) If the Work includes a "NOTICE" text file as part of its
 *           distribution, then any Derivative Works that You distribute must
 *           include a readable copy of the attribution notices contained
 *           within such NOTICE file, excluding those notices that do not
 *           pertain to any part of the Derivative Works, in at least one
 *           of the following places: within a NOTICE text file distributed
 *           as part of the Derivative Works; within the Source form or
 *           documentation, if provided along with the Derivative Works; or,
 *           within a display generated by the Derivative Works, if and
 *           wherever such third-party notices normally appear. The contents
 *           of the NOTICE file are for informational purposes only and
 *           do not modify the License. You may add Your own attribution
 *           notices within Derivative Works that You distribute, alongside
 *           or as an addendum to the NOTICE text from the Work, provided
 *           that such additional attribution notices cannot be construed
 *           as modifying the License.
 *
 *       You may add Your own copyright statement to Your modifications and
 *       may provide additional or different license terms and conditions
 *       for use, reproduction, or distribution of Your modifications, or
 *       for any such Derivative Works as a whole, provided Your use,
 *       reproduction, and distribution of the Work otherwise complies with
 *       the conditions stated in this License.
 *
 *    5. Submission of Contributions. Unless You explicitly state otherwise,
 *       any Contribution intentionally submitted for inclusion in the Work
 *       by You to the Licensor shall be under the terms and conditions of
 *       this License, without any additional terms or conditions.
 *       Notwithstanding the above, nothing herein shall supersede or modify
 *       the terms of any separate license agreement you may have executed
 *       with Licensor regarding such Contributions.
 *
 *    6. Trademarks. This License does not grant permission to use the trade
 *       names, trademarks, service marks, or product names of the Licensor,
 *       except as required for reasonable and customary use in describing the
 *       origin of the Work and reproducing the content of the NOTICE file.
 *
 *    7. Disclaimer of Warranty. Unless required by applicable law or
 *       agreed to in writing, Licensor provides the Work (and each
 *       Contributor provides its Contributions) on an "AS IS" BASIS,
 *       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *       implied, including, without limitation, any warranties or conditions
 *       of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *       PARTICULAR PURPOSE. You are solely responsible for determining the
 *       appropriateness of using or redistributing the Work and assume any
 *       risks associated with Your exercise of permissions under this License.
 *
 *    8. Limitation of Liability. In no event and under no legal theory,
 *       whether in tort (including negligence), contract, or otherwise,
 *       unless required by applicable law (such as deliberate and grossly
 *       negligent acts) or agreed to in writing, shall any Contributor be
 *       liable to You for damages, including any direct, indirect, special,
 *       incidental, or consequential damages of any character arising as a
 *       result of this License or out of the use or inability to use the
 *       Work (including but not limited to damages for loss of goodwill,
 *       work stoppage, computer failure or malfunction, or any and all
 *       other commercial damages or losses), even if such Contributor
 *       has been advised of the possibility of such damages.
 *
 *    9. Accepting Warranty or Additional Liability. While redistributing
 *       the Work or Derivative Works thereof, You may choose to offer,
 *       and charge a fee for, acceptance of support, warranty, indemnity,
 *       or other liability obligations and/or rights consistent with this
 *       License. However, in accepting such obligations, You may act only
 *       on Your own behalf and on Your sole responsibility, not on behalf
 *       of any other Contributor, and only if You agree to indemnify,
 *       defend, and hold each Contributor harmless for any liability
 *       incurred by, or claims asserted against, such Contributor by reason
 *       of your accepting any such warranty or additional liability.
 *
 *    END OF TERMS AND CONDITIONS
 *
 *    APPENDIX: How to apply the Apache License to your work.
 *
 *       To apply the Apache License to your work, attach the following
 *       boilerplate notice, with the fields enclosed by brackets "{}"
 *       replaced with your own identifying information. (Don't include
 *       the brackets!)  The text should be enclosed in the appropriate
 *       comment syntax for the file format. We also recommend that a
 *       file or class name and description of purpose be included on the
 *       same "printed page" as the copyright notice for easier
 *       identification within third-party archives.
 *
 *    Copyright 2014 Edgar Espina
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package org.jooby.assets;

import com.eclipsesource.v8.V8;
import com.eclipsesource.v8.V8Function;
import static com.eclipsesource.v8.utils.V8ObjectUtils.toV8Array;
import static com.eclipsesource.v8.utils.V8ObjectUtils.toV8Object;
import com.google.common.hash.Hasher;
import com.google.common.hash.Hashing;
import com.google.common.io.BaseEncoding;
import com.typesafe.config.Config;
import org.apache.batik.transcoder.TranscoderInput;
import org.apache.batik.transcoder.TranscoderOutput;
import org.apache.batik.transcoder.image.PNGTranscoder;
import org.jooby.funzy.Try;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.OutputStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * <h1>svg-sprites</h1>
 * <p>
 * An {@link AssetAggregator} that creates SVG sprites with PNG fallbacks at needed sizes via
 * <a href="https://github.com/drdk/dr-svg-sprites">dr-svg-sprites</a>.
 * </p>
 *
 * <h2>usage</h2>
 *
 * <pre>
 * assets {
 *   fileset {
 *     sprite: svg-sprites
 *     home: home.scss
 *   }
 *
 *   svg-sprites {
 *     spriteElementPath: "images/svg-source",
 *     spritePath: "css"
 *   }
 * }
 * </pre>
 *
 * <p>
 * The <code>spriteElementPath</code> contains all the <code>*.svg</code> files you want to process.
 * The <code>spritePath</code> indicates where to save the sprite, here you will find the following
 * generated files: <code>css/sprite.css</code>, <code>css/sprite.svg</code> and
 * <code>css/sprite.png</code>.
 * </p>
 *
 * <h2>options</h2>
 *
 * <pre>
 * assets {
 *   fileset {
 *     sprite: svg-sprites
 *     home: home.scss
 *   }
 *
 *   svg-sprites {
 *     spriteElementPath: "images/svg-source",
 *     spritePath: "css",
 *     layout: "vertical",
 *     sizes: {
 *       large: 24,
 *       small: 16
 *     },
 *     refSize: "large"
 *   }
 * }
 * </pre>
 *
 * <p>
 * Please refer to <a href="https://github.com/drdk/dr-svg-sprites">dr-svg-sprites</a> for more
 * details.
 * </p>
 *
 * @author edgar
 *
 */
public class SvgSprites extends AssetAggregator {

  /** The logging system. */
  private final Logger log = LoggerFactory.getLogger(getClass());

  @Override
  public SvgSprites set(final Config options) {
    super.set(options);
    return this;
  }

  @Override
  public SvgSprites set(final String name, final Object value) {
    super.set(name, value);
    return this;
  }

  @Override
  public List<String> fileset() {
    return Arrays.asList(cssPath());
  }

  @Override
  public void run(final Config conf) throws Exception {
    File spriteElementPath = resolve(get("spriteElementPath").toString());
    if (!spriteElementPath.exists()) {
      throw new FileNotFoundException(spriteElementPath.toString());
    }
    // TODO: sanitization of arguments
    File workdir = new File(Try.apply(() -> conf.getString("application.tmpdir"))
        .orElse(System.getProperty("java.io.tmpdir")));

    File spritePath = resolve(spritePath());
    File cssPath = resolve(cssPath());
    String sha1 = new File(spritePath()).getName()
        .replace(".svg", "-" + sha1(spriteElementPath, spritePath, cssPath) + ".sha1");
    File uptodate = workdir.toPath().resolve("svg-sprites").resolve(sha1).toFile();

    if (uptodate.exists()) {
      log.info("svg-sprites is up-to-date: {}", uptodate);
      return;
    }

    Nodejs.run(workdir, node -> {
      node.overwrite(conf.hasPath("_overwrite") ? conf.getBoolean("_overwrite") : false)
          .exec("dr-svg-sprites", v8 -> {
            Map<String, Object> options = options();
            // rewrite paths
            options.put("spritePath", spritePath.toString());
            options.put("cssPath", cssPath.toString());
            options.put("spriteElementPath", spriteElementPath.toString());

            log.debug("svg-sprites options {}  ", options.entrySet().stream()
                .map(e -> e.getKey() + ": " + e.getValue())
                .collect(Collectors.joining("\n  ", "{\n  ", "\n}")));

            v8.add("$options", toV8Object(v8, options));

            /**
             * Hook sv2png and remove panthomjs dependency.
             */
            v8.add("svg2png", new V8Function(v8, (receiver, params) -> {
              String svgPath = params.get(0).toString();
              String pngPath = params.get(1).toString();
              Float w = new Float(params.getDouble(2));
              Float h = new Float(params.getDouble(3));
              V8Function callback = (V8Function) params.get(4);
              Try.run(() -> {
                try (FileReader in = new FileReader(svgPath);
                    OutputStream out = new FileOutputStream(pngPath)) {
                  PNGTranscoder transcoder = new PNGTranscoder();
                  transcoder.addTranscodingHint(PNGTranscoder.KEY_WIDTH, w);
                  transcoder.addTranscodingHint(PNGTranscoder.KEY_HEIGHT, h);
                  transcoder.transcode(new TranscoderInput(in), new TranscoderOutput(out));
                }
              })
                  .onSuccess(() -> callback.call(null, null))
                  .onFailure(x -> {
                    log.debug("png-fallback resulted in exception", x);
                    callback.call(null, toV8Array(v8, Arrays.asList(x.getMessage())));
                  });
              return V8.UNDEFINED;
            }));
          });
    });
    log.debug("creating sha1: {}", uptodate);
    uptodate.getParentFile().mkdirs();
    // clean old/previous *.sha1 files
    try (Stream<Path> sha1files = Files.walk(uptodate.getParentFile().toPath())
        .filter(it -> it.toString().endsWith(".sha1"))) {
      sha1files.forEach(it -> Try.run(() -> Files.delete(it)));
    }
    Files.createFile(uptodate.toPath());
    uptodate.deleteOnExit();
  }

  private String sha1(final File dir, final File sprite, final File css) throws IOException {
    try (Stream<Path> stream = Files.walk(dir.toPath())) {
      Hasher sha1 = Hashing.sha1().newHasher();
      stream.filter(p -> !Files.isDirectory(p))
          .forEach(p -> Try.run(() -> sha1.putBytes(Files.readAllBytes(p))));
      if (sprite.exists()) {
        sha1.putBytes(Files.readAllBytes(sprite.toPath()));
      }
      if (css.exists()) {
        sha1.putBytes(Files.readAllBytes(css.toPath()));
      }
      return BaseEncoding.base16().encode(sha1.hash().asBytes()).toLowerCase();
    }
  }

  private File resolve(final String path) {
    // basedir is set to public from super class
    return new File(get("basedir").toString(), path);
  }

  public String cssPath() {
    try {
      return nameFor("cssPath", ".css");
    } catch (IllegalArgumentException x) {
      return spritePath().replace(".svg", ".css");
    }
  }

  public String spritePath() {
    return nameFor("spritePath", ".svg");
  }

  private String nameFor(final String property, final String ext) {
    String spritePath = get(property);
    if (spritePath == null) {
      throw new IllegalArgumentException(
          "Required option 'svg-sprites." + property + "' not present");
    }
    if (spritePath.endsWith(ext)) {
      return spritePath;
    } else {
      return spritePath + "/" + prefix("prefix") + prefix("name") + "sprite" + ext;
    }
  }

  private String prefix(final String name) {
    return Optional.ofNullable(get(name))
        .map(it -> it + "-")
        .orElse("");
  }

}
