/**
 * Apache License
 * Version 2.0, January 2004
 * http://www.apache.org/licenses/
 * <p>
 * TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 * <p>
 * 1. Definitions.
 * <p>
 * "License" shall mean the terms and conditions for use, reproduction,
 * and distribution as defined by Sections 1 through 9 of this document.
 * <p>
 * "Licensor" shall mean the copyright owner or entity authorized by
 * the copyright owner that is granting the License.
 * <p>
 * "Legal Entity" shall mean the union of the acting entity and all
 * other entities that control, are controlled by, or are under common
 * control with that entity. For the purposes of this definition,
 * "control" means (i) the power, direct or indirect, to cause the
 * direction or management of such entity, whether by contract or
 * otherwise, or (ii) ownership of fifty percent (50%) or more of the
 * outstanding shares, or (iii) beneficial ownership of such entity.
 * <p>
 * "You" (or "Your") shall mean an individual or Legal Entity
 * exercising permissions granted by this License.
 * <p>
 * "Source" form shall mean the preferred form for making modifications,
 * including but not limited to software source code, documentation
 * source, and configuration files.
 * <p>
 * "Object" form shall mean any form resulting from mechanical
 * transformation or translation of a Source form, including but
 * not limited to compiled object code, generated documentation,
 * and conversions to other media types.
 * <p>
 * "Work" shall mean the work of authorship, whether in Source or
 * Object form, made available under the License, as indicated by a
 * copyright notice that is included in or attached to the work
 * (an example is provided in the Appendix below).
 * <p>
 * "Derivative Works" shall mean any work, whether in Source or Object
 * form, that is based on (or derived from) the Work and for which the
 * editorial revisions, annotations, elaborations, or other modifications
 * represent, as a whole, an original work of authorship. For the purposes
 * of this License, Derivative Works shall not include works that remain
 * separable from, or merely link (or bind by name) to the interfaces of,
 * the Work and Derivative Works thereof.
 * <p>
 * "Contribution" shall mean any work of authorship, including
 * the original version of the Work and any modifications or additions
 * to that Work or Derivative Works thereof, that is intentionally
 * submitted to Licensor for inclusion in the Work by the copyright owner
 * or by an individual or Legal Entity authorized to submit on behalf of
 * the copyright owner. For the purposes of this definition, "submitted"
 * means any form of electronic, verbal, or written communication sent
 * to the Licensor or its representatives, including but not limited to
 * communication on electronic mailing lists, source code control systems,
 * and issue tracking systems that are managed by, or on behalf of, the
 * Licensor for the purpose of discussing and improving the Work, but
 * excluding communication that is conspicuously marked or otherwise
 * designated in writing by the copyright owner as "Not a Contribution."
 * <p>
 * "Contributor" shall mean Licensor and any individual or Legal Entity
 * on behalf of whom a Contribution has been received by Licensor and
 * subsequently incorporated within the Work.
 * <p>
 * 2. Grant of Copyright License. Subject to the terms and conditions of
 * this License, each Contributor hereby grants to You a perpetual,
 * worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 * copyright license to reproduce, prepare Derivative Works of,
 * publicly display, publicly perform, sublicense, and distribute the
 * Work and such Derivative Works in Source or Object form.
 * <p>
 * 3. Grant of Patent License. Subject to the terms and conditions of
 * this License, each Contributor hereby grants to You a perpetual,
 * worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 * (except as stated in this section) patent license to make, have made,
 * use, offer to sell, sell, import, and otherwise transfer the Work,
 * where such license applies only to those patent claims licensable
 * by such Contributor that are necessarily infringed by their
 * Contribution(s) alone or by combination of their Contribution(s)
 * with the Work to which such Contribution(s) was submitted. If You
 * institute patent litigation against any entity (including a
 * cross-claim or counterclaim in a lawsuit) alleging that the Work
 * or a Contribution incorporated within the Work constitutes direct
 * or contributory patent infringement, then any patent licenses
 * granted to You under this License for that Work shall terminate
 * as of the date such litigation is filed.
 * <p>
 * 4. Redistribution. You may reproduce and distribute copies of the
 * Work or Derivative Works thereof in any medium, with or without
 * modifications, and in Source or Object form, provided that You
 * meet the following conditions:
 * <p>
 * (a) You must give any other recipients of the Work or
 * Derivative Works a copy of this License; and
 * <p>
 * (b) You must cause any modified files to carry prominent notices
 * stating that You changed the files; and
 * <p>
 * (c) You must retain, in the Source form of any Derivative Works
 * that You distribute, all copyright, patent, trademark, and
 * attribution notices from the Source form of the Work,
 * excluding those notices that do not pertain to any part of
 * the Derivative Works; and
 * <p>
 * (d) If the Work includes a "NOTICE" text file as part of its
 * distribution, then any Derivative Works that You distribute must
 * include a readable copy of the attribution notices contained
 * within such NOTICE file, excluding those notices that do not
 * pertain to any part of the Derivative Works, in at least one
 * of the following places: within a NOTICE text file distributed
 * as part of the Derivative Works; within the Source form or
 * documentation, if provided along with the Derivative Works; or,
 * within a display generated by the Derivative Works, if and
 * wherever such third-party notices normally appear. The contents
 * of the NOTICE file are for informational purposes only and
 * do not modify the License. You may add Your own attribution
 * notices within Derivative Works that You distribute, alongside
 * or as an addendum to the NOTICE text from the Work, provided
 * that such additional attribution notices cannot be construed
 * as modifying the License.
 * <p>
 * You may add Your own copyright statement to Your modifications and
 * may provide additional or different license terms and conditions
 * for use, reproduction, or distribution of Your modifications, or
 * for any such Derivative Works as a whole, provided Your use,
 * reproduction, and distribution of the Work otherwise complies with
 * the conditions stated in this License.
 * <p>
 * 5. Submission of Contributions. Unless You explicitly state otherwise,
 * any Contribution intentionally submitted for inclusion in the Work
 * by You to the Licensor shall be under the terms and conditions of
 * this License, without any additional terms or conditions.
 * Notwithstanding the above, nothing herein shall supersede or modify
 * the terms of any separate license agreement you may have executed
 * with Licensor regarding such Contributions.
 * <p>
 * 6. Trademarks. This License does not grant permission to use the trade
 * names, trademarks, service marks, or product names of the Licensor,
 * except as required for reasonable and customary use in describing the
 * origin of the Work and reproducing the content of the NOTICE file.
 * <p>
 * 7. Disclaimer of Warranty. Unless required by applicable law or
 * agreed to in writing, Licensor provides the Work (and each
 * Contributor provides its Contributions) on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 * implied, including, without limitation, any warranties or conditions
 * of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 * PARTICULAR PURPOSE. You are solely responsible for determining the
 * appropriateness of using or redistributing the Work and assume any
 * risks associated with Your exercise of permissions under this License.
 * <p>
 * 8. Limitation of Liability. In no event and under no legal theory,
 * whether in tort (including negligence), contract, or otherwise,
 * unless required by applicable law (such as deliberate and grossly
 * negligent acts) or agreed to in writing, shall any Contributor be
 * liable to You for damages, including any direct, indirect, special,
 * incidental, or consequential damages of any character arising as a
 * result of this License or out of the use or inability to use the
 * Work (including but not limited to damages for loss of goodwill,
 * work stoppage, computer failure or malfunction, or any and all
 * other commercial damages or losses), even if such Contributor
 * has been advised of the possibility of such damages.
 * <p>
 * 9. Accepting Warranty or Additional Liability. While redistributing
 * the Work or Derivative Works thereof, You may choose to offer,
 * and charge a fee for, acceptance of support, warranty, indemnity,
 * or other liability obligations and/or rights consistent with this
 * License. However, in accepting such obligations, You may act only
 * on Your own behalf and on Your sole responsibility, not on behalf
 * of any other Contributor, and only if You agree to indemnify,
 * defend, and hold each Contributor harmless for any liability
 * incurred by, or claims asserted against, such Contributor by reason
 * of your accepting any such warranty or additional liability.
 * <p>
 * END OF TERMS AND CONDITIONS
 * <p>
 * APPENDIX: How to apply the Apache License to your work.
 * <p>
 * To apply the Apache License to your work, attach the following
 * boilerplate notice, with the fields enclosed by brackets "{}"
 * replaced with your own identifying information. (Don't include
 * the brackets!)  The text should be enclosed in the appropriate
 * comment syntax for the file format. We also recommend that a
 * file or class name and description of purpose be included on the
 * same "printed page" as the copyright notice for easier
 * identification within third-party archives.
 * <p>
 * Copyright 2014 Edgar Espina
 * <p>
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jooby.assets;

import com.google.common.base.Strings;
import com.google.common.collect.ImmutableList;
import com.typesafe.config.Config;
import org.jooby.funzy.Throwing;
import org.jsoup.Jsoup;
import org.jsoup.nodes.Element;
import org.jsoup.parser.Tag;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Stream;

/**
 * <h1>svg-symbol</h1>
 * <p>
 * SVG <code>symbol</code> for icons: merge svg files from a folder and generates a
 * <code>sprite.svg</code> and <code>sprite.css</code> files.
 * </p>
 *
 * <h2>usage</h2>
 *
 * <pre>{@code
 * assets {
 *   fileset {
 *     sprite: svg-symbol
 *   }
 *
 *   svg-symbol {
 *     input: "images/svg"
 *   }
 * }
 * }</pre>
 *
 * <p>
 * Previous example looks for <code>*.svg</code> files inside the <code>images/svg</code> folder and
 * generate a <code>sprite.svg</code> and <code>sprite.css</code> files.
 * </p>
 *
 * <p>
 * You can display the svg icons using id reference:
 * </p>
 *
 * <pre>{@code
 * <svg>
 *   <use xlink:href="#approved" />
 * </svg>
 * }</pre>
 *
 * <p>
 * This technique is described here:
 * <a href="https://css-tricks.com/svg-symbol-good-choice-icons">SVG symbol a Good Choice for
 * Icons</a>
 * </p>
 *
 * <h2>options</h2>
 *
 * <h3>output</h3>
 * <p>
 * Defines where to write the <code>svg</code> and <code>css</code> files. Default value is:
 * <code>sprite</code>.
 * </p>
 *
 * <pre>{@code
 * svg-symbol {
 *   output: "folder/symbols"
 * }
 * }</pre>
 *
 * <p>
 * There are two more specific output options: <code>svg.output</code> and <code>css.output</code>
 * if any of these options are present the <code>output</code> option is ignored:
 * </p>
 *
 * <pre>{@code
 * svg-symbol {
 *   css {
 *     output: "css/sprite.css"
 *   },
 *   svg {
 *     output: "img/sprite.svg"
 *   }
 * }
 * }</pre>
 *
 * <h3>id prefix and suffix</h3>
 * <p>
 * ID is generated from <code>svg file names</code>. These options prepend or append something to
 * the generated id.
 * </p>
 *
 * <pre>{@code
 * svg-symbol {
 *   output: "sprite"
 *   id {
 *     prefix: "icon-"
 *   }
 * }
 * }</pre>
 *
 * <p>
 * Generates IDs like: <code>icon-approved</code>, while:
 * </p>
 *
 * <pre>{@code
 * svg-symbol {
 *   output: "sprite"
 *   id {
 *     suffix: "-icon"
 *   }
 * }
 * }</pre>
 *
 * <p>
 * Generates IDs like: <code>approved-icon</code>
 * </p>
 *
 * <h3>css prefix</h3>
 * <p>
 * Prepend a string to a generated css class. Here is the css class for <code>approved.svg</code>:
 * </p>
 *
 * <pre>{@code
 * .approved {
 *   width: 18px;
 *   height: 18px;
 * }
 * }</pre>
 *
 * <p>
 * If we set a <code>svg</code> css prefix:
 *
 * <pre>{@code
 * {
 *   svg-symbol: {
 *     css {
 *       prefix: "svg"
 *     }
 *   }
 * }
 * }</pre>
 *
 * <p>
 * The generated css class will be:
 * <pre>{@code
 * svg.approved {
 *   width: 18px;
 *   height: 18px;
 * }
 * }</pre>
 *
 * <p>
 * This option is useful for generating more specific css class selectors.
 * </p>
 *
 * @author edgar
 */
public class SvgSymbol extends AssetAggregator {

  public static class Tuple<T1, T2> {
    public final T1 _1;
    public final T2 _2;

    public Tuple(final T1 _1, final T2 _2) {
      this._1 = _1;
      this._2 = _2;
    }
  }

  /**
   * #3.
   */
  private static final int _3 = 3;

  /**
   * Handle svg dimension: 0 0 20 20.
   */
  static final Pattern SIZE = Pattern.compile("(\\d+(\\.\\d+)?)(\\w+)");

  /**
   * The logging system.
   */
  private final Logger log = LoggerFactory.getLogger(getClass());

  /**
   * Creates a new {@link SvgSymbol}.
   */
  public SvgSymbol() {
    set("output", "sprite");

    set("id.prefix", "");
    set("id.suffix", "");

    set("css.prefix", "");
    set("css.round", true);

    set("svg.xmlns", "http://www.w3.org/2000/svg");
  }

  @Override
  public List<String> fileset() {
    return ImmutableList.of(cssPath());
  }

  @Override
  public void run(final Config conf) throws Exception {
    String input = get("input");
    if (input == null) {
      throw new IllegalArgumentException("Required option 'svg-symbol.input' not present");
    }
    Path basedir = Paths.get(get("basedir").toString());
    Path dir = basedir.resolve(input);

    List<CharSequence> cssout = new ArrayList<>();

    Element svg = new Element(Tag.valueOf("svg"), "");
    attrs("svg", "output").forEach((n, v) -> svg.attr(n, v.toString()));

    files(dir, file -> {
      log.debug("{}", file);
      String id = get("id.prefix") + file.getFileName().toString().replace(".svg", "")
          + get("id.suffix");

      Tuple<Element, Element> rewrite = symbol(file, id);
      svg.appendChild(rewrite._2);

      cssout.add(css(id, rewrite._1));
    });

    write(basedir.resolve(svgPath()), ImmutableList.of(svg.outerHtml()));
    write(basedir.resolve(cssPath()), cssout);
  }

  private void files(Path dir, Throwing.Consumer<Path> consumer) throws Exception {
    try (Stream<Path> stream = Files.walk(dir)) {
      Iterator<Path> files = stream.filter(Files::isRegularFile)
          .filter(it -> it.toString().endsWith(".svg"))
          .sorted()
          .iterator();
      while (files.hasNext()) {
        consumer.accept(files.next());
      }
    }
  }

  /**
   * Read an object path and optionally filter some child paths.
   *
   * @param path    Path to read.
   * @param without Properties to filter.
   * @return Properties.
   */
  private Map<String, Object> attrs(final String path, final String... without) {
    Map<String, Object> attrs = new LinkedHashMap<>(get(path));
    Arrays.asList(without).forEach(attrs::remove);
    return attrs;
  }

  /**
   * Read a svg file and return the svg element (original) and a new symbol element (created from
   * original).
   *
   * @param file Svg file.
   * @param id   ID to use.
   * @return Svg element (original) and a symbol element (converted).
   * @throws IOException If something goes wrong.
   */
  private Tuple<Element, Element> symbol(final Path file, final String id) throws IOException {
    Element svg = Jsoup.parse(file.toFile(), "UTF-8").select("svg").first();
    Element symbol = new Element(Tag.valueOf("symbol"), "")
        .attr("id", id)
        .attr("viewBox", svg.attr("viewBox"));
    new ArrayList<>(svg.childNodes()).forEach(symbol::appendChild);
    return new Tuple(svg, symbol);
  }

  /**
   * Generate a CSS rule, it reads the width and height attributes of the svg element or fallback to
   * viewBox attribute.
   *
   * @param id  ID to use.
   * @param svg Svg element to convert.
   * @return A css rule.
   */
  private CharSequence css(final String id, final Element svg) {
    Throwing.Function<String, Tuple<Tuple<Number, String>, Tuple<Number, String>>> viewBox = Throwing
        .<String, Tuple<Tuple<Number, String>, Tuple<Number, String>>>throwingFunction(name -> {
          String vbox = svg.attr(name);
          String[] dimension = vbox.split("\\s+");
          return new Tuple(parse(dimension[2]), parse(dimension[_3]));
        }).memoized();
    Tuple<Number, String> w = Optional.ofNullable(Strings.emptyToNull(svg.attr("width")))
        .map(this::parse)
        .orElseGet(() -> viewBox.apply("viewBox")._1);
    Tuple<Number, String> h = Optional.ofNullable(Strings.emptyToNull(svg.attr("height")))
        .map(this::parse)
        .orElseGet(() -> viewBox.apply("viewBox")._2);
    StringBuilder css = new StringBuilder();
    css.append(get("css.prefix").toString()).append(".").append(id)
        .append(" {\n  width: ").append(w._1).append(w._2).append(";\n")
        .append("  height: ").append(h._1).append(h._2).append(";\n}");
    return css;
  }

  /**
   * Parse a css size unit value, like 10px or 18.919px and optionally round the value to the
   * closest integer.
   *
   * @param value Value to parse.
   * @return A tuple with a number and unit(px, em, etc...)
   */
  private Tuple<Number, String> parse(final String value) {
    Matcher matcher = SIZE.matcher(value);
    if (matcher.find()) {
      String number = matcher.group(1);
      String unit = matcher.group(_3);
      boolean round = get("css.round");
      Number num = Double.parseDouble(number);
      return new Tuple(round ? Math.round(num.doubleValue()) : num, unit);
    }
    return null;
  }

  /**
   * Write content to file.
   *
   * @param path     Target file.
   * @param sequence File content.
   * @throws IOException If something goes wrong.
   */
  private void write(final Path path, final List<CharSequence> sequence) throws IOException {
    log.debug("writing: {}", path.normalize().toAbsolutePath());
    path.toFile().getParentFile().mkdirs();
    Files.write(path, sequence);
  }

  /**
   * @return Generate a css path from <code>css.output</code> or fallback to <code>output</code>
   */
  private String cssPath() {
    String css = Optional.<String>ofNullable(get("css.output")).orElse(get("output"));
    return css.endsWith(".css") ? css : css + ".css";
  }

  /**
   * @return Generate a css path from <code>svg.output</code> or fallback to <code>output</code>
   */
  private String svgPath() {
    String svg = Optional.<String>ofNullable(get("svg.output")).orElse(get("output"));
    return svg.endsWith(".svg") ? svg : svg + ".svg";
  }

}
