/**
 *                                  Apache License
 *                            Version 2.0, January 2004
 *                         http://www.apache.org/licenses/
 *
 *    TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *    1. Definitions.
 *
 *       "License" shall mean the terms and conditions for use, reproduction,
 *       and distribution as defined by Sections 1 through 9 of this document.
 *
 *       "Licensor" shall mean the copyright owner or entity authorized by
 *       the copyright owner that is granting the License.
 *
 *       "Legal Entity" shall mean the union of the acting entity and all
 *       other entities that control, are controlled by, or are under common
 *       control with that entity. For the purposes of this definition,
 *       "control" means (i) the power, direct or indirect, to cause the
 *       direction or management of such entity, whether by contract or
 *       otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *       outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *       "You" (or "Your") shall mean an individual or Legal Entity
 *       exercising permissions granted by this License.
 *
 *       "Source" form shall mean the preferred form for making modifications,
 *       including but not limited to software source code, documentation
 *       source, and configuration files.
 *
 *       "Object" form shall mean any form resulting from mechanical
 *       transformation or translation of a Source form, including but
 *       not limited to compiled object code, generated documentation,
 *       and conversions to other media types.
 *
 *       "Work" shall mean the work of authorship, whether in Source or
 *       Object form, made available under the License, as indicated by a
 *       copyright notice that is included in or attached to the work
 *       (an example is provided in the Appendix below).
 *
 *       "Derivative Works" shall mean any work, whether in Source or Object
 *       form, that is based on (or derived from) the Work and for which the
 *       editorial revisions, annotations, elaborations, or other modifications
 *       represent, as a whole, an original work of authorship. For the purposes
 *       of this License, Derivative Works shall not include works that remain
 *       separable from, or merely link (or bind by name) to the interfaces of,
 *       the Work and Derivative Works thereof.
 *
 *       "Contribution" shall mean any work of authorship, including
 *       the original version of the Work and any modifications or additions
 *       to that Work or Derivative Works thereof, that is intentionally
 *       submitted to Licensor for inclusion in the Work by the copyright owner
 *       or by an individual or Legal Entity authorized to submit on behalf of
 *       the copyright owner. For the purposes of this definition, "submitted"
 *       means any form of electronic, verbal, or written communication sent
 *       to the Licensor or its representatives, including but not limited to
 *       communication on electronic mailing lists, source code control systems,
 *       and issue tracking systems that are managed by, or on behalf of, the
 *       Licensor for the purpose of discussing and improving the Work, but
 *       excluding communication that is conspicuously marked or otherwise
 *       designated in writing by the copyright owner as "Not a Contribution."
 *
 *       "Contributor" shall mean Licensor and any individual or Legal Entity
 *       on behalf of whom a Contribution has been received by Licensor and
 *       subsequently incorporated within the Work.
 *
 *    2. Grant of Copyright License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       copyright license to reproduce, prepare Derivative Works of,
 *       publicly display, publicly perform, sublicense, and distribute the
 *       Work and such Derivative Works in Source or Object form.
 *
 *    3. Grant of Patent License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       (except as stated in this section) patent license to make, have made,
 *       use, offer to sell, sell, import, and otherwise transfer the Work,
 *       where such license applies only to those patent claims licensable
 *       by such Contributor that are necessarily infringed by their
 *       Contribution(s) alone or by combination of their Contribution(s)
 *       with the Work to which such Contribution(s) was submitted. If You
 *       institute patent litigation against any entity (including a
 *       cross-claim or counterclaim in a lawsuit) alleging that the Work
 *       or a Contribution incorporated within the Work constitutes direct
 *       or contributory patent infringement, then any patent licenses
 *       granted to You under this License for that Work shall terminate
 *       as of the date such litigation is filed.
 *
 *    4. Redistribution. You may reproduce and distribute copies of the
 *       Work or Derivative Works thereof in any medium, with or without
 *       modifications, and in Source or Object form, provided that You
 *       meet the following conditions:
 *
 *       (a) You must give any other recipients of the Work or
 *           Derivative Works a copy of this License; and
 *
 *       (b) You must cause any modified files to carry prominent notices
 *           stating that You changed the files; and
 *
 *       (c) You must retain, in the Source form of any Derivative Works
 *           that You distribute, all copyright, patent, trademark, and
 *           attribution notices from the Source form of the Work,
 *           excluding those notices that do not pertain to any part of
 *           the Derivative Works; and
 *
 *       (d) If the Work includes a "NOTICE" text file as part of its
 *           distribution, then any Derivative Works that You distribute must
 *           include a readable copy of the attribution notices contained
 *           within such NOTICE file, excluding those notices that do not
 *           pertain to any part of the Derivative Works, in at least one
 *           of the following places: within a NOTICE text file distributed
 *           as part of the Derivative Works; within the Source form or
 *           documentation, if provided along with the Derivative Works; or,
 *           within a display generated by the Derivative Works, if and
 *           wherever such third-party notices normally appear. The contents
 *           of the NOTICE file are for informational purposes only and
 *           do not modify the License. You may add Your own attribution
 *           notices within Derivative Works that You distribute, alongside
 *           or as an addendum to the NOTICE text from the Work, provided
 *           that such additional attribution notices cannot be construed
 *           as modifying the License.
 *
 *       You may add Your own copyright statement to Your modifications and
 *       may provide additional or different license terms and conditions
 *       for use, reproduction, or distribution of Your modifications, or
 *       for any such Derivative Works as a whole, provided Your use,
 *       reproduction, and distribution of the Work otherwise complies with
 *       the conditions stated in this License.
 *
 *    5. Submission of Contributions. Unless You explicitly state otherwise,
 *       any Contribution intentionally submitted for inclusion in the Work
 *       by You to the Licensor shall be under the terms and conditions of
 *       this License, without any additional terms or conditions.
 *       Notwithstanding the above, nothing herein shall supersede or modify
 *       the terms of any separate license agreement you may have executed
 *       with Licensor regarding such Contributions.
 *
 *    6. Trademarks. This License does not grant permission to use the trade
 *       names, trademarks, service marks, or product names of the Licensor,
 *       except as required for reasonable and customary use in describing the
 *       origin of the Work and reproducing the content of the NOTICE file.
 *
 *    7. Disclaimer of Warranty. Unless required by applicable law or
 *       agreed to in writing, Licensor provides the Work (and each
 *       Contributor provides its Contributions) on an "AS IS" BASIS,
 *       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *       implied, including, without limitation, any warranties or conditions
 *       of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *       PARTICULAR PURPOSE. You are solely responsible for determining the
 *       appropriateness of using or redistributing the Work and assume any
 *       risks associated with Your exercise of permissions under this License.
 *
 *    8. Limitation of Liability. In no event and under no legal theory,
 *       whether in tort (including negligence), contract, or otherwise,
 *       unless required by applicable law (such as deliberate and grossly
 *       negligent acts) or agreed to in writing, shall any Contributor be
 *       liable to You for damages, including any direct, indirect, special,
 *       incidental, or consequential damages of any character arising as a
 *       result of this License or out of the use or inability to use the
 *       Work (including but not limited to damages for loss of goodwill,
 *       work stoppage, computer failure or malfunction, or any and all
 *       other commercial damages or losses), even if such Contributor
 *       has been advised of the possibility of such damages.
 *
 *    9. Accepting Warranty or Additional Liability. While redistributing
 *       the Work or Derivative Works thereof, You may choose to offer,
 *       and charge a fee for, acceptance of support, warranty, indemnity,
 *       or other liability obligations and/or rights consistent with this
 *       License. However, in accepting such obligations, You may act only
 *       on Your own behalf and on Your sole responsibility, not on behalf
 *       of any other Contributor, and only if You agree to indemnify,
 *       defend, and hold each Contributor harmless for any liability
 *       incurred by, or claims asserted against, such Contributor by reason
 *       of your accepting any such warranty or additional liability.
 *
 *    END OF TERMS AND CONDITIONS
 *
 *    APPENDIX: How to apply the Apache License to your work.
 *
 *       To apply the Apache License to your work, attach the following
 *       boilerplate notice, with the fields enclosed by brackets "{}"
 *       replaced with your own identifying information. (Don't include
 *       the brackets!)  The text should be enclosed in the appropriate
 *       comment syntax for the file format. We also recommend that a
 *       file or class name and description of purpose be included on the
 *       same "printed page" as the copyright notice for easier
 *       identification within third-party archives.
 *
 *    Copyright 2014 Edgar Espina
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package org.jooby.camel;

import com.google.common.base.CaseFormat;
import com.google.common.collect.Lists;
import com.google.inject.Binder;
import com.google.inject.multibindings.Multibinder;
import com.google.inject.name.Names;
import com.typesafe.config.Config;
import com.typesafe.config.ConfigFactory;
import static java.util.Objects.requireNonNull;
import org.apache.camel.CamelContext;
import org.apache.camel.ConsumerTemplate;
import org.apache.camel.FluentProducerTemplate;
import org.apache.camel.ProducerTemplate;
import org.apache.camel.builder.RouteBuilder;
import org.apache.camel.component.properties.PropertiesComponent;
import org.apache.camel.impl.DefaultCamelContext;
import org.apache.camel.spi.ShutdownStrategy;
import org.apache.camel.spi.StreamCachingStrategy;
import org.apache.camel.spi.ThreadPoolProfile;
import org.jooby.Env;
import org.jooby.Jooby;
import org.jooby.internal.camel.CamelFinalizer;
import org.jooby.funzy.Try;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.Properties;

/**
 * Camel for Jooby. Exposes a {@link CamelContext}, {@link ProducerTemplate} and
 * {@link ConsumerTemplate}.
 *
 * <p>
 * NOTE: This module was designed to provide a better integration with Jooby. This module doesn't
 * depend on <a href="http://camel.apache.org/guice.html">camel-guice</a>, but it provides similar
 * features.
 * </p>
 *
 * <h1>usage</h1>
 *
 * <pre>
 * {
 *   use(new Camel()
 *     .routes((rb, config) {@literal ->} {
 *       rb.from("direct:noop").to("mock:out");
 *     })
 *   );
 *
 *   get("/noop", req {@literal ->} {
 *     req.require(ProducerTemplate.class).sendBody("direct:noop", "NOOP");
 *     return "/noop";
 *   });
 *
 * }
 * </pre>
 *
 * <p>
 * Previous example, add a direct route using the Java DSL. A route builder can be created and
 * injected by Guice, see next section.
 * </p>
 *
 * <h1>camel routes</h1>
 * <pre>
 * public class MyRoutes extends RouteBuilder {
 *
 *   &#64;Inject
 *   public MyRoutes(Service service) {
 *     this.service = service;
 *   }
 *
 *   public void configure() {
 *     from("direct:noop").to("mock:out").bean(service);
 *   }
 * }
 *
 * ...
 * {
 *   use(new Camel().routes(MyRoutes.class));
 * }
 * </pre>
 *
 * <p>
 * or without extending RouteBuilder:
 * </p>
 *
 * <pre>
 * public class MyRoutes {
 *
 *   &#64;Inject
 *   public MyRoutes(RouteBuilder router, Service service) {
 *     router.from("direct:noop").to("mock:out").bean(service);
 *   }
 *
 * }
 *
 * ...
 * {
 *   use(new Camel().routes(MyRoutes.class));
 * }
 * </pre>
 *
 * <h1>configuration</h1>
 * <p>
 * Custom configuration is achieved in two ways:
 * </p>
 *
 * <h2>application.conf</h2>
 * <p>
 * A {@link CamelContext} can be configured from your <code>application.conf</code>:
 * </p>
 *
 * <pre>
 * camel.handleFault = false
 *
 * camel.shutdownRoute = Default
 *
 * camel.shutdownRunningTask = CompleteCurrentTaskOnly
 *
 * camel.streamCaching.enabled = false
 *
 * camel.tracing = false
 *
 * camel.autoStartup = true
 *
 * camel.allowUseOriginalMessage = false
 *
 * camel.jmx = false
 * </pre>
 *
 * <p>
 * Same for {@link ShutdownStrategy}:
 * </p>
 *
 * <pre>
 * camel.shutdown.shutdownRoutesInReverseOrder = true
 *
 * camel.shutdown.timeUnit = SECONDS
 *
 * camel.shutdown.timeout = 10
 * </pre>
 *
 * <p>
 * {@link ThreadPoolProfile}:
 * </p>
 *
 * <pre>
 * camel.threads.poolSize = ${runtime.processors-plus1}
 * camel.threads.maxPoolSize = ${runtime.processors-x2}
 * camel.threads.keepAliveTime = 60
 * camel.threads.timeUnit = SECONDS
 * camel.threads.rejectedPolicy = CallerRuns
 * camel.threads.maxQueueSize = 1000
 * camel.threads.id = default
 * camel.threads.defaultProfile = true
 * </pre>
 *
 * And {@link StreamCachingStrategy}:
 *
 * <pre>
 * camel.streamCaching.enabled = false
 * camel.streamCaching.spoolDirectory = ${application.tmpdir}${file.separator}"camel"${file.separator}"#uuid#"
 * </pre>
 *
 * <h2>programmatically</h2>
 * <p>
 * Using the {@link #doWith(Configurer)} method:
 * </p>
 *
 * <pre>
 * {
 *   use(new Camel().doWith((ctx, config) {@literal ->} {
 *     // set/override any other property.
 *   }));
 * }
 * </pre>
 * <p>
 * That's all folks! Enjoy it!!!
 * </p>
 *
 * @author edgar
 * @since 0.5.0
 */
public class Camel implements Jooby.Module {

  public interface Configurer<T> {

    void configure(T ctx, Config config) throws Exception;

  }

  /** The logging system. */
  private final Logger log = LoggerFactory.getLogger(getClass());

  private Configurer<CamelContext> configurer;

  private Configurer<RouteBuilder> routes;

  private List<Class<?>> routeList = new ArrayList<>();

  /**
   * Add a route builder (or alike) which will be injected by Guice and call it before starting the
   * {@link CamelContext}.
   *
   * <pre>
   * public class MyRoutes extends RouteBuilder {
   *
   *   &#64;Inject
   *   public MyRoutes(Service service) {
   *     this.service = service;
   *   }
   *
   *   public void configure() {
   *     from("direct:noop").to("mock:out").bean(service);
   *   }
   * }
   *
   * ...
   * {
   *   use(new Camel().routes(MyRoutes.class));
   * }
   * </pre>
   *
   * <p>
   * or without extending RouteBuilder:
   * </p>
   *
   * <pre>
   * public class MyRoutes {
   *
   *   &#64;Inject
   *   public MyRoutes(RouteBuilder router, Service service) {
   *     router.from("direct:noop").to("mock:out").bean(service);
   *   }
   *
   * }
   *
   * ...
   * {
   *   use(new Camel().routes(MyRoutes.class));
   * }
   * </pre>
   *
   * @param routeClass Type of routes.
   * @return This camel module.
   */
  public Camel routes(final Class<?> routeClass) {
    routeList.add(routeClass);
    return this;
  }

  /**
   * Hook to customize a {@link CamelContext}.
   *
   * @param configurer A configurer callback.
   * @return This instance.
   */
  public Camel doWith(final Configurer<CamelContext> configurer) {
    this.configurer = requireNonNull(configurer, "A configurer is required.");
    return this;
  }

  /**
   * Register one or more routes:
   *
   * <pre>
   * {
   *   use(new Camel()
   *     .routes((rb, config) {@literal ->} {
   *       rb.from("direct:noop").to("mock:out");
   *     })
   *   );
   *
   *   get("/noop", req {@literal ->} {
   *     req.require(ProducerTemplate.class).sendBody("direct:noop", "NOOP");
   *     return "/noop";
   *   });
   *
   * }
   * </pre>
   *
   * @param routes Route callback.
   * @return This camel module.
   */
  public Camel routes(final Configurer<RouteBuilder> routes) {
    this.routes = requireNonNull(routes, "Route configurer is required.");
    return this;
  }

  @Override
  public void configure(final Env env, final Config config, final Binder binder) throws Throwable {
    Config $camel = config.getConfig("camel");
    DefaultCamelContext ctx = configure(new DefaultCamelContext(), $camel
        .withoutPath("shutdown")
        .withoutPath("threads")
        .withoutPath("jmx")
        .withoutPath("streamCaching"));

    if (!$camel.getBoolean("jmx")) {
      ctx.disableJMX();
    }

    /**
     * Executor and thread poll
     */
    ThreadPoolProfile threadPool = configure(new ThreadPoolProfile(), $camel.getConfig("threads"));
    ctx.getExecutorServiceManager().setDefaultThreadPoolProfile(threadPool);

    /**
     * Shutdown options.
     */
    configure(ctx.getShutdownStrategy(), $camel.getConfig("shutdown"));

    if ($camel.getBoolean("streamCaching.enabled")) {
      ctx.setStreamCaching(true);
      configure(ctx.getStreamCachingStrategy(), $camel.getConfig("streamCaching"));
    } else {
      ctx.setStreamCaching(false);
    }

    /**
     * Components etc..
     */
    if (configurer != null) {
      configurer.configure(ctx, config);
    }

    /**
     * Routes
     */
    if (routes != null) {
      routes(routes, ctx, config);
    }

    /**
     * Properties
     */
    PropertiesComponent properties = new PropertiesComponent(config.root().origin().description());
    properties.setIgnoreMissingLocation(true);
    properties.setPropertiesResolver((context, ignoreMissingLocation, location) -> {
      Properties props = new Properties();
      config.entrySet()
          .forEach(e -> props.setProperty(e.getKey(), e.getValue().unwrapped().toString()));
      return props;
    });
    properties.setPrefixToken("${");
    properties.setSuffixToken("}");
    ctx.addComponent("properties", properties);

    env.lifeCycle(CamelFinalizer.class);

    /**
     * Guice!
     */
    binder.bind(CamelContext.class).toInstance(ctx);
    binder.bind(DefaultCamelContext.class).toInstance(ctx);
    binder.bind(ProducerTemplate.class).toInstance(ctx.createProducerTemplate());
    binder.bind(FluentProducerTemplate.class).toInstance(ctx.createFluentProducerTemplate());
    binder.bind(ConsumerTemplate.class).toInstance(ctx.createConsumerTemplate());

    binder.bind(CamelFinalizer.class).asEagerSingleton();

    binder.bind(RouteBuilder.class).toInstance(rb());

    Multibinder<Object> routesBinder = Multibinder
        .newSetBinder(binder, Object.class, Names.named("camel.routes"));

    routeList.forEach(routeType -> routesBinder.addBinding().to(routeType));
  }

  private static RouteBuilder rb() {
    return new RouteBuilder() {

      @Override
      public void configure() throws Exception {
      }
    };
  }

  private static void routes(final Configurer<RouteBuilder> callback,
      final DefaultCamelContext ctx, final Config config) throws Exception {
    ctx.addRoutes(new RouteBuilder() {

      @Override
      public void configure() throws Exception {
        callback.configure(this, config);
      }
    });
  }

  @Override
  public Config config() {
    return ConfigFactory.parseResources(getClass(), "camel.conf");
  }

  @SuppressWarnings({"rawtypes", "unchecked"})
  private <T> T configure(final T source, final Config config) {
    List<Method> methods = Lists.newArrayList(source.getClass().getMethods());
    config.entrySet().forEach(o -> {
      String key = o.getKey();
      String setter = "set" + CaseFormat.LOWER_CAMEL.to(CaseFormat.UPPER_CAMEL, key);
      Object raw = o.getValue().unwrapped();
      Optional<Method> result = methods.stream()
          .filter(m -> m.getName().equals(setter))
          .findFirst();
      if (result.isPresent()) {
        Method method = result.get();
        Class type = method.getParameterTypes()[0];
        Object value = cast(type, raw);
        Try.run(() -> method.invoke(source, value)).onFailure(ex -> {
          throw new IllegalArgumentException("Bad option: <" + raw + "> for: " + method, ex);
        });
      } else {
        log.error("Unknown option camel.{} = {}", key, raw);
      }
    });
    return source;
  }

  private Object cast(final Class type, Object value) {
    if (type.isEnum()) {
      return Enum.valueOf(type, value.toString());
    }
    if (type == Long.class) {
      return ((Number) value).longValue();
    }
    if (type == File.class) {
      return new File(value.toString());
    }
    return value;
  }

}
