/**
 *                                  Apache License
 *                            Version 2.0, January 2004
 *                         http://www.apache.org/licenses/
 *
 *    TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *    1. Definitions.
 *
 *       "License" shall mean the terms and conditions for use, reproduction,
 *       and distribution as defined by Sections 1 through 9 of this document.
 *
 *       "Licensor" shall mean the copyright owner or entity authorized by
 *       the copyright owner that is granting the License.
 *
 *       "Legal Entity" shall mean the union of the acting entity and all
 *       other entities that control, are controlled by, or are under common
 *       control with that entity. For the purposes of this definition,
 *       "control" means (i) the power, direct or indirect, to cause the
 *       direction or management of such entity, whether by contract or
 *       otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *       outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *       "You" (or "Your") shall mean an individual or Legal Entity
 *       exercising permissions granted by this License.
 *
 *       "Source" form shall mean the preferred form for making modifications,
 *       including but not limited to software source code, documentation
 *       source, and configuration files.
 *
 *       "Object" form shall mean any form resulting from mechanical
 *       transformation or translation of a Source form, including but
 *       not limited to compiled object code, generated documentation,
 *       and conversions to other media types.
 *
 *       "Work" shall mean the work of authorship, whether in Source or
 *       Object form, made available under the License, as indicated by a
 *       copyright notice that is included in or attached to the work
 *       (an example is provided in the Appendix below).
 *
 *       "Derivative Works" shall mean any work, whether in Source or Object
 *       form, that is based on (or derived from) the Work and for which the
 *       editorial revisions, annotations, elaborations, or other modifications
 *       represent, as a whole, an original work of authorship. For the purposes
 *       of this License, Derivative Works shall not include works that remain
 *       separable from, or merely link (or bind by name) to the interfaces of,
 *       the Work and Derivative Works thereof.
 *
 *       "Contribution" shall mean any work of authorship, including
 *       the original version of the Work and any modifications or additions
 *       to that Work or Derivative Works thereof, that is intentionally
 *       submitted to Licensor for inclusion in the Work by the copyright owner
 *       or by an individual or Legal Entity authorized to submit on behalf of
 *       the copyright owner. For the purposes of this definition, "submitted"
 *       means any form of electronic, verbal, or written communication sent
 *       to the Licensor or its representatives, including but not limited to
 *       communication on electronic mailing lists, source code control systems,
 *       and issue tracking systems that are managed by, or on behalf of, the
 *       Licensor for the purpose of discussing and improving the Work, but
 *       excluding communication that is conspicuously marked or otherwise
 *       designated in writing by the copyright owner as "Not a Contribution."
 *
 *       "Contributor" shall mean Licensor and any individual or Legal Entity
 *       on behalf of whom a Contribution has been received by Licensor and
 *       subsequently incorporated within the Work.
 *
 *    2. Grant of Copyright License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       copyright license to reproduce, prepare Derivative Works of,
 *       publicly display, publicly perform, sublicense, and distribute the
 *       Work and such Derivative Works in Source or Object form.
 *
 *    3. Grant of Patent License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       (except as stated in this section) patent license to make, have made,
 *       use, offer to sell, sell, import, and otherwise transfer the Work,
 *       where such license applies only to those patent claims licensable
 *       by such Contributor that are necessarily infringed by their
 *       Contribution(s) alone or by combination of their Contribution(s)
 *       with the Work to which such Contribution(s) was submitted. If You
 *       institute patent litigation against any entity (including a
 *       cross-claim or counterclaim in a lawsuit) alleging that the Work
 *       or a Contribution incorporated within the Work constitutes direct
 *       or contributory patent infringement, then any patent licenses
 *       granted to You under this License for that Work shall terminate
 *       as of the date such litigation is filed.
 *
 *    4. Redistribution. You may reproduce and distribute copies of the
 *       Work or Derivative Works thereof in any medium, with or without
 *       modifications, and in Source or Object form, provided that You
 *       meet the following conditions:
 *
 *       (a) You must give any other recipients of the Work or
 *           Derivative Works a copy of this License; and
 *
 *       (b) You must cause any modified files to carry prominent notices
 *           stating that You changed the files; and
 *
 *       (c) You must retain, in the Source form of any Derivative Works
 *           that You distribute, all copyright, patent, trademark, and
 *           attribution notices from the Source form of the Work,
 *           excluding those notices that do not pertain to any part of
 *           the Derivative Works; and
 *
 *       (d) If the Work includes a "NOTICE" text file as part of its
 *           distribution, then any Derivative Works that You distribute must
 *           include a readable copy of the attribution notices contained
 *           within such NOTICE file, excluding those notices that do not
 *           pertain to any part of the Derivative Works, in at least one
 *           of the following places: within a NOTICE text file distributed
 *           as part of the Derivative Works; within the Source form or
 *           documentation, if provided along with the Derivative Works; or,
 *           within a display generated by the Derivative Works, if and
 *           wherever such third-party notices normally appear. The contents
 *           of the NOTICE file are for informational purposes only and
 *           do not modify the License. You may add Your own attribution
 *           notices within Derivative Works that You distribute, alongside
 *           or as an addendum to the NOTICE text from the Work, provided
 *           that such additional attribution notices cannot be construed
 *           as modifying the License.
 *
 *       You may add Your own copyright statement to Your modifications and
 *       may provide additional or different license terms and conditions
 *       for use, reproduction, or distribution of Your modifications, or
 *       for any such Derivative Works as a whole, provided Your use,
 *       reproduction, and distribution of the Work otherwise complies with
 *       the conditions stated in this License.
 *
 *    5. Submission of Contributions. Unless You explicitly state otherwise,
 *       any Contribution intentionally submitted for inclusion in the Work
 *       by You to the Licensor shall be under the terms and conditions of
 *       this License, without any additional terms or conditions.
 *       Notwithstanding the above, nothing herein shall supersede or modify
 *       the terms of any separate license agreement you may have executed
 *       with Licensor regarding such Contributions.
 *
 *    6. Trademarks. This License does not grant permission to use the trade
 *       names, trademarks, service marks, or product names of the Licensor,
 *       except as required for reasonable and customary use in describing the
 *       origin of the Work and reproducing the content of the NOTICE file.
 *
 *    7. Disclaimer of Warranty. Unless required by applicable law or
 *       agreed to in writing, Licensor provides the Work (and each
 *       Contributor provides its Contributions) on an "AS IS" BASIS,
 *       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *       implied, including, without limitation, any warranties or conditions
 *       of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *       PARTICULAR PURPOSE. You are solely responsible for determining the
 *       appropriateness of using or redistributing the Work and assume any
 *       risks associated with Your exercise of permissions under this License.
 *
 *    8. Limitation of Liability. In no event and under no legal theory,
 *       whether in tort (including negligence), contract, or otherwise,
 *       unless required by applicable law (such as deliberate and grossly
 *       negligent acts) or agreed to in writing, shall any Contributor be
 *       liable to You for damages, including any direct, indirect, special,
 *       incidental, or consequential damages of any character arising as a
 *       result of this License or out of the use or inability to use the
 *       Work (including but not limited to damages for loss of goodwill,
 *       work stoppage, computer failure or malfunction, or any and all
 *       other commercial damages or losses), even if such Contributor
 *       has been advised of the possibility of such damages.
 *
 *    9. Accepting Warranty or Additional Liability. While redistributing
 *       the Work or Derivative Works thereof, You may choose to offer,
 *       and charge a fee for, acceptance of support, warranty, indemnity,
 *       or other liability obligations and/or rights consistent with this
 *       License. However, in accepting such obligations, You may act only
 *       on Your own behalf and on Your sole responsibility, not on behalf
 *       of any other Contributor, and only if You agree to indemnify,
 *       defend, and hold each Contributor harmless for any liability
 *       incurred by, or claims asserted against, such Contributor by reason
 *       of your accepting any such warranty or additional liability.
 *
 *    END OF TERMS AND CONDITIONS
 *
 *    APPENDIX: How to apply the Apache License to your work.
 *
 *       To apply the Apache License to your work, attach the following
 *       boilerplate notice, with the fields enclosed by brackets "{}"
 *       replaced with your own identifying information. (Don't include
 *       the brackets!)  The text should be enclosed in the appropriate
 *       comment syntax for the file format. We also recommend that a
 *       file or class name and description of purpose be included on the
 *       same "printed page" as the copyright notice for easier
 *       identification within third-party archives.
 *
 *    Copyright 2014 Edgar Espina
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package org.jooby.cassandra;

import com.datastax.driver.core.Cluster;
import com.datastax.driver.core.CodecRegistry;
import com.datastax.driver.core.Configuration;
import com.datastax.driver.core.Session;
import com.datastax.driver.extras.codecs.jdk8.InstantCodec;
import com.datastax.driver.extras.codecs.jdk8.LocalDateCodec;
import com.datastax.driver.extras.codecs.jdk8.LocalTimeCodec;
import com.datastax.driver.mapping.Mapper;
import com.datastax.driver.mapping.MappingManager;
import com.datastax.driver.mapping.annotations.Accessor;
import com.google.inject.Binder;
import com.typesafe.config.Config;
import static java.util.Objects.requireNonNull;
import org.jooby.Env;
import org.jooby.Env.ServiceKey;
import org.jooby.Jooby.Module;
import org.jooby.funzy.Throwing;
import org.jooby.funzy.Try;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;
import java.util.function.BiConsumer;
import java.util.function.Consumer;
import java.util.function.Supplier;

/**
 * <h1>cassandra</h1>
 * <p>
 * <a href="http://cassandra.apache.org">The Apache Cassandra</a> database is the right choice when
 * you need scalability and high availability without compromising performance. Linear scalability
 * and proven fault-tolerance on commodity hardware or cloud infrastructure make it the perfect
 * platform for mission-critical data. Cassandra's support for replicating across multiple
 * datacenters is best-in-class, providing lower latency for your users and the peace of mind of
 * knowing that you can survive regional outages.
 * </p>
 *
 * <p>
 * This module offers <a href="http://cassandra.apache.org">cassandra</a> database features via
 * <a href="http://datastax.github.io/java-driver">Datastax Java Driver</a>.
 * </p>
 *
 * <h2>exports</h2>
 * <ul>
 * <li>Cluster</li>
 * <li>Session</li>
 * <li>MappingManager</li>
 * <li>Datastore</li>
 * <li>Optionally a Cassandra {@link org.jooby.Session.Store}</li>
 * </ul>
 *
 * <h2>usage</h2>
 * <p>
 * Via connection string:
 * </p>
 * <pre>{@code
 * {
 *   use(new Cassandra("cassandra://localhost/db"));
 * }
 * }</pre>
 *
 * <p>
 * Via connection property:
 * </p>
 * <pre>{@code
 * {
 *   use(new Cassandra("db"));
 * }
 * }</pre>
 *
 * <p>
 * After you install the module {@link Session}, {@link MappingManager} and {@link Datastore} are
 * ready to use.
 * </p>
 *
 * <pre>{@code
 * {
 *   use(new Cassandra("cassandra://localhost/db"));
 *
 *   get("/doWithSession", req -> {
 *     Session session = require(Session.class);
 *     // work with session
 *   });
 *
 *   get("/doWithMappingManager", req -> {
 *     MappingManager manager = require(MappingManager.class);
 *     Mapper<Beer> mapper = manager.mapper(Beer.class);
 *     // work with mapper;
 *   });
 *
 *   get("/doWithDatastore", req -> {
 *     Datastore ds = require(Datastore.class);
 *     // work with datastore;
 *   });
 * }
 * }</pre>
 *
 * <h2>basic crud</h2>
 * <p>
 * This module exports {@link MappingManager} so you are free to use a {@link Mapper}. Jooby also
 * offers the {@link Datastore} service which basically wrap a {@link Mapper} and provides
 * query/read operations.
 * </p>
 * <p>
 * The main advantage of {@link Datastore} over {@link Mapper} is that you require just once
 * instance
 * regardless of your number of entities, but also it provides some useful <code>query*</code>
 * methods.
 * </p>
 *
 * <p>
 * Here is a basic API on top of {@link Datastore}:
 * </p>
 *
 * <pre>{@code
 * {
 *
 *   use("/api/beer")
 *     .post(req -> {
 *       Datastore ds = req.require(Datastore.class);
 *       Beer beer = req.body().to(Beer.class);
 *       ds.save(beer);
 *       return beer;
 *     })
 *     .get("/:id", req -> {
 *       Datastore ds = req.require(Datastore.class);
 *       Beer beer = ds.get(Beer.class, req.param("id").value());
 *       return beer;
 *     })
 *     .get(req -> {
 *       Datastore ds = req.require(Datastore.class);
 *       return ds.query(Beer.class, "select * from beer").all();
 *     })
 *     .delete("/:id", req -> {
 *       Datastore ds = req.require(Datastore.class);
 *       ds.delete(Beer.class, req.param("id").value());
 *       return Results.noContent();
 *     });
 * }
 * }</pre>
 *
 * <p>
 * Keep in mind your entities must be mapped as usual or as required by {@link Mapper}. A great
 * example is available
 * <a href="http://datastax.github.io/java-driver/manual/object_mapper/creating">here</a>
 * </p>
 *
 * <h2>accessors</h2>
 * <p>
 * Accessors provide a way to map custom queries not supported by the default entity mappers.
 * Accessors are created at application startup time via {@link #accesor(Class)} method:
 * </p>
 *
 * <pre>{@code
 * {
 *   use(new Cassandra("cassandra://localhost/db")
 *    .accessor(UserAccessor.class)
 *   );
 *
 *   get("/users", req -> {
 *     return require(UserAccessor.class).getAll();
 *   });
 * }
 * }</pre>
 *
 * <p>
 * The accessor can be required or injected in a MVC route.
 * </p>
 *
 * <h2>dse-driver</h2>
 * <p>
 * Add the <code>dse-driver</code> dependency to your classpath and then:
 * </p>
 *
 * <pre>
 * {
 *    use(new Cassandra(DseCluster::build));
 * }
 * </pre>
 *
 * <p>
 * That's all! Now you can <code>require/inject</code> a <code>DseSession</code>.
 * </p>
 *
 * <h2>async</h2>
 * <p>
 * Async? Of course!!! just use the Datastax async API:
 * </p>
 *
 * <pre>{@code
 * {
 *   use(new Cassandra("cassandra://localhost/db"));
 *
 *   use("/api/beer")
 *     .post(req -> {
 *       Datastore ds = req.require(Datastore.class);
 *       Beer beer = req.body().to(Beer.class);
 *       ds.saveAsync(beer);
 *       return beer;
 *     })
 *     .get("/:id", req -> {
 *       Datastore ds = req.require(Datastore.class);
 *       ListeneableFuture<Beer> beer = ds.getAsync(Beer.class, req.param("id").value());
 *       return beer;
 *     })
 *     .get(req -> {
 *       Datastore ds = req.require(Datastore.class);
 *       return ds.queryAsync(Beer.class, "select * from beer").all();
 *     })
 *     .delete("/:id", req -> {
 *       Datastore ds = req.require(Datastore.class);
 *       ds.deleteAsync(Beer.class, req.param("id").value());
 *       return Results.noContent();
 *     });
 * }
 * }</pre>
 *
 * <h2>multiple contact points</h2>
 * <p>
 * Multiple contact points are separated by a comma:
 * </p>
 * <pre>{@code
 * {
 *   use(new Casssandra("cassandra://host1,host2/db");
 * }
 * }</pre>
 *
 * <h2>advanced configuration</h2>
 * <p>
 * Advanced configuration is available via cluster builder callback:
 * </p>
 *
 * <pre>{@code
 * {
 *   use(new Casssandra("cassandra://localhost/db")
 *     .doWithClusterBuilder(builder -> {
 *       builder.withClusterName("mycluster");
 *     }));
 * }
 * }</pre>
 *
 * <p>
 * Or via cluster callback:
 * </p>
 *
 * <pre>{@code
 * {
 *   use(new Casssandra("cassandra://localhost/db")
 *     .doWithCluster(cluster -> {
 *       Configuration configuration = cluster.getConfiguration();
 *       // set option
 *     }));
 * }
 * }</pre>
 *
 * @author edgar
 * @since 1.0.0.CR7
 */
public class Cassandra implements Module {

  /** The logging system. */
  private final Logger log = LoggerFactory.getLogger(getClass());

  private static final Supplier<Cluster.Builder> BUILDER = Cluster::builder;

  private BiConsumer<Cluster.Builder, Config> ccbuilder;

  private BiConsumer<Cluster, Config> cc;

  private final String db;

  @SuppressWarnings("rawtypes")

  private final List<Class> accesors = new ArrayList<>();

  private Supplier<Cluster.Builder> builder = Cluster::builder;

  /**
   * Creates a new {@link Cassandra} module.
   *
   * Conenct via connection string:
   *
   * <pre>{@code
   * {
   *   use(new Cassandra("cassandra://localhost/db", DseCluster::builder));
   * }
   * }</pre>
   *
   * Or via property:
   *
   * <pre>{@code
   * {
   *   use(new Cassandra("db", DseCluster::builder));
   * }
   * }</pre>
   *
   * @param db Connection string or a property with a connection String.
   * @param builder Cluster builder.
   */
  public Cassandra(final String db, final Supplier<Cluster.Builder> builder) {
    this.db = requireNonNull(db, "ContactPoint/db key required.");
    this.builder = requireNonNull(builder, "Cluster.Builder required.");
  }

  /**
   * Creates a new {@link Cassandra} module.
   *
   * Conenct via connection string:
   *
   * <pre>{@code
   * {
   *   use(new Cassandra("cassandra://localhost/db"));
   * }
   * }</pre>
   *
   * Or via property:
   *
   * <pre>{@code
   * {
   *   use(new Cassandra("db"));
   * }
   * }</pre>
   *
   * @param db Connection string or a property with a connection String.
   */
  public Cassandra(final String db) {
    this(db, BUILDER);
  }

  /**
   * Creates a new {@link Cassandra} module. A property <code>db</code> property must be present and
   * have a valid connection string.
   */
  public Cassandra(final Supplier<Cluster.Builder> builder) {
    this("db", builder);
  }

  /**
   * Creates a new {@link Cassandra} module. A property <code>db</code> property must be present and
   * have a valid connection string.
   */
  public Cassandra() {
    this("db");
  }

  /**
   * Register an {@link Accessor} which is accessible via require calls or dependency injection.
   *
   * @param accessor An accessor.
   * @return This module.
   */
  public Cassandra accesor(final Class<?> accessor) {
    accesors.add(accessor);
    return this;
  }

  /**
   * Configure a cluster before creating it.
   *
   * @param configurer Configurer callback.
   * @return This module.
   */
  public Cassandra doWithClusterBuilder(final BiConsumer<Cluster.Builder, Config> configurer) {
    this.ccbuilder = requireNonNull(configurer, "ClusterBuilder conf callback required.");
    return this;
  }

  /**
   * Configure a cluster before creating it.
   *
   * @param configurer Configurer callback.
   * @return This module.
   */
  public Cassandra doWithClusterBuilder(final Consumer<Cluster.Builder> configurer) {
    requireNonNull(configurer, "ClusterBuilder conf callback required.");
    return doWithClusterBuilder((b, c) -> configurer.accept(b));
  }

  /**
   * Configure a cluster after creation.
   *
   * @param configurer Configurer callback.
   * @return This module.
   */
  public Cassandra doWithCluster(final BiConsumer<Cluster, Config> configurer) {
    this.cc = requireNonNull(configurer, "Cluster conf callbackrequired.");
    return this;
  }

  /**
   * Configure a cluster after creation.
   *
   * @param configurer Configurer callback.
   * @return This module.
   */
  public Cassandra doWithCluster(final Consumer<Cluster> configurer) {
    requireNonNull(configurer, "Cluster conf callbackrequired.");
    return doWithCluster((cc, c) -> configurer.accept(cc));
  }

  @SuppressWarnings({"rawtypes", "unchecked"})
  @Override
  public void configure(final Env env, final Config conf, final Binder binder) {
    ConnectionString cstr = Try.apply(() -> ConnectionString.parse(db))
        .orElseGet(() -> ConnectionString.parse(conf.getString(db)));

    ServiceKey serviceKey = env.serviceKey();

    Throwing.Function3<Class, String, Object, Void> bind = (type, name, value) -> {
      serviceKey.generate(type, name, k -> binder.bind(k).toInstance(value));
      return null;
    };

    Cluster.Builder builder = this.builder.get()
        .addContactPoints(cstr.contactPoints())
        .withPort(cstr.port());

    // allow user configure cluster builder
    if (ccbuilder != null) {
      ccbuilder.accept(builder, conf);
    }

    log.debug("Starting {}", cstr);

    Cluster cluster = builder.build();

    // allow user configure cluster
    if (cc != null) {
      cc.accept(cluster, conf);
    }

    /** codecs */
    Configuration configuration = cluster.getConfiguration();
    CodecRegistry codecRegistry = configuration.getCodecRegistry();
    // java 8 codecs
    codecRegistry.register(
        InstantCodec.instance,
        LocalDateCodec.instance,
        LocalTimeCodec.instance);

    hierarchy(cluster.getClass(), type -> bind.apply(type, cstr.keyspace(), cluster));

    /** Session + Mapper */
    Session session = cluster.connect(cstr.keyspace());
    hierarchy(session.getClass(), type -> bind.apply(type, cstr.keyspace(), session));

    MappingManager manager = new MappingManager(session);
    bind.apply(MappingManager.class, cstr.keyspace(), manager);
    bind.apply(Datastore.class, cstr.keyspace(), new Datastore(manager));

    /** accessors */
    accesors.forEach(c -> {
      Object accessor = manager.createAccessor(c);
      binder.bind(c).toInstance(accessor);
    });

    env.router()
        .map(new CassandraMapper());

    env.onStop(() -> {
      log.debug("Stopping {}", cstr);
      Try.run(session::close)
          .onFailure(x -> log.error("session.close() resulted in exception", x));

      cluster.close();

      log.info("Stopped {}", cstr);
    });
  }

  @SuppressWarnings("rawtypes")
  static void hierarchy(final Class type, final Consumer<Class> consumer) {
    if (type != Object.class) {
      if (type.getName().startsWith("com.datastax")) {
        consumer.accept(type);
      }
      for (Class i : type.getInterfaces()) {
        hierarchy(i, consumer);
      }
      if (!type.isInterface()) {
        hierarchy(type.getSuperclass(), consumer);
      }
    }
  }

}
