/**
 *                                  Apache License
 *                            Version 2.0, January 2004
 *                         http://www.apache.org/licenses/
 *
 *    TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *    1. Definitions.
 *
 *       "License" shall mean the terms and conditions for use, reproduction,
 *       and distribution as defined by Sections 1 through 9 of this document.
 *
 *       "Licensor" shall mean the copyright owner or entity authorized by
 *       the copyright owner that is granting the License.
 *
 *       "Legal Entity" shall mean the union of the acting entity and all
 *       other entities that control, are controlled by, or are under common
 *       control with that entity. For the purposes of this definition,
 *       "control" means (i) the power, direct or indirect, to cause the
 *       direction or management of such entity, whether by contract or
 *       otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *       outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *       "You" (or "Your") shall mean an individual or Legal Entity
 *       exercising permissions granted by this License.
 *
 *       "Source" form shall mean the preferred form for making modifications,
 *       including but not limited to software source code, documentation
 *       source, and configuration files.
 *
 *       "Object" form shall mean any form resulting from mechanical
 *       transformation or translation of a Source form, including but
 *       not limited to compiled object code, generated documentation,
 *       and conversions to other media types.
 *
 *       "Work" shall mean the work of authorship, whether in Source or
 *       Object form, made available under the License, as indicated by a
 *       copyright notice that is included in or attached to the work
 *       (an example is provided in the Appendix below).
 *
 *       "Derivative Works" shall mean any work, whether in Source or Object
 *       form, that is based on (or derived from) the Work and for which the
 *       editorial revisions, annotations, elaborations, or other modifications
 *       represent, as a whole, an original work of authorship. For the purposes
 *       of this License, Derivative Works shall not include works that remain
 *       separable from, or merely link (or bind by name) to the interfaces of,
 *       the Work and Derivative Works thereof.
 *
 *       "Contribution" shall mean any work of authorship, including
 *       the original version of the Work and any modifications or additions
 *       to that Work or Derivative Works thereof, that is intentionally
 *       submitted to Licensor for inclusion in the Work by the copyright owner
 *       or by an individual or Legal Entity authorized to submit on behalf of
 *       the copyright owner. For the purposes of this definition, "submitted"
 *       means any form of electronic, verbal, or written communication sent
 *       to the Licensor or its representatives, including but not limited to
 *       communication on electronic mailing lists, source code control systems,
 *       and issue tracking systems that are managed by, or on behalf of, the
 *       Licensor for the purpose of discussing and improving the Work, but
 *       excluding communication that is conspicuously marked or otherwise
 *       designated in writing by the copyright owner as "Not a Contribution."
 *
 *       "Contributor" shall mean Licensor and any individual or Legal Entity
 *       on behalf of whom a Contribution has been received by Licensor and
 *       subsequently incorporated within the Work.
 *
 *    2. Grant of Copyright License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       copyright license to reproduce, prepare Derivative Works of,
 *       publicly display, publicly perform, sublicense, and distribute the
 *       Work and such Derivative Works in Source or Object form.
 *
 *    3. Grant of Patent License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       (except as stated in this section) patent license to make, have made,
 *       use, offer to sell, sell, import, and otherwise transfer the Work,
 *       where such license applies only to those patent claims licensable
 *       by such Contributor that are necessarily infringed by their
 *       Contribution(s) alone or by combination of their Contribution(s)
 *       with the Work to which such Contribution(s) was submitted. If You
 *       institute patent litigation against any entity (including a
 *       cross-claim or counterclaim in a lawsuit) alleging that the Work
 *       or a Contribution incorporated within the Work constitutes direct
 *       or contributory patent infringement, then any patent licenses
 *       granted to You under this License for that Work shall terminate
 *       as of the date such litigation is filed.
 *
 *    4. Redistribution. You may reproduce and distribute copies of the
 *       Work or Derivative Works thereof in any medium, with or without
 *       modifications, and in Source or Object form, provided that You
 *       meet the following conditions:
 *
 *       (a) You must give any other recipients of the Work or
 *           Derivative Works a copy of this License; and
 *
 *       (b) You must cause any modified files to carry prominent notices
 *           stating that You changed the files; and
 *
 *       (c) You must retain, in the Source form of any Derivative Works
 *           that You distribute, all copyright, patent, trademark, and
 *           attribution notices from the Source form of the Work,
 *           excluding those notices that do not pertain to any part of
 *           the Derivative Works; and
 *
 *       (d) If the Work includes a "NOTICE" text file as part of its
 *           distribution, then any Derivative Works that You distribute must
 *           include a readable copy of the attribution notices contained
 *           within such NOTICE file, excluding those notices that do not
 *           pertain to any part of the Derivative Works, in at least one
 *           of the following places: within a NOTICE text file distributed
 *           as part of the Derivative Works; within the Source form or
 *           documentation, if provided along with the Derivative Works; or,
 *           within a display generated by the Derivative Works, if and
 *           wherever such third-party notices normally appear. The contents
 *           of the NOTICE file are for informational purposes only and
 *           do not modify the License. You may add Your own attribution
 *           notices within Derivative Works that You distribute, alongside
 *           or as an addendum to the NOTICE text from the Work, provided
 *           that such additional attribution notices cannot be construed
 *           as modifying the License.
 *
 *       You may add Your own copyright statement to Your modifications and
 *       may provide additional or different license terms and conditions
 *       for use, reproduction, or distribution of Your modifications, or
 *       for any such Derivative Works as a whole, provided Your use,
 *       reproduction, and distribution of the Work otherwise complies with
 *       the conditions stated in this License.
 *
 *    5. Submission of Contributions. Unless You explicitly state otherwise,
 *       any Contribution intentionally submitted for inclusion in the Work
 *       by You to the Licensor shall be under the terms and conditions of
 *       this License, without any additional terms or conditions.
 *       Notwithstanding the above, nothing herein shall supersede or modify
 *       the terms of any separate license agreement you may have executed
 *       with Licensor regarding such Contributions.
 *
 *    6. Trademarks. This License does not grant permission to use the trade
 *       names, trademarks, service marks, or product names of the Licensor,
 *       except as required for reasonable and customary use in describing the
 *       origin of the Work and reproducing the content of the NOTICE file.
 *
 *    7. Disclaimer of Warranty. Unless required by applicable law or
 *       agreed to in writing, Licensor provides the Work (and each
 *       Contributor provides its Contributions) on an "AS IS" BASIS,
 *       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *       implied, including, without limitation, any warranties or conditions
 *       of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *       PARTICULAR PURPOSE. You are solely responsible for determining the
 *       appropriateness of using or redistributing the Work and assume any
 *       risks associated with Your exercise of permissions under this License.
 *
 *    8. Limitation of Liability. In no event and under no legal theory,
 *       whether in tort (including negligence), contract, or otherwise,
 *       unless required by applicable law (such as deliberate and grossly
 *       negligent acts) or agreed to in writing, shall any Contributor be
 *       liable to You for damages, including any direct, indirect, special,
 *       incidental, or consequential damages of any character arising as a
 *       result of this License or out of the use or inability to use the
 *       Work (including but not limited to damages for loss of goodwill,
 *       work stoppage, computer failure or malfunction, or any and all
 *       other commercial damages or losses), even if such Contributor
 *       has been advised of the possibility of such damages.
 *
 *    9. Accepting Warranty or Additional Liability. While redistributing
 *       the Work or Derivative Works thereof, You may choose to offer,
 *       and charge a fee for, acceptance of support, warranty, indemnity,
 *       or other liability obligations and/or rights consistent with this
 *       License. However, in accepting such obligations, You may act only
 *       on Your own behalf and on Your sole responsibility, not on behalf
 *       of any other Contributor, and only if You agree to indemnify,
 *       defend, and hold each Contributor harmless for any liability
 *       incurred by, or claims asserted against, such Contributor by reason
 *       of your accepting any such warranty or additional liability.
 *
 *    END OF TERMS AND CONDITIONS
 *
 *    APPENDIX: How to apply the Apache License to your work.
 *
 *       To apply the Apache License to your work, attach the following
 *       boilerplate notice, with the fields enclosed by brackets "{}"
 *       replaced with your own identifying information. (Don't include
 *       the brackets!)  The text should be enclosed in the appropriate
 *       comment syntax for the file format. We also recommend that a
 *       file or class name and description of purpose be included on the
 *       same "printed page" as the copyright notice for easier
 *       identification within third-party archives.
 *
 *    Copyright 2014 Edgar Espina
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package org.jooby.cassandra;

import com.datastax.driver.core.BoundStatement;
import com.datastax.driver.core.DataType;
import com.datastax.driver.core.PreparedStatement;
import com.datastax.driver.core.ResultSet;
import com.datastax.driver.core.querybuilder.Insert;
import com.datastax.driver.core.querybuilder.QueryBuilder;
import static com.datastax.driver.core.querybuilder.QueryBuilder.eq;
import static com.datastax.driver.core.querybuilder.QueryBuilder.insertInto;
import static com.datastax.driver.core.querybuilder.QueryBuilder.raw;
import static com.datastax.driver.core.querybuilder.QueryBuilder.select;
import static com.datastax.driver.core.querybuilder.QueryBuilder.ttl;
import com.datastax.driver.core.schemabuilder.Create;
import com.datastax.driver.core.schemabuilder.SchemaBuilder;
import com.google.common.util.concurrent.FutureCallback;
import com.google.common.util.concurrent.Futures;
import com.typesafe.config.Config;
import com.typesafe.config.ConfigFactory;
import com.typesafe.config.ConfigValueFactory;
import static java.util.Objects.requireNonNull;
import org.jooby.Session;
import org.jooby.Session.Builder;
import org.jooby.Session.Store;
import org.jooby.funzy.Throwing;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import javax.inject.Named;
import java.util.Date;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.TimeUnit;

/**
 * <h1>cassandra session store</h1>
 * <p>
 * A {@link Session.Store} powered by <a href="http://cassandra.apache.org">Cassandra</a>.
 * </p>
 *
 * <h2>usage</h2>
 *
 * <pre>{@code
 * {
 *   use(new Cassandra("cassandra://localhost/db"));
 *
 *   session(CassandraSessionStore.class);
 *
 *   get("/", req -> {
 *     Session session = req.session();
 *     session.put("foo", "bar");
 *
 *     ..
 *   });
 * }
 * }</pre>
 *
 * <p>
 * Session data is persisted in Cassandra using a <code>session</code> table.
 * </p>
 *
 * <h2>options</h2>
 *
 * <h3>timeout</h3>
 * <p>
 * By default, a session will expire after <code>30 minutes</code>. Changing the default timeout is
 * as simple as:
 * </p>
 *
 * <pre>
 * # 8 hours
 * session.timeout = 8h
 *
 * # 15 seconds
 * session.timeout = 15
 *
 * # 120 minutes
 * session.timeout = 120m
 * </pre>
 *
 * <p>
 * Expiration is done via Cassandra ttl option.
 * </p>
 *
 * <p>
 * If no timeout is required, use <code>-1</code>.
 * </p>
 *
 * @author edgar
 * @since 1.0.0.CR7
 */
public class CassandraSessionStore implements Store {

  private static final String TIMEOUT = "timeout";

  private static final String ID = "id";

  private static final String ATTRIBUTES = "attributes";

  private static final String SAVED_AT = "savedAt";

  private static final String ACCESSED_AT = "accessedAt";

  private static final String CREATED_AT = "createdAt";

  /** The logging system. */
  private final Logger log = LoggerFactory.getLogger(getClass());

  private final com.datastax.driver.core.Session session;

  private final int timeout;

  private final String tableName = "session";

  private final Throwing.Function<String, PreparedStatement> insertSQL;

  private final Throwing.Function<String, PreparedStatement> selectSQL;

  private final Throwing.Function<String, PreparedStatement> deleteSQL;

  public CassandraSessionStore(final com.datastax.driver.core.Session session, final int timeout) {
    this.session = requireNonNull(session, "Session required.");
    this.timeout = timeout;

    createTableIfNotExists(session, tableName, log);

    this.insertSQL = Throwing.<String, PreparedStatement>throwingFunction(table ->
        session.prepare(insertSQL(table, timeout))
    ).memoized();
    this.selectSQL = Throwing.<String, PreparedStatement>throwingFunction(table ->
        session.prepare(selectSQL(table))
    ).memoized();
    this.deleteSQL = Throwing.<String, PreparedStatement>throwingFunction(table ->
        session.prepare(deleteSQL(table))
    ).memoized();
  }

  @Inject
  public CassandraSessionStore(final com.datastax.driver.core.Session session,
      final @Named("session.timeout") String timeout) {
    this(session, seconds(timeout));
  }

  @Override
  public Session get(final Builder builder) {
    ResultSet rs = session
        .execute(new BoundStatement(selectSQL.apply(tableName)).bind(builder.sessionId()));
    return Optional.ofNullable(rs.one())
        .map(row -> {
          long createdAt = row.getTimestamp(CREATED_AT).getTime();
          long accessedAt = row.getTimestamp(ACCESSED_AT).getTime();
          long savedAt = row.getTimestamp(SAVED_AT).getTime();
          Map<String, String> attributes = row.getMap(ATTRIBUTES, String.class, String.class);
          Session session = builder
              .accessedAt(accessedAt)
              .createdAt(createdAt)
              .savedAt(savedAt)
              .set(attributes)
              .build();
          // touch ttl
          if (timeout > 0) {
            save(session);
          }
          return session;
        })
        .orElse(null);
  }

  @Override
  public void save(final Session session) {
    this.session.execute(new BoundStatement(insertSQL.apply(tableName))
        .bind(
            session.id(),
            new Date(session.createdAt()),
            new Date(session.accessedAt()),
            new Date(session.savedAt()),
            session.attributes()));
  }

  @Override
  public void create(final Session session) {
    save(session);
  }

  @Override
  public void delete(final String id) {
    session.execute(new BoundStatement(deleteSQL.apply(tableName)).bind(id));
  }

  private static int seconds(final String value) {
    try {
      return Integer.parseInt(value);
    } catch (NumberFormatException ex) {
      Config config = ConfigFactory.empty()
          .withValue(TIMEOUT, ConfigValueFactory.fromAnyRef(value));
      return (int) config.getDuration(TIMEOUT, TimeUnit.SECONDS);
    }
  }

  private static void createTableIfNotExists(final com.datastax.driver.core.Session session,
      final String table, final Logger log) {
    Create createTable = SchemaBuilder.createTable(table)
        .addPartitionKey(ID, DataType.varchar())
        .addColumn(CREATED_AT, DataType.timestamp())
        .addColumn(ACCESSED_AT, DataType.timestamp())
        .addColumn(SAVED_AT, DataType.timestamp())
        .addColumn(ATTRIBUTES, DataType.map(DataType.varchar(), DataType.varchar()))
        .ifNotExists();

    Futures.addCallback(session.executeAsync(createTable), new FutureCallback<ResultSet>() {
      @Override
      public void onSuccess(final ResultSet result) {
        log.debug("Session table successfully created");
      }

      @Override
      public void onFailure(final Throwable x) {
        log.error("Create session table resulted in exception", x);
      }
    });
  }

  private static String selectSQL(final String table) {
    return select().from(table).where(eq(ID, raw("?"))).getQueryString();
  }

  private static String deleteSQL(final String table) {
    return QueryBuilder.delete().from(table).where(eq(ID, raw("?"))).getQueryString();
  }

  private static String insertSQL(final String table, final int timeout) {
    Insert insertInto = insertInto(table)
        .value(ID, raw("?"))
        .value(CREATED_AT, raw("?"))
        .value(ACCESSED_AT, raw("?"))
        .value(SAVED_AT, raw("?"))
        .value(ATTRIBUTES, raw("?"));
    if (timeout > 0) {
      insertInto.using(ttl(timeout));
    }
    return insertInto.getQueryString();
  }

}
