/**
 *                                  Apache License
 *                            Version 2.0, January 2004
 *                         http://www.apache.org/licenses/
 *
 *    TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *    1. Definitions.
 *
 *       "License" shall mean the terms and conditions for use, reproduction,
 *       and distribution as defined by Sections 1 through 9 of this document.
 *
 *       "Licensor" shall mean the copyright owner or entity authorized by
 *       the copyright owner that is granting the License.
 *
 *       "Legal Entity" shall mean the union of the acting entity and all
 *       other entities that control, are controlled by, or are under common
 *       control with that entity. For the purposes of this definition,
 *       "control" means (i) the power, direct or indirect, to cause the
 *       direction or management of such entity, whether by contract or
 *       otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *       outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *       "You" (or "Your") shall mean an individual or Legal Entity
 *       exercising permissions granted by this License.
 *
 *       "Source" form shall mean the preferred form for making modifications,
 *       including but not limited to software source code, documentation
 *       source, and configuration files.
 *
 *       "Object" form shall mean any form resulting from mechanical
 *       transformation or translation of a Source form, including but
 *       not limited to compiled object code, generated documentation,
 *       and conversions to other media types.
 *
 *       "Work" shall mean the work of authorship, whether in Source or
 *       Object form, made available under the License, as indicated by a
 *       copyright notice that is included in or attached to the work
 *       (an example is provided in the Appendix below).
 *
 *       "Derivative Works" shall mean any work, whether in Source or Object
 *       form, that is based on (or derived from) the Work and for which the
 *       editorial revisions, annotations, elaborations, or other modifications
 *       represent, as a whole, an original work of authorship. For the purposes
 *       of this License, Derivative Works shall not include works that remain
 *       separable from, or merely link (or bind by name) to the interfaces of,
 *       the Work and Derivative Works thereof.
 *
 *       "Contribution" shall mean any work of authorship, including
 *       the original version of the Work and any modifications or additions
 *       to that Work or Derivative Works thereof, that is intentionally
 *       submitted to Licensor for inclusion in the Work by the copyright owner
 *       or by an individual or Legal Entity authorized to submit on behalf of
 *       the copyright owner. For the purposes of this definition, "submitted"
 *       means any form of electronic, verbal, or written communication sent
 *       to the Licensor or its representatives, including but not limited to
 *       communication on electronic mailing lists, source code control systems,
 *       and issue tracking systems that are managed by, or on behalf of, the
 *       Licensor for the purpose of discussing and improving the Work, but
 *       excluding communication that is conspicuously marked or otherwise
 *       designated in writing by the copyright owner as "Not a Contribution."
 *
 *       "Contributor" shall mean Licensor and any individual or Legal Entity
 *       on behalf of whom a Contribution has been received by Licensor and
 *       subsequently incorporated within the Work.
 *
 *    2. Grant of Copyright License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       copyright license to reproduce, prepare Derivative Works of,
 *       publicly display, publicly perform, sublicense, and distribute the
 *       Work and such Derivative Works in Source or Object form.
 *
 *    3. Grant of Patent License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       (except as stated in this section) patent license to make, have made,
 *       use, offer to sell, sell, import, and otherwise transfer the Work,
 *       where such license applies only to those patent claims licensable
 *       by such Contributor that are necessarily infringed by their
 *       Contribution(s) alone or by combination of their Contribution(s)
 *       with the Work to which such Contribution(s) was submitted. If You
 *       institute patent litigation against any entity (including a
 *       cross-claim or counterclaim in a lawsuit) alleging that the Work
 *       or a Contribution incorporated within the Work constitutes direct
 *       or contributory patent infringement, then any patent licenses
 *       granted to You under this License for that Work shall terminate
 *       as of the date such litigation is filed.
 *
 *    4. Redistribution. You may reproduce and distribute copies of the
 *       Work or Derivative Works thereof in any medium, with or without
 *       modifications, and in Source or Object form, provided that You
 *       meet the following conditions:
 *
 *       (a) You must give any other recipients of the Work or
 *           Derivative Works a copy of this License; and
 *
 *       (b) You must cause any modified files to carry prominent notices
 *           stating that You changed the files; and
 *
 *       (c) You must retain, in the Source form of any Derivative Works
 *           that You distribute, all copyright, patent, trademark, and
 *           attribution notices from the Source form of the Work,
 *           excluding those notices that do not pertain to any part of
 *           the Derivative Works; and
 *
 *       (d) If the Work includes a "NOTICE" text file as part of its
 *           distribution, then any Derivative Works that You distribute must
 *           include a readable copy of the attribution notices contained
 *           within such NOTICE file, excluding those notices that do not
 *           pertain to any part of the Derivative Works, in at least one
 *           of the following places: within a NOTICE text file distributed
 *           as part of the Derivative Works; within the Source form or
 *           documentation, if provided along with the Derivative Works; or,
 *           within a display generated by the Derivative Works, if and
 *           wherever such third-party notices normally appear. The contents
 *           of the NOTICE file are for informational purposes only and
 *           do not modify the License. You may add Your own attribution
 *           notices within Derivative Works that You distribute, alongside
 *           or as an addendum to the NOTICE text from the Work, provided
 *           that such additional attribution notices cannot be construed
 *           as modifying the License.
 *
 *       You may add Your own copyright statement to Your modifications and
 *       may provide additional or different license terms and conditions
 *       for use, reproduction, or distribution of Your modifications, or
 *       for any such Derivative Works as a whole, provided Your use,
 *       reproduction, and distribution of the Work otherwise complies with
 *       the conditions stated in this License.
 *
 *    5. Submission of Contributions. Unless You explicitly state otherwise,
 *       any Contribution intentionally submitted for inclusion in the Work
 *       by You to the Licensor shall be under the terms and conditions of
 *       this License, without any additional terms or conditions.
 *       Notwithstanding the above, nothing herein shall supersede or modify
 *       the terms of any separate license agreement you may have executed
 *       with Licensor regarding such Contributions.
 *
 *    6. Trademarks. This License does not grant permission to use the trade
 *       names, trademarks, service marks, or product names of the Licensor,
 *       except as required for reasonable and customary use in describing the
 *       origin of the Work and reproducing the content of the NOTICE file.
 *
 *    7. Disclaimer of Warranty. Unless required by applicable law or
 *       agreed to in writing, Licensor provides the Work (and each
 *       Contributor provides its Contributions) on an "AS IS" BASIS,
 *       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *       implied, including, without limitation, any warranties or conditions
 *       of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *       PARTICULAR PURPOSE. You are solely responsible for determining the
 *       appropriateness of using or redistributing the Work and assume any
 *       risks associated with Your exercise of permissions under this License.
 *
 *    8. Limitation of Liability. In no event and under no legal theory,
 *       whether in tort (including negligence), contract, or otherwise,
 *       unless required by applicable law (such as deliberate and grossly
 *       negligent acts) or agreed to in writing, shall any Contributor be
 *       liable to You for damages, including any direct, indirect, special,
 *       incidental, or consequential damages of any character arising as a
 *       result of this License or out of the use or inability to use the
 *       Work (including but not limited to damages for loss of goodwill,
 *       work stoppage, computer failure or malfunction, or any and all
 *       other commercial damages or losses), even if such Contributor
 *       has been advised of the possibility of such damages.
 *
 *    9. Accepting Warranty or Additional Liability. While redistributing
 *       the Work or Derivative Works thereof, You may choose to offer,
 *       and charge a fee for, acceptance of support, warranty, indemnity,
 *       or other liability obligations and/or rights consistent with this
 *       License. However, in accepting such obligations, You may act only
 *       on Your own behalf and on Your sole responsibility, not on behalf
 *       of any other Contributor, and only if You agree to indemnify,
 *       defend, and hold each Contributor harmless for any liability
 *       incurred by, or claims asserted against, such Contributor by reason
 *       of your accepting any such warranty or additional liability.
 *
 *    END OF TERMS AND CONDITIONS
 *
 *    APPENDIX: How to apply the Apache License to your work.
 *
 *       To apply the Apache License to your work, attach the following
 *       boilerplate notice, with the fields enclosed by brackets "{}"
 *       replaced with your own identifying information. (Don't include
 *       the brackets!)  The text should be enclosed in the appropriate
 *       comment syntax for the file format. We also recommend that a
 *       file or class name and description of purpose be included on the
 *       same "printed page" as the copyright notice for easier
 *       identification within third-party archives.
 *
 *    Copyright 2014 Edgar Espina
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package org.jooby.couchbase;

import java.util.List;

import com.couchbase.client.core.message.kv.MutationToken;
import com.couchbase.client.java.AsyncBucket;
import com.couchbase.client.java.PersistTo;
import com.couchbase.client.java.ReplicaMode;
import com.couchbase.client.java.ReplicateTo;
import com.couchbase.client.java.document.Document;
import com.couchbase.client.java.document.JsonDocument;
import com.couchbase.client.java.error.TemporaryLockFailureException;
import com.couchbase.client.java.query.N1qlQuery;
import com.couchbase.client.java.query.Statement;
import com.couchbase.client.java.repository.AsyncRepository;
import com.couchbase.client.java.repository.annotation.Id;
import com.couchbase.client.java.view.ViewQuery;

import rx.Observable;

/**
 * <h1>datastore</h1>
 * <p>
 * Create, read, update and delete entities from an {@link AsyncBucket}. It is similar to
 * {@link AsyncRepository} but less verbose or more ready to use, but also add support for
 * {@link ViewQuery} and {@link N1qlQuery}.
 * </p>
 *
 * <h2>design/implementation choices</h2>
 * <p>
 * In order to abstract developers from doing basic CRUD operations, the following design has been
 * made:
 * </p>
 *
 * <ul>
 * <li>IDs look like: <code>classname::id</code>.</li>
 * <li>ID is also stored within the document as an attribute</li>
 * <li>A <code>class</code> attribute is also stored within the document</li>
 * </ul>
 *
 * <p>
 * Here is an example of a document for a class: <code>model.Beer</code>:
 * </p>
 * <pre>{@code
 * {
 *   "model.Beer::678": {
 *     "name": "IPA",
 *     "id": 678,
 *     "class": "model.Beer"
 *   }
 * }
 * }</pre>
 *
 * <h2>ID selection</h2>
 * <p>
 * In order to mark a class field as document ID, you must:
 * </p>
 * <ul>
 * <li>Name the field: <code>id</code>, or</li>
 * <li>Annotated the field with {@link Id}</li>
 * </ul>
 *
 * <p>
 * Auto-increment ID are supported by annotating the field with {@link GeneratedValue} and declare
 * the field as {@link Long}. Then whenever you make a call to {@link #insert(Object)} or
 * {@link #upsert(Object)} a new ID will be generated if need it.
 * </p>
 *
 * @author edgar
 * @since 1.0.0.CR7
 */
public interface AsyncDatastore {

  /**
   * Result of {@link AsyncDatastore#query(ViewQuery)} contains a list of entities plus totalRows in
   * the view.
   *
   * @author edgar
   * @param <T> Entity type.
   */
  class AsyncViewQueryResult<T> {

    /** Total number of rows in the view. */
    private final int totalRows;

    /** List of rows from current execution. */
    private final Observable<List<T>> rows;

    /**
     * Creates a new {@link AsyncViewQueryResult}.
     *
     * @param totalRows Total number of rows in the view.
     * @param rows Resultset from current execution.
     */
    public AsyncViewQueryResult(final int totalRows, final Observable<List<T>> rows) {
      this.totalRows = totalRows;
      this.rows = rows;
    }

    /**
     * @return Total number of rows in the view.
     */
    public int getTotalRows() {
      return totalRows;
    }

    /**
     * @return Resultset from current execution.
     */
    public Observable<List<T>> getRows() {
      return rows;
    }

  }

  /**
   * Provides advanced options for couchbase operation.
   *
   * @author edgar
   */
  interface AsyncCommand {
    /**
     * Set the expiry/ttl entity option.
     *
     * @param expiry The expiration time expressed as relative seconds from now.
     * @return This command.
     */
    AsyncCommand expiry(int expiry);

    /**
     * Set a CAS value for the entity (0 if not set).
     *
     * @param cas Cas value.
     * @return This command.
     */
    AsyncCommand cas(long cas);

    /**
     * The optional, opaque mutation token set after a successful mutation and if enabled on
     * the environment.
     *
     * Note that the mutation token is always null, unless they are explicitly enabled on the
     * environment, the server version is supported (&gt;= 4.0.0) and the mutation operation succeeded.
     *
     * If set, it can be used for enhanced durability requirements, as well as optimized consistency
     * for N1QL queries.
     *
     * @param mutationToken the mutation token if set, otherwise null.
     * @return This command.
     */
    AsyncCommand mutationToken(MutationToken mutationToken);

    /**
     * Execute this command.
     *
     * @param entity Entity to use.
     * @param <R> Entity type.
     * @return Command result.
     */
    default <R> Observable<R> execute(final Object entity) {
      return execute(entity, PersistTo.NONE);
    }

    /**
     * Execute this command.
     *
     * @param entity Entity to use.
     * @param persistTo Persist to value.
     * @param <R> Entity type.
     * @return Command result.
     */
    default <R> Observable<R> execute(final Object entity, final PersistTo persistTo) {
      return execute(entity, persistTo, ReplicateTo.NONE);
    }

    /**
     * Execute this command.
     *
     * @param entity Entity to use.
     * @param replicateTo Replicate to value.
     * @param <R> Entity type.
     * @return Command result.
     */
    default <R> Observable<R> execute(final Object entity, final ReplicateTo replicateTo) {
      return execute(entity, PersistTo.NONE, replicateTo);
    }

    /**
     * Execute this command.
     *
     * @param entity Entity to use.
     * @param persistTo Persist to value.
     * @param replicateTo Replicate to value.
     * @param <R> Entity type.
     * @return Command result.
     */
    <R> Observable<R> execute(Object entity, PersistTo persistTo, ReplicateTo replicateTo);
  }

  /**
   * Provides advanced options for couchbase operation.
   *
   * @author edgar
   * @since 1.0.0.CR7
   */
  @SuppressWarnings("unchecked")
  interface AsyncRemoveCommand extends AsyncCommand {

    /**
     * Execute a remove document operation
     *
     * @param entity Entity to remove.
     * @param persistTo Persist to option.
     * @param replicateTo Replicate to option.
     * @return CAS value.
     */
    @Override
    Observable<Long> execute(Object entity, PersistTo persistTo, ReplicateTo replicateTo);

    /**
     * Execute a remove document operation
     *
     * @param entityClass Entity class to remove.
     * @param id Entity id to remove.
     * @return CAS value.
     */
    default Observable<Long> execute(final Class<?> entityClass, final Object id) {
      return execute(entityClass, id, PersistTo.NONE);
    }

    /**
     * Execute a remove document operation
     *
     * @param entityClass Entity class to remove.
     * @param id Entity id to remove.
     * @param persistTo Persist to option.
     * @return CAS value.
     */
    default Observable<Long> execute(final Class<?> entityClass, final Object id,
        final PersistTo persistTo) {
      return execute(entityClass, id, persistTo, ReplicateTo.NONE);
    }

    /**
     * Execute a remove document operation
     *
     * @param entityClass Entity class to remove.
     * @param id Entity id to remove.
     * @param replicateTo Replicate to option.
     * @return CAS value.
     */
    default Observable<Long> execute(final Class<?> entityClass, final Object id,
        final ReplicateTo replicateTo) {
      return execute(entityClass, id, PersistTo.NONE, replicateTo);
    }

    /**
     * Execute a remove document operation
     *
     * @param entityClass Entity class to remove.
     * @param id Entity id to remove.
     * @param persistTo Persist to option.
     * @param replicateTo Replicate to option.
     * @return CAS value.
     */
    Observable<Long> execute(final Class<?> entityClass, Object id, PersistTo persistTo,
        ReplicateTo replicateTo);
  }

  /**
   * Get an entity/document by ID. The unique ID is constructed via
   * {@link N1Q#qualifyId(Class, Object)}.
   *
   * @param entityClass Entity class.
   * @param id Entity id.
   * @param <T> Entity type.
   * @return An observable entity matching the id or an empty observable.
   */
  <T> Observable<T> get(Class<T> entityClass, Object id);

  /**
   * Get an entity/document by ID. The unique ID is constructed via
   * {@link N1Q#qualifyId(Class, Object)}.
   *
   * @param entityClass Entity class.
   * @param id Entity id.
   * @param mode Replica mode.
   * @param <T> Entity type.
   * @return An observable entity matching the id or an empty observable.
   */
  <T> Observable<T> getFromReplica(Class<T> entityClass, Object id, ReplicaMode mode);

  /**
   * Retrieve and lock a entity by its unique ID.
   *
   * If the entity is found, a entity is returned. If the entity is not found, the
   * {@link Observable} completes without an item emitted (empty).
   *
   * This method works similar to {@link #get(Class, Object)}, but in addition it (write) locks the
   * entity for the given lock time interval. Note that this lock time is hard capped to 30
   * seconds, even if provided with a higher value and is not configurable. The entity will unlock
   * afterwards automatically.
   *
   * Detecting an already locked entity is done by checking for
   * {@link TemporaryLockFailureException}. Note that this exception can also be raised in other
   * conditions, always when the error is transient and retrying may help.
   *
   * @param entityClass Entity class.
   * @param id id the unique ID of the entity.
   * @param lockTime the time to write lock the entity (max. 30 seconds).
   * @param <T> Entity type.
   * @return an {@link Observable} eventually containing the found {@link JsonDocument}.
   */
  <T> Observable<T> getAndLock(Class<T> entityClass, Object id, int lockTime);

  /**
   * Retrieve and touch an entity by its unique ID.
   *
   * If the entity is found, an entity is returned. If the entity is not found, the
   * {@link Observable} completes without an item emitted (empty).
   *
   * This method works similar to {@link #get(Class, Object)}, but in addition it touches the
   * entity, which will reset its configured expiration time to the value provided.
   *
   * @param entityClass Entity class.
   * @param id id the unique ID of the entity.
   * @param expiry the new expiration time for the entity (in seconds).
   * @param <T> Entity type.
   * @return an {@link Observable} eventually containing the found {@link JsonDocument}.
   */
  <T> Observable<T> getAndTouch(Class<T> entityClass, Object id, int expiry);

  /**
   * Check whether a entity with the given ID does exist in the bucket.
   *
   * @param entityClass Entity class.
   * @param id Entity id.
   * @return True, if exists.
   */
  Observable<Boolean> exists(Class<?> entityClass, Object id);

  /**
   * @return A new upsert command.
   */
  AsyncCommand upsert();

  /**
   * Insert or overwrite an entity. If the entity already has an ID, then that ID is selected. If
   * the entity doesn't have an ID and the field is annotated with {@link GeneratedValue} this
   * method will generate a new ID and insert the entity.
   *
   * @param entity Entity to insert or overwrite.
   * @param <T> Entity type.
   * @return Updated entity.
   */
  default <T> Observable<T> upsert(final T entity) {
    return upsert().execute(entity);
  }

  /**
   * @return A new insert command.
   */
  AsyncCommand insert();

  /**
   * Insert an entity. If the entity already has an ID, then that ID is selected. If
   * the entity doesn't have an ID and the field is annotated with {@link GeneratedValue} this
   * method will generate a new ID and insert the entity.
   *
   * @param entity Entity to insert.
   * @param <T> Entity type.
   * @return Updated entity.
   */
  default <T> Observable<T> insert(final T entity) {
    return insert().execute(entity);
  }

  /**
   * Replace an entity if it does exist and watch for durability constraints.
   *
   * @return A new replace command.
   */
  AsyncCommand replace();

  /**
   * Replace an entity if it does exist and watch for durability constraints.
   *
   * This method works exactly like {@link AsyncBucket#replace(Document)}, but afterwards watches
   * the server states if the given durability constraints are met. If this is the case, a new
   * document is returned which contains the original properties, but has the refreshed CAS value
   * set.
   *
   * @param entity An entity to replace.
   * @param <T> Entity type.
   * @return The replace entity.
   */
  default <T> Observable<T> replace(final T entity) {
    return replace().execute(entity);
  }

  /**
   * Removes an entity from the Server.
   *
   * The an entity returned just has the document ID and its CAS value set, since the value and all
   * other associated properties have been removed from the server.
   *
   * @return A new remove command.
   */
  AsyncRemoveCommand remove();

  /**
   * Removes an entity from the Server.
   *
   * The an entity returned just has the document ID and its CAS value set, since the value and all
   * other associated properties have been removed from the server.
   *
   * @param entity Entity to remove.
   * @return The cas value.
   */
  default Observable<Long> remove(final Object entity) {
    return remove().execute(entity);
  }

  /**
   * Removes an entity from the Server.
   *
   * The an entity returned just has the document ID and its CAS value set, since the value and all
   * other associated properties have been removed from the server.
   *
   * @param entityClass Entity class to remove.
   * @param id Entity id.
   * @return The cas value.
   */
  default Observable<Long> remove(final Class<?> entityClass, final Object id) {
    return remove().execute(entityClass, id);
  }

  /**
   * Run a {@link N1qlQuery#simple(Statement)} query.
   *
   * @param statement Statement.
   * @param <T> Entity type.
   * @return A list of results.
   * @see N1Q#from(Class)
   */
  default <T> Observable<List<T>> query(final Statement statement) {
    return query(N1qlQuery.simple(statement));
  }

  /**
   * Run a {@link N1qlQuery} query.
   *
   * @param query Query.
   * @param <T> Entity type.
   * @return A list of results.
   * @see N1Q#from(Class)
   */
  <T> Observable<List<T>> query(N1qlQuery query);

  /**
   * Run a {@link ViewQuery} query.
   *
   * @param query View query.
   * @param <T> Entity type.
   * @return Results.
   */
  <T> Observable<AsyncViewQueryResult<T>> query(ViewQuery query);

}
