/**
 * Apache License
 * Version 2.0, January 2004
 * http://www.apache.org/licenses/
 *
 * TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 * 1. Definitions.
 *
 * "License" shall mean the terms and conditions for use, reproduction,
 * and distribution as defined by Sections 1 through 9 of this document.
 *
 * "Licensor" shall mean the copyright owner or entity authorized by
 * the copyright owner that is granting the License.
 *
 * "Legal Entity" shall mean the union of the acting entity and all
 * other entities that control, are controlled by, or are under common
 * control with that entity. For the purposes of this definition,
 * "control" means (i) the power, direct or indirect, to cause the
 * direction or management of such entity, whether by contract or
 * otherwise, or (ii) ownership of fifty percent (50%) or more of the
 * outstanding shares, or (iii) beneficial ownership of such entity.
 *
 * "You" (or "Your") shall mean an individual or Legal Entity
 * exercising permissions granted by this License.
 *
 * "Source" form shall mean the preferred form for making modifications,
 * including but not limited to software source code, documentation
 * source, and configuration files.
 *
 * "Object" form shall mean any form resulting from mechanical
 * transformation or translation of a Source form, including but
 * not limited to compiled object code, generated documentation,
 * and conversions to other media types.
 *
 * "Work" shall mean the work of authorship, whether in Source or
 * Object form, made available under the License, as indicated by a
 * copyright notice that is included in or attached to the work
 * (an example is provided in the Appendix below).
 *
 * "Derivative Works" shall mean any work, whether in Source or Object
 * form, that is based on (or derived from) the Work and for which the
 * editorial revisions, annotations, elaborations, or other modifications
 * represent, as a whole, an original work of authorship. For the purposes
 * of this License, Derivative Works shall not include works that remain
 * separable from, or merely link (or bind by name) to the interfaces of,
 * the Work and Derivative Works thereof.
 *
 * "Contribution" shall mean any work of authorship, including
 * the original version of the Work and any modifications or additions
 * to that Work or Derivative Works thereof, that is intentionally
 * submitted to Licensor for inclusion in the Work by the copyright owner
 * or by an individual or Legal Entity authorized to submit on behalf of
 * the copyright owner. For the purposes of this definition, "submitted"
 * means any form of electronic, verbal, or written communication sent
 * to the Licensor or its representatives, including but not limited to
 * communication on electronic mailing lists, source code control systems,
 * and issue tracking systems that are managed by, or on behalf of, the
 * Licensor for the purpose of discussing and improving the Work, but
 * excluding communication that is conspicuously marked or otherwise
 * designated in writing by the copyright owner as "Not a Contribution."
 *
 * "Contributor" shall mean Licensor and any individual or Legal Entity
 * on behalf of whom a Contribution has been received by Licensor and
 * subsequently incorporated within the Work.
 *
 * 2. Grant of Copyright License. Subject to the terms and conditions of
 * this License, each Contributor hereby grants to You a perpetual,
 * worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 * copyright license to reproduce, prepare Derivative Works of,
 * publicly display, publicly perform, sublicense, and distribute the
 * Work and such Derivative Works in Source or Object form.
 *
 * 3. Grant of Patent License. Subject to the terms and conditions of
 * this License, each Contributor hereby grants to You a perpetual,
 * worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 * (except as stated in this section) patent license to make, have made,
 * use, offer to sell, sell, import, and otherwise transfer the Work,
 * where such license applies only to those patent claims licensable
 * by such Contributor that are necessarily infringed by their
 * Contribution(s) alone or by combination of their Contribution(s)
 * with the Work to which such Contribution(s) was submitted. If You
 * institute patent litigation against any entity (including a
 * cross-claim or counterclaim in a lawsuit) alleging that the Work
 * or a Contribution incorporated within the Work constitutes direct
 * or contributory patent infringement, then any patent licenses
 * granted to You under this License for that Work shall terminate
 * as of the date such litigation is filed.
 *
 * 4. Redistribution. You may reproduce and distribute copies of the
 * Work or Derivative Works thereof in any medium, with or without
 * modifications, and in Source or Object form, provided that You
 * meet the following conditions:
 *
 * (a) You must give any other recipients of the Work or
 * Derivative Works a copy of this License; and
 *
 * (b) You must cause any modified files to carry prominent notices
 * stating that You changed the files; and
 *
 * (c) You must retain, in the Source form of any Derivative Works
 * that You distribute, all copyright, patent, trademark, and
 * attribution notices from the Source form of the Work,
 * excluding those notices that do not pertain to any part of
 * the Derivative Works; and
 *
 * (d) If the Work includes a "NOTICE" text file as part of its
 * distribution, then any Derivative Works that You distribute must
 * include a readable copy of the attribution notices contained
 * within such NOTICE file, excluding those notices that do not
 * pertain to any part of the Derivative Works, in at least one
 * of the following places: within a NOTICE text file distributed
 * as part of the Derivative Works; within the Source form or
 * documentation, if provided along with the Derivative Works; or,
 * within a display generated by the Derivative Works, if and
 * wherever such third-party notices normally appear. The contents
 * of the NOTICE file are for informational purposes only and
 * do not modify the License. You may add Your own attribution
 * notices within Derivative Works that You distribute, alongside
 * or as an addendum to the NOTICE text from the Work, provided
 * that such additional attribution notices cannot be construed
 * as modifying the License.
 *
 * You may add Your own copyright statement to Your modifications and
 * may provide additional or different license terms and conditions
 * for use, reproduction, or distribution of Your modifications, or
 * for any such Derivative Works as a whole, provided Your use,
 * reproduction, and distribution of the Work otherwise complies with
 * the conditions stated in this License.
 *
 * 5. Submission of Contributions. Unless You explicitly state otherwise,
 * any Contribution intentionally submitted for inclusion in the Work
 * by You to the Licensor shall be under the terms and conditions of
 * this License, without any additional terms or conditions.
 * Notwithstanding the above, nothing herein shall supersede or modify
 * the terms of any separate license agreement you may have executed
 * with Licensor regarding such Contributions.
 *
 * 6. Trademarks. This License does not grant permission to use the trade
 * names, trademarks, service marks, or product names of the Licensor,
 * except as required for reasonable and customary use in describing the
 * origin of the Work and reproducing the content of the NOTICE file.
 *
 * 7. Disclaimer of Warranty. Unless required by applicable law or
 * agreed to in writing, Licensor provides the Work (and each
 * Contributor provides its Contributions) on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 * implied, including, without limitation, any warranties or conditions
 * of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 * PARTICULAR PURPOSE. You are solely responsible for determining the
 * appropriateness of using or redistributing the Work and assume any
 * risks associated with Your exercise of permissions under this License.
 *
 * 8. Limitation of Liability. In no event and under no legal theory,
 * whether in tort (including negligence), contract, or otherwise,
 * unless required by applicable law (such as deliberate and grossly
 * negligent acts) or agreed to in writing, shall any Contributor be
 * liable to You for damages, including any direct, indirect, special,
 * incidental, or consequential damages of any character arising as a
 * result of this License or out of the use or inability to use the
 * Work (including but not limited to damages for loss of goodwill,
 * work stoppage, computer failure or malfunction, or any and all
 * other commercial damages or losses), even if such Contributor
 * has been advised of the possibility of such damages.
 *
 * 9. Accepting Warranty or Additional Liability. While redistributing
 * the Work or Derivative Works thereof, You may choose to offer,
 * and charge a fee for, acceptance of support, warranty, indemnity,
 * or other liability obligations and/or rights consistent with this
 * License. However, in accepting such obligations, You may act only
 * on Your own behalf and on Your sole responsibility, not on behalf
 * of any other Contributor, and only if You agree to indemnify,
 * defend, and hold each Contributor harmless for any liability
 * incurred by, or claims asserted against, such Contributor by reason
 * of your accepting any such warranty or additional liability.
 *
 * END OF TERMS AND CONDITIONS
 *
 * APPENDIX: How to apply the Apache License to your work.
 *
 * To apply the Apache License to your work, attach the following
 * boilerplate notice, with the fields enclosed by brackets "{}"
 * replaced with your own identifying information. (Don't include
 * the brackets!)  The text should be enclosed in the appropriate
 * comment syntax for the file format. We also recommend that a
 * file or class name and description of purpose be included on the
 * same "printed page" as the copyright notice for easier
 * identification within third-party archives.
 *
 * Copyright 2014 Edgar Espina
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jooby.frontend;

import com.github.eirslett.maven.plugins.frontend.lib.CacheResolver;
import com.github.eirslett.maven.plugins.frontend.lib.FrontendPluginFactory;
import com.github.eirslett.maven.plugins.frontend.lib.InstallationException;
import com.github.eirslett.maven.plugins.frontend.lib.NodeInstaller;
import com.github.eirslett.maven.plugins.frontend.lib.ProxyConfig;
import com.google.common.hash.Hashing;
import com.google.inject.Binder;
import com.typesafe.config.Config;
import com.typesafe.config.ConfigFactory;
import static java.util.Objects.requireNonNull;
import org.jooby.Env;
import org.jooby.Jooby;
import org.jooby.funzy.Throwing;
import org.jooby.funzy.Try;

import static org.jooby.funzy.Throwing.throwingConsumer;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

/**
 * <h1>frontend</h1>
 * <p>
 * Run <a href="https://webpack.js.org/">webpack</a>,
 * <a href="https://github.com/rollup/rollup">rollup</a>, <a href="https://gulpjs.com/">gulp</a>
 * , etc...
 * </p>
 * <p>
 * All this is possible thanks to the
 * <a href="https://github.com/eirslett/frontend-maven-plugin">frontend</a> library, which
 * download and install <a href="https://nodejs.org">node.js</a> locally at your project root.
 * This simplify development and make sure the build works in any environment.
 * </p>
 *
 * <h2>usage</h2>
 * <p>
 * Using <a href="https://www.npmjs.com">npm</a>:
 * </p>
 * <pre>{@code
 * {
 *   on("dev", () -> {
 *     use(new Npm("v8.6.0"));
 *   });
 * }
 * }</pre>
 *
 * <p>
 * Using <a href="https://yarnpkg.com">yarn</a>:
 * </p>
 *
 * <pre>{@code
 * {
 *   on("dev", () -> {
 *     use(new Yarn("v8.6.0", "v1.1.0"));
 *   });
 * }
 * }</pre>
 *
 * <p>
 * <strong>NOTE</strong>: The module must be used in development only. That's why we wrap the
 * module installation using the environment predicate. In order to build a production version,
 * you need to configure a build time process.
 * The <a href="https://github.com/eirslett/frontend-maven-plugin">frontend</a> maven plugin does
 * this job for maven projects.
 * If you are a Gradle user, you might want to try the
 * <a href="https://github.com/srs/gradle-node-plugin">gradle-node-plugin</a>.
 * </p>
 *
 * <h2>install phase</h2>
 * <p>
 * The module automatically sync and install dependencies at startup time.
 * </p>
 *
 * <h2>default task</h2>
 * <p>
 * The module runs <code>npm run build</code> or <code>yarn run build</code> at startup time
 * (after install phase).
 * </p>
 * <p>
 * This mean your <code>package.json</code> must defines a <code>build</code> script, like:
 * </p>
 *
 * <pre>
 * {
 *   "scripts": {
 *     "build": "webpack"
 *   }
 * }
 * </pre>
 *
 * <p>The default task is always executed, unless you define one or more custom tasks (see next).</p>
 *
 * <h2>custom task</h2>
 * <p>
 * Execution of arbitrary scripts is available via {@link #onStart(Throwing.Consumer)} and
 * {@link #onStarted(Throwing.Consumer)}.
 * </p>
 *
 * <p>
 * The next example executes two scripts:
 * </p>
 *
 * <pre>{@code
 * {
 *
 *   use(new Npm("v8.6.0"))
 *     .onStart(npm -> {
 *       npm.executeSync("run", "local");
 *       npm.executeSync("run", "next");
 *     });
 *   );
 * }
 * }</pre>
 *
 * <p>
 * The next example execute a script without waiting to finish (useful to run background task):
 * </p>
 *
 * <pre>{@code
 * {
 *
 *   use(new Npm("v8.6.0"))
 *     .onStart(npm -> {
 *       npm.execute("run", "webserver");
 *     });
 *   );
 * }
 * }</pre>
 *
 * @author edgar
 * @since 1.2.0
 */
public abstract class Frontend implements Jooby.Module {

  private Path workDirectory;

  private Path installDirectory;

  private Throwing.Consumer<NodeTask> onStart;

  private Throwing.Consumer<NodeTask> onStarted;

  private final String nodeVersion;

  public Frontend(String nodeVersion) {
    this.nodeVersion = requireNonNull(nodeVersion, "Node version required.");
  }

  /**
   * Set working directory (location of package.json).
   *
   * @param workDirectory Work directory.
   * @return This module.
   */
  public Frontend workDirectory(Path workDirectory) {
    this.workDirectory = requireNonNull(workDirectory, "Work directory required.");
    return this;
  }

  /**
   * Set install directory where to save the node executable (node/node).
   *
   * @param installDirectory Location where to save the node executable (node/node).
   * @return This module.
   */
  public Frontend installDirectory(Path installDirectory) {
    this.installDirectory = requireNonNull(installDirectory, "Install directory required.");
    return this;
  }

  /**
   * Run a node task at application start up time.
   *
   * @param callback Callback.
   * @return This module.
   */
  public Frontend onStart(Throwing.Consumer<NodeTask> callback) {
    this.onStart = callback;
    return this;
  }

  /**
   * Run a node task after application has been initialized and is ready to serve.
   *
   * @param callback Callback.
   * @return This module.
   */
  public Frontend onStarted(Throwing.Consumer<NodeTask> callback) {
    this.onStarted = callback;
    return this;
  }

  @Override public void configure(final Env env, final Config conf, final Binder binder)
      throws Throwable {
    Path userdir = Paths.get(System.getProperty("user.dir"));
    Path workDirectory = Optional.ofNullable(this.workDirectory)
        .orElse(userdir);

    Path installDirectory = Optional.ofNullable(this.installDirectory).orElse(workDirectory);

    ProxyConfig proxy = new ProxyConfig(proxies(conf));

    CacheResolver cache = Files.exists(userdir.resolve("pom.xml"))
        ? new MavenCacheResolver()
        : null;

    FrontendPluginFactory factory = new FrontendPluginFactory(workDirectory.toFile(),
        installDirectory.toFile(), cache);

    installNode(conf, nodeVersion, proxy, factory);

    Map<String, String> environment = environment(conf);

    NodeTask task = newTask(factory, conf, proxy, environment, nodeVersion);

    onSyncPackageJson(conf, workDirectory, task::executeSync);

    if (onStart == null && onStarted == null) {
      onStart = cmd -> cmd.execute("run", "build");
    }
    if (onStart != null) {
      env.onStart(() -> onStart.accept(task));
    }
    if (onStarted != null) {
      env.onStarted(() -> onStarted.accept(task));
    }
  }

  private Map<String, String> environment(Config conf) {
    Map<String, String> vars = new HashMap<>(System.getenv());
    conf.entrySet().forEach(e -> vars.put(e.getKey(), e.getValue().unwrapped().toString()));
    vars.put("NODE_ENV",
        conf.getString("application.env").equals("dev") ? "development" : "production");
    return vars;
  }

  private void onSyncPackageJson(Config conf, Path workDirectory, Throwing.Consumer<String> action)
      throws IOException {
    Path tmp = Paths.get(conf.getString("application.tmpdir"), "package.json");
    Files.createDirectories(tmp);
    String sha1 = Hashing.sha256()
        .hashBytes(Files.readAllBytes(workDirectory.resolve("package.json")))
        .toString();
    Path lastSha1 = tmp.resolve(sha1);
    if (!Files.exists(lastSha1) || !Files.exists(workDirectory.resolve("node_modules"))) {
      action.accept("install");
      Try.of(Files.walk(tmp))
          .run(files -> files.filter(f -> !f.equals(tmp)).forEach(throwingConsumer(Files::deleteIfExists)));
      Files.write(tmp.resolve(lastSha1), Arrays.asList(""));
    }
  }

  private void installNode(Config conf, String version, ProxyConfig proxy,
      FrontendPluginFactory factory) throws InstallationException {
    NodeInstaller installer = factory.getNodeInstaller(proxy)
        .setNodeVersion(version)
        .setNodeDownloadRoot(conf.getString("node.downloadRoot"));

    installNpm(installer);
    installer.install();
  }

  protected void installNpm(NodeInstaller node) {
    // NOOP
  }

  @Override public Config config() {
    return ConfigFactory.parseResources(getClass(), "frontend.conf");
  }

  private List<ProxyConfig.Proxy> proxies(Config conf) {
    if (conf.hasPath("proxy")) {
      return Arrays.asList(new ProxyConfig.Proxy(conf.getString("proxy.id"),
          conf.getString("proxy.protocol"),
          conf.getString("proxy.host"),
          conf.getInt("proxy.port"),
          conf.getString("proxy.username"),
          conf.getString("proxy.password"),
          conf.getString("proxy.nonProxyHosts")));
    }
    return Collections.emptyList();
  }

  protected abstract NodeTask newTask(FrontendPluginFactory factory, Config conf,
      ProxyConfig proxy, Map<String, String> env, String nodeVersion) throws InstallationException;

}
