/**
 *                                  Apache License
 *                            Version 2.0, January 2004
 *                         http://www.apache.org/licenses/
 *
 *    TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *    1. Definitions.
 *
 *       "License" shall mean the terms and conditions for use, reproduction,
 *       and distribution as defined by Sections 1 through 9 of this document.
 *
 *       "Licensor" shall mean the copyright owner or entity authorized by
 *       the copyright owner that is granting the License.
 *
 *       "Legal Entity" shall mean the union of the acting entity and all
 *       other entities that control, are controlled by, or are under common
 *       control with that entity. For the purposes of this definition,
 *       "control" means (i) the power, direct or indirect, to cause the
 *       direction or management of such entity, whether by contract or
 *       otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *       outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *       "You" (or "Your") shall mean an individual or Legal Entity
 *       exercising permissions granted by this License.
 *
 *       "Source" form shall mean the preferred form for making modifications,
 *       including but not limited to software source code, documentation
 *       source, and configuration files.
 *
 *       "Object" form shall mean any form resulting from mechanical
 *       transformation or translation of a Source form, including but
 *       not limited to compiled object code, generated documentation,
 *       and conversions to other media types.
 *
 *       "Work" shall mean the work of authorship, whether in Source or
 *       Object form, made available under the License, as indicated by a
 *       copyright notice that is included in or attached to the work
 *       (an example is provided in the Appendix below).
 *
 *       "Derivative Works" shall mean any work, whether in Source or Object
 *       form, that is based on (or derived from) the Work and for which the
 *       editorial revisions, annotations, elaborations, or other modifications
 *       represent, as a whole, an original work of authorship. For the purposes
 *       of this License, Derivative Works shall not include works that remain
 *       separable from, or merely link (or bind by name) to the interfaces of,
 *       the Work and Derivative Works thereof.
 *
 *       "Contribution" shall mean any work of authorship, including
 *       the original version of the Work and any modifications or additions
 *       to that Work or Derivative Works thereof, that is intentionally
 *       submitted to Licensor for inclusion in the Work by the copyright owner
 *       or by an individual or Legal Entity authorized to submit on behalf of
 *       the copyright owner. For the purposes of this definition, "submitted"
 *       means any form of electronic, verbal, or written communication sent
 *       to the Licensor or its representatives, including but not limited to
 *       communication on electronic mailing lists, source code control systems,
 *       and issue tracking systems that are managed by, or on behalf of, the
 *       Licensor for the purpose of discussing and improving the Work, but
 *       excluding communication that is conspicuously marked or otherwise
 *       designated in writing by the copyright owner as "Not a Contribution."
 *
 *       "Contributor" shall mean Licensor and any individual or Legal Entity
 *       on behalf of whom a Contribution has been received by Licensor and
 *       subsequently incorporated within the Work.
 *
 *    2. Grant of Copyright License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       copyright license to reproduce, prepare Derivative Works of,
 *       publicly display, publicly perform, sublicense, and distribute the
 *       Work and such Derivative Works in Source or Object form.
 *
 *    3. Grant of Patent License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       (except as stated in this section) patent license to make, have made,
 *       use, offer to sell, sell, import, and otherwise transfer the Work,
 *       where such license applies only to those patent claims licensable
 *       by such Contributor that are necessarily infringed by their
 *       Contribution(s) alone or by combination of their Contribution(s)
 *       with the Work to which such Contribution(s) was submitted. If You
 *       institute patent litigation against any entity (including a
 *       cross-claim or counterclaim in a lawsuit) alleging that the Work
 *       or a Contribution incorporated within the Work constitutes direct
 *       or contributory patent infringement, then any patent licenses
 *       granted to You under this License for that Work shall terminate
 *       as of the date such litigation is filed.
 *
 *    4. Redistribution. You may reproduce and distribute copies of the
 *       Work or Derivative Works thereof in any medium, with or without
 *       modifications, and in Source or Object form, provided that You
 *       meet the following conditions:
 *
 *       (a) You must give any other recipients of the Work or
 *           Derivative Works a copy of this License; and
 *
 *       (b) You must cause any modified files to carry prominent notices
 *           stating that You changed the files; and
 *
 *       (c) You must retain, in the Source form of any Derivative Works
 *           that You distribute, all copyright, patent, trademark, and
 *           attribution notices from the Source form of the Work,
 *           excluding those notices that do not pertain to any part of
 *           the Derivative Works; and
 *
 *       (d) If the Work includes a "NOTICE" text file as part of its
 *           distribution, then any Derivative Works that You distribute must
 *           include a readable copy of the attribution notices contained
 *           within such NOTICE file, excluding those notices that do not
 *           pertain to any part of the Derivative Works, in at least one
 *           of the following places: within a NOTICE text file distributed
 *           as part of the Derivative Works; within the Source form or
 *           documentation, if provided along with the Derivative Works; or,
 *           within a display generated by the Derivative Works, if and
 *           wherever such third-party notices normally appear. The contents
 *           of the NOTICE file are for informational purposes only and
 *           do not modify the License. You may add Your own attribution
 *           notices within Derivative Works that You distribute, alongside
 *           or as an addendum to the NOTICE text from the Work, provided
 *           that such additional attribution notices cannot be construed
 *           as modifying the License.
 *
 *       You may add Your own copyright statement to Your modifications and
 *       may provide additional or different license terms and conditions
 *       for use, reproduction, or distribution of Your modifications, or
 *       for any such Derivative Works as a whole, provided Your use,
 *       reproduction, and distribution of the Work otherwise complies with
 *       the conditions stated in this License.
 *
 *    5. Submission of Contributions. Unless You explicitly state otherwise,
 *       any Contribution intentionally submitted for inclusion in the Work
 *       by You to the Licensor shall be under the terms and conditions of
 *       this License, without any additional terms or conditions.
 *       Notwithstanding the above, nothing herein shall supersede or modify
 *       the terms of any separate license agreement you may have executed
 *       with Licensor regarding such Contributions.
 *
 *    6. Trademarks. This License does not grant permission to use the trade
 *       names, trademarks, service marks, or product names of the Licensor,
 *       except as required for reasonable and customary use in describing the
 *       origin of the Work and reproducing the content of the NOTICE file.
 *
 *    7. Disclaimer of Warranty. Unless required by applicable law or
 *       agreed to in writing, Licensor provides the Work (and each
 *       Contributor provides its Contributions) on an "AS IS" BASIS,
 *       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *       implied, including, without limitation, any warranties or conditions
 *       of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *       PARTICULAR PURPOSE. You are solely responsible for determining the
 *       appropriateness of using or redistributing the Work and assume any
 *       risks associated with Your exercise of permissions under this License.
 *
 *    8. Limitation of Liability. In no event and under no legal theory,
 *       whether in tort (including negligence), contract, or otherwise,
 *       unless required by applicable law (such as deliberate and grossly
 *       negligent acts) or agreed to in writing, shall any Contributor be
 *       liable to You for damages, including any direct, indirect, special,
 *       incidental, or consequential damages of any character arising as a
 *       result of this License or out of the use or inability to use the
 *       Work (including but not limited to damages for loss of goodwill,
 *       work stoppage, computer failure or malfunction, or any and all
 *       other commercial damages or losses), even if such Contributor
 *       has been advised of the possibility of such damages.
 *
 *    9. Accepting Warranty or Additional Liability. While redistributing
 *       the Work or Derivative Works thereof, You may choose to offer,
 *       and charge a fee for, acceptance of support, warranty, indemnity,
 *       or other liability obligations and/or rights consistent with this
 *       License. However, in accepting such obligations, You may act only
 *       on Your own behalf and on Your sole responsibility, not on behalf
 *       of any other Contributor, and only if You agree to indemnify,
 *       defend, and hold each Contributor harmless for any liability
 *       incurred by, or claims asserted against, such Contributor by reason
 *       of your accepting any such warranty or additional liability.
 *
 *    END OF TERMS AND CONDITIONS
 *
 *    APPENDIX: How to apply the Apache License to your work.
 *
 *       To apply the Apache License to your work, attach the following
 *       boilerplate notice, with the fields enclosed by brackets "{}"
 *       replaced with your own identifying information. (Don't include
 *       the brackets!)  The text should be enclosed in the appropriate
 *       comment syntax for the file format. We also recommend that a
 *       file or class name and description of purpose be included on the
 *       same "printed page" as the copyright notice for easier
 *       identification within third-party archives.
 *
 *    Copyright 2014 Edgar Espina
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package org.jooby.livereload;

import com.google.common.base.CaseFormat;
import com.google.common.collect.ImmutableSet;
import com.google.inject.Binder;
import com.typesafe.config.Config;
import org.jooby.Env;
import org.jooby.Jooby.Module;
import org.jooby.MediaType;
import org.jooby.Request;
import org.jooby.Route;
import org.jooby.Router;
import org.jooby.WebSocket;
import org.jooby.filewatcher.FileWatcher;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.Predicate;

/**
 * <h1>liveReload</h1>
 * <p>
 * <a href="http://livereload.com">LiveReload</a> monitors changes in the file system. As soon as
 * you save a file, it is preprocessed as
 * needed, and the browser is refreshed.
 * </p>
 * <p>
 * Even cooler, when you change a CSS file or an image, the browser is updated instantly without
 * reloading the page.
 * </p>
 *
 * <h2>usage</h2>
 * <pre>{@code
 * {
 *   use(new Jackson());
 *
 *   use(new LiveReload());
 * }
 * }</pre>
 *
 * <p>
 * This module is available as long you run in development: <code>application.env=dev</code>.
 * </p>
 *
 * <h2>configuration</h2>
 *
 * <h3>browser extension</h3>
 * <p>
 * Install the <a href="http://livereload.com/extensions/">LiveReload browser extension.</a>
 * </p>
 *
 * <h3>livereload.js</h3>
 * <p>
 * Add the <code>livereload.js</code> to your web page.
 * </p>
 *
 * <p>
 * You can do this manually:
 * </p>
 *
 * <pre>{@code
 *  <script src="/livereload.js"></script>
 * }</pre>
 *
 * <p>
 * Or use the <code>liveReload</code> local variable with your preferred template engine. Here is an
 * example using <code>Handlebars</code>:
 * </p>
 *
 * <pre>
 * {{liveReload}}
 * </pre>
 *
 * <h3>json module</h3>
 * <p>
 * The <a href=
 * "http://feedback.livereload.com/knowledgebase/articles/86174-livereload-protocol">LiveReload
 * Protocol</a> run on top of a <code>WebSocket</code> and uses <code>JSON</code> as protocol
 * format. That's is why we also need to install a <code>JSON module</code>, like <a
 * href="http://jooby.org/doc/jackson">Jackson</a>.
 * </p>
 *
 * <h2>watcher</h2>
 * <p>
 * It automatically reload static resources from <code>public</code>, <code>target</code> (Maven
 * projects) or <code>build</code> folders (Gradle projects).
 * </p>
 *
 * <p>
 * Every time a change is detected the websocket send a <code>reload command</code>.
 * </p>
 *
 * <p>
 * That's all folks!!
 * </p>
 *
 * @author edgar
 * @since 1.1.0
 */
public class LiveReload implements Module {

  public static final String V7 = "http://livereload.com/protocols/official-7";

  /** The logging system. */
  private final Logger log = LoggerFactory.getLogger(getClass());

  private static final Set<String> CSS = ImmutableSet.of(".css", ".scss", ".sass", ".less");

  private Predicate<Path> css = path -> CSS.stream()
      .filter(ext -> path.toString().endsWith(ext))
      .findFirst()
      .isPresent();

  private List<Object[]> paths = new ArrayList<>();

  /**
   * Add the given path to the watcher.
   *
   * @param path Path to watch.
   * @param includes Glob pattern of matching files.
   * @return This module.
   */
  public LiveReload register(final Path path, final String... includes) {
    if (Files.exists(path)) {
      paths.add(new Object[]{path, Arrays.asList(includes)});
    }
    return this;
  }

  /**
   * Test if a path is a css like file (less, sass, etc.).
   *
   * @param predicate CSS predicate.
   * @return This module.
   */
  public LiveReload liveCss(final Predicate<Path> predicate) {
    this.css = predicate;
    return this;
  }

  @SuppressWarnings("unchecked")
  @Override
  public void configure(final Env env, final Config conf, final Binder binder) throws Throwable {
    boolean enabled = conf.hasPath("livereload.enabled")
        ? conf.getBoolean("livereload.enabled")
        : "dev".equals(env.name());
    if (enabled) {
      Router router = env.router();
      /**
       * Livereload client:
       */
      String livereloadjs = "/" + LiveReload.class.getPackage().getName().replace(".", "/")
          + "/livereload.js";
      router.assets("/livereload.js", livereloadjs);
      /** {{liveReload}} local variable */
      router.use("*", (req, rsp) -> req.set("liveReload", template(req)))
          .name("livereload");

      String serverName = CaseFormat.LOWER_CAMEL
          .to(CaseFormat.UPPER_CAMEL, conf.getString("application.name"));

      Queue<WebSocket> broadcast = new ConcurrentLinkedQueue<>();
      AtomicBoolean first = new AtomicBoolean(true);
      /**
       * Websocket:
       */
      router.ws("/livereload", ws -> {
        // add to broadcast
        broadcast.add(ws);

        ws.onMessage(msg -> {
          Map<String, Object> cmd = msg.to(Map.class);
          log.debug("command: {}", cmd);

          Map<String, Object> rsp = handshake(cmd, serverName, V7);
          if (rsp != null) {
            log.debug("sending: {}", rsp);
            ws.send(rsp);
          } else {
            log.trace("ignoring command: {}", cmd);
            // resync app state after a jooby:run restart
            if (first.compareAndSet(true, false)) {
              int counter = Integer.parseInt(System.getProperty("joobyRun.counter", "0"));
              if (counter > 0) {
                ws.send(reload("/", true));
              }
            }
          }
        });

        // remove from broadcast
        ws.onClose(reason -> broadcast.remove(ws));
      }).consumes(MediaType.json).produces(MediaType.json);

      if (paths.isEmpty()) {
        Path basedir = Paths.get(System.getProperty("user.dir"));
        Path assetsDir;
        if (conf.hasPath("assets.outputDir")) {
          assetsDir = Paths.get(conf.getString("assets.outputDir"));
        } else {
          assetsDir = basedir.resolve("public");
        }
        register(basedir.resolve("public"),
            "**/*.html",
            "**/*.ftl",
            "**/*.hbs",
            "**/*.jade"
            );
        register(assetsDir,
            "**/*.css",
            "**/*.scss",
            "**/*.sass",
            "**/*.less",
            "**/*.js",
            "**/*.coffee",
            "**/*.ts");
        register(basedir.resolve("target"),
            "**/*.class",
            "**/*.conf",
            "**/*.properties");
        register(basedir.resolve("build"),
            "**/*.class",
            "**/*.conf",
            "**/*.properties");
      }

      if (paths.size() > 0) {
        FileWatcher watcher = new FileWatcher();
        paths.forEach(it -> watcher.register((Path) it[0], (kind, path) -> {
          Path relative = relative(paths, path.toAbsolutePath());
          log.debug("file changed {}: {}", relative, File.separator);
          Map<String, Object> reload = reload(
              Route.normalize("/" + relative.toString().replace(File.separator, "/")),
              css.test(relative));
          for (WebSocket ws : broadcast) {
            try {
              log.debug("sending: {}", reload);
              ws.send(reload);
            } catch (Exception x) {
              log.debug("execution of {} resulted in exception", reload, x);
            }
          }
        }, options -> ((List<String>) it[1]).forEach(options::includes)));
        watcher.configure(env, conf, binder);
      } else {
        log.warn("File watcher is off");
      }
    }
  }

  private String template(Request req) {
    String contextPath = req.contextPath();
    return "<script>"
        + "window.LiveReloadOptions = {"
        + "host: '" + req.hostname() + "',"
        + "port: '" + req.port() + contextPath + "'"
        + "};"
        + "</script>\n"
        + "<script src=\"" + contextPath + "/livereload.js\"></script>";
  }

  private boolean isHello(final Map<String, Object> message) {
    return "hello".equals(message.get("command"));
  }

  @SuppressWarnings("unchecked")
  private Map<String, Object> handshake(final Map<String, Object> client,
      final String serverName, final String version) {
    if (isHello(client)) {
      List<String> protocols = (List<String>) client.get("protocols");
      return protocols.stream()
          .filter(protocol -> version.equalsIgnoreCase(protocol))
          .map(protocol -> {
            Map<String, Object> server = new LinkedHashMap<>();
            server.put("command", "hello");
            server.put("protocols", new String[]{protocol});
            server.put("serverName", serverName);
            return server;
          })
          .findFirst()
          .orElse(null);
    }
    return null;
  }

  private Path relative(final List<Object[]> paths, final Path path) {
    for (Object[] meta : paths) {
      Path root = ((Path) meta[0]).toAbsolutePath();
      Path relative = root.relativize(path);
      if (Files.exists(root.resolve(relative))) {
        return relative;
      }
    }
    return path;
  }

  private Map<String, Object> reload(final String path, final boolean liveCss) {
    Map<String, Object> cmd = new LinkedHashMap<>();
    cmd.put("command", "reload");
    cmd.put("path", path);
    cmd.put("liveCSS", liveCss);
    return cmd;
  }

}
