/**
 *                                  Apache License
 *                            Version 2.0, January 2004
 *                         http://www.apache.org/licenses/
 *
 *    TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *    1. Definitions.
 *
 *       "License" shall mean the terms and conditions for use, reproduction,
 *       and distribution as defined by Sections 1 through 9 of this document.
 *
 *       "Licensor" shall mean the copyright owner or entity authorized by
 *       the copyright owner that is granting the License.
 *
 *       "Legal Entity" shall mean the union of the acting entity and all
 *       other entities that control, are controlled by, or are under common
 *       control with that entity. For the purposes of this definition,
 *       "control" means (i) the power, direct or indirect, to cause the
 *       direction or management of such entity, whether by contract or
 *       otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *       outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *       "You" (or "Your") shall mean an individual or Legal Entity
 *       exercising permissions granted by this License.
 *
 *       "Source" form shall mean the preferred form for making modifications,
 *       including but not limited to software source code, documentation
 *       source, and configuration files.
 *
 *       "Object" form shall mean any form resulting from mechanical
 *       transformation or translation of a Source form, including but
 *       not limited to compiled object code, generated documentation,
 *       and conversions to other media types.
 *
 *       "Work" shall mean the work of authorship, whether in Source or
 *       Object form, made available under the License, as indicated by a
 *       copyright notice that is included in or attached to the work
 *       (an example is provided in the Appendix below).
 *
 *       "Derivative Works" shall mean any work, whether in Source or Object
 *       form, that is based on (or derived from) the Work and for which the
 *       editorial revisions, annotations, elaborations, or other modifications
 *       represent, as a whole, an original work of authorship. For the purposes
 *       of this License, Derivative Works shall not include works that remain
 *       separable from, or merely link (or bind by name) to the interfaces of,
 *       the Work and Derivative Works thereof.
 *
 *       "Contribution" shall mean any work of authorship, including
 *       the original version of the Work and any modifications or additions
 *       to that Work or Derivative Works thereof, that is intentionally
 *       submitted to Licensor for inclusion in the Work by the copyright owner
 *       or by an individual or Legal Entity authorized to submit on behalf of
 *       the copyright owner. For the purposes of this definition, "submitted"
 *       means any form of electronic, verbal, or written communication sent
 *       to the Licensor or its representatives, including but not limited to
 *       communication on electronic mailing lists, source code control systems,
 *       and issue tracking systems that are managed by, or on behalf of, the
 *       Licensor for the purpose of discussing and improving the Work, but
 *       excluding communication that is conspicuously marked or otherwise
 *       designated in writing by the copyright owner as "Not a Contribution."
 *
 *       "Contributor" shall mean Licensor and any individual or Legal Entity
 *       on behalf of whom a Contribution has been received by Licensor and
 *       subsequently incorporated within the Work.
 *
 *    2. Grant of Copyright License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       copyright license to reproduce, prepare Derivative Works of,
 *       publicly display, publicly perform, sublicense, and distribute the
 *       Work and such Derivative Works in Source or Object form.
 *
 *    3. Grant of Patent License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       (except as stated in this section) patent license to make, have made,
 *       use, offer to sell, sell, import, and otherwise transfer the Work,
 *       where such license applies only to those patent claims licensable
 *       by such Contributor that are necessarily infringed by their
 *       Contribution(s) alone or by combination of their Contribution(s)
 *       with the Work to which such Contribution(s) was submitted. If You
 *       institute patent litigation against any entity (including a
 *       cross-claim or counterclaim in a lawsuit) alleging that the Work
 *       or a Contribution incorporated within the Work constitutes direct
 *       or contributory patent infringement, then any patent licenses
 *       granted to You under this License for that Work shall terminate
 *       as of the date such litigation is filed.
 *
 *    4. Redistribution. You may reproduce and distribute copies of the
 *       Work or Derivative Works thereof in any medium, with or without
 *       modifications, and in Source or Object form, provided that You
 *       meet the following conditions:
 *
 *       (a) You must give any other recipients of the Work or
 *           Derivative Works a copy of this License; and
 *
 *       (b) You must cause any modified files to carry prominent notices
 *           stating that You changed the files; and
 *
 *       (c) You must retain, in the Source form of any Derivative Works
 *           that You distribute, all copyright, patent, trademark, and
 *           attribution notices from the Source form of the Work,
 *           excluding those notices that do not pertain to any part of
 *           the Derivative Works; and
 *
 *       (d) If the Work includes a "NOTICE" text file as part of its
 *           distribution, then any Derivative Works that You distribute must
 *           include a readable copy of the attribution notices contained
 *           within such NOTICE file, excluding those notices that do not
 *           pertain to any part of the Derivative Works, in at least one
 *           of the following places: within a NOTICE text file distributed
 *           as part of the Derivative Works; within the Source form or
 *           documentation, if provided along with the Derivative Works; or,
 *           within a display generated by the Derivative Works, if and
 *           wherever such third-party notices normally appear. The contents
 *           of the NOTICE file are for informational purposes only and
 *           do not modify the License. You may add Your own attribution
 *           notices within Derivative Works that You distribute, alongside
 *           or as an addendum to the NOTICE text from the Work, provided
 *           that such additional attribution notices cannot be construed
 *           as modifying the License.
 *
 *       You may add Your own copyright statement to Your modifications and
 *       may provide additional or different license terms and conditions
 *       for use, reproduction, or distribution of Your modifications, or
 *       for any such Derivative Works as a whole, provided Your use,
 *       reproduction, and distribution of the Work otherwise complies with
 *       the conditions stated in this License.
 *
 *    5. Submission of Contributions. Unless You explicitly state otherwise,
 *       any Contribution intentionally submitted for inclusion in the Work
 *       by You to the Licensor shall be under the terms and conditions of
 *       this License, without any additional terms or conditions.
 *       Notwithstanding the above, nothing herein shall supersede or modify
 *       the terms of any separate license agreement you may have executed
 *       with Licensor regarding such Contributions.
 *
 *    6. Trademarks. This License does not grant permission to use the trade
 *       names, trademarks, service marks, or product names of the Licensor,
 *       except as required for reasonable and customary use in describing the
 *       origin of the Work and reproducing the content of the NOTICE file.
 *
 *    7. Disclaimer of Warranty. Unless required by applicable law or
 *       agreed to in writing, Licensor provides the Work (and each
 *       Contributor provides its Contributions) on an "AS IS" BASIS,
 *       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *       implied, including, without limitation, any warranties or conditions
 *       of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *       PARTICULAR PURPOSE. You are solely responsible for determining the
 *       appropriateness of using or redistributing the Work and assume any
 *       risks associated with Your exercise of permissions under this License.
 *
 *    8. Limitation of Liability. In no event and under no legal theory,
 *       whether in tort (including negligence), contract, or otherwise,
 *       unless required by applicable law (such as deliberate and grossly
 *       negligent acts) or agreed to in writing, shall any Contributor be
 *       liable to You for damages, including any direct, indirect, special,
 *       incidental, or consequential damages of any character arising as a
 *       result of this License or out of the use or inability to use the
 *       Work (including but not limited to damages for loss of goodwill,
 *       work stoppage, computer failure or malfunction, or any and all
 *       other commercial damages or losses), even if such Contributor
 *       has been advised of the possibility of such damages.
 *
 *    9. Accepting Warranty or Additional Liability. While redistributing
 *       the Work or Derivative Works thereof, You may choose to offer,
 *       and charge a fee for, acceptance of support, warranty, indemnity,
 *       or other liability obligations and/or rights consistent with this
 *       License. However, in accepting such obligations, You may act only
 *       on Your own behalf and on Your sole responsibility, not on behalf
 *       of any other Contributor, and only if You agree to indemnify,
 *       defend, and hold each Contributor harmless for any liability
 *       incurred by, or claims asserted against, such Contributor by reason
 *       of your accepting any such warranty or additional liability.
 *
 *    END OF TERMS AND CONDITIONS
 *
 *    APPENDIX: How to apply the Apache License to your work.
 *
 *       To apply the Apache License to your work, attach the following
 *       boilerplate notice, with the fields enclosed by brackets "{}"
 *       replaced with your own identifying information. (Don't include
 *       the brackets!)  The text should be enclosed in the appropriate
 *       comment syntax for the file format. We also recommend that a
 *       file or class name and description of purpose be included on the
 *       same "printed page" as the copyright notice for easier
 *       identification within third-party archives.
 *
 *    Copyright 2014 Edgar Espina
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package org.jooby.metrics;

import static java.util.Objects.requireNonNull;

import java.io.Closeable;
import java.util.ArrayList;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import java.util.function.BiFunction;
import java.util.function.Consumer;
import java.util.function.Function;

import org.jooby.Env;
import org.jooby.Jooby;
import org.jooby.Router;
import org.jooby.internal.metrics.HealthCheckRegistryInitializer;
import org.jooby.internal.metrics.MetricRegistryInitializer;

import com.codahale.metrics.Metric;
import com.codahale.metrics.MetricRegistry;
import com.codahale.metrics.Reporter;
import com.codahale.metrics.health.HealthCheck;
import com.codahale.metrics.health.HealthCheckRegistry;
import com.google.inject.Binder;
import com.google.inject.multibindings.MapBinder;
import com.google.inject.multibindings.Multibinder;
import com.typesafe.config.Config;

/**
 * <h1>metrics</h1>
 * <p>
 * Metrics provides a powerful toolkit of ways to measure the behavior of critical components in
 * your production environment.
 * </p>
 *
 * <h2>usage</h2>
 * <pre>
 * {
 *   use(new Metrics()
 *      .request()
 *      .threadDump()
 *      .ping()
 *      .healthCheck("db", new DatabaseHealthCheck())
 *      .metric("memory", new MemoryUsageGaugeSet())
 *      .metric("threads", new ThreadStatesGaugeSet())
 *      .metric("gc", new GarbageCollectorMetricSet())
 *      .metric("fs", new FileDescriptorRatioGauge())
 *      );
 * }
 * </pre>
 * <p>
 * Let's see what all these means.
 * </p>
 *
 * <h2>metrics</h2>
 * <p>
 * Metrics are available at <code>/sys/metrics</code> or <code>/sys/metrics/:type</code> via:
 * </p>
 * <pre>
 *   use(new Metrics()
 *      .metric("memory", new MemoryUsageGaugeSet())
 *      .metric("threads", new ThreadStatesGaugeSet())
 *      .metric("gc", new GarbageCollectorMetricSet())
 *      .metric("fs", new FileDescriptorRatioGauge()));
 * </pre>
 * <p>
 * The <code>/:type</code> parameter is optional and let you filter metrics by type
 * <code>counters</code>, <code>guages</code>, etc..
 * </p>
 * <p>
 * There is a <code>name</code> filter too: <code>/sys/metrics?name=memory</code> or
 * <code>/sys/metrics/guages?name=memory</code>. The <code>name</code> parameter filter all the
 * metrics where the name starts with the given <code>name</code>.
 * </p>
 *
 * <h2>health checks</h2>
 * <p>
 * Health checks are available at <code>/sys/healthCheck</code> via:
 * </p>
 * <pre>
 *   use(new Metrics()
 *      .healthCheck("db", new DatabaseHealthCheck()));
 * </pre>
 *
 * <h2>instrumented requests</h2>
 * <p>
 * Captures request information (like active requests or min/mean/max execution time) and a
 * breakdown of the response codes being returned: {@link InstrumentedHandler}.
 * </p>
 * <pre>
 *   use(new Metrics()
 *      .request());
 * </pre>
 *
 * <h2>thread dump</h2>
 * <p>
 * A thread dump is available at <code>/sys/threadDump</code> via:
 * </p>
 * <pre>
 *   use(new Metrics()
 *      .threadDump());
 * </pre>
 *
 * <h2>reporting</h2>
 * <p>
 * Reporters are appended via a callback API:
 * </p>
 * <pre>
 *   {
 *     use(new Metrics()
 *        .reporter(registry {@literal ->} {
 *          ConsoleReporter reporter = ConsoleReporter.forRegistry(registry).build();
 *          reporter.start(1, TimeUnit.MINUTES);
 *          return reporter;
 *        });
 *   }
 * </pre>
 * <p>
 * You can add all the reporters you want. Keep in mind you have to start them (if need it), but you
 * don't have to stop them as long they implements the {@link Closeable} interface.
 * </p>
 *
 * <p>
 * That's all folks!
 * </p>
 *
 * @author edgar
 * @since 0.13.0
 */
public class Metrics implements Jooby.Module {

  static interface Bindings {
    void bind(Binder binder, Router routes, Config conf);
  }

  private String pattern;

  private List<Bindings> bindings = new ArrayList<>();

  private List<Consumer<Router>> routes = new ArrayList<>();

  private Set<BiFunction<MetricRegistry, Config, Reporter>> reporters = new LinkedHashSet<>();

  private MetricRegistry metricRegistry;
  private HealthCheckRegistry healthCheckRegistry;

  /**
   * Creates a new {@link Metric} module.
   *
   * @param metricRegistry Use the given metricRegistry.
   * @param healthCheckRegistry Use the given healthCheckRegistry.
   * @param pattern A root pattern where to publish all the services. Default is: <code>/sys</code>.
   */
  public Metrics(final MetricRegistry metricRegistry,
    final HealthCheckRegistry healthCheckRegistry, final String pattern) {
    this.metricRegistry = requireNonNull(metricRegistry, "Metric registry is required.");
    this.healthCheckRegistry = requireNonNull(healthCheckRegistry, "Health check registry is required.");
    this.pattern = requireNonNull(pattern, "A pattern is required.");
  }

  /**
   * Creates a new {@link Metric} module.
   *
   * @param metricRegistry Use the given metricRegistry.
   * @param pattern A root pattern where to publish all the services. Default is: <code>/sys</code>.
   */
  public Metrics(final MetricRegistry metricRegistry, final String pattern) {
    this(metricRegistry, new HealthCheckRegistry(), pattern);
  }

  /**
   * Creates a new {@link Metric} module.
   *
   * @param pattern A root pattern where to publish all the services. Default is: <code>/sys</code>.
   */
  public Metrics(final String pattern) {
    this(new MetricRegistry(), pattern);
  }

  /**
   * Creates a new {@link Metric} module. Services will be available at: <code>/sys</code>.
   *
   * @param healthCheckRegistry Use the given healthCheckRegistry.
   */
  public Metrics(final HealthCheckRegistry healthCheckRegistry) {
    this(new MetricRegistry(), healthCheckRegistry, "/sys");
  }

  /**
   * Creates a new {@link Metric} module. Services will be available at: <code>/sys</code>.
   *
   * @param metricRegistry Use the given metricRegistry.
   */
  public Metrics(final MetricRegistry metricRegistry) {
    this(metricRegistry, "/sys");
  }

  /**
   * Creates a new {@link Metric} module. Services will be available at: <code>/sys</code>.
   */
  public Metrics() {
    this("/sys");
  }

  /**
   * Instrument request using {@link InstrumentedHandler}.
   *
   * @param method Method to filter for. Default is: <code>GET</code>.
   * @param pattern A pattern to filter for. Default is: <code>*</code> (all the requests).
   * @return This metrics module.
   */
  public Metrics request(final String method, final String pattern) {
    routes.add(r -> r.use(method, pattern, new InstrumentedHandler()));
    return this;
  }

  /**
   * Instrument request using {@link InstrumentedHandler}.
   *
   * @param pattern A pattern to filter for. Default is: <code>*</code> (all the requests).
   * @return This metrics module.
   */
  public Metrics request(final String pattern) {
    return request("GET", pattern);
  }

  /**
   * Instrument request using {@link InstrumentedHandler}. It will intercept all the
   * <code>GET</code> calls.
   *
   * @return This metrics module.
   */
  public Metrics request() {
    return request("*");
  }

  /**
   * Append a simple ping handler that results in a <code>200</code> responses with a
   * <code>pong</code> body. See {@link PingHandler}
   *
   * @return This metrics module.
   */
  public Metrics ping() {
    bindings.add((binder, routes, conf) -> {
      routes.use("GET", this.pattern + "/ping", new PingHandler());
    });
    return this;
  }

  /**
   * Append a handler that prints thread states (a.k.a thread dump). See {@link ThreadDumpHandler}.
   *
   * @return This metrics module.
   */
  public Metrics threadDump() {
    bindings.add((binder, routes, conf) -> {
      routes.use("GET", this.pattern + "/thread-dump", new ThreadDumpHandler());
    });
    return this;
  }

  /**
   * Append a metric to the {@link MetricRegistry}, this call is identical to
   * {@link MetricRegistry#register(String, Metric)}.
   *
   * @param name Name of the metric.
   * @param metric A metric object
   * @return This metrics module.
   */
  public Metrics metric(final String name, final Metric metric) {
    bindings.add((binder, routes, conf) -> {
      MapBinder.newMapBinder(binder, String.class, Metric.class).addBinding(name)
          .toInstance(metric);
    });
    return this;
  }

  /**
   * Append a metric to the {@link MetricRegistry}. The metric will be resolved by Guice. This call
   * is identical to {@link MetricRegistry#register(String, Metric)}.
   *
   * @param name Name of the metric.
   * @param metric A metric object.
   * @param <M> Metric type.
   * @return This metrics module.
   */
  public <M extends Metric> Metrics metric(final String name, final Class<M> metric) {
    bindings.add((binder, routes, conf) -> {
      MapBinder.newMapBinder(binder, String.class, Metric.class).addBinding(name)
          .to(metric);
    });
    return this;
  }

  /**
   * Append a health check to the {@link HealthCheckRegistry}. This call is identical to
   * {@link HealthCheckRegistry#register(String, HealthCheck)}.
   *
   * @param name Name of the check.
   * @param check A check object.
   * @return This metrics module.
   */
  public Metrics healthCheck(final String name, final HealthCheck check) {
    bindings.add((binder, routes, conf) -> {
      MapBinder.newMapBinder(binder, String.class, HealthCheck.class).addBinding(name)
          .toInstance(check);
    });
    return this;
  }

  /**
   * Append a health check to the {@link HealthCheckRegistry}. The metric will be resolved by Guice.
   * This call is identical to {@link HealthCheckRegistry#register(String, HealthCheck)}.
   *
   * @param name Name of the check.
   * @param check A check object.
   * @param <H> {@link HealthCheck} type.
   * @return This metrics module.
   */
  public <H extends HealthCheck> Metrics healthCheck(final String name, final Class<H> check) {
    bindings.add((binder, routes, conf) -> {
      MapBinder.newMapBinder(binder, String.class, HealthCheck.class)
          .addBinding(name)
          .to(check);
    });
    return this;
  }

  /**
   * Append a {@link Reporter} to the {@link MetricRegistry}.
   *
   * @param callback Reporter callback.
   * @return This metrics module.
   */
  public Metrics reporter(final BiFunction<MetricRegistry, Config, Reporter> callback) {
    this.reporters.add(requireNonNull(callback, "Callback is required."));
    return this;
  }

  /**
   * Append a {@link Reporter} to the {@link MetricRegistry}.
   *
   * @param callback Reporter callback.
   * @return This metrics module.
   */
  public Metrics reporter(final Function<MetricRegistry, Reporter> callback) {
    return reporter((registry, conf) -> callback.apply(registry));
  }

  @Override
  public void configure(final Env env, final Config conf, final Binder binder) {
    // empty metric & checks
    MapBinder.newMapBinder(binder, String.class, Metric.class);
    MapBinder.newMapBinder(binder, String.class, HealthCheck.class);

    Router routes = env.router();

    MetricHandler mhandler = new MetricHandler();
    routes.use("GET", this.pattern + "/metrics", mhandler);
    routes.use("GET", this.pattern + "/metrics/:type", mhandler);
    routes.use("GET", this.pattern + "/healthcheck", new HealthCheckHandler());

    Multibinder<Reporter> reporters = Multibinder.newSetBinder(binder, Reporter.class);

    binder.bind(MetricRegistry.class).toInstance(metricRegistry);

    this.reporters.forEach(it -> reporters.addBinding().toInstance(it.apply(metricRegistry, conf)));

    binder.bind(MetricRegistryInitializer.class).asEagerSingleton();

    env.onStop(app -> app.require(MetricRegistryInitializer.class).close());

    binder.bind(HealthCheckRegistry.class).toInstance(healthCheckRegistry);

    binder.bind(HealthCheckRegistryInitializer.class).asEagerSingleton();

    bindings.forEach(it -> it.bind(binder, routes, conf));

    this.routes.forEach(it -> it.accept(routes));
  }

}
