/**
 *                                  Apache License
 *                            Version 2.0, January 2004
 *                         http://www.apache.org/licenses/
 *
 *    TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *    1. Definitions.
 *
 *       "License" shall mean the terms and conditions for use, reproduction,
 *       and distribution as defined by Sections 1 through 9 of this document.
 *
 *       "Licensor" shall mean the copyright owner or entity authorized by
 *       the copyright owner that is granting the License.
 *
 *       "Legal Entity" shall mean the union of the acting entity and all
 *       other entities that control, are controlled by, or are under common
 *       control with that entity. For the purposes of this definition,
 *       "control" means (i) the power, direct or indirect, to cause the
 *       direction or management of such entity, whether by contract or
 *       otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *       outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *       "You" (or "Your") shall mean an individual or Legal Entity
 *       exercising permissions granted by this License.
 *
 *       "Source" form shall mean the preferred form for making modifications,
 *       including but not limited to software source code, documentation
 *       source, and configuration files.
 *
 *       "Object" form shall mean any form resulting from mechanical
 *       transformation or translation of a Source form, including but
 *       not limited to compiled object code, generated documentation,
 *       and conversions to other media types.
 *
 *       "Work" shall mean the work of authorship, whether in Source or
 *       Object form, made available under the License, as indicated by a
 *       copyright notice that is included in or attached to the work
 *       (an example is provided in the Appendix below).
 *
 *       "Derivative Works" shall mean any work, whether in Source or Object
 *       form, that is based on (or derived from) the Work and for which the
 *       editorial revisions, annotations, elaborations, or other modifications
 *       represent, as a whole, an original work of authorship. For the purposes
 *       of this License, Derivative Works shall not include works that remain
 *       separable from, or merely link (or bind by name) to the interfaces of,
 *       the Work and Derivative Works thereof.
 *
 *       "Contribution" shall mean any work of authorship, including
 *       the original version of the Work and any modifications or additions
 *       to that Work or Derivative Works thereof, that is intentionally
 *       submitted to Licensor for inclusion in the Work by the copyright owner
 *       or by an individual or Legal Entity authorized to submit on behalf of
 *       the copyright owner. For the purposes of this definition, "submitted"
 *       means any form of electronic, verbal, or written communication sent
 *       to the Licensor or its representatives, including but not limited to
 *       communication on electronic mailing lists, source code control systems,
 *       and issue tracking systems that are managed by, or on behalf of, the
 *       Licensor for the purpose of discussing and improving the Work, but
 *       excluding communication that is conspicuously marked or otherwise
 *       designated in writing by the copyright owner as "Not a Contribution."
 *
 *       "Contributor" shall mean Licensor and any individual or Legal Entity
 *       on behalf of whom a Contribution has been received by Licensor and
 *       subsequently incorporated within the Work.
 *
 *    2. Grant of Copyright License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       copyright license to reproduce, prepare Derivative Works of,
 *       publicly display, publicly perform, sublicense, and distribute the
 *       Work and such Derivative Works in Source or Object form.
 *
 *    3. Grant of Patent License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       (except as stated in this section) patent license to make, have made,
 *       use, offer to sell, sell, import, and otherwise transfer the Work,
 *       where such license applies only to those patent claims licensable
 *       by such Contributor that are necessarily infringed by their
 *       Contribution(s) alone or by combination of their Contribution(s)
 *       with the Work to which such Contribution(s) was submitted. If You
 *       institute patent litigation against any entity (including a
 *       cross-claim or counterclaim in a lawsuit) alleging that the Work
 *       or a Contribution incorporated within the Work constitutes direct
 *       or contributory patent infringement, then any patent licenses
 *       granted to You under this License for that Work shall terminate
 *       as of the date such litigation is filed.
 *
 *    4. Redistribution. You may reproduce and distribute copies of the
 *       Work or Derivative Works thereof in any medium, with or without
 *       modifications, and in Source or Object form, provided that You
 *       meet the following conditions:
 *
 *       (a) You must give any other recipients of the Work or
 *           Derivative Works a copy of this License; and
 *
 *       (b) You must cause any modified files to carry prominent notices
 *           stating that You changed the files; and
 *
 *       (c) You must retain, in the Source form of any Derivative Works
 *           that You distribute, all copyright, patent, trademark, and
 *           attribution notices from the Source form of the Work,
 *           excluding those notices that do not pertain to any part of
 *           the Derivative Works; and
 *
 *       (d) If the Work includes a "NOTICE" text file as part of its
 *           distribution, then any Derivative Works that You distribute must
 *           include a readable copy of the attribution notices contained
 *           within such NOTICE file, excluding those notices that do not
 *           pertain to any part of the Derivative Works, in at least one
 *           of the following places: within a NOTICE text file distributed
 *           as part of the Derivative Works; within the Source form or
 *           documentation, if provided along with the Derivative Works; or,
 *           within a display generated by the Derivative Works, if and
 *           wherever such third-party notices normally appear. The contents
 *           of the NOTICE file are for informational purposes only and
 *           do not modify the License. You may add Your own attribution
 *           notices within Derivative Works that You distribute, alongside
 *           or as an addendum to the NOTICE text from the Work, provided
 *           that such additional attribution notices cannot be construed
 *           as modifying the License.
 *
 *       You may add Your own copyright statement to Your modifications and
 *       may provide additional or different license terms and conditions
 *       for use, reproduction, or distribution of Your modifications, or
 *       for any such Derivative Works as a whole, provided Your use,
 *       reproduction, and distribution of the Work otherwise complies with
 *       the conditions stated in this License.
 *
 *    5. Submission of Contributions. Unless You explicitly state otherwise,
 *       any Contribution intentionally submitted for inclusion in the Work
 *       by You to the Licensor shall be under the terms and conditions of
 *       this License, without any additional terms or conditions.
 *       Notwithstanding the above, nothing herein shall supersede or modify
 *       the terms of any separate license agreement you may have executed
 *       with Licensor regarding such Contributions.
 *
 *    6. Trademarks. This License does not grant permission to use the trade
 *       names, trademarks, service marks, or product names of the Licensor,
 *       except as required for reasonable and customary use in describing the
 *       origin of the Work and reproducing the content of the NOTICE file.
 *
 *    7. Disclaimer of Warranty. Unless required by applicable law or
 *       agreed to in writing, Licensor provides the Work (and each
 *       Contributor provides its Contributions) on an "AS IS" BASIS,
 *       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *       implied, including, without limitation, any warranties or conditions
 *       of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *       PARTICULAR PURPOSE. You are solely responsible for determining the
 *       appropriateness of using or redistributing the Work and assume any
 *       risks associated with Your exercise of permissions under this License.
 *
 *    8. Limitation of Liability. In no event and under no legal theory,
 *       whether in tort (including negligence), contract, or otherwise,
 *       unless required by applicable law (such as deliberate and grossly
 *       negligent acts) or agreed to in writing, shall any Contributor be
 *       liable to You for damages, including any direct, indirect, special,
 *       incidental, or consequential damages of any character arising as a
 *       result of this License or out of the use or inability to use the
 *       Work (including but not limited to damages for loss of goodwill,
 *       work stoppage, computer failure or malfunction, or any and all
 *       other commercial damages or losses), even if such Contributor
 *       has been advised of the possibility of such damages.
 *
 *    9. Accepting Warranty or Additional Liability. While redistributing
 *       the Work or Derivative Works thereof, You may choose to offer,
 *       and charge a fee for, acceptance of support, warranty, indemnity,
 *       or other liability obligations and/or rights consistent with this
 *       License. However, in accepting such obligations, You may act only
 *       on Your own behalf and on Your sole responsibility, not on behalf
 *       of any other Contributor, and only if You agree to indemnify,
 *       defend, and hold each Contributor harmless for any liability
 *       incurred by, or claims asserted against, such Contributor by reason
 *       of your accepting any such warranty or additional liability.
 *
 *    END OF TERMS AND CONDITIONS
 *
 *    APPENDIX: How to apply the Apache License to your work.
 *
 *       To apply the Apache License to your work, attach the following
 *       boilerplate notice, with the fields enclosed by brackets "{}"
 *       replaced with your own identifying information. (Don't include
 *       the brackets!)  The text should be enclosed in the appropriate
 *       comment syntax for the file format. We also recommend that a
 *       file or class name and description of purpose be included on the
 *       same "printed page" as the copyright notice for easier
 *       identification within third-party archives.
 *
 *    Copyright 2014 Edgar Espina
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package org.jooby.neo4j;

import static iot.jcypher.database.DBProperties.DATABASE_DIR;
import static iot.jcypher.database.DBProperties.SERVER_ROOT_URI;
import static java.util.Objects.requireNonNull;

import java.io.IOException;
import java.lang.reflect.Field;
import java.nio.file.FileVisitResult;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.SimpleFileVisitor;
import java.nio.file.attribute.BasicFileAttributes;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Objects;
import java.util.Properties;
import java.util.concurrent.atomic.AtomicInteger;

import org.jooby.Env;
import org.jooby.Env.ServiceKey;
import org.jooby.Jooby.Module;
import org.neo4j.driver.v1.AuthTokens;
import org.neo4j.driver.v1.Driver;
import org.neo4j.driver.v1.GraphDatabase;
import org.neo4j.driver.v1.Session;
import org.neo4j.graphdb.GraphDatabaseService;
import org.neo4j.graphdb.factory.GraphDatabaseBuilder;
import org.neo4j.graphdb.factory.GraphDatabaseFactory;
import org.neo4j.logging.slf4j.Slf4jLogProvider;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.bridge.SLF4JBridgeHandler;

import com.google.inject.Binder;
import com.typesafe.config.Config;
import com.typesafe.config.ConfigException;
import com.typesafe.config.ConfigFactory;
import com.typesafe.config.ConfigValue;
import com.typesafe.config.ConfigValueFactory;
import com.typesafe.config.ConfigValueType;

import iot.jcypher.database.DBProperties;
import iot.jcypher.database.IDBAccess;
import iot.jcypher.database.embedded.AbstractEmbeddedDBAccess;
import iot.jcypher.database.embedded.EmbeddedDBAccess;
import iot.jcypher.database.remote.BoltDBAccess;

/**
 * <h1>neo4j</h1>
 * <p>
 * <a href="https://neo4j.com">Neo4j</a> is a highly scalable native graph database that leverages
 * data relationships as first-class entities, helping enterprises build intelligent applications to
 * meet today’s evolving data challenges.
 * </p>
 *
 * <p>
 * This module give you access to <a href="https://neo4j.com">neo4j</a> and
 * <a href="https://github.com/Wolfgang-Schuetzelhofer/jcypher">jcypher</a> APIs.
 * </p>
 *
 * <h2>exports</h2>
 * <ul>
 * <li>GraphDatabaseService for embedded neo4j instances.</li>
 * <li>Driver and Session objects for remote instances</li>
 * <li>IDBAccess object</li>
 * </ul>
 *
 * <h2>usage</h2>
 *
 * <pre>{@code
 * {
 *   use(new Neo4j());
 *
 *   get("/driver", () -> {
 *     // work with driver
 *     Driver driver = require(Driver.class);
 *   });
 *
 *   get("/session", () -> {
 *     // work with session
 *     Session session = require(Session.class);
 *   });
 *
 *   get("/dbaccess", () -> {
 *     // work with driver
 *     BoltDBAccess dbaccess = require(BoltDBAccess.class);
 *   });
 * }
 * }</pre>
 *
 * <p>
 * application.conf
 * </p>
 * <pre>
 *  db.url = "bolt://localhost:7687"
 *  db.user = myuser
 *  db.password = mypassword
 * </pre>
 *
 * <h2>embedded</h2>
 * <p>
 * In addition to remote access using <code>bolt</code> protocol, this module provide
 * access to <code>embedded</code> neo4j instances:
 * </p>
 *
 * <p>
 * In memory mode:
 * </p>
 * <pre>{@code
 * {
 *   use(new Neo4j("mem"));
 * }
 * }</pre>
 *
 * <p>
 * File system mode:
 * </p>
 * <pre>{@code
 * {
 *   use(new Neo4j("fs"));
 * }
 * }</pre>
 *
 * <p>
 * Optionally you can specify the desired path:
 * </p>
 * <pre>{@code
 * {
 *   use(new Neo4j(Paths.get("path", "mydb")));
 * }
 * }</pre>
 *
 * <p>
 * The embedded mode allow you to access {@link GraphDatabaseService} instances:
 * </p>
 *
 * <pre>{@code
 * {
 *   use(new Neo4j("mem"));
 *
 *   get("/", () -> {
 *     GraphDatabaseService db = require(GraphDatabaseService.class);
 *   });
 * }
 * }</pre>
 *
 * <p>
 * As well as {@link EmbeddedDBAccess}:
 * </p>
 * <pre>{@code
 * {
 *   use(new Neo4j("mem"));
 *
 *   get("/", () -> {
 *     EmbeddedDBAccess db = require(EmbeddedDBAccess.class);
 *   });
 * }
 * }</pre>
 *
 * <h2>runtime modules</h2>
 * <p>
 * This option is available for <code>embedded</code> Neo4j instances and we allow to configure one
 * or more runtime modules via <code>.conf</code> file:
 * </p>
 *
 * <pre>
 * com.graphaware.runtime.enabled = true
 *
 * com.graphaware.module = [{
 *   class: com.graphaware.neo4j.expire.ExpirationModuleBootstrapper
 *   nodeExpirationProperty: _expire
 * }, {
 *   class: com.graphaware.neo4j.expire.AnotherModule
 *   modProp: modValue
 * }]
 * </pre>
 *
 * <p>
 * You first need to <code>enabled</code> the graph runtime framework by setting the
 * <code>com.graphaware.runtime.enabled</code> property.
 * </p>
 *
 * <p>
 * Then you need to add one or more modules under the <code>com.graphaware.module</code>
 * property path.
 * </p>
 *
 * <h2>two or more connections</h2>
 * <p>
 * Two or more connection is available by setting and installing multiples {@link Neo4j} modules:
 * </p>
 *
 * <pre>{@code
 * {
 *   use(new Neo4j("db1"));
 *
 *   use(new Neo4j("db2"));
 *
 *   get("/", () -> {
 *     Driver db1 = require("db1", Driver.class);
 *     BoltDBAccess bolt1 = require("db1", BoltDBAccess.class);
 *
 *     Driver db2 = require("db2", Driver.class);
 *     BoltDBAccess bolt2 = require("db2", BoltDBAccess.class);
 *   });
 * }
 * }</pre>
 *
 * <p>
 * application.conf:
 * </p>
 *
 * <pre>
 *  db1.url = "bolt://localhost:7687"
 *  db1.user = db1user
 *  db1.password = db1pass
 *
 *  db2.url = "bolt://localhost:7687"
 *  db2.user = db2user
 *  db2.password = db2pass
 * </pre>
 *
 * <h2>options</h2>
 * <p>
 * <a href="https://neo4j.com">Neo4j</a> options are available via <code>.conf</code> file:
 * </p>
 *
 * <pre>
 * neo4j.dbms.read_only = true
 * neo4j.unsupported.dbms.block_size.array_properties = 120
 * </pre>
 *
 * @author edgar
 * @author sbcd90
 */
public class Neo4j implements Module {

  static {
    if (!SLF4JBridgeHandler.isInstalled()) {
      SLF4JBridgeHandler.removeHandlersForRootLogger();
      SLF4JBridgeHandler.install();
    }
  }

  /** The logging system. */
  private final Logger log = LoggerFactory.getLogger(getClass());

  private String db;

  /**
   * Creates a new {@link Neo4j} module.
   *
   * @param db Database type: <code>mem</code>, <code>fs</code>,
   *        <code>path</code> for embedded database or <code>property</code> pointing the
   *        a database server or one of the previously described values.
   */
  public Neo4j(final String db) {
    this.db = requireNonNull(db, "Database required.");
  }

  /**
   * Creates a new embedded {@link Neo4j} module.
   *
   * @param db Path to store the database.
   */
  public Neo4j(final Path db) {
    this(db.toAbsolutePath().toString());
  }

  /**
   * Creates a new {@link Neo4j} module a <code>db</code> property must be defined in your
   * <code>.conf</code> file. Property value must be one of: <code>mem</code>, <code>fs</code>,
   * <code>file system path</code> or <code>bolt url</code>.
   *
   */
  public Neo4j() {
    this("db");
  }

  @SuppressWarnings({"unchecked", "rawtypes" })
  @Override
  public void configure(final Env env, final Config conf, final Binder binder) throws Throwable {
    String db = database(conf, this.db);
    Properties props = props(neo4j(conf, this.db));

    ServiceKey keys = env.serviceKey();
    IDBAccess dbaccess = dbaccess(conf, this.db, db, props, keys, binder);

    Arrays.asList(props.getProperty(SERVER_ROOT_URI), props.getProperty(DATABASE_DIR))
        .stream()
        .filter(Objects::nonNull)
        .findFirst()
        .ifPresent(it -> log.info("Starting neo4j: {}", it));

    Class dbaccessType = dbaccess.getClass();
    keys.generate(IDBAccess.class, this.db, k -> binder.bind(k).toInstance(dbaccess));
    keys.generate(dbaccessType, this.db, k -> binder.bind(k).toInstance(dbaccess));

    env.onStop(dbaccess::close);
  }

  @Override
  public Config config() {
    return ConfigFactory.empty(getClass().getName().toLowerCase() + ".conf")
        .withValue("neo4j.session.label", ConfigValueFactory.fromAnyRef("session"));
  }

  private Config neo4j(final Config conf, final String db) {
    Config result = conf.hasPath("com.graphaware")
        ? ConfigFactory.empty().withValue("com.graphaware", conf.getConfig("com.graphaware").root())
        : ConfigFactory.empty();
    String[] paths = {"neo4j", db, "neo4j." + db };
    for (String path : paths) {
      try {
        if (conf.hasPath(path)) {
          Config it = conf.getConfig(path);
          result = it.withFallback(result);
        }
      } catch (ConfigException x) {
        // Skip/ignore bad path
      }
    }
    return result;
  }

  @SuppressWarnings("deprecation")
  private IDBAccess dbaccess(final Config conf, final String dbkey, final String db,
      final Properties props, final Env.ServiceKey keys, final Binder binder) throws Exception {
    // remote
    if (db.startsWith("bolt")) {
      String username = conf.getString(dbkey + ".user");
      String password = conf.getString(dbkey + ".password");
      Driver driver = GraphDatabase.driver(db, AuthTokens.basic(username, password));
      Session session = driver.session();
      props.setProperty(DBProperties.SERVER_ROOT_URI, db);
      keys.generate(Driver.class, dbkey, k -> binder.bind(k).toInstance(driver));
      keys.generate(Session.class, dbkey, k -> binder.bind(k).toInstance(session));
      BoltDBAccess dbaccess = new BoltDBAccess();
      dbaccess.initialize(props);
      driverhack(dbaccess, driver);
      return dbaccess;
    }
    boolean mem = db.equals("mem");
    Path path = (db.equals("fs") || mem
        ? Paths.get(conf.getString("application.tmpdir")).resolve("neo4j" + db)
        : Paths.get(db)).toAbsolutePath();
    props.setProperty(DBProperties.DATABASE_DIR, path.toString());
    if (mem) {
      // clean up existing database
      rm(path);
    }
    GraphDatabaseBuilder builder = new GraphDatabaseFactory()
        .setUserLogProvider(new Slf4jLogProvider())
        .newEmbeddedDatabaseBuilder(path.toFile());
    props.forEach((k, v) -> builder.setConfig(k.toString(), v.toString()));
    GraphDatabaseService graphDatabaseService = builder.newGraphDatabase();
    keys.generate(GraphDatabaseService.class, dbkey,
        k -> binder.bind(k).toInstance(graphDatabaseService));

    EmbeddedDBAccess dbaccess = new EmbeddedDBAccess();
    dbaccess.initialize(props);
    dbServicehack(dbaccess, graphDatabaseService);

    return dbaccess;
  }

  private String database(final Config conf, final String db) {
    String[] paths = {db + ".url", db };
    for (String path : paths) {
      if (conf.hasPath(path)) {
        return conf.getString(path);
      }
    }
    if (db.equals("db")) {
      throw new ConfigException.Missing(db);
    }
    // path
    return db;
  }

  private Properties props(final Config conf) {
    Properties props = new Properties();
    conf.entrySet().stream()
        .filter(it -> !it.getKey().startsWith("com.graphaware.module"))
        .forEach(prop -> {
          String key = prop.getKey();
          Object value = prop.getValue().unwrapped();
          props.put(key, value);
        });
    // write module:
    if (conf.hasPath("com.graphaware.module")) {
      ConfigValue raw = conf.getValue("com.graphaware.module");
      List<Config> modules = new ArrayList<>();
      if (raw.valueType() == ConfigValueType.LIST) {
        modules.addAll(conf.getConfigList("com.graphaware.module"));
      } else {
        modules.add(conf.getConfig("com.graphaware.module"));
      }
      AtomicInteger nextOrder = new AtomicInteger(0);
      modules.forEach(module -> {
        String type = module.getString("class");
        int order = nextOrder.incrementAndGet();
        String key = "com.graphaware.module.m" + order;
        props.put(key + "." + order, type);
        module.withoutPath("class").root()
            .forEach((k, v) -> props.put(key + "." + k, v.unwrapped().toString()));
      });
    }
    log.debug("neo4j properties: {}", props);
    return props;
  }

  static void driverhack(final BoltDBAccess dbaccess, final Driver value) throws Exception {
    fieldhack(BoltDBAccess.class, dbaccess, "driver", value);
  }

  static void dbServicehack(final EmbeddedDBAccess dbaccess, final GraphDatabaseService value)
      throws Exception {
    fieldhack(AbstractEmbeddedDBAccess.class, dbaccess, "graphDb", value);
  }

  @SuppressWarnings("rawtypes")
  static void fieldhack(final Class owner, final Object instance, final String name,
      final Object value) throws Exception {
    Field field = owner.getDeclaredField(name);
    field.setAccessible(true);
    field.set(instance, value);
  }

  static void rm(final Path path) throws IOException {
    if (Files.exists(path)) {
      Files.walkFileTree(path, new SimpleFileVisitor<Path>() {
        @Override
        public FileVisitResult visitFile(final Path file, final BasicFileAttributes attrs)
            throws IOException {
          Files.delete(file);
          return FileVisitResult.CONTINUE;
        }

        @Override
        public FileVisitResult postVisitDirectory(final Path dir, final IOException exc)
            throws IOException {
          Files.delete(dir);
          return FileVisitResult.CONTINUE;
        }

      });
    }
  }
}
