/**
 *                                  Apache License
 *                            Version 2.0, January 2004
 *                         http://www.apache.org/licenses/
 *
 *    TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *    1. Definitions.
 *
 *       "License" shall mean the terms and conditions for use, reproduction,
 *       and distribution as defined by Sections 1 through 9 of this document.
 *
 *       "Licensor" shall mean the copyright owner or entity authorized by
 *       the copyright owner that is granting the License.
 *
 *       "Legal Entity" shall mean the union of the acting entity and all
 *       other entities that control, are controlled by, or are under common
 *       control with that entity. For the purposes of this definition,
 *       "control" means (i) the power, direct or indirect, to cause the
 *       direction or management of such entity, whether by contract or
 *       otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *       outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *       "You" (or "Your") shall mean an individual or Legal Entity
 *       exercising permissions granted by this License.
 *
 *       "Source" form shall mean the preferred form for making modifications,
 *       including but not limited to software source code, documentation
 *       source, and configuration files.
 *
 *       "Object" form shall mean any form resulting from mechanical
 *       transformation or translation of a Source form, including but
 *       not limited to compiled object code, generated documentation,
 *       and conversions to other media types.
 *
 *       "Work" shall mean the work of authorship, whether in Source or
 *       Object form, made available under the License, as indicated by a
 *       copyright notice that is included in or attached to the work
 *       (an example is provided in the Appendix below).
 *
 *       "Derivative Works" shall mean any work, whether in Source or Object
 *       form, that is based on (or derived from) the Work and for which the
 *       editorial revisions, annotations, elaborations, or other modifications
 *       represent, as a whole, an original work of authorship. For the purposes
 *       of this License, Derivative Works shall not include works that remain
 *       separable from, or merely link (or bind by name) to the interfaces of,
 *       the Work and Derivative Works thereof.
 *
 *       "Contribution" shall mean any work of authorship, including
 *       the original version of the Work and any modifications or additions
 *       to that Work or Derivative Works thereof, that is intentionally
 *       submitted to Licensor for inclusion in the Work by the copyright owner
 *       or by an individual or Legal Entity authorized to submit on behalf of
 *       the copyright owner. For the purposes of this definition, "submitted"
 *       means any form of electronic, verbal, or written communication sent
 *       to the Licensor or its representatives, including but not limited to
 *       communication on electronic mailing lists, source code control systems,
 *       and issue tracking systems that are managed by, or on behalf of, the
 *       Licensor for the purpose of discussing and improving the Work, but
 *       excluding communication that is conspicuously marked or otherwise
 *       designated in writing by the copyright owner as "Not a Contribution."
 *
 *       "Contributor" shall mean Licensor and any individual or Legal Entity
 *       on behalf of whom a Contribution has been received by Licensor and
 *       subsequently incorporated within the Work.
 *
 *    2. Grant of Copyright License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       copyright license to reproduce, prepare Derivative Works of,
 *       publicly display, publicly perform, sublicense, and distribute the
 *       Work and such Derivative Works in Source or Object form.
 *
 *    3. Grant of Patent License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       (except as stated in this section) patent license to make, have made,
 *       use, offer to sell, sell, import, and otherwise transfer the Work,
 *       where such license applies only to those patent claims licensable
 *       by such Contributor that are necessarily infringed by their
 *       Contribution(s) alone or by combination of their Contribution(s)
 *       with the Work to which such Contribution(s) was submitted. If You
 *       institute patent litigation against any entity (including a
 *       cross-claim or counterclaim in a lawsuit) alleging that the Work
 *       or a Contribution incorporated within the Work constitutes direct
 *       or contributory patent infringement, then any patent licenses
 *       granted to You under this License for that Work shall terminate
 *       as of the date such litigation is filed.
 *
 *    4. Redistribution. You may reproduce and distribute copies of the
 *       Work or Derivative Works thereof in any medium, with or without
 *       modifications, and in Source or Object form, provided that You
 *       meet the following conditions:
 *
 *       (a) You must give any other recipients of the Work or
 *           Derivative Works a copy of this License; and
 *
 *       (b) You must cause any modified files to carry prominent notices
 *           stating that You changed the files; and
 *
 *       (c) You must retain, in the Source form of any Derivative Works
 *           that You distribute, all copyright, patent, trademark, and
 *           attribution notices from the Source form of the Work,
 *           excluding those notices that do not pertain to any part of
 *           the Derivative Works; and
 *
 *       (d) If the Work includes a "NOTICE" text file as part of its
 *           distribution, then any Derivative Works that You distribute must
 *           include a readable copy of the attribution notices contained
 *           within such NOTICE file, excluding those notices that do not
 *           pertain to any part of the Derivative Works, in at least one
 *           of the following places: within a NOTICE text file distributed
 *           as part of the Derivative Works; within the Source form or
 *           documentation, if provided along with the Derivative Works; or,
 *           within a display generated by the Derivative Works, if and
 *           wherever such third-party notices normally appear. The contents
 *           of the NOTICE file are for informational purposes only and
 *           do not modify the License. You may add Your own attribution
 *           notices within Derivative Works that You distribute, alongside
 *           or as an addendum to the NOTICE text from the Work, provided
 *           that such additional attribution notices cannot be construed
 *           as modifying the License.
 *
 *       You may add Your own copyright statement to Your modifications and
 *       may provide additional or different license terms and conditions
 *       for use, reproduction, or distribution of Your modifications, or
 *       for any such Derivative Works as a whole, provided Your use,
 *       reproduction, and distribution of the Work otherwise complies with
 *       the conditions stated in this License.
 *
 *    5. Submission of Contributions. Unless You explicitly state otherwise,
 *       any Contribution intentionally submitted for inclusion in the Work
 *       by You to the Licensor shall be under the terms and conditions of
 *       this License, without any additional terms or conditions.
 *       Notwithstanding the above, nothing herein shall supersede or modify
 *       the terms of any separate license agreement you may have executed
 *       with Licensor regarding such Contributions.
 *
 *    6. Trademarks. This License does not grant permission to use the trade
 *       names, trademarks, service marks, or product names of the Licensor,
 *       except as required for reasonable and customary use in describing the
 *       origin of the Work and reproducing the content of the NOTICE file.
 *
 *    7. Disclaimer of Warranty. Unless required by applicable law or
 *       agreed to in writing, Licensor provides the Work (and each
 *       Contributor provides its Contributions) on an "AS IS" BASIS,
 *       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *       implied, including, without limitation, any warranties or conditions
 *       of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *       PARTICULAR PURPOSE. You are solely responsible for determining the
 *       appropriateness of using or redistributing the Work and assume any
 *       risks associated with Your exercise of permissions under this License.
 *
 *    8. Limitation of Liability. In no event and under no legal theory,
 *       whether in tort (including negligence), contract, or otherwise,
 *       unless required by applicable law (such as deliberate and grossly
 *       negligent acts) or agreed to in writing, shall any Contributor be
 *       liable to You for damages, including any direct, indirect, special,
 *       incidental, or consequential damages of any character arising as a
 *       result of this License or out of the use or inability to use the
 *       Work (including but not limited to damages for loss of goodwill,
 *       work stoppage, computer failure or malfunction, or any and all
 *       other commercial damages or losses), even if such Contributor
 *       has been advised of the possibility of such damages.
 *
 *    9. Accepting Warranty or Additional Liability. While redistributing
 *       the Work or Derivative Works thereof, You may choose to offer,
 *       and charge a fee for, acceptance of support, warranty, indemnity,
 *       or other liability obligations and/or rights consistent with this
 *       License. However, in accepting such obligations, You may act only
 *       on Your own behalf and on Your sole responsibility, not on behalf
 *       of any other Contributor, and only if You agree to indemnify,
 *       defend, and hold each Contributor harmless for any liability
 *       incurred by, or claims asserted against, such Contributor by reason
 *       of your accepting any such warranty or additional liability.
 *
 *    END OF TERMS AND CONDITIONS
 *
 *    APPENDIX: How to apply the Apache License to your work.
 *
 *       To apply the Apache License to your work, attach the following
 *       boilerplate notice, with the fields enclosed by brackets "{}"
 *       replaced with your own identifying information. (Don't include
 *       the brackets!)  The text should be enclosed in the appropriate
 *       comment syntax for the file format. We also recommend that a
 *       file or class name and description of purpose be included on the
 *       same "printed page" as the copyright notice for easier
 *       identification within third-party archives.
 *
 *    Copyright 2014 Edgar Espina
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package org.jooby.internal.netty;

import com.google.common.collect.ArrayListMultimap;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Multimap;
import io.netty.buffer.ByteBuf;
import io.netty.buffer.ByteBufHolder;
import io.netty.buffer.ByteBufInputStream;
import io.netty.channel.Channel;
import static io.netty.channel.ChannelFutureListener.FIRE_EXCEPTION_ON_FAILURE;
import io.netty.channel.ChannelHandlerContext;
import io.netty.handler.codec.http.FullHttpRequest;
import io.netty.handler.codec.http.HttpContent;
import io.netty.handler.codec.http.HttpHeaderNames;
import io.netty.handler.codec.http.HttpHeaders;
import io.netty.handler.codec.http.HttpMethod;
import io.netty.handler.codec.http.HttpRequest;
import io.netty.handler.codec.http.QueryStringDecoder;
import io.netty.handler.codec.http.cookie.Cookie;
import io.netty.handler.codec.http.cookie.ServerCookieDecoder;
import io.netty.handler.codec.http.multipart.DefaultHttpDataFactory;
import io.netty.handler.codec.http.multipart.FileUpload;
import io.netty.handler.codec.http.multipart.HttpData;
import io.netty.handler.codec.http.multipart.HttpPostRequestDecoder;
import io.netty.handler.codec.http.multipart.InterfaceHttpData.HttpDataType;
import io.netty.handler.codec.http.websocketx.WebSocketServerHandshaker;
import io.netty.handler.codec.http.websocketx.WebSocketServerHandshakerFactory;
import io.netty.handler.codec.http2.HttpConversionUtil;
import io.netty.util.AttributeKey;
import io.netty.util.ReferenceCounted;
import org.jooby.MediaType;
import org.jooby.Router;
import org.jooby.Sse;
import org.jooby.spi.NativePushPromise;
import org.jooby.spi.NativeRequest;
import org.jooby.spi.NativeUpload;
import org.jooby.spi.NativeWebSocket;

import java.io.IOException;
import java.io.InputStream;
import java.net.InetSocketAddress;
import java.util.Collections;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.Executor;
import java.util.function.Function;
import java.util.stream.Collectors;

public class NettyRequest implements NativeRequest {

  public static final AttributeKey<String> PROTOCOL = AttributeKey
      .newInstance(NettyRequest.class.getName() + ".protol");

  public static final AttributeKey<Boolean> NEED_FLUSH = AttributeKey
      .newInstance(NettyRequest.class.getName() + ".needFlush");

  public static final AttributeKey<Boolean> ASYNC = AttributeKey
      .newInstance(NettyRequest.class.getName() + ".async");

  public static final AttributeKey<Boolean> SECURE = AttributeKey
      .newInstance(NettyRequest.class.getName() + ".secure");

  private HttpRequest req;

  private final HttpHeaders responseHeaders;

  private QueryStringDecoder query;

  private List<org.jooby.Cookie> cookies;

  private Multimap<String, String> params;

  private Multimap<String, NativeUpload> files;

  private String tmpdir;

  private String path;

  private ChannelHandlerContext ctx;

  private int wsMaxMessageSize;

  public NettyRequest(final ChannelHandlerContext ctx, final HttpRequest req,
      final HttpHeaders responseHeaders, final String tmpdir, final int wsMaxMessageSize) {
    this.ctx = ctx;
    this.req = req;
    this.responseHeaders = responseHeaders;
    this.tmpdir = tmpdir;
    this.query = new QueryStringDecoder(req.uri());
    this.path = Router.decode(query.path());
    this.wsMaxMessageSize = wsMaxMessageSize;
    Channel channel = ctx.channel();
    channel.attr(ASYNC).set(false);
  }

  @Override
  public Optional<String> queryString() {
    String uri = req.uri();
    int at = uri.indexOf('?') + 1;
    return at > 0 && at < uri.length() ? Optional.of(uri.substring(at)) : Optional.empty();
  }

  @Override
  public String method() {
    return req.method().name();
  }

  @Override
  public String rawPath() {
    String uri = req.uri();
    int at = uri.indexOf('?');
    return at > 0 ? uri.substring(0, at) : uri;
  }

  @Override
  public String path() {
    return path;
  }

  @Override
  public List<String> paramNames() throws IOException {
    return ImmutableList.copyOf(decodeParams().keySet());
  }

  @Override
  public List<String> params(final String name) throws Exception {
    return (List<String>) decodeParams().get(name);
  }

  @Override
  public List<String> headers(final String name) {
    return req.headers().getAll(name);
  }

  @Override
  public Optional<String> header(final String name) {
    String value = req.headers().get(name);
    return Optional.ofNullable(value);
  }

  @Override
  public List<String> headerNames() {
    ImmutableList.Builder<String> builder = ImmutableList.builder();
    req.headers().names().forEach(it -> builder.add(it.toString()));
    return builder.build();
  }

  @Override
  public List<org.jooby.Cookie> cookies() {
    if (this.cookies == null) {
      String cookieString = req.headers().get(HttpHeaderNames.COOKIE);
      if (cookieString != null) {
        this.cookies = ServerCookieDecoder.STRICT.decode(cookieString).stream()
            .map(this::cookie)
            .collect(Collectors.toList());

      } else {
        this.cookies = Collections.emptyList();
      }
    }
    return this.cookies;
  }

  @Override
  public List<NativeUpload> files(final String name) throws IOException {
    decodeParams();
    return ImmutableList.copyOf(this.files.get(name));
  }

  @Override
  public List<NativeUpload> files() throws IOException {
    decodeParams();
    return ImmutableList.copyOf(this.files.values());
  }

  @Override
  public InputStream in() {
    ByteBuf content = ((HttpContent) req).content();
    return new ByteBufInputStream(content);
  }

  @Override
  public String ip() {
    InetSocketAddress remoteAddress = (InetSocketAddress) ctx.channel().remoteAddress();
    return remoteAddress.getAddress().getHostAddress();
  }

  @Override
  public String protocol() {
    return ctx.pipeline().get("h2") == null
        ? req.protocolVersion().text()
        : "HTTP/2.0";
  }

  @Override
  public boolean secure() {
    return ifSecure(Boolean.TRUE, Boolean.FALSE).booleanValue();
  }

  @SuppressWarnings("unchecked")
  @Override
  public <T> T upgrade(final Class<T> type) {
    if (type == NativeWebSocket.class) {
      String protocol = ifSecure("wss", "ws");
      String webSocketURL = protocol + "://" + req.headers().get(HttpHeaderNames.HOST) + path;

      WebSocketServerHandshakerFactory wsFactory = new WebSocketServerHandshakerFactory(
          webSocketURL, null, true, wsMaxMessageSize);
      WebSocketServerHandshaker handshaker = wsFactory.newHandshaker(req);
      NettyWebSocket result = new NettyWebSocket(ctx, handshaker, (ws) -> {
        handshaker.handshake(ctx.channel(), (FullHttpRequest) req)
            .addListener(FIRE_EXCEPTION_ON_FAILURE)
            .addListener(payload -> ws.connect())
            .addListener(FIRE_EXCEPTION_ON_FAILURE);
      });
      ctx.channel().attr(NettyWebSocket.KEY).set(result);
      return (T) result;
    } else if (type == Sse.class) {
      NettySse sse = new NettySse(ctx, responseHeaders);
      return (T) sse;
    } else if (type == NativePushPromise.class) {
      return (T) new NettyPush(ctx,
          req.headers().getInt(HttpConversionUtil.ExtensionHeaderNames.STREAM_ID.text()),
          header("host").orElse(ip()), ifSecure("https", "http"));
    }
    throw new UnsupportedOperationException("Not Supported: " + type);
  }

  @Override
  public void startAsync(final Executor executor, final Runnable runnable) {
    Channel channel = ctx.channel();
    channel.attr(NEED_FLUSH).set(false);
    channel.attr(ASYNC).set(true);

    ReferenceCounted body = ((ByteBufHolder) req).content();
    body.retain();
    executor.execute(() -> {
      try {
        runnable.run();
      } finally {
        body.release();
      }
    });
  }

  private org.jooby.Cookie cookie(final Cookie c) {
    org.jooby.Cookie.Definition cookie = new org.jooby.Cookie.Definition(c.name(), c.value());
    Optional.ofNullable(c.domain()).ifPresent(cookie::domain);
    Optional.ofNullable(c.path()).ifPresent(cookie::path);

    return cookie.toCookie();
  }

  private Multimap<String, String> decodeParams() throws IOException {
    if (params == null) {
      params = ArrayListMultimap.create();
      files = ArrayListMultimap.create();

      query.parameters()
          .forEach((name, values) -> values.forEach(value -> params.put(name, value)));

      HttpMethod method = req.method();
      boolean hasBody = method.equals(HttpMethod.POST) || method.equals(HttpMethod.PUT)
          || method.equals(HttpMethod.PATCH);
      boolean formLike = false;
      if (req.headers().contains("Content-Type")) {
        String contentType = req.headers().get("Content-Type").toLowerCase();
        formLike = (contentType.startsWith(MediaType.multipart.name())
            || contentType.startsWith(MediaType.form.name()));
      }
      if (hasBody && formLike) {
        HttpPostRequestDecoder decoder = new HttpPostRequestDecoder(
            new DefaultHttpDataFactory(), req);
        try {
          Function<HttpPostRequestDecoder, Boolean> hasNext = it -> {
            try {
              return it.hasNext();
            } catch (HttpPostRequestDecoder.EndOfDataDecoderException ex) {
              return false;
            }
          };
          while (hasNext.apply(decoder)) {
            HttpData field = (HttpData) decoder.next();
            String name = field.getName();
            if (field.getHttpDataType() == HttpDataType.FileUpload) {
              files.put(name, new NettyUpload((FileUpload) field, tmpdir));
            } else {
              params.put(name, field.getString());
            }
          }
        } finally {
          decoder.destroy();
        }
      }
    }
    return params;
  }

  private <T> T ifSecure(final T then, final T otherwise) {
    return ctx.pipeline().get("ssl") != null ? then : otherwise;
  }
}
