/**
 *                                  Apache License
 *                            Version 2.0, January 2004
 *                         http://www.apache.org/licenses/
 *
 *    TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *    1. Definitions.
 *
 *       "License" shall mean the terms and conditions for use, reproduction,
 *       and distribution as defined by Sections 1 through 9 of this document.
 *
 *       "Licensor" shall mean the copyright owner or entity authorized by
 *       the copyright owner that is granting the License.
 *
 *       "Legal Entity" shall mean the union of the acting entity and all
 *       other entities that control, are controlled by, or are under common
 *       control with that entity. For the purposes of this definition,
 *       "control" means (i) the power, direct or indirect, to cause the
 *       direction or management of such entity, whether by contract or
 *       otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *       outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *       "You" (or "Your") shall mean an individual or Legal Entity
 *       exercising permissions granted by this License.
 *
 *       "Source" form shall mean the preferred form for making modifications,
 *       including but not limited to software source code, documentation
 *       source, and configuration files.
 *
 *       "Object" form shall mean any form resulting from mechanical
 *       transformation or translation of a Source form, including but
 *       not limited to compiled object code, generated documentation,
 *       and conversions to other media types.
 *
 *       "Work" shall mean the work of authorship, whether in Source or
 *       Object form, made available under the License, as indicated by a
 *       copyright notice that is included in or attached to the work
 *       (an example is provided in the Appendix below).
 *
 *       "Derivative Works" shall mean any work, whether in Source or Object
 *       form, that is based on (or derived from) the Work and for which the
 *       editorial revisions, annotations, elaborations, or other modifications
 *       represent, as a whole, an original work of authorship. For the purposes
 *       of this License, Derivative Works shall not include works that remain
 *       separable from, or merely link (or bind by name) to the interfaces of,
 *       the Work and Derivative Works thereof.
 *
 *       "Contribution" shall mean any work of authorship, including
 *       the original version of the Work and any modifications or additions
 *       to that Work or Derivative Works thereof, that is intentionally
 *       submitted to Licensor for inclusion in the Work by the copyright owner
 *       or by an individual or Legal Entity authorized to submit on behalf of
 *       the copyright owner. For the purposes of this definition, "submitted"
 *       means any form of electronic, verbal, or written communication sent
 *       to the Licensor or its representatives, including but not limited to
 *       communication on electronic mailing lists, source code control systems,
 *       and issue tracking systems that are managed by, or on behalf of, the
 *       Licensor for the purpose of discussing and improving the Work, but
 *       excluding communication that is conspicuously marked or otherwise
 *       designated in writing by the copyright owner as "Not a Contribution."
 *
 *       "Contributor" shall mean Licensor and any individual or Legal Entity
 *       on behalf of whom a Contribution has been received by Licensor and
 *       subsequently incorporated within the Work.
 *
 *    2. Grant of Copyright License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       copyright license to reproduce, prepare Derivative Works of,
 *       publicly display, publicly perform, sublicense, and distribute the
 *       Work and such Derivative Works in Source or Object form.
 *
 *    3. Grant of Patent License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       (except as stated in this section) patent license to make, have made,
 *       use, offer to sell, sell, import, and otherwise transfer the Work,
 *       where such license applies only to those patent claims licensable
 *       by such Contributor that are necessarily infringed by their
 *       Contribution(s) alone or by combination of their Contribution(s)
 *       with the Work to which such Contribution(s) was submitted. If You
 *       institute patent litigation against any entity (including a
 *       cross-claim or counterclaim in a lawsuit) alleging that the Work
 *       or a Contribution incorporated within the Work constitutes direct
 *       or contributory patent infringement, then any patent licenses
 *       granted to You under this License for that Work shall terminate
 *       as of the date such litigation is filed.
 *
 *    4. Redistribution. You may reproduce and distribute copies of the
 *       Work or Derivative Works thereof in any medium, with or without
 *       modifications, and in Source or Object form, provided that You
 *       meet the following conditions:
 *
 *       (a) You must give any other recipients of the Work or
 *           Derivative Works a copy of this License; and
 *
 *       (b) You must cause any modified files to carry prominent notices
 *           stating that You changed the files; and
 *
 *       (c) You must retain, in the Source form of any Derivative Works
 *           that You distribute, all copyright, patent, trademark, and
 *           attribution notices from the Source form of the Work,
 *           excluding those notices that do not pertain to any part of
 *           the Derivative Works; and
 *
 *       (d) If the Work includes a "NOTICE" text file as part of its
 *           distribution, then any Derivative Works that You distribute must
 *           include a readable copy of the attribution notices contained
 *           within such NOTICE file, excluding those notices that do not
 *           pertain to any part of the Derivative Works, in at least one
 *           of the following places: within a NOTICE text file distributed
 *           as part of the Derivative Works; within the Source form or
 *           documentation, if provided along with the Derivative Works; or,
 *           within a display generated by the Derivative Works, if and
 *           wherever such third-party notices normally appear. The contents
 *           of the NOTICE file are for informational purposes only and
 *           do not modify the License. You may add Your own attribution
 *           notices within Derivative Works that You distribute, alongside
 *           or as an addendum to the NOTICE text from the Work, provided
 *           that such additional attribution notices cannot be construed
 *           as modifying the License.
 *
 *       You may add Your own copyright statement to Your modifications and
 *       may provide additional or different license terms and conditions
 *       for use, reproduction, or distribution of Your modifications, or
 *       for any such Derivative Works as a whole, provided Your use,
 *       reproduction, and distribution of the Work otherwise complies with
 *       the conditions stated in this License.
 *
 *    5. Submission of Contributions. Unless You explicitly state otherwise,
 *       any Contribution intentionally submitted for inclusion in the Work
 *       by You to the Licensor shall be under the terms and conditions of
 *       this License, without any additional terms or conditions.
 *       Notwithstanding the above, nothing herein shall supersede or modify
 *       the terms of any separate license agreement you may have executed
 *       with Licensor regarding such Contributions.
 *
 *    6. Trademarks. This License does not grant permission to use the trade
 *       names, trademarks, service marks, or product names of the Licensor,
 *       except as required for reasonable and customary use in describing the
 *       origin of the Work and reproducing the content of the NOTICE file.
 *
 *    7. Disclaimer of Warranty. Unless required by applicable law or
 *       agreed to in writing, Licensor provides the Work (and each
 *       Contributor provides its Contributions) on an "AS IS" BASIS,
 *       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *       implied, including, without limitation, any warranties or conditions
 *       of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *       PARTICULAR PURPOSE. You are solely responsible for determining the
 *       appropriateness of using or redistributing the Work and assume any
 *       risks associated with Your exercise of permissions under this License.
 *
 *    8. Limitation of Liability. In no event and under no legal theory,
 *       whether in tort (including negligence), contract, or otherwise,
 *       unless required by applicable law (such as deliberate and grossly
 *       negligent acts) or agreed to in writing, shall any Contributor be
 *       liable to You for damages, including any direct, indirect, special,
 *       incidental, or consequential damages of any character arising as a
 *       result of this License or out of the use or inability to use the
 *       Work (including but not limited to damages for loss of goodwill,
 *       work stoppage, computer failure or malfunction, or any and all
 *       other commercial damages or losses), even if such Contributor
 *       has been advised of the possibility of such damages.
 *
 *    9. Accepting Warranty or Additional Liability. While redistributing
 *       the Work or Derivative Works thereof, You may choose to offer,
 *       and charge a fee for, acceptance of support, warranty, indemnity,
 *       or other liability obligations and/or rights consistent with this
 *       License. However, in accepting such obligations, You may act only
 *       on Your own behalf and on Your sole responsibility, not on behalf
 *       of any other Contributor, and only if You agree to indemnify,
 *       defend, and hold each Contributor harmless for any liability
 *       incurred by, or claims asserted against, such Contributor by reason
 *       of your accepting any such warranty or additional liability.
 *
 *    END OF TERMS AND CONDITIONS
 *
 *    APPENDIX: How to apply the Apache License to your work.
 *
 *       To apply the Apache License to your work, attach the following
 *       boilerplate notice, with the fields enclosed by brackets "{}"
 *       replaced with your own identifying information. (Don't include
 *       the brackets!)  The text should be enclosed in the appropriate
 *       comment syntax for the file format. We also recommend that a
 *       file or class name and description of purpose be included on the
 *       same "printed page" as the copyright notice for easier
 *       identification within third-party archives.
 *
 *    Copyright 2014 Edgar Espina
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package org.jooby.quartz;

import static java.util.Objects.requireNonNull;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.function.BiConsumer;
import java.util.function.Consumer;

import org.jooby.Env;
import org.jooby.Jooby;
import org.jooby.internal.quartz.JobExpander;
import org.jooby.internal.quartz.QuartzProvider;
import org.quartz.Job;
import org.quartz.JobBuilder;
import org.quartz.JobDetail;
import org.quartz.JobExecutionContext;
import org.quartz.JobKey;
import org.quartz.Scheduler;
import org.quartz.Trigger;
import org.quartz.TriggerBuilder;
import org.quartz.TriggerKey;

import com.google.common.collect.Lists;
import com.google.inject.Binder;
import com.google.inject.TypeLiteral;
import com.google.inject.name.Names;
import com.typesafe.config.Config;
import com.typesafe.config.ConfigFactory;

/**
 * Build and create a {@link Scheduler Quartz Scheduler} and {@link Job jobs}.
 *
 * <h1>usage</h1>
 *
 * <pre>
 *  {
 *    use(new Quartz().with(MyJob.class));
 *  }
 * </pre>
 *
 * Previous example will startup Quartz and schedule MyJob.
 *
 * <h1>jobs</h1>
 * <p>
 * A job can implement the {@link Job} interface as described in the <a
 * href="http://quartz-scheduler.org/documentation">Quartz documentation</a>
 * </p>
 *
 * <p>
 * If you prefer to not implement the {@link Job} interface, all you have to do is to annotated a
 * method with the {@link Scheduled} annotation.
 * </p>
 * <p>
 * By default, job name is set the class name or to the method name. Default group is set to the
 * package name of the job class.
 * </p>
 *
 * <h2>job methods</h2>
 * <p>
 * A job method must follow this rules:
 * </p>
 *
 * <ul>
 * <li>It must be a public method</li>
 * <li>Without a return value</li>
 * <li>Have ZERO arguments</li>
 * <li>or just ONE argument of type {@link JobExecutionContext}</li>
 * </ul>
 *
 * The next section will you show how to add a trigger to a job and some examples too.
 *
 * <h1>triggers</h1>
 * <p>
 * Trigger are defined by the {@link Scheduled} annotation. The annotation defines a single and
 * required attribute, which is basically a trigger expression or a reference to one.
 * </p>
 *
 * Examples:
 * <p>
 * Run every 5 minutes, start immediately and repeat for ever:
 * </p>
 *
 * <pre>
 * &#64;Scheduled("5m")
 *
 * &#64;Scheduled("5m; delay=0")
 *
 * &#64;Scheduled("5m; delay=0; repeat=*")
 * </pre>
 *
 * Previous, expressions are identical.
 *
 * <p>
 * Run every 1 hour with an initial delay of 15 minutes for 10 times
 * </p>
 *
 * <pre>
 * &#64;Scheduled("1h; delay=15m; repeat=10")
 * </pre>
 *
 * <p>
 * Fire at 12pm (noon) every day
 * </p>
 *
 * <pre>
 * 0 0 12 * * ?
 * </pre>
 *
 * <p>
 * Fire at 10:15am every day
 * </p>
 *
 * <pre>
 * 0 15 10 ? * *
 * </pre>
 *
 * <h1>grouping jobs together</h1>
 * <p>
 * If you have two or more jobs doing something similar, it is possible to group all them into one
 * single class:
 * </p>
 *
 * <pre>
 *  public class MyJobs {
 *    &#64;Scheduled("5minutes")
 *    public void job1() {
 *      ...
 *    }
 *
 *    &#64;Scheduled("1h")
 *    public void job2() {
 *      ...
 *    }
 *  }
 * </pre>
 *
 * <h1>dependency injection</h1>
 * <p>
 * Not much to add here, just let you know jobs are created by Guice.
 * </p>
 *
 * <pre>
 *  public class MyJob {
 *
 *    private A a;
 *
 *    &#64;Inject
 *    public MyJob(A a) {
 *      this.a = a;
 *    }
 *
 *    &#64;Scheduled("5minutes")
 *    public void doWork() {
 *      this.a.doWork();
 *    }
 *  }
 * </pre>
 *
 * <p>
 * Injecting a {@link Scheduler}
 * </p>
 *
 * <pre>
 *  public class MyJobManager {
 *
 *    private Scheduler scheduler;
 *
 *    &#64;Inject
 *    public MyJobManager(Scheduler scheduler) {
 *      this.scheduler = scheduler;
 *    }
 *  }
 * </pre>
 *
 * <h1>configuration</h1>
 * <p>
 * Example: Setting max number of threads
 * </p>
 *
 * <pre>
 *  # application.conf
 *  org.quartz.threadPool.threadCount = 1 # default is number of available processors
 * </pre>
 * <p>
 * Configuration follows the <a href="http://quartz-scheduler.org/documentation">Quartz
 * documentation</a>. The only difference is that you need to put add the properties on your
 * <code>*.conf</code> file, NOT in a custom <code>quartz.properties</code> file.
 * </p>
 *
 * <h2>jdbc store</h2>
 * <p>
 * Jdbc Store is fully supported but it depends on the <code>jooby-jdbc</code> module. So, in order
 * to use the Jdbc Store you need to follow these steps:
 * </p>
 *
 * <p>
 * 1. Install the Jdbc module:
 * </p>
 *
 * <pre>
 * {
 *   use(new Jdbc());
 *   use(new Quartz(MyJob.class));
 * }
 * </pre>
 *
 * <p>
 * 2. Set the quartz properties:
 * </p>
 *
 * <pre>
 *  org.quartz.jobStore.class = org.quartz.impl.jdbcjobstore.JobStoreTX
 *  org.quartz.jobStore.driverDelegateClass = org.quartz.impl.jdbcjobstore.StdJDBCDelegate
 *  org.quartz.jobStore.dataSource = db
 * </pre>
 *
 * <h1>adding jobs programmatically</h1>
 * <p>
 * When {@link Scheduled} isn't not enough and/or if you prefer to build jobs manually, you can try
 * one of the available alternatives.
 * </p>
 *
 * <p>
 * Example 1: build the trigger and use default job naming
 * </p>
 *
 * <pre>
 *   {
 *     use(new Quartz()
 *      .with(MyJob.class, trigger {@literal ->} {
 *        trigger
 *          .withSchedule(withIntervalInDays(3))
 *          .startAt(futureDate(10, MINUTES));
 *      })
 *     );
 *   }
 * </pre>
 *
 * <p>
 * Example 2: build the job, the trigger and use default job naming
 * </p>
 *
 * <pre>
 *   {
 *     use(new Quartz()
 *      .with(MyJob.class, (job, trigger) {@literal ->} {
 *        job.withDescription("etc...");
 *
 *        trigger
 *          .withSchedule(withIntervalInDays(3))
 *          .startAt(futureDate(10, MINUTES));
 *      })
 *     );
 *   }
 * </pre>
 *
 * <p>
 * Example 3: build and set everything from scratch
 * </p>
 *
 * <pre>
 *   {
 *     use(new Quartz()
 *      .with(
 *        newJob(MyJob.class).withDescription("etc...")
 *          .build(),
 *        newTrigger()
 *          .withSchedule(withIntervalInDays(3))
 *          .startAt(futureDate(10, MINUTES))
 *          .build()
 *      })
 *     );
 *   }
 * </pre>
 *
 * Enjoy it!
 *
 * @author edgar
 * @since 0.5.0
 */
public class Quartz implements Jooby.Module {

  private final List<Class<?>> jobs;

  private final Map<JobDetail, Trigger> jobMap = new HashMap<>();

  /**
   * Creates a new {@link Quartz} module. Optionally add some jobs.
   *
   * @param jobs Jobs to setup. Optional.
   * @see #with(Class)
   */
  public Quartz(final Class<?>... jobs) {
    this.jobs = Lists.newArrayList(jobs);
  }

  /**
   * Schedule the provided job and trigger.
   *
   * @param job A job to schedule.
   * @param trigger A trigger for provided job.
   * @return This quartz instance.
   */
  public Quartz with(final JobDetail job, final Trigger trigger) {
    requireNonNull(job, "Job is required.");
    requireNonNull(trigger, "Trigger is required.");
    jobMap.put(job, trigger);
    return this;
  }

  /**
   * Setup and schedule the provided job, it might implement a {@link Job} and the {@link Scheduled}
   * annotation must be present.
   *
   * @param jobClass A jobClass to setup and schedule.
   * @return This quartz instance.
   */
  public Quartz with(final Class<?> jobClass) {
    jobs.add(jobClass);
    return this;
  }

  /**
   * Schedule the provided job and trigger. This method will setup a default name and group for
   * both.
   *
   * @param jobClass A jobClass to setup and schedule.
   * @param configurer A callback to setup the job and trigger.
   * @return This quartz instance.
   */
  public Quartz with(final Class<? extends Job> jobClass,
      final BiConsumer<JobBuilder, TriggerBuilder<Trigger>> configurer) {
    requireNonNull(jobClass, "Job class is required.");
    JobBuilder job = JobBuilder.newJob(jobClass)
        .withIdentity(
            JobKey.jobKey(jobClass.getSimpleName(), jobClass.getPackage().getName())
        );
    TriggerBuilder<Trigger> trigger = TriggerBuilder.newTrigger()
        .withIdentity(
            TriggerKey.triggerKey(jobClass.getSimpleName(), jobClass.getPackage().getName())
        );
    configurer.accept(job, trigger);
    return with(job.build(), trigger.build());
  }

  /**
   * Schedule the provided job and trigger. This method will setup a default name and group for
   * both.
   *
   * @param jobClass A jobClass to setup and schedule.
   * @param configurer A callback to setup the trigger.
   * @return This quartz instance.
   */
  public Quartz with(final Class<? extends Job> jobClass,
      final Consumer<TriggerBuilder<Trigger>> configurer) {
    requireNonNull(jobClass, "Job class is required.");
    JobBuilder job = JobBuilder.newJob(jobClass)
        .withIdentity(
            JobKey.jobKey(jobClass.getSimpleName(), jobClass.getPackage().getName())
        );
    TriggerBuilder<Trigger> trigger = TriggerBuilder.newTrigger()
        .withIdentity(
            TriggerKey.triggerKey(jobClass.getSimpleName(), jobClass.getPackage().getName())
        );
    configurer.accept(trigger);
    return with(job.build(), trigger.build());
  }

  @Override
  public void configure(final Env env, final Config config, final Binder binder) {
    jobMap.putAll(JobExpander.jobs(config, jobs));
    binder.bind(Scheduler.class).toProvider(QuartzProvider.class).asEagerSingleton();
    env.lifeCycle(QuartzProvider.class);
    binder.bind(new TypeLiteral<Map<JobDetail, Trigger>>() {
    }).annotatedWith(Names.named("org.quartz.jobs")).toInstance(jobMap);
  }

  @Override
  public Config config() {
    return ConfigFactory.parseResources(getClass(), "quartz.conf")
        .withFallback(ConfigFactory.parseResources(Job.class, "quartz.properties"));
  }
}
