/**
 *                                  Apache License
 *                            Version 2.0, January 2004
 *                         http://www.apache.org/licenses/
 *
 *    TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *    1. Definitions.
 *
 *       "License" shall mean the terms and conditions for use, reproduction,
 *       and distribution as defined by Sections 1 through 9 of this document.
 *
 *       "Licensor" shall mean the copyright owner or entity authorized by
 *       the copyright owner that is granting the License.
 *
 *       "Legal Entity" shall mean the union of the acting entity and all
 *       other entities that control, are controlled by, or are under common
 *       control with that entity. For the purposes of this definition,
 *       "control" means (i) the power, direct or indirect, to cause the
 *       direction or management of such entity, whether by contract or
 *       otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *       outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *       "You" (or "Your") shall mean an individual or Legal Entity
 *       exercising permissions granted by this License.
 *
 *       "Source" form shall mean the preferred form for making modifications,
 *       including but not limited to software source code, documentation
 *       source, and configuration files.
 *
 *       "Object" form shall mean any form resulting from mechanical
 *       transformation or translation of a Source form, including but
 *       not limited to compiled object code, generated documentation,
 *       and conversions to other media types.
 *
 *       "Work" shall mean the work of authorship, whether in Source or
 *       Object form, made available under the License, as indicated by a
 *       copyright notice that is included in or attached to the work
 *       (an example is provided in the Appendix below).
 *
 *       "Derivative Works" shall mean any work, whether in Source or Object
 *       form, that is based on (or derived from) the Work and for which the
 *       editorial revisions, annotations, elaborations, or other modifications
 *       represent, as a whole, an original work of authorship. For the purposes
 *       of this License, Derivative Works shall not include works that remain
 *       separable from, or merely link (or bind by name) to the interfaces of,
 *       the Work and Derivative Works thereof.
 *
 *       "Contribution" shall mean any work of authorship, including
 *       the original version of the Work and any modifications or additions
 *       to that Work or Derivative Works thereof, that is intentionally
 *       submitted to Licensor for inclusion in the Work by the copyright owner
 *       or by an individual or Legal Entity authorized to submit on behalf of
 *       the copyright owner. For the purposes of this definition, "submitted"
 *       means any form of electronic, verbal, or written communication sent
 *       to the Licensor or its representatives, including but not limited to
 *       communication on electronic mailing lists, source code control systems,
 *       and issue tracking systems that are managed by, or on behalf of, the
 *       Licensor for the purpose of discussing and improving the Work, but
 *       excluding communication that is conspicuously marked or otherwise
 *       designated in writing by the copyright owner as "Not a Contribution."
 *
 *       "Contributor" shall mean Licensor and any individual or Legal Entity
 *       on behalf of whom a Contribution has been received by Licensor and
 *       subsequently incorporated within the Work.
 *
 *    2. Grant of Copyright License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       copyright license to reproduce, prepare Derivative Works of,
 *       publicly display, publicly perform, sublicense, and distribute the
 *       Work and such Derivative Works in Source or Object form.
 *
 *    3. Grant of Patent License. Subject to the terms and conditions of
 *       this License, each Contributor hereby grants to You a perpetual,
 *       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *       (except as stated in this section) patent license to make, have made,
 *       use, offer to sell, sell, import, and otherwise transfer the Work,
 *       where such license applies only to those patent claims licensable
 *       by such Contributor that are necessarily infringed by their
 *       Contribution(s) alone or by combination of their Contribution(s)
 *       with the Work to which such Contribution(s) was submitted. If You
 *       institute patent litigation against any entity (including a
 *       cross-claim or counterclaim in a lawsuit) alleging that the Work
 *       or a Contribution incorporated within the Work constitutes direct
 *       or contributory patent infringement, then any patent licenses
 *       granted to You under this License for that Work shall terminate
 *       as of the date such litigation is filed.
 *
 *    4. Redistribution. You may reproduce and distribute copies of the
 *       Work or Derivative Works thereof in any medium, with or without
 *       modifications, and in Source or Object form, provided that You
 *       meet the following conditions:
 *
 *       (a) You must give any other recipients of the Work or
 *           Derivative Works a copy of this License; and
 *
 *       (b) You must cause any modified files to carry prominent notices
 *           stating that You changed the files; and
 *
 *       (c) You must retain, in the Source form of any Derivative Works
 *           that You distribute, all copyright, patent, trademark, and
 *           attribution notices from the Source form of the Work,
 *           excluding those notices that do not pertain to any part of
 *           the Derivative Works; and
 *
 *       (d) If the Work includes a "NOTICE" text file as part of its
 *           distribution, then any Derivative Works that You distribute must
 *           include a readable copy of the attribution notices contained
 *           within such NOTICE file, excluding those notices that do not
 *           pertain to any part of the Derivative Works, in at least one
 *           of the following places: within a NOTICE text file distributed
 *           as part of the Derivative Works; within the Source form or
 *           documentation, if provided along with the Derivative Works; or,
 *           within a display generated by the Derivative Works, if and
 *           wherever such third-party notices normally appear. The contents
 *           of the NOTICE file are for informational purposes only and
 *           do not modify the License. You may add Your own attribution
 *           notices within Derivative Works that You distribute, alongside
 *           or as an addendum to the NOTICE text from the Work, provided
 *           that such additional attribution notices cannot be construed
 *           as modifying the License.
 *
 *       You may add Your own copyright statement to Your modifications and
 *       may provide additional or different license terms and conditions
 *       for use, reproduction, or distribution of Your modifications, or
 *       for any such Derivative Works as a whole, provided Your use,
 *       reproduction, and distribution of the Work otherwise complies with
 *       the conditions stated in this License.
 *
 *    5. Submission of Contributions. Unless You explicitly state otherwise,
 *       any Contribution intentionally submitted for inclusion in the Work
 *       by You to the Licensor shall be under the terms and conditions of
 *       this License, without any additional terms or conditions.
 *       Notwithstanding the above, nothing herein shall supersede or modify
 *       the terms of any separate license agreement you may have executed
 *       with Licensor regarding such Contributions.
 *
 *    6. Trademarks. This License does not grant permission to use the trade
 *       names, trademarks, service marks, or product names of the Licensor,
 *       except as required for reasonable and customary use in describing the
 *       origin of the Work and reproducing the content of the NOTICE file.
 *
 *    7. Disclaimer of Warranty. Unless required by applicable law or
 *       agreed to in writing, Licensor provides the Work (and each
 *       Contributor provides its Contributions) on an "AS IS" BASIS,
 *       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *       implied, including, without limitation, any warranties or conditions
 *       of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *       PARTICULAR PURPOSE. You are solely responsible for determining the
 *       appropriateness of using or redistributing the Work and assume any
 *       risks associated with Your exercise of permissions under this License.
 *
 *    8. Limitation of Liability. In no event and under no legal theory,
 *       whether in tort (including negligence), contract, or otherwise,
 *       unless required by applicable law (such as deliberate and grossly
 *       negligent acts) or agreed to in writing, shall any Contributor be
 *       liable to You for damages, including any direct, indirect, special,
 *       incidental, or consequential damages of any character arising as a
 *       result of this License or out of the use or inability to use the
 *       Work (including but not limited to damages for loss of goodwill,
 *       work stoppage, computer failure or malfunction, or any and all
 *       other commercial damages or losses), even if such Contributor
 *       has been advised of the possibility of such damages.
 *
 *    9. Accepting Warranty or Additional Liability. While redistributing
 *       the Work or Derivative Works thereof, You may choose to offer,
 *       and charge a fee for, acceptance of support, warranty, indemnity,
 *       or other liability obligations and/or rights consistent with this
 *       License. However, in accepting such obligations, You may act only
 *       on Your own behalf and on Your sole responsibility, not on behalf
 *       of any other Contributor, and only if You agree to indemnify,
 *       defend, and hold each Contributor harmless for any liability
 *       incurred by, or claims asserted against, such Contributor by reason
 *       of your accepting any such warranty or additional liability.
 *
 *    END OF TERMS AND CONDITIONS
 *
 *    APPENDIX: How to apply the Apache License to your work.
 *
 *       To apply the Apache License to your work, attach the following
 *       boilerplate notice, with the fields enclosed by brackets "{}"
 *       replaced with your own identifying information. (Don't include
 *       the brackets!)  The text should be enclosed in the appropriate
 *       comment syntax for the file format. We also recommend that a
 *       file or class name and description of purpose be included on the
 *       same "printed page" as the copyright notice for easier
 *       identification within third-party archives.
 *
 *    Copyright 2014 Edgar Espina
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package org.jooby.rx;

import com.google.inject.Binder;
import com.typesafe.config.Config;
import com.typesafe.config.ConfigFactory;
import static java.util.Objects.requireNonNull;
import org.jooby.Deferred;
import org.jooby.Env;
import org.jooby.Route;
import org.jooby.exec.Exec;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import rx.Completable;
import rx.Observable;
import rx.Scheduler;
import rx.Single;
import rx.Subscriber;
import rx.plugins.RxJavaPlugins;
import rx.plugins.RxJavaSchedulersHook;
import rx.schedulers.Schedulers;

import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.Executor;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.Function;

/**
 * <h1>rxjava</h1>
 * <p>
 * Reactive programming via <a href="https://github.com/ReactiveX/RxJava">RxJava library</a>.
 * </p>
 * <p>
 * RxJava is a Java VM implementation of <a href="http://reactivex.io">Reactive Extensions</a>: a
 * library for composing asynchronous and event-based programs by using observable sequences.
 * </p>
 *
 * <h2>exports</h2>
 * <ul>
 * <li>Route mapper that converts {@link Observable} (and family) into {@link Deferred} API.
 * </li>
 * <li>
 * manage the lifecycle of {@link Scheduler schedulers} and make sure they go down on application
 * shutdown time.
 * </li>
 * </ul>
 *
 * <h2>usage</h2>
 * <pre>{@code
 * ...
 * import org.jooby.rx.Rx;
 * ...
 *
 * {
 *   use(new Rx());
 *
 *   get("/", req -> Observable.from("reactive programming in jooby!"));
 * }
 * }</pre>
 *
 * <h2>how it works?</h2>
 * <p>
 * Previous example is translated to:
 * </p>
 * <pre>{@code
 * {
 *   use(new Rx());
 *
 *   get("/", req -> {
 *
 *    return new Deferred(deferred -> {
 *      Observable.from("reactive programming in jooby!")
 *        .subscribe(deferred::resolve, deferred::reject);
 *    });
 *
 *   });
 * }
 * }</pre>
 *
 * <p>
 * Translation is done via {@link Rx#rx()} route mapper. If you are a
 * <a href="https://github.com/ReactiveX/RxJava">rxjava</a> programmer then you don't need to worry
 * for learning a new API and semantic. The {@link Rx#rx()} route mapper deal and take cares of
 * the {@link Deferred} API.
 * </p>
 *
 * <h2>rx mapper</h2>
 * <p>
 * Advanced observable configuration is allowed via function adapter:
 * </p>
 *
 * <pre>{@code
 * ...
 * import org.jooby.rx.Rx;
 * ...
 *
 * {
 *   use(new Rx()
 *     .withObservable(observable -> observable.observeOn(Scheduler.io()),
 *     .withSingle(single -> single.observeOn(Scheduler.io()),
 *     .withCompletable(completable -> completable.observeOn(Scheduler.io()));
 *
 *   get("/observable", req -> Observable...);
 *
 *   get("/single", req -> Single...);
 *
 *   ....
 *
 *   get("/completable", req -> Completable...);
 *
 * }
 * }</pre>
 *
 * <p>
 * Here every Observable/Single/Completable from a route handler will observe on the <code>io</code>
 * scheduler.
 * </p>
 *
 * <h2>schedulers</h2>
 * <p>
 * This module provides the default {@link Scheduler} from
 * <a href="https://github.com/ReactiveX/RxJava">rxjava</a>. But also let you define your own
 * {@link Scheduler scheduler} using the {@link Exec} module.
 * </p>
 *
 * <pre>
 * rx.schedulers.io = forkjoin
 * rx.schedulers.computation = fixed
 * rx.schedulers.newThread = "fixed = 10"
 * </pre>
 *
 * <p>
 * The previous example defines a:
 * </p>
 * <ul>
 * <li>forkjoin pool for {@link Schedulers#io()}</li>
 * <li>fixed thread pool equals to the number of available processors for
 * {@link Schedulers#computation()}</li>
 * <li>fixed thread pool with a max of 10 for {@link Schedulers#newThread()}</li>
 * </ul>
 *
 * <p>
 * Of course, you can define/override all, some or none of them. In any case the {@link Scheduler}
 * will be shutdown at application shutdown time.
 * </p>
 *
 * @author edgar
 * @since 1.0.0.CR3
 */
@SuppressWarnings("rawtypes")
public class Rx extends Exec {

  static class DeferredSubscriber extends Subscriber<Object> {

    private Deferred deferred;

    private final AtomicBoolean done = new AtomicBoolean(false);

    public DeferredSubscriber(final Deferred deferred) {
      this.deferred = deferred;
    }

    @Override
    public void onCompleted() {
      if (done.compareAndSet(false, true)) {
        deferred.resolve(null);
      }
      deferred = null;
    }

    @Override
    public void onError(final Throwable cause) {
      done.set(true);
      deferred.reject(cause);
    }

    @Override
    public void onNext(final Object value) {
      if (done.compareAndSet(false, true)) {
        deferred.resolve(value);
      }
    }
  }

  /** The logging system. */
  private final Logger log = LoggerFactory.getLogger(getClass());

  private Function<Observable, Observable> observable = Function.identity();

  private Function<Single, Single> single = Function.identity();

  private Function<Completable, Completable> completable = Function.identity();

  /**
   * Creates a new {@link Rx} module.
   */
  public Rx() {
    // daemon by default.
    daemon(true);
  }

  /**
   * Map a rx object like {@link Observable}, {@link Single} or {@link Completable} into a
   * {@link Deferred} object.
   *
   * <pre>{@code
   * ...
   * import org.jooby.rx.Rx;
   * ...
   *
   * {
   *   with(() -> {
   *     get("/1", req -> Observable...);
   *
   *     get("/2", req -> Single...);
   *
   *     ....
   *
   *     get("/N", req -> Completable...);
   *
   *   }).map(Rx.rx());
   * }
   * }</pre>
   *
   * @return A new mapper.
   */
  public static Route.Mapper<Object> rx() {
    return rx(Function.identity(), Function.identity());
  }

  /**
   * Map a rx object like {@link Observable}, {@link Single} or {@link Completable} into a
   * {@link Deferred} object.
   *
   * <pre>{@code
   * ...
   * import org.jooby.rx.Rx;
   * ...
   *
   * {
   *   use(new Rx());
   *
   *   with(() -> {
   *     get("/1", req -> Observable...);
   *
   *     get("/2", req -> Observable...);
   *
   *     ....
   *
   *     get("/N", req -> Observable...);
   *
   *   }).map(Rx.rx(
   *       observable -> observable.observeOn(Scheduler.io()),
   *       single -> single.observeOn(Scheduler.io()),
   *       completable -> completable.observeOn(Scheduler.io())));
   * }
   * }</pre>
   *
   * @param observable Observable adapter.
   * @param single Single adapter.
   * @return A new mapper.
   */
  public static Route.Mapper<Object> rx(final Function<Observable, Observable> observable,
    final Function<Single, Single> single) {
    return rx(observable, single, Function.identity());
  }

  /**
   * Map a rx object like {@link Observable}, {@link Single} or {@link Completable} into a
   * {@link Deferred} object.
   *
   * <pre>{@code
   * ...
   * import org.jooby.rx.Rx;
   * ...
   *
   * {
   *   use(new Rx());
   *
   *   with(() -> {
   *     get("/1", req -> Observable...);
   *
   *     get("/2", req -> Observable...);
   *
   *     ....
   *
   *     get("/N", req -> Observable...);
   *
   *   }).map(Rx.rx(
   *       observable -> observable.observeOn(Scheduler.io()),
   *       single -> single.observeOn(Scheduler.io()),
   *       completable -> completable.observeOn(Scheduler.io())));
   * }
   * }</pre>
   *
   * @param observable Observable adapter.
   * @param single Single adapter.
   * @param completable Completable adapter.
   * @return A new mapper.
   */
  @SuppressWarnings("unchecked")
  public static Route.Mapper<Object> rx(final Function<Observable, Observable> observable,
    final Function<Single, Single> single, final Function<Completable, Completable> completable) {
    requireNonNull(observable, "Observable's adapter is required.");
    requireNonNull(single, "Single's adapter is required.");
    requireNonNull(completable, "Completable's adapter is required.");

    return Route.Mapper.create("rx", v -> {
      if (v instanceof Observable) {
        return new Deferred(
          deferred -> observable.apply((Observable) v).subscribe(new DeferredSubscriber(deferred)));
      } else if (v instanceof Single) {
        return new Deferred(
          deferred -> single.apply((Single) v).subscribe(new DeferredSubscriber(deferred)));
      } else if (v instanceof Completable) {
        return new Deferred(deferred -> completable.apply((Completable) v)
          .subscribe(new DeferredSubscriber(deferred)));
      }
      return v;
    });
  }

  /**
   * Apply the given function adapter to observables returned by routes:
   *
   * <pre>{@code
   * {
   *   use(new Rx().withObservable(observable -> observable.observeOn(Schedulers.io())));
   *
   *   get("observable", -> {
   *     return Observable...
   *   });
   * }
   * }</pre>
   *
   * @param adapter Observable adapter.
   * @return This module.
   */
  public Rx withObservable(final Function<Observable, Observable> adapter) {
    this.observable = requireNonNull(adapter, "Observable's adapter is required.");
    return this;
  }

  /**
   * Apply the given function adapter to single returned by routes:
   *
   * <pre>{@code
   * {
   *   use(new Rx().withSingle(observable -> observable.observeOn(Schedulers.io())));
   *
   *   get("single", -> {
   *     return Single...
   *   });
   * }
   * }</pre>
   *
   * @param adapter Single adapter.
   * @return This module.
   */
  public Rx withSingle(final Function<Single, Single> adapter) {
    this.single = requireNonNull(adapter, "Single's adapter is required.");
    return this;
  }

  /**
   * Apply the given function adapter to completable returned by routes:
   *
   * <pre>{@code
   * {
   *   use(new Rx().withObservable(observable -> observable.observeOn(Schedulers.io())));
   *
   *   get("completable", -> {
   *     return Completable...
   *   });
   * }
   * }</pre>
   *
   * @param adapter Completable adapter.
   * @return This module.
   */
  public Rx withCompletable(final Function<Completable, Completable> adapter) {
    this.completable = requireNonNull(adapter, "Completable's adapter is required.");
    return this;
  }

  @Override
  public void configure(final Env env, final Config conf, final Binder binder) {
    // dump rx.* as system properties
    conf.getConfig("rx")
      .withoutPath("schedulers").entrySet()
      .forEach(
        e -> System.setProperty("rx." + e.getKey(), e.getValue().unwrapped().toString()));
    Map<String, Executor> executors = new HashMap<>();
    super.configure(env, conf, binder, executors::put);

    env.router()
      .map(rx(observable, single, completable));

    /**
     * Side effects of global/evil static state. Hack to turn off some of this errors.
     */
    trySchedulerHook(executors);

    // shutdown schedulers: silent shutdown in dev mode between app reloads
    env.onStop(() -> {
      try {
        Schedulers.shutdown();
      } catch (Throwable ex) {
        log.debug("Schedulers.shutdown() resulted in error", ex);
      }
    });
  }

  @Override
  public Config config() {
    return ConfigFactory.parseResources(getClass(), "rx.conf");
  }

  private void trySchedulerHook(final Map<String, Executor> executors) {
    RxJavaPlugins plugins = RxJavaPlugins.getInstance();
    try {
      plugins.registerSchedulersHook(new ExecSchedulerHook(executors));
    } catch (IllegalStateException ex) {
      // there is a scheduler hook already, check if ours and ignore the exception
      RxJavaSchedulersHook hook = plugins.getSchedulersHook();
      if (!(hook instanceof ExecSchedulerHook)) {
        throw ex;
      }
    }
  }

}
